Package: clue
Version: 0.3-68
Encoding: UTF-8
Title: Cluster Ensembles
Description: CLUster Ensembles.
Authors@R: c(person("Kurt", "Hornik", role = c("aut", "cre"),
                    email = "Kurt.Hornik@R-project.org",
		    comment = c(ORCID = "0000-0003-4198-9911")),
             person("Walter", "Böhm", role = "ctb"))
License: GPL-2
Depends: R (>= 3.2.0)
Imports: stats, cluster, graphics, methods
Suggests: e1071, lpSolve (>= 5.5.7), quadprog (>= 1.4-8), relations
Enhances: RWeka, ape, cba, cclust, flexclust, flexmix, kernlab, mclust,
        movMF, modeltools
NeedsCompilation: yes
Packaged: 2026-03-26 14:54:57 UTC; hornik
Author: Kurt Hornik [aut, cre] (ORCID: <https://orcid.org/0000-0003-4198-9911>),
  Walter Böhm [ctb]
Maintainer: Kurt Hornik <Kurt.Hornik@R-project.org>
Repository: CRAN
Date/Publication: 2026-03-26 15:56:36 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 10:36:23 UTC; unix
Archs: clue.so.dSYM

Package: clue
Version: 0.3-65
Encoding: UTF-8
Title: Cluster Ensembles
Description: CLUster Ensembles.
Authors@R: c(person("Kurt", "Hornik", role = c("aut", "cre"),
                    email = "Kurt.Hornik@R-project.org",
		    comment = c(ORCID = "0000-0003-4198-9911")),
             person("Walter", "Böhm", role = "ctb"))
License: GPL-2
Depends: R (>= 3.2.0)
Imports: stats, cluster, graphics, methods
Suggests: e1071, lpSolve (>= 5.5.7), quadprog (>= 1.4-8), relations
Enhances: RWeka, ape, cba, cclust, flexclust, flexmix, kernlab, mclust,
        movMF, modeltools
NeedsCompilation: yes
Packaged: 2023-09-23 11:02:49 UTC; hornik
Author: Kurt Hornik [aut, cre] (<https://orcid.org/0000-0003-4198-9911>),
  Walter Böhm [ctb]
Maintainer: Kurt Hornik <Kurt.Hornik@R-project.org>
Repository: CRAN
Date/Publication: 2023-09-23 11:25:57 UTC
Built: R 4.2.0; aarch64-apple-darwin20; 2023-09-24 20:56:43 UTC; unix
Archs: clue.so.dSYM

useDynLib("clue", .registration = TRUE, .fixes = "C_")

import("stats")
importFrom("graphics", "par", "plot")
importFrom("cluster", "clara", "pam", "silhouette")

export("cl_agreement",
       "cl_bag",
       "cl_boot",
       "cl_consensus",
       "cl_dissimilarity",
       "cl_ensemble", "as.cl_ensemble", "is.cl_ensemble",
       "cl_fuzziness",
       "cl_join",
       "cl_margin",
       "cl_meet",
       "cl_medoid",
       "cl_membership", "as.cl_membership",
       "cl_object_names",
       "cl_pam",
       "cl_pclust",
       "cl_predict",
       "cl_prototypes",
       "cl_tabulate",
       "cl_ultrametric", "as.cl_ultrametric",
       "ls_fit_addtree",
       "ls_fit_centroid",
       "ls_fit_ultrametric",       
       "ls_fit_sum_of_ultrametrics",
       "ls_fit_ultrametric_target",
       "l1_fit_ultrametric",       
       "l1_fit_ultrametric_target",
       "cl_validity",
       "n_of_objects",
       "n_of_classes",
       "cl_class_ids", "as.cl_class_ids",
       "cl_classes",
       "is.cl_partition", "as.cl_partition",
       "is.cl_hard_partition", "as.cl_hard_partition",
       "is.cl_soft_partition", "is.cl_dendrogram",
       "as.cl_dendrogram",
       "is.cl_hierarchy", "as.cl_hierarchy",
       "as.cl_addtree",
       "kmedoids",
       "pclust", "pclust_family", "pclust_object",
       "solve_LSAP",
       "sumt")

## S3 methods, sorted alphabetically.

S3method("[", "cl_agreement")
S3method("[", "cl_dissimilarity")
S3method("[", "cl_ensemble")
S3method("[", "cl_proximity")

S3method("Complex", "cl_hierarchy")
S3method("Complex", "cl_partition")

S3method("Math", "cl_hierarchy")
S3method("Math", "cl_partition")

S3method("Ops", "cl_dendrogram")
S3method("Ops", "cl_hierarchy")
S3method("Ops", "cl_partition")

S3method("Summary", "cl_hierarchy")
S3method("Summary", "cl_partition")
S3method("Summary", "cl_partition_ensemble")

S3method("as.cl_addtree", "default")
S3method("as.cl_addtree", "phylo")

S3method("as.cl_membership", "default")
S3method("as.cl_membership", "matrix")

S3method("as.cl_ultrametric", "default")
S3method("as.cl_ultrametric", "matrix")

S3method("as.dendrogram", "cl_ultrametric")

S3method("as.dist", "cl_dissimilarity")

S3method("as.hclust", "cl_hierarchy")
S3method("as.hclust", "cl_ultrametric")
S3method("as.hclust", "mona")

S3method("as.matrix", "cl_proximity")
S3method("as.matrix", "hclust")

S3method("c", "cl_ensemble")

S3method("cl_class_ids", "Mclust")
S3method("cl_class_ids", "Weka_clusterer")
S3method("cl_class_ids", "bclust")
S3method("cl_class_ids", "ccfkms")
S3method("cl_class_ids", "cclust")
S3method("cl_class_ids", "cl_class_ids")
S3method("cl_class_ids", "cl_membership")
S3method("cl_class_ids", "cl_pam")
S3method("cl_class_ids", "cl_partition")
S3method("cl_class_ids", "cl_partition_by_class_ids")
S3method("cl_class_ids", "cshell")
S3method("cl_class_ids", "default")
S3method("cl_class_ids", "fclust")
S3method("cl_class_ids", "flexmix")
S3method("cl_class_ids", "kcca")
S3method("cl_class_ids", "kmeans")
S3method("cl_class_ids", "kmedoids")
S3method("cl_class_ids", "movMF")
S3method("cl_class_ids", "partition")
S3method("cl_class_ids", "pclust")
S3method("cl_class_ids", "relation")
S3method("cl_class_ids", "rock")
S3method("cl_class_ids", "specc")

S3method("cl_classes", "cl_hierarchy")
S3method("cl_classes", "cl_partition")
S3method("cl_classes", "cl_ultrametric")
S3method("cl_classes", "default")

S3method("cl_membership", "Mclust")
S3method("cl_membership", "bclust")
S3method("cl_membership", "cclust")
S3method("cl_membership", "cl_membership")
S3method("cl_membership", "cl_partition")
S3method("cl_membership", "cshell")
S3method("cl_membership", "default")
S3method("cl_membership", "fanny")
S3method("cl_membership", "fclust")
S3method("cl_membership", "flexmix")
S3method("cl_membership", "kmeans")
S3method("cl_membership", "movMF")
S3method("cl_membership", "partition")
S3method("cl_membership", "pclust")

S3method("cl_object_names", "cl_ensemble")
S3method("cl_object_names", "cl_hierarchy")
S3method("cl_object_names", "cl_membership")
S3method("cl_object_names", "cl_partition")
S3method("cl_object_names", "cl_ultrametric")
S3method("cl_object_names", "default")
S3method("cl_object_names", "dist")
S3method("cl_object_names", "hclust")
S3method("cl_object_names", "mona")
S3method("cl_object_names", "phylo")
S3method("cl_object_names", "twins")

S3method("cl_predict", "Mclust")
S3method("cl_predict", "Weka_clusterer")
S3method("cl_predict", "bclust")
S3method("cl_predict", "cclust")
S3method("cl_predict", "cl_partition")
S3method("cl_predict", "ccfkms")
S3method("cl_predict", "clara")
S3method("cl_predict", "cshell")
S3method("cl_predict", "default")
S3method("cl_predict", "fanny")
S3method("cl_predict", "fclust")
S3method("cl_predict", "flexmix")
S3method("cl_predict", "kcca")
S3method("cl_predict", "kmeans")
S3method("cl_predict", "movMF")
S3method("cl_predict", "pam")
S3method("cl_predict", "pclust")
S3method("cl_predict", "rock")

S3method("cl_prototypes", "Mclust")
S3method("cl_prototypes", "bclust")
S3method("cl_prototypes", "ccfkms")
S3method("cl_prototypes", "cclust")
S3method("cl_prototypes", "cl_pam")
S3method("cl_prototypes", "cl_partition")
S3method("cl_prototypes", "clara")
S3method("cl_prototypes", "cshell")
S3method("cl_prototypes", "fclust")
S3method("cl_prototypes", "kcca")
S3method("cl_prototypes", "kmeans")
S3method("cl_prototypes", "pam")
S3method("cl_prototypes", "pclust")
S3method("cl_prototypes", "specc")

S3method("cl_validity", "agnes")
S3method("cl_validity", "default")
S3method("cl_validity", "diana")
S3method("cl_validity", "cl_partition")
S3method("cl_validity", "pclust")

S3method("cophenetic", "cl_ultrametric")
S3method("cophenetic", "mona")

S3method("cut", "cl_dendrogram")

S3method("is.cl_dendrogram", "cl_dendrogram")
S3method("is.cl_dendrogram", "default")
S3method("is.cl_dendrogram", "hclust")
S3method("is.cl_dendrogram", "mona")
S3method("is.cl_dendrogram", "phylo")
S3method("is.cl_dendrogram", "twins")

S3method("is.cl_hard_partition", "Mclust")
S3method("is.cl_hard_partition", "Weka_clusterer")
S3method("is.cl_hard_partition", "bclust")
S3method("is.cl_hard_partition", "ccfkms")
S3method("is.cl_hard_partition", "cclust")
S3method("is.cl_hard_partition", "cl_hard_partition")
S3method("is.cl_hard_partition", "cl_partition")
S3method("is.cl_hard_partition", "cshell")
S3method("is.cl_hard_partition", "default")
S3method("is.cl_hard_partition", "fanny")
S3method("is.cl_hard_partition", "fclust")
S3method("is.cl_hard_partition", "flexmix")
S3method("is.cl_hard_partition", "kcca")
S3method("is.cl_hard_partition", "kmeans")
S3method("is.cl_hard_partition", "kmedoids")
S3method("is.cl_hard_partition", "movMF")
S3method("is.cl_hard_partition", "partition")
S3method("is.cl_hard_partition", "pclust")
S3method("is.cl_hard_partition", "rock")
S3method("is.cl_hard_partition", "specc")

S3method("is.cl_hierarchy", "cl_hierarchy")
S3method("is.cl_hierarchy", "default")
S3method("is.cl_hierarchy", "hclust")
S3method("is.cl_hierarchy", "mona")
S3method("is.cl_hierarchy", "phylo")
S3method("is.cl_hierarchy", "twins")

S3method("is.cl_partition", "Mclust")
S3method("is.cl_partition", "Weka_clusterer")
S3method("is.cl_partition", "bclust")
S3method("is.cl_partition", "ccfkms")
S3method("is.cl_partition", "cclust")
S3method("is.cl_partition", "cl_partition")
S3method("is.cl_partition", "cshell")
S3method("is.cl_partition", "default")
S3method("is.cl_partition", "fclust")
S3method("is.cl_partition", "flexmix")
S3method("is.cl_partition", "kcca")
S3method("is.cl_partition", "kmeans")
S3method("is.cl_partition", "kmedoids")
S3method("is.cl_partition", "movMF")
S3method("is.cl_partition", "partition")
S3method("is.cl_partition", "pclust")
S3method("is.cl_partition", "rock")
S3method("is.cl_partition", "specc")

S3method("n_of_classes", "Mclust")
S3method("n_of_classes", "bclust")
S3method("n_of_classes", "cclust")
S3method("n_of_classes", "cl_membership")
S3method("n_of_classes", "cl_partition")
S3method("n_of_classes", "cshell")
S3method("n_of_classes", "default")
S3method("n_of_classes", "fanny")
S3method("n_of_classes", "fclust")
S3method("n_of_classes", "kmeans")
S3method("n_of_classes", "partition")
S3method("n_of_classes", "pclust")

S3method("n_of_objects", "Mclust")
S3method("n_of_objects", "bclust")
S3method("n_of_objects", "cclust")
S3method("n_of_objects", "cl_ensemble")
S3method("n_of_objects", "cl_hierarchy")
S3method("n_of_objects", "cl_membership")
S3method("n_of_objects", "cl_partition")
S3method("n_of_objects", "cl_ultrametric")
S3method("n_of_objects", "cshell")
S3method("n_of_objects", "default")
S3method("n_of_objects", "dist")
S3method("n_of_objects", "fclust")
S3method("n_of_objects", "hclust")
S3method("n_of_objects", "kmeans")
S3method("n_of_objects", "mona")
S3method("n_of_objects", "partition")
S3method("n_of_objects", "pclust")
S3method("n_of_objects", "phylo")
S3method("n_of_objects", "twins")

S3method("names", "cl_proximity")

S3method("plot", "cl_addtree")
S3method("plot", "cl_ensemble")
S3method("plot", "cl_dendrogram")
S3method("plot", "cl_ultrametric")

S3method("print", "cl_class_ids")
S3method("print", "cl_classes_of_partition_of_objects")
S3method("print", "cl_classes_of_hierarchy_of_objects")
S3method("print", "cl_cross_proximity")
S3method("print", "cl_dendrogram")
S3method("print", "cl_dendrogram_ensemble")
S3method("print", "cl_ensemble")
S3method("print", "cl_fuzziness")
S3method("print", "cl_hierarchy")
S3method("print", "cl_hierarchy_ensemble")
S3method("print", "cl_membership")
S3method("print", "cl_pam")
S3method("print", "cl_partition")
S3method("print", "cl_partition_by_class_ids")
S3method("print", "cl_partition_by_memberships")
S3method("print", "cl_partition_ensemble")
S3method("print", "cl_proximity")
S3method("print", "cl_pclust")
S3method("print", "cl_validity")
S3method("print", "kmedoids")
S3method("print", "pclust")
S3method("print", "solve_LSAP")

S3method("rep", "cl_ensemble")

S3method("silhouette", "cl_partition")
S3method("silhouette", "cl_pclust")

S3method("unique", "cl_ensemble")

S3method(".maybe_is_proper_soft_partition", "Mclust")
S3method(".maybe_is_proper_soft_partition", "cl_partition")
S3method(".maybe_is_proper_soft_partition", "cshell")
S3method(".maybe_is_proper_soft_partition", "default") 
S3method(".maybe_is_proper_soft_partition", "fanny")
S3method(".maybe_is_proper_soft_partition", "fclust")
S3method(".maybe_is_proper_soft_partition", "flexmix")
S3method(".maybe_is_proper_soft_partition", "pclust")

# The Clue programming language

![image](https://user-images.githubusercontent.com/87673997/156028540-7a94db51-dd90-4bc6-9718-96e056d24cab.png)  
[![Crates.io](https://img.shields.io/crates/v/clue?logo=rust&style=for-the-badge)](https://crates.io/crates/clue)
[![Crates.io](https://img.shields.io/crates/d/clue?logo=rust&style=for-the-badge)](https://crates.io/crates/clue)
[![GitHub](https://img.shields.io/github/v/release/ClueLang/Clue?logo=github&color=181717&display_name=release&include_prereleases&sort=semver&style=for-the-badge)](https://github.com/ClueLang/Clue)
[![AUR](https://img.shields.io/aur/version/clue?color=1793d1&logo=arch-linux&style=for-the-badge)](https://aur.archlinux.org/packages/clue/)

Clue is a programming language that compiles blazingly fast into Lua code with a syntax similar to languages like C or Rust.

Clue tries to be almost as simple as Lua (with only a slightly more advanced syntax) but adds many optional features that can make code look better or make some things (like metatables) easier to code.

Clue does not compile to a specfic version of Lua: flags can be toggled to alter the output to allow most if not all versions or modifications of Lua to be compiled to with Clue.

## General syntax differences
- Code blocks are now inside `{}` instead of `then`/`do`/`repeat` and `end`/`until`
- Comments are made with `// ...` or `/* ... */`

If you want a complete documentation of every change and addition in Clue check [the wiki](https://github.com/ClueLang/Clue/wiki).

## Example code
```rs
@ifos linux {
	@define GREETING "Hello, Linux user "
} @else_ifos macos {
	@define GREETING "Hello, MacOS user "
} @else {
	@define GREETING "Hello, Windows user "
}
  
@macro GREET(target) { $GREETING .. $target .. "!" }
  
print($GREET!("Maiori"))

local fn add(x = 0, y = 0) {
    return x + y
}

global n = 1

while n < 10 {
    n += add($, $)
    match n {
        3 => {
            continue
        }
        4 if x => {
            break
        }
        default => {
            print(n < 3 ? n : -n)
        }
    }
}
```
More examples can be found in:
- the [wiki](https://github.com/ClueLang/Clue/wiki)
- the [Examples directory](https://github.com/ClueLang/Clue/tree/main/examples)
- a [game made with LÖVE](https://github.com/Maiori44/ip-please)
- Clue's [MessagePack library](https://github.com/Maiori44/msgpack-clue)
- a [simple graph visualizer made with LÖVE](https://github.com/ClueLang/Clue-example)

## How to install

### Using Cargo
1. Paste and run this command in the console: `cargo install clue`
2. Type `clue` in the console to run the compiler, it will explain the rest

Clue supports extra features that can be toggled when installing:
* `interpreter`: adds the `--execute` flag to let Clue run the generated output using [mlua](https://github.com/khvzak/mlua)
* `rpmalloc`: uses [rpmalloc](https://github.com/EmbarkStudios/rpmalloc-rs) to improve performance, not available on all platforms

By default Clue enables both features.

### Using Linux packages
These can be downloaded in the [latest release](https://github.com/ClueLang/Clue/releases/latest).
* .deb
```
sudo dpkg -i clue_<version>_<arch>.deb
```
* .rpm
```
sudo rpm -i clue-<version>.<arch>.rpm
```

### Using the [AUR](https://aur.archlinux.org/clue)
* With paru
```sh
paru -S clue
```
* With yay
```
yay -S clue
```
* With makepkg
```sh
git clone https://aur.archlinux.org/clue.git
cd clue
makepkg -si
```

### Manual insallation
1. Download the [latest release](https://github.com/ClueLang/Clue/releases/latest) and save it somewhere
2. Open your system environment variables
3. Add the path to the directory that contains `clue.exe` in the PATH variable
4. Type `clue` in your cmd/PowerShell to run the compiler, it will explain the rest

## More coming soon!
There are still features that I'm considering adding and others that will be added soon.
The most likely ones to be added in the future are:
- type system (coming in 4.0)
- better error messages (comming in 4.0)
- proper language server support (coming in 4.0)

For any suggestion or bug you can make a github issue.
If you need help with the language itself, you can check out the [Discord server](https://discord.gg/EQsnWpqN3C).

I hope Clue will be useful to you :)

## Why is Clue named Clue?
~~I have no *clue*.~~

bibentry(bibtype = "Book",
         key = "Arabie+Carroll+DeSarbo:1987",
         author = c(person(given = "Phipps",
                           family = "Arabie"),
                    person(given = c("J.", "Douglas"),
                           family = "Carroll"),
                    person(given = c("Wayne", "S."),
                           family = "DeSarbo")),
         title = "Three-way Scaling and Clustering",
         publisher = "Sage Publications",
         year = "1987",
         series = "Quantitative Applications in the Social Sciences",
         address = "Newbury Park, CA",
         doi = "10.4135/9781412986359")

bibentry(bibtype = "Book",
         key = "Barthélemy+Guénoche:1991",
         author = c(person(given = "Jean-Pierre",
                           family = "Barthélemy"),
                    person(given = "Alain",
                           family = "Guénoche")),
         title = "Trees and Proximity Representations",
         publisher = "Wiley",
         year = "1991",
         series = "Wiley Series in Discrete Mathematics and Optimization",
         address = "Chicester",
         isbn = "978-0471922636")

bibentry(bibtype = "Article",
         key = "Barthélemy+Leclerc+Monjardet:1986",
         title = "On the Use of Ordered Sets in Problems of Comparison and Consensus of Classifications",
         volume = "3",
         .issn = "1432-1343",
         doi = "10.1007/bf01894188",
         number = "2",
         journal = "Journal of Classification",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = "Jean-Pierre",
                           family = "Barthélemy"),
                    person(given = "Bruno",
                           family = "Leclerc"),
                    person(given = "Bernard",
                           family = "Monjardet")),
         year = "1986",
         month = "sep",
         pages = "187--224")

bibentry(bibtype = "Book",
         key = "Bezdek:1981",
         author = person(given = c("James", "C."),
                         family = "Bezdek"),
         title = "Pattern Recognition with Fuzzy Objective Function Algorithms",
         publisher = "Plenum",
         year = "1981",
         address = "New York",
         doi = "10.1007/978-1-4757-0450-1")

bibentry(bibtype = "InCollection",
         key = "Boorman+Arabie:1972",
         author = c(person(given = c("Scott", "A."),
                           family = "Boorman"),
                    person(given = "Phipps",
                           family = "Arabie")),
         title = "Structural Measures and the Method of Sorting",
         booktitle = "Multidimensional Scaling: Theory and Applications in the Behavioral Sciences, 1: Theory",
         publisher = "Seminar Press",
         year = "1972",
         editor = c(person(given = c("R.", "N."),
                           family = "Shepard"),
                    person(given = c("A.", "K."),
                           family = "Romney"),
                    person(given = c("S.", "B."),
                           family = "Nerlove")),
         pages = "225--249",
         address = "New York")

bibentry(bibtype = "Article",
         key = "Boorman+Olivier:1973",
         title = "Metrics on Spaces of Finite Trees",
         volume = "10",
         .issn = "0022-2496",
         doi = "10.1016/0022-2496(73)90003-5",
         number = "1",
         journal = "Journal of Mathematical Psychology",
         publisher = "Elsevier BV",
         author = c(person(given = c("Scott", "A."),
                           family = "Boorman"),
                    person(given = c("Donald", "C."),
                           family = "Olivier")),
         year = "1973",
         month = "feb",
         pages = "26--59")

bibentry(bibtype = "InCollection",
         key = "Carroll+Pruzansky:1980",
         author = c(person(given = c("J.", "D."),
                           family = "Carroll"),
                    person(given = "S.",
                           family = "Pruzansky")),
         title = "Discrete and Hybrid Scaling Models",
         booktitle = "Similarity and Choice",
         publisher = "Huber",
         year = "1980",
         editor = c(person(given = c("E.", "D."),
                           family = "Lantermann"),
                    person(given = "H.",
                           family = "Feger")),
         address = "Bern, Switzerland")

bibentry(bibtype = "Article",
         key = "Charon+Denoeud+Guénoche:2006",
         title = "Maximum Transfer Distance Between Partitions",
         volume = "23",
         .issn = "1432-1343",
         doi = "10.1007/s00357-006-0006-2",
         number = "1",
         journal = "Journal of Classification",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = "Irene",
                           family = "Charon"),
                    person(given = "Lucile",
                           family = "Denoeud"),
                    person(given = "Alain",
                           family = "Guénoche"),
                    person(given = "Olivier",
                           family = "Hudry")),
         year = "2006",
         month = "jun",
         pages = "103--121")

bibentry(bibtype = "Article",
         key = "De_Soete:1983",
         title = "A Least Squares Algorithm for Fitting Additive Trees to Proximity Data",
         volume = "48",
         .issn = "1860-0980",
         doi = "10.1007/bf02293884",
         number = "4",
         journal = "Psychometrika",
         publisher = "Cambridge University Press (CUP)",
         author = person(given = "Geert",
                         family = "De Soete"),
         year = "1983",
         month = "dec",
         pages = "621--626")

bibentry(bibtype = "Article",
         key = "Day:1981",
         title = "The Complexity of Computing Metric Distances Between Partitions",
         volume = "1",
         .issn = "0165-4896",
         doi = "10.1016/0165-4896(81)90042-1",
         number = "3",
         journal = "Mathematical Social Sciences",
         publisher = "Elsevier BV",
         author = person(given = c("William", "H.", "E."),
                         family = "Day"),
         year = "1981",
         month = "may",
         pages = "269--287")

bibentry(bibtype = "Article",
         key = "De_Soete:1984a",
         title = "A Least Squares Algorithm for Fitting an Ultrametric Tree to a Dissimilarity Matrix",
         volume = "2",
         .issn = "0167-8655",
         doi = "10.1016/0167-8655(84)90036-9",
         number = "3",
         journal = "Pattern Recognition Letters",
         publisher = "Elsevier BV",
         author = person(given = "Geert",
                         family = "De Soete"),
         year = "1984",
         month = "mar",
         pages = "133--137")

bibentry(bibtype = "Article",
         key = "De_Soete:1984b",
         title = "Ultrametric Tree Representations of Incomplete Dissimilarity Data",
         volume = "1",
         .issn = "1432-1343",
         doi = "10.1007/bf01890124",
         number = "1",
         journal = "Journal of Classification",
         publisher = "Springer Science and Business Media LLC",
         author = person(given = "Geert",
                         family = "De Soete"),
         year = "1984",
         month = "dec",
         pages = "235--242")

bibentry(bibtype = "Article",
         key = "Dimitriadou+Weingessel+Hornik:2002",
         title = "A Combination Scheme for Fuzzy Clustering",
         volume = "16",
         .issn = "1793-6381",
         doi = "10.1142/s0218001402002052",
         number = "07",
         journal = "International Journal of Pattern Recognition and Artificial Intelligence",
         publisher = "World Scientific Pub Co Pte Lt",
         author = c(person(given = "Evgenia",
                           family = "Dimitriadou"),
                    person(given = "Andreas",
                           family = "Weingessel"),
                    person(given = "Kurt",
                           family = "Hornik")),
         year = "2002",
         month = "nov",
         pages = "901--912")

bibentry(bibtype = "Article",
         key = "Dudoit+Fridlyand:2003",
         title = "Bagging to Improve the Accuracy of a Clustering Procedure",
         volume = "19",
         .issn = "1367-4803",
         doi = "10.1093/bioinformatics/btg038",
         number = "9",
         journal = "Bioinformatics",
         publisher = "Oxford University Press (OUP)",
         author = c(person(given = "Sandrine",
                           family = "Dudoit"),
                    person(given = "Jane",
                           family = "Fridlyand")),
         year = "2003",
         month = "jun",
         pages = "1090--1099")

bibentry(bibtype = "Book",
         key = "Fiacco+McCormick:1968",
         author = c(person(given = c("Anthony", "V."),
                           family = "Fiacco"),
                    person(given = c("Garth", "P."),
                           family = "McCormick")),
         title = "Nonlinear Programming: Sequential Unconstrained Minimization Techniques",
         publisher = "John Wiley & Sons",
         year = "1968",
         address = "New York")

bibentry(bibtype = "Article",
         key = "Fowlkes+Mallows:1983",
         title = "A Method for Comparing Two Hierarchical Clusterings",
         volume = "78",
         .issn = "1537-274X",
         doi = "10.1080/01621459.1983.10478008",
         number = "383",
         journal = "Journal of the American Statistical Association",
         publisher = "Informa UK Limited",
         author = c(person(given = c("E.", "B."),
                           family = "Fowlkes"),
                    person(given = c("C.", "L."),
                           family = "Mallows")),
         year = "1983",
         month = "sep",
         pages = "553--569")

bibentry(bibtype = "Article",
         key = "Gaul+Schader:1988",
         title = "Clusterwise Aggregation of Relations",
         volume = "4",
         .issn = "1099-0747",
         doi = "10.1002/asm.3150040406",
         number = "4",
         journal = "Applied Stochastic Models and Data Analysis",
         publisher = "Wiley",
         author = c(person(given = "Wolfgang",
                           family = "Gaul"),
                    person(given = "Martin",
                           family = "Schader")),
         year = "1988",
         month = "dec",
         pages = "273--282")

bibentry(bibtype = "Article",
         key = "Gusfield:2002",
         title = "Partition-Distance: A Problem and Class of Perfect Graphs Arising in Clustering",
         volume = "82",
         .issn = "0020-0190",
         doi = "10.1016/s0020-0190(01)00263-0",
         number = "3",
         journal = "Information Processing Letters",
         publisher = "Elsevier BV",
         author = person(given = "Dan",
                         family = "Gusfield"),
         year = "2002",
         month = "may",
         pages = "159--164")

bibentry(bibtype = "Article",
         key = "Gordon+Vichi:1998",
         title = "Partitions of Partitions",
         volume = "15",
         .issn = "1432-1343",
         doi = "10.1007/s003579900034",
         number = "2",
         journal = "Journal of Classification",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = c("Allan", "Drummond"),
                           family = "Gordon"),
                    person(given = "Maurizio",
                           family = "Vichi")),
         year = "1998",
         month = "feb",
         pages = "265--285")

bibentry(bibtype = "Article",
         key = "Gordon+Vichi:2001",
         title = "Fuzzy Partition Models for Fitting a Set of Partitions",
         volume = "66",
         .issn = "1860-0980",
         doi = "10.1007/bf02294837",
         number = "2",
         journal = "Psychometrika",
         publisher = "Cambridge University Press (CUP)",
         author = c(person(given = c("Allan", "Drummond"),
                           family = "Gordon"),
                    person(given = "Maurizio",
                           family = "Vichi")),
         year = "2001",
         month = "jun",
         pages = "229--247")

bibentry(bibtype = "Article",
         key = "Hubert+Arabie:1985",
         title = "Comparing Partitions",
         volume = "2",
         .issn = "1432-1343",
         doi = "10.1007/bf01908075",
         number = "1",
         journal = "Journal of Classification",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = "Lawrence",
                           family = "Hubert"),
                    person(given = "Phipps",
                           family = "Arabie")),
         year = "1985",
         month = "dec",
         pages = "193--218")

bibentry(bibtype = "Article",
         key = "Hubert+Arabie:1995",
         title = "Iterative Projection Strategies for the Least-Squares Fitting of Tree Structures to Proximity Data",
         volume = "48",
         .issn = "2044-8317",
         doi = "10.1111/j.2044-8317.1995.tb01065.x",
         number = "2",
         journal = "British Journal of Mathematical and Statistical Psychology",
         publisher = "Wiley",
         author = c(person(given = "Lawrence",
                           family = "Hubert"),
                    person(given = "Phipps",
                           family = "Arabie")),
         year = "1995",
         month = "nov",
         pages = "281--317")

bibentry(bibtype = "Book",
         key = "Hubert+Arabie+Meulman:2006",
         author = c(person(given = "Lawrence",
                           family = "Hubert"),
                    person(given = "Phipps",
                           family = "Arabie"),
                    person(given = "Jacqueline",
                           family = "Meulman")),
         title = "The Structural Representation of Proximity Matrices with {MATLAB}",
         publisher = "SIAM",
         year = "2006",
         series = "ASA-SIAM Series on Statistics and Applied Mathematics",
         address = "Philadelphia, PA",
         doi = "10.1137/1.9780898718355")

bibentry(bibtype = "Book",
         key = "Jardine+Sibson:1971",
         author = c(person(given = "Nicholas",
                           family = "Jardine"),
                    person(given = "Robin",
                           family = "Sibson")),
         title = "Mathematical Taxonomy",
         publisher = "Wiley",
         year = "1971",
         address = "London")

bibentry(bibtype = "Article",
         key = "Katz+Powell:1953",
         title = "A Proposed Index of the Conformity of One Sociometric Measurement to Another",
         volume = "18",
         .issn = "1860-0980",
         doi = "10.1007/bf02289063",
         number = "3",
         journal = "Psychometrika",
         publisher = "Cambridge University Press (CUP)",
         author = c(person(given = "Leo",
                           family = "Katz"),
                    person(given = c("James", "H."),
                           family = "Powell")),
         year = "1953",
         month = "sep",
         pages = "249--256")

bibentry(bibtype = "Book",
         key = "Kaufman+Rousseeuw:1990",
         author = c(person(given = "Leonard",
                           family = "Kaufman"),
                    person(given = c("Peter", "J."),
                           family = "Rousseeuw")),
         title = "Finding Groups in Data: An Introduction to Cluster Analysis",
         publisher = "Wiley",
         year = "1990",
         address = "New York",
         doi = "10.1002/9780470316801")

bibentry(bibtype = "Article",
         key = "Křivánek+Morávek:1986",
         title = "{NP}-hard Problems in Hierarchical-Tree Clustering",
         volume = "23",
         .issn = "1432-0525",
         doi = "10.1007/bf00289116",
         number = "3",
         journal = "Acta Informatica",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = "Mirko",
                           family = "Křivánek"),
                    person(given = "Jaroslav",
                           family = "Morávek")),
         year = "1986",
         month = "jun",
         pages = "311--323")

bibentry(bibtype = "Misc",
         key = "Leisch:1999",
         doi = "10.57938/9B129F95-B53B-44CE-A129-5B7A1168D832",
         author = person(given = "Friedrich",
                         family = "Leisch"),
         language = "en",
         title = "Bagged Clustering",
         publisher = "Vienna University of Economics and Business",
         year = "1999")

bibentry(bibtype = "Article",
         key = "Margush+McMorris:1981",
         title = "Consensus $n$-Trees",
         volume = "43",
         .issn = "1522-9602",
         doi = "10.1007/bf02459446",
         number = "2",
         journal = "Bulletin of Mathematical Biology",
         publisher = "Springer Science and Business Media LLC",
         author = c(person(given = "T.",
                           family = "Margush"),
                    person(given = c("F.", "R."),
                           family = "McMorris")),
         year = "1981",
         month = "mar",
         pages = "239--244")

bibentry(bibtype = "InProceedings",
         key = "Meila:2003",
         author = person(given = "Marina",
                         family = "Meilă"),
         editor = c(person(given = "Bernhard",
                           family = "Schölkopf"),
                    person(given = c("Manfred", "K."),
                           family = "Warmuth")),
         title = "Comparing Clusterings by the Variation of Information",
         booktitle = "Learning Theory and Kernel Machines",
         year = "2003",
         publisher = "Springer Berlin Heidelberg",
         address = "Berlin, Heidelberg",
         pages = "173--187",
         isbn = "978-3-540-45167-9",
         doi = "10.1007/978-3-540-45167-9_14")

bibentry(bibtype = "Article",
         key = "Mérigot+Durbec+Gaertner:2010",
         title = "On Goodness-of-Fit Measure for Dendrogram‐Based Analyses",
         volume = "91",
         .issn = "1939-9170",
         doi = "10.1890/09-1387.1",
         number = "6",
         journal = "Ecology",
         publisher = "Wiley",
         author = c(person(given = "Bastien",
                           family = "Mérigot"),
                    person(given = "Jean-Pierre",
                           family = "Durbec"),
                    person(given = "Jean-Claude",
                           family = "Gaertner")),
         year = "2010",
         month = "jun",
         pages = "1850--1859")

bibentry(bibtype = "Article",
         key = "Miller+Nicely:1955",
         title = "An Analysis of Perceptual Confusions Among Some English Consonants",
         volume = "27",
         .issn = "1520-8524",
         doi = "10.1121/1.1907526",
         number = "2",
         journal = "The Journal of the Acoustical Society of America",
         publisher = "Acoustical Society of America (ASA)",
         author = c(person(given = c("George", "A."),
                           family = "Miller"),
                    person(given = c("Patricia", "E."),
                           family = "Nicely")),
         year = "1955",
         month = "mar",
         pages = "338--352")

bibentry(bibtype = "Book",
         key = "Papadimitriou+Steiglitz:1982",
         author = c(person(given = c("Christos", "H."),
                           family = "Papadimitriou"),
                    person(given = "Kenneth",
                           family = "Steiglitz")),
         title = "Combinatorial Optimization: Algorithms and Complexity",
         publisher = "Prentice Hall",
         year = "1982",
         address = "Englewood Cliffs")

bibentry(bibtype = "Article",
         key = "Rajski:1961",
         title = "A Metric Space of Discrete Probability Distributions",
         volume = "4",
         .issn = "0019-9958",
         doi = "10.1016/s0019-9958(61)80055-7",
         number = "4",
         journal = "Information and Control",
         publisher = "Elsevier BV",
         author = person(given = "C.",
                         family = "Rajski"),
         year = "1961",
         month = "dec",
         pages = "371--377")

bibentry(bibtype = "Article",
         key = "Rand:1971",
         title = "Objective Criteria for the Evaluation of Clustering Methods",
         volume = "66",
         .issn = "1537-274X",
         doi = "10.1080/01621459.1971.10482356",
         number = "336",
         journal = "Journal of the American Statistical Association",
         publisher = "Informa UK Limited",
         author = person(given = c("William", "M."),
                         family = "Rand"),
         year = "1971",
         month = "dec",
         pages = "846--850")

bibentry(bibtype = "InCollection",
         key = "Rosenberg:1982",
         author = person(given = "Seymour",
                         family = "Rosenberg"),
         title = "The Method of Sorting in Multivariate Research with Applications Selected from Cognitive Psychology and Person Perception",
         booktitle = "Multivariate Applications in the Social Sciences",
         publisher = "Erlbaum",
         year = "1982",
         editor = c(person(given = "N.",
                           family = "Hirschberg"),
                    person(given = c("L.", "G."),
                           family = "Humphreys")),
         pages = "117--142",
         address = "Hillsdale, NJ",
         doi = "10.4324/9781315802831")

bibentry(bibtype = "Article",
         key = "Rosenberg+Kim:1975",
         title = "The Method of Sorting as a Data-Gathering Procedure in Multivariate Research",
         volume = "10",
         .issn = "1532-7906",
         doi = "10.1207/s15327906mbr1004_7",
         number = "4",
         journal = "Multivariate Behavioral Research",
         publisher = "Informa UK Limited",
         author = c(person(given = "Seymour",
                           family = "Rosenberg"),
                    person(given = c("Moonja", "Park"),
                           family = "Kim")),
         year = "1975",
         month = "oct",
         pages = "489--502")

bibentry(bibtype = "InCollection",
         key = "Roux:1988",
         title = "Techniques of Approximation for Building Two Tree Structures",
         editor = c(person(given = "Chikio",
                           family = "Hayashi"),
                    person(given = "Michel",
                           family = "Jambu"),
                    person(given = "Edwin",
                           family = "Diday"),
                    person(given = "Noboru",
                           family = "Ohsumi")),
         booktitle = "Recent Developments in Clustering and Data Analysis",
         publisher = "Academic Press",
         pages = "151--170",
         year = "1988",
         isbn = "978-0-12-215485-0",
         doi = "10.1016/B978-0-12-215485-0.50017-X",
         author = person(given = "M.",
                         family = "Roux"))

bibentry(bibtype = "Article",
         key = "Rubin:1967",
         title = "Optimal Classification into Groups: An Approach for Solving The Taxonomy Problem",
         volume = "15",
         .issn = "0022-5193",
         doi = "10.1016/0022-5193(67)90046-x",
         number = "1",
         journal = "Journal of Theoretical Biology",
         publisher = "Elsevier BV",
         author = person(given = "Jerrold",
                         family = "Rubin"),
         year = "1967",
         month = "apr",
         pages = "103--144")

bibentry(bibtype = "Article",
         key = "Sato+Sato:1994",
         title = "On a Multicriteria Fuzzy Clustering Method for 3-Way Data",
         volume = "02",
         .issn = "1793-6411",
         doi = "10.1142/s0218488594000122",
         number = "02",
         journal = "International Journal of Uncertainty, Fuzziness and Knowledge-Based Systems",
         publisher = "World Scientific Pub Co Pte Lt",
         author = c(person(given = "Mika",
                           family = "Sato"),
                    person(given = "Yoshiharu",
                           family = "Sato")),
         year = "1994",
         month = "jun",
         pages = "127--142")

bibentry(bibtype = "Article",
         key = "Smith:2001",
         author = person(given = c("Thomas", "J."),
                         family = "Smith"),
         title = "Constructing Ultrametric and Additive Trees Based on the $L_1$ Norm",
         journal = "Journal of Classification",
         year = "2001",
         volume = "18",
         number = "2",
         pages = "185--207",
         doi = "10.1007/s00357-001-0015-0")

bibentry(bibtype = "Article",
         key = "Strehl+Ghosh:2002",
         author = c(person(given = "Alexander",
                           family = "Strehl"),
                    person(given = "Joydeep",
                           family = "Ghosh")),
         title = "Cluster Ensembles --- A Knowledge Reuse Framework for Combining Multiple Partition",
         journal = "Journal of Machine Learning Research",
         year = "2002",
         volume = "3",
         pages = "583--617",
         url = "https://jmlr.org/papers/v3/strehl02a.html")

bibentry(bibtype = "Article",
         key = "Tibshirani+Walther:2005",
         title = "Cluster Validation by Prediction Strength",
         volume = "14",
         .issn = "1537-2715",
         doi = "10.1198/106186005x59243",
         number = "3",
         journal = "Journal of Computational and Graphical Statistics",
         publisher = "Informa UK Limited",
         author = c(person(given = "Robert",
                           family = "Tibshirani"),
                    person(given = "Guenther",
                           family = "Walther")),
         year = "2005",
         month = "sep",
         pages = "511--528")

bibentry(bibtype = "InProceedings",
         key = "Zhou+Li+Zha:2005",
         series = "ICML ’05",
         title = "A New {Mallows} Distance Based Metric for Comparing Clusterings",
         doi = "10.1145/1102351.1102481",
         booktitle = "Proceedings of the 22nd international conference on Machine learning -- ICML ’05",
         publisher = "ACM Press",
         author = c(person(given = "Ding",
                           family = "Zhou"),
                    person(given = "Jia",
                           family = "Li"),                    
                    person(given = "Hongyuan",
                           family = "Zha")),
         year = "2005",
         pages = "1028--1035",
         collection = "ICML ’05")

cl_ensemble <-
function(..., list = NULL)
{
    clusterings <- c(list(...), list)

    if(!length(clusterings)) {
        ## Return an empty cl_ensemble.
        ## In this case, we cannot additionally know whether it contains
        ## partitions or hierarchies ...
        attr(clusterings, "n_of_objects") <- as.integer(NA)
        class(clusterings) <- "cl_ensemble"
        return(clusterings)
    }

    ## Previously, we used to require that the elements of the ensemble
    ## either all be partitions, or all be hierarchies.  We no longer do
    ## this, as it makes sense to also allow e.g. object dissimilarities
    ## (raw "dist" objects or additive distances) as elements (e.g.,
    ## when computing proximities), and it is rather cumbersome to
    ## decide in advance which combinations of elements might be useful
    ## and hence should be allowed.  All we enforce is that all elements
    ## correspond to the same number of objects (as we typically cannot
    ## verify that they relate to the *same* objects).  For "pure"
    ## ensembles of partitions or hierarchies we add additional class
    ## information.

    if(all(vapply(clusterings, is.cl_partition, NA)))
        class(clusterings) <- c("cl_partition_ensemble", "cl_ensemble")
    else if(all(vapply(clusterings, is.cl_dendrogram, NA)))
        class(clusterings) <- c("cl_dendrogram_ensemble",
                                "cl_hierarchy_ensemble", "cl_ensemble")
    else if(all(vapply(clusterings, is.cl_hierarchy, NA)))
        class(clusterings) <- c("cl_hierarchy_ensemble", "cl_ensemble")
    else
        class(clusterings) <- "cl_ensemble"

    n <- sapply(clusterings, n_of_objects)
    if(any(diff(n)))
        stop("All elements must have the same number of objects.")
    attr(clusterings, "n_of_objects") <- as.integer(n[1L])

    clusterings
}

is.cl_ensemble <-
function(x)
    inherits(x, "cl_ensemble")

## <NOTE>
## In the old days, kmeans() results were unclassed lists, hence such
## objects were taken as representing a single clustering.  Nowadays, we
## take these as lists of clusterings.
as.cl_ensemble <-
function(x)
{
    if(is.cl_ensemble(x)) x
    else if(is.list(x) && !is.object(x)) cl_ensemble(list = x)
    else cl_ensemble(x)
}
## </NOTE>

c.cl_ensemble <-
function(..., recursive = FALSE)
{
    clusterings <- unlist(lapply(list(...), as.cl_ensemble),
                          recursive = FALSE)
    cl_ensemble(list = clusterings)
}

"[.cl_ensemble" <-
function(x, i)
{
    ## Make subscripting empty ensembles a noop.
    if(length(x) == 0L) return(x)
    cl_ensemble(list = NextMethod("["))
}

rep.cl_ensemble <-
function(x, times, ...)
    cl_ensemble(list = NextMethod("rep"))

print.cl_partition_ensemble <-
function(x, ...)
{
    msg <- sprintf(ngettext(length(x),
                            "An ensemble of %d partition of %d objects.",
                            "An ensemble of %d partitions of %d objects."),
                   length(x), n_of_objects(x))
    writeLines(strwrap(msg))
    invisible(x)
}

Summary.cl_partition_ensemble <-
function(..., na.rm = FALSE)
{
    ok <- switch(.Generic, max = , min = , range = TRUE, FALSE)
    if(!ok)
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)
    args <- list(...)
    ## Combine the given partition ensembles.
    x <- do.call(c, args)
    switch(.Generic,
           "min" = cl_meet(x),
           "max" = cl_join(x),
           "range" = cl_ensemble(min = cl_meet(x), max = cl_join(x)))
}

print.cl_dendrogram_ensemble <-
function(x, ...)
{
    msg <- sprintf(ngettext(length(x),
                            "An ensemble of %d dendrogram of %d objects.",
                            "An ensemble of %d dendrograms of %d objects."),
                   length(x), n_of_objects(x))
    writeLines(strwrap(msg))
    invisible(x)
}

print.cl_hierarchy_ensemble <-
function(x, ...)
{
    msg <- sprintf(ngettext(length(x),
                            "An ensemble of %d hierarchy of %d objects.",
                            "An ensemble of %d hierarchies of %d objects."),
                   length(x), n_of_objects(x))
    writeLines(strwrap(msg))
    invisible(x)
}

print.cl_ensemble <-
function(x, ...)
{
    writeLines(sprintf(ngettext(length(x),
                                "An ensemble with %d element.",
                                "An ensemble with %d elements."),
                       length(x)))
    invisible(x)
}

plot.cl_ensemble <-
function(x, ..., main = NULL, layout = NULL)
{
    if(!is.cl_ensemble(x))
        stop("Wrong class.")

    ## What we can definitely plot is are cl_addtree, cl_dendrogram and
    ## cl_ultrametric objects.  (We could also add simple methods for
    ## plotting raw dissimilarities, but of course seriation::dissplot()
    ## would be the thing to use.)  What we cannot reasonably plot is
    ## partitions (in particular, as these do not know about the
    ## underlying dissimilarities.  But then we could perhaps provide
    ## silhoutte plots etc for ensembles of partitions ...
    ## <FIXME>
    ## Think about this.
    ## </FIXME>
    ## So let us check for the things we can plot.

    ## (Note that currently there is neither is.cl_ultrametric() nor
    ## is.cl_addtree().)
    ok <- vapply(x,
                 function(e)
                 (is.cl_dendrogram(e) ||
                  inherits(e, c("cl_addtree", "cl_ultrametric"))),
                 NA)
    if(!all(ok))
        stop(gettextf("Plotting not available for elements %s of the ensemble.",
                      paste(which(!ok), collapse = " ")),
             domain = NA)

    ## Prefer dendrogram plot methods to those for hclust objects.
    ind <- vapply(x, is.cl_dendrogram, NA)
    if(any(ind))
        x[ind] <- lapply(x, as.cl_dendrogram)

    ## Now the usual layouting ... same as for plotting relation
    ## ensembles.

    ## Number of elements.
    n <- length(x)
    ## Layout.
    byrow <- TRUE
    if(is.null(layout)) {
        nc <- ceiling(sqrt(n))
        nr <- ceiling(n / nc)
    }
    else {
        layout <- c(as.list(layout), byrow)[seq_len(3)]
        if(is.null(names(layout)))
            names(layout) <- c("nr", "nc", "byrow")
        nr <- layout[["nr"]]
        nc <- layout[["nc"]]
        byrow <- layout[["byrow"]]
    }
    op <- if(byrow)
        par(mfrow = c(nr, nc))
    else
        par(mfcol = c(nr, nc))
    on.exit(par(op))

    ## Try recycling main (might want the same for others as well).
    if(!is.list(main)) {
        main <- if(is.null(main))
            vector("list", length = n)
        else
            rep.int(as.list(main), n)
    }
    
    for(i in seq_along(x))
        plot(x[[i]], main = main[[i]], ...)
}
                   
unique.cl_ensemble <-
function(x, incomparables = FALSE, ...)
    cl_ensemble(list = NextMethod("unique"))

.cl_ensemble_type <-
function(x)
{
    if(inherits(x, "cl_partition_ensemble"))
        "partition"
    else if(inherits(x, "cl_hierarchy_ensemble"))
        "hierarchy"
    else
        NULL
}

### * cl_consensus

cl_consensus <-
function(x, method = NULL, weights = 1, control = list())
{
    ## <NOTE>
    ## Interfaces are a matter of taste.
    ## E.g., one might want to have a 'type' argument indication whether
    ## hard or soft partitions are sought.  One could then do
    ##   cl_consensus(x, method = "euclidean", type = "hard")
    ## to look for an optimal median (or least squares) hard partition
    ## (for euclidean dissimilarity).
    ## For us, "method" really indicates a certain algorithm, with its
    ## bells and whistles accessed via the 'control' argument.
    ## </NOTE>

    clusterings <- as.cl_ensemble(x)

    if(!length(clusterings))
        stop("Cannot compute consensus of empty ensemble.")

    weights <- rep_len(weights, length(clusterings))
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")

    if(!is.function(method)) {
        if(!inherits(method, "cl_consensus_method")) {
            ## Get the method definition from the registry.
            type <- .cl_ensemble_type(clusterings)
            if(is.null(method))
                method <- .cl_consensus_method_default(type)                
            method <- get_cl_consensus_method(method, type)            
        }
        method <- method$definition
    }

    method(clusterings, weights, control)
}

### * .cl_consensus_partition_DWH

.cl_consensus_partition_DWH <-
function(clusterings, weights, control)
{
    ## <TODO>
    ## Could make things more efficient by subscripting on positive
    ## weights.
    ## (Note that this means control$order has to be subscripted as
    ## well.)
    ## </TODO>
    
    max_n_of_classes <- max(sapply(clusterings, n_of_classes))

    ## Control parameters.
    k <- control$k
    if(is.null(k))
        k <- max_n_of_classes
    order <- control$order
    if(is.null(order))
        order <- sample(seq_along(clusterings))

    clusterings <- clusterings[order]
    weights <- weights[order]

    k_max <- max(k, max_n_of_classes)
    s <- weights / cumsum(weights)
    s[is.na(s)] <- 0                    # Division by zero ...
    
    M <- cl_membership(clusterings[[1L]], k_max)
    for(b in seq_along(clusterings)[-1L]) {
        mem <- cl_membership(clusterings[[b]], k_max)
        ## Match classes from conforming memberships.
        ind <- solve_LSAP(crossprod(M, mem), maximum = TRUE)
        M <- (1 - s[b]) * M + s[b] * mem[, ind]
        if(k < k_max)
            M <- .project_to_leading_columns(M, k)
    }

    M <- .cl_membership_from_memberships(M[, seq_len(k), drop = FALSE], k)

    as.cl_partition(M)
}

### * .cl_consensus_partition_AOS

.cl_consensus_partition_AOS <-
function(clusterings, weights, control,
         type = c("SE", "HE", "SM", "HM"))
{
    ## The start of a general purpose optimizer for determining
    ## consensus partitions by minimizing
    ##   \sum_b w_b d(M, M_b) ^ e
    ##     = \sum_b \min_{P_b} w_b f(M, M_b P_b) ^ e
    ## for the special case where the criterion function is based on
    ## M and M_b P_b (i.e., column permutations of M_b), as opposed to
    ## the general case where d(M, M_b) = \min_{P_b} f(M, P_b, M_b)
    ## handled by .cl_consensus_partition_AOG().
    ##
    ## The AO ("alternative optimization") proceeds by alternatively
    ## matching the M_b to M by minimizing f(M, M_b P_b) over P_b, and
    ## fitting M by minimizing \sum_b w_b f(M, M_b P_b) ^ e for fixed
    ## matchings.
    ##
    ## Such a procedure requires three ingredients: a function for
    ## matching M_b to M (in fact simply replacing M_b by the matched
    ## M_b P_b); a function for fitting M to the \{M_b P_b\}, and a
    ## function for computing the value of the criterion function
    ## corresponding to this fit (so that one can stop if the relative
    ## improvement is small enough).
    ##
    ## For the time being, we only use this to determine soft and hard
    ## Euclidean least squares consensus partitions (soft and hard
    ## Euclidean means), so the interface does not yet reflect the
    ## generality of the approach (which would either pass the three
    ## functions, or even set up family objects encapsulating the three
    ## functions).
    ##
    ## This special case is provided for efficiency and convenience.
    ## Using the special form of the criterion function, we can simply
    ## always work memberships with the same maximal number of columns,
    ## and with the permuted \{ M_b P_b \}.

    ## For the time being ...
    type <- match.arg(type)

    w <- weights / sum(weights)    
    n <- n_of_objects(clusterings)
    k_max <- max(sapply(clusterings, n_of_classes))

    ## Control parameters.
    k <- control$k
    if(is.null(k))
        k <- k_max
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 100
    nruns <- control$nruns
    reltol <- control$reltol
    if(is.null(reltol))
        reltol <- sqrt(.Machine$double.eps)
    start <- control$start
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
        nruns <- length(start)
    } else {
        if(is.null(nruns)) {
            ## Use nruns only if start is not given.
            nruns <- 1L
        }
        start <- replicate(nruns,
                           .random_stochastic_matrix(n, k),
                           simplify = FALSE)
    }

    ## The maximal (possible) number of classes in M and the \{ M_b \}.
    k_all <- max(k, k_max)

    value <-
        switch(type,
               SE = , HE = function(M, memberships, w) {
                   sum(w * sapply(memberships,
                                  function(u) sum((u - M) ^ 2)))
               },
               SM = , HM = function(M, memberships, w) {
                   sum(w * sapply(memberships,
                                  function(u) sum(abs(u - M))))
               })
    ## Return the M[, ind] column permutation of M optimally matching N.
    match_memberships <-
        switch(type,
               SE = , HE = function(M, N) {
                   M[, solve_LSAP(crossprod(N, M), maximum = TRUE),
                     drop = FALSE]
               },
               SM = , HM = function(M, N) {
                   M[, solve_LSAP(.cxdist(N, M, "manhattan")),
                     drop = FALSE]
               })
    ## Function for fitting M to (fixed) memberships \{ M_b P_b \}.
    ## As we use a common number of columns for all membership matrices
    ## involved, we need to pass the desired 'k' ...
    fit_M <-
        switch(type,
               SE = function(memberships, w, k) {
                   ## Update M as \sum w_b M_b P_b.
                   M <- .weighted_sum_of_matrices(memberships, w, nrow(M))
                   ## If k < k_all, "project" as indicated in Gordon &
                   ## Vichi (2001), p. 238.
                   if(k < ncol(M))
                       M <- .project_to_leading_columns(M, k)
                   M
               },
               HE = , HM = function(memberships, w, k) {
                   ## Compute M as \sum w_b M_b P_b.
                   M <- .weighted_sum_of_matrices(memberships, w, nrow(M))
                   ## And compute a closest hard partition H(M) from
                   ## that, using the first k columns of M.
                   ids <- max.col(M[ , seq_len(k), drop = FALSE])
                   .cl_membership_from_class_ids(ids, ncol(M))
               },
               SM = .l1_fit_M)

    memberships <- lapply(clusterings, cl_membership, k_all)

    V_opt <- Inf
    M_opt <- NULL
    for(run in seq_along(start)) {
        if(verbose && (nruns > 1L))
            message(gettextf("AOS run: %d", run))
        M <- start[[run]]
        if(k < k_all)
            M <- cbind(M, matrix(0, nrow(M), k_all - k))
        memberships <- lapply(memberships, match_memberships, M)
        old_value <- value(M, memberships, w)
        if(verbose)
            message(gettextf("Iteration: 0 *** value: %g", old_value))
        iter <- 1L
        while(iter <= maxiter) {
            ## Fit M to the M_b P_b.
            M <- fit_M(memberships, w, k)
            ## Match the \{ M_b P_b \} to M.
            memberships <- lapply(memberships, match_memberships, M)
            ## Update value.
            new_value <- value(M, memberships, w)
            if(verbose)
                message(gettextf("Iteration: %d *** value: %g",
                                 iter, new_value))
            if(abs(old_value - new_value)
               < reltol * (abs(old_value) + reltol))
                break
            old_value <- new_value
            iter <- iter + 1L
        }
        if(new_value < V_opt) {
            converged <- (iter <= maxiter)
            V_opt <- new_value
            M_opt <- M
        }
        if(verbose)
            message(gettextf("Minimum: %g", V_opt))
    }

    M <- .stochastify(M_opt)
    rownames(M) <- rownames(memberships[[1L]])
    meta <- list(objval = value(M, memberships, w),
                 converged = converged)
    M <- .cl_membership_from_memberships(M[, seq_len(k), drop = FALSE],
                                         k, meta)
    
    as.cl_partition(M)
}

.random_stochastic_matrix <-
function(n, k)
{
    M <- matrix(runif(n * k), n, k)
    M / rowSums(M)
}

.l1_fit_M <-
function(memberships, w, k)
{
    ## Determine stochastic matrix M with at most k leading nonzero
    ## columns such that
    ##
    ##    \sum_b w_b \sum_{i,j} | m_{ij}(b) - m_{ij} | => min
    ##
    ## where the sum over j goes from 1 to k.
    ##
    ## Clearly, this can be done separately for each row, where we need
    ## to minimize
    ##
    ##    \sum_b w_b \sum_j | y_j(b) - x_j | => min
    ##
    ## over all probability vectors x.  Such problems can e.g. be solved
    ## via the following linear program:
    ##
    ##    \sum_b \sum_j w_b e'(u(b) + v(b)) => min
    ##
    ## subject to
    ##
    ##    u(1), v(1), ..., u(B), v(B), x >= 0
    ##                   x + u(b) - v(b)  = y(b),    b = 1, ..., B
    ##                               e'x  = 1
    ##
    ## (where e = [1, ..., 1]).
    ##
    ## So we have one long vector z of "variables":
    ##
    ##    z = [u(1)', v(1)', ..., u(B)', v(B)', x']'
    ##
    ## of length (2B + 1) k, with x the object of interest.
    
    ## Rather than providing a separate function for weighted L1 fitting
    ## of probability vectors we prefer doing "everything" at once, in
    ## order to avoid recomputing the coefficients and constraints of
    ## the associated linear program.

    B <- length(memberships)
    L <- (2 * B + 1) * k

    ## Set up associated linear program.

    ## Coefficients in the objective function.
    objective_in <- c(rep(w, each = 2 * k), rep.int(0, k))
    
    ## Constraints.
    constr_mat <- rbind(diag(1, L),
                        cbind(kronecker(diag(1, B),
                                        cbind(diag(1, k),
                                              diag(-1, k))),
                              kronecker(rep.int(1, B),
                                        diag(1, k))),
                        c(rep.int(0, 2 * B * k), rep.int(1, k)))
    constr_dir <- c(rep.int(">=", L), rep.int("==", B * k + 1L))

    ind <- seq.int(from = 2 * B * k + 1L, length.out = k)
    nr <- NROW(memberships[[1L]])
    nc <- NCOL(memberships[[1L]])
    M <- matrix(0, nrow = nr, ncol = k)

    ## Put the memberships into one big array so that we can get their
    ## rows more conveniently (and efficiently):

    memberships <- array(unlist(memberships), c(nr, nc, B))

    for(i in seq_len(nr)) {
        out <- lpSolve::lp("min",
                           objective_in,
                           constr_mat,
                           constr_dir,
                           c(rep.int(0, L), memberships[i, seq_len(k), ], 1))
        M[i, ] <- out$solution[ind]
    }

    ## Add zero columns if necessary.
    if(k < nc)
        M <- cbind(M, matrix(0, nr, nc - k))

    M
}

### ** .cl_consensus_partition_soft_euclidean

.cl_consensus_partition_soft_euclidean <-
function(clusterings, weights, control)
    .cl_consensus_partition_AOS(clusterings, weights, control, "SE")

### ** .cl_consensus_partition_hard_euclidean

.cl_consensus_partition_hard_euclidean <-
function(clusterings, weights, control)
    .cl_consensus_partition_AOS(clusterings, weights, control, "HE")

### ** .cl_consensus_partition_soft_manhattan

.cl_consensus_partition_soft_manhattan <-
function(clusterings, weights, control)
    .cl_consensus_partition_AOS(clusterings, weights, control, "SM")

### ** .cl_consensus_partition_hard_manhattan

.cl_consensus_partition_hard_manhattan <-
function(clusterings, weights, control)
    .cl_consensus_partition_AOS(clusterings, weights, control, "HM")

### * .cl_consensus_partition_AOG

.cl_consensus_partition_AOG <-
function(clusterings, weights, control, type = c("GV1"))
{
    ## The start of a general purpose optimizer for determining
    ## consensus partitions by minimizing
    ##   \sum_b w_b d(M, M_b) ^ p
    ##     = \sum_b \min_{P_b} w_b f(M, M_b, P_b) ^ e
    ## for general dissimilarity matrices which involve class matching
    ## via permutation matrices P_b.
    ##
    ## The AO ("Alternative Optimization") proceeds by alternating
    ## between determining the optimal permutations P_b by minimizing
    ##   f(M, M_b, P_b)
    ## for fixed M, and fitting M by minimizing
    ##   \sum_b w_b f(M, M_b, P_b) ^ e
    ## for fixed \{ P_b \}.
    ##
    ## We encapsulate this into functions fit_P() and fit_M() (and a
    ## value() function for the criterion function to be minimized with
    ## respect to both M and \{ P_b \}, even though the current
    ## interface does not yet reflect the generality of the approach.
    ##
    ## Note that rather than passing on information about the numbers of
    ## classes (e.g., needed for GV1) and representing all involved
    ## membership matrices with the same maximal number of columns, we
    ## use "minimal" representations with no dummy classes (strictly
    ## speaking, with the possible exception of M, for which the given k
    ## is used).

    ## For the time being ...
    type <- match.arg(type)

    w <- weights / sum(weights)    
    n <- n_of_objects(clusterings)
    k_max <- max(sapply(clusterings, n_of_classes))
    
    ## Control parameters.
    k <- control$k
    if(is.null(k))
        k <- k_max
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 100L
    nruns <- control$nruns
    reltol <- control$reltol
    if(is.null(reltol))
        reltol <- sqrt(.Machine$double.eps)
    start <- control$start
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
        nruns <- length(start)
    } else {
        if(is.null(nruns)) {
            ## Use nruns only if start is not given.
            nruns <- 1L
        }
        start <- replicate(nruns,
                           .random_stochastic_matrix(n, k),
                           simplify = FALSE)
    }

    ## <NOTE>
    ## For the given memberships, we can simply use ncol() in the
    ## computations (rather than n_of_classes(), because we used
    ## cl_membership() to create them.  For M, the number of classes
    ## could be smaller than the given k "target".
    ## </NOTE>
    
    value <- function(M, permutations, memberships, w) {

        k <- .n_of_nonzero_columns(M)
        d <- function(u, p) {
            ## Compute the squared GV1 dissimilarity between M and u
            ## based on the M->u class matching p.
            nc_u <- ncol(u)
            if(nc_u == k) {
                ## Simple case: all classes are matched.
                sum((u[, p] - M) ^ 2)
            }
            else {
                ## Only include the matched non-dummy classes of M ..
                ind <- seq_len(k)
                ## ... which are matched to non-dummy classes of u.
                ind <- ind[p[ind] <= nc_u]
                sum((u[, p[ind]] - M[, ind]) ^ 2)
            }
        }
        
        sum(w * mapply(d, memberships, permutations))
    }

    fit_P <- function(u, M) {
        
        ## Return a permutation representing a GV1 optimal matching of
        ## the columns of M to the columns of u (note the order of the
        ## arguments), using a minimal number of dummy classes (i.e., p
        ## has max(.n_of_nonzero_columns(M), n_of_classes(u)) entries).

        ## See also .cl_dissimilarity_partition_GV1().
        
        C <- outer(colSums(M ^ 2), colSums(u ^ 2), `+`) -
            2 * crossprod(M, u)
        nc_M <- .n_of_nonzero_columns(M)
        nc_u <- ncol(u)
        ## (See above for ncol() vs n_of_classes().)
        if(nc_M < nc_u)
            C <- rbind(C, matrix(0, nrow = nc_u - nc_M, ncol = nc_u))
        else if(nc_M > nc_u)
            C <- cbind(C, matrix(0, nrow = nc_M, ncol = nc_M - nc_u))
        
        solve_LSAP(C)
    }

    fit_M <- function(permutations, memberships, w) {
        
        ## Here comes the trickiest part ...
        ##
        ## In general, M = [m_{iq}] is determined as follows.
        ## Write value(M, permutations, memberships, w) as
        ##   \sum_b \sum_i \sum_{p=1}^{k_b} \sum_{q=1}^k
        ##      w_b (u_{ip}(b) - m_{iq})^2 x_{pq}(b)
        ## where U(b) and X(b) are the b-th membership matrix and the
        ## permutation matrix representing the M->U(b) non-dummy class
        ## matching (as always, note the order of the arguments).
        ##
        ## Let
        ##   \beta_{iq} = \sum_b \sum_{p=1}^{k_b} w_b u_{ip}(b) x_{pq}(b)
        ##   \alpha_q   = \sum_b \sum_{p=1}^{k_b} w_b x_{pq}(b)
        ## and
        ##   \bar{m}_{iq} =
        ##     \cases{\beta_{iq}/\alpha_q, & $\alpha_q > 0$ \cr
        ##            0                    & otherwise}.
        ## Then, as the cross-product terms cancel out, the value
        ## function rewrites as
        ##   \sum_b \sum_i \sum_{p=1}^{k_b} \sum_{q=1}^k
        ##      w_b (u_{ip}(b) - \bar{m}_{iq})^2 x_{pq}(b)
        ##   + \sum_i \sum_q \alpha_q (\bar{m}_{iq} - m_{iq}) ^ 2,
        ## where the first term is a constant, and the minimum is found
        ## by solving
        ##   \sum_q \alpha_q (\bar{m}_{iq} - m_{iq}) ^ 2 => min!
        ## s.t.
        ##   m_{i1}, ..., m_{ik} >= 0, \sum_{iq} m_{iq} = 1.
        ##
        ## We can distinguish three cases.
        ## A. If S_i = \sum_q \bar{m}_{iq} = 1, things are trivial.
        ## B. If S_i = \sum_q \bar{m}_{iq} < 1.
        ##    B1. If some \alpha_q are zero, then we can choose
        ##        m_{iq} = \bar{m}_{iq} for those q with \alpha_q = 0;
        ##        m_{iq} = 1 / number of zero \alpha's, otherwise.
        ##    B2. If all \alpha_q are positive, we can simply
        ##        equidistribute 1 - S_i over all classes as written
        ##        in G&V.
        ## C. If S_i > 1, things are not so clear (as equidistributing
        ##    will typically result in violations of the non-negativity
        ##    constraint).  We currently revert to using solve.QP() from
        ##    package quadprog, as constrOptim() already failed in very
        ##    simple test cases.
        ##
        ## Now consider \sum_{p=1}^{k_b} x_{pq}(b).  If k <= k_b for all
        ## b, all M classes from 1 to k are matched to one of the k_b
        ## classes in U(b), hence the sum and also \alpha_q are one.
        ## But then
        ##    \sum_q \bar{m}_{iq}
        ##      =  \sum_b \sum_{p=1}^{k_b} w_b u_{ip}(b) x_{pq}(b)
        ##      <= \sum_b \sum_{p=1}^{k_b} w_b u_{ip}(b)
        ##      =  1
        ## with equality if k = k_b for all b.  I.e.,
        ## * If k = \min_b k_b = \max k_b, we are in case A.
        ## * If k <= \min_b k_b, we are in case B2.
        ## And it makes sense to handle these cases explicitly for
        ## efficiency reasons.

        ## And now for something completely different ... the code.

        k <- .n_of_nonzero_columns(M)
        nr_M <- nrow(M)
        nc_M <- ncol(M)
        nc_memberships <- sapply(memberships, ncol)

        if(k <= min(nc_memberships)) {
            ## Compute the weighted means \bar{M}.
            M <- .weighted_sum_of_matrices(mapply(function(u, p)
                                                  u[ , p[seq_len(k)]],
                                                  memberships,
                                                  permutations,
                                                  SIMPLIFY = FALSE),
                                           w, nr_M)
            ## And add dummy classes if necessary.
            if(k < nc_M)
                M <- cbind(M, matrix(0, nr_M, nc_M - k))
            ## If we always got the same number of classes, we are
            ## done.  Otherwise, equidistribute ...
            if(k < max(nc_memberships))
                M <- pmax(M + (1 - rowSums(M)) / nc_M, 0)
            return(M)
        }

        ## Here comes the general case.

        ## First, compute the \alpha and \beta.
        alpha <- rowSums(rep(w, each = k) *
                         mapply(function(p, n) p[seq_len(k)] <= n,
                                permutations, nc_memberships))
        ## Alternatively (more literally):
        ##   X <- lapply(permutations, .make_X_from_p)
        ##   alpha1 <- double(length = k)
        ##   for(b in seq_along(permutations)) {
        ##       alpha1 <- alpha1 +
        ##           w[b] * colSums(X[[b]][seq_len(nc_memberships[b]), ])
        ##   }
        
        ## A helper function giving suitably permuted memberships.
        pmem <- function(u, p) {
            ## Only matched classes, similar to the one used in value(),
            ## maybe merge eventually ...
            v <- matrix(0, nr_M, k)
            ind <- seq_len(k)
            ind <- ind[p[ind] <= ncol(u)]
            if(any(ind))
                v[ , ind] <- u[ , p[ind]]
            v
        }
        beta <- .weighted_sum_of_matrices(mapply(pmem,
                                                 memberships,
                                                 permutations,
                                                 SIMPLIFY = FALSE),
                                          w, nr_M)
        ## Alternatively (more literally):
        ##   beta1 <- matrix(0, nr_M, nc_M)
        ##   for(b in seq_along(permutations)) {
        ##     ind <- seq_len(nc_memberships[b])
        ##     beta1 <- beta1 +
        ##       w[b] * memberships[[b]][, ind] %*% X[[b]][ind, ]
        ##   }
        
        ## Compute the weighted means \bar{M}.
        M <- .cscale(beta, ifelse(alpha > 0, 1 / alpha, 0))
        ## Alternatively (see comments for .cscale()):
        ##   M1 <- beta %*% diag(ifelse(alpha > 0, 1 / alpha, 0))
        
        ## And add dummy classes if necessary.
        if(k < nc_M)
            M <- cbind(M, matrix(0, nr_M, nc_M - k))

        S <- rowSums(M)
        ## Take care of those rows with row sums < 1.
        ind <- (S < 1)
        if(any(ind)) {
            i_0 <- alpha == 0
            if(any(i_0))
                M[ind, i_0] <- 1 / sum(i_0)
            else
                M[ind, ] <- pmax(M[ind, ] + (1 - S[ind]) / nc_M, 0)
        }
        ## Take care of those rows with row sums > 1.
        ind <- (S > 1)
        if(any(ind)) {
            ## Argh.  Call solve.QP() for each such i.  Alternatively,
            ## could set up on very large QP, but is this any better?
            Dmat <- diag(alpha, nc_M)
            Amat <- t(rbind(rep.int(-1, nc_M), diag(1, nc_M)))
            bvec <- c(-1, rep.int(0, nc_M))
            for(i in which(ind))
                M[i, ] <- quadprog::solve.QP(Dmat, alpha * M[i, ],
                                             Amat, bvec)$solution
        }

        M
    }

    memberships <- lapply(clusterings, cl_membership)

    V_opt <- Inf
    M_opt <- NULL
    for(run in seq_along(start)) {
        if(verbose && (nruns > 1L))
            message(gettextf("AOG run: %d", run))
        M <- start[[run]]
        permutations <- lapply(memberships, fit_P, M)
        old_value <- value(M, permutations, memberships, w)
        message(gettextf("Iteration: 0 *** value: %g", old_value))
        iter <- 1L
        while(iter <= maxiter) {
            ## Fit M.
            M <- fit_M(permutations, memberships, w)
            ## Fit \{ P_b \}.
            permutations <- lapply(memberships, fit_P, M)
            ## Update value.
            new_value <- value(M, permutations, memberships, w)
            if(verbose)
                message(gettextf("Iteration: %d *** value: %g",
                                 iter, new_value))
            if(abs(old_value - new_value)
               < reltol * (abs(old_value) + reltol))
                break
            old_value <- new_value
            iter <- iter + 1L
        }
        if(new_value < V_opt) {
            converged <- (iter <= maxiter)
            V_opt <- new_value
            M_opt <- M
        }
        if(verbose)
            message(gettextf("Minimum: %g", V_opt))
    }

    M <- .stochastify(M_opt)
    ## Seems that M is always kept a k columns ... if not, use
    ##   M <- .stochastify(M_opt[, seq_len(k), drop = FALSE])
    rownames(M) <- rownames(memberships[[1L]])
    ## Recompute the value, just making sure ...
    permutations <- lapply(memberships, fit_P, M)
    meta <- list(objval = value(M, permutations, memberships, w),
                 converged = converged)
    M <- .cl_membership_from_memberships(M, k, meta)
    
    as.cl_partition(M)
}

### ** .cl_consensus_partition_GV1

.cl_consensus_partition_GV1 <-
function(clusterings, weights, control)
    .cl_consensus_partition_AOG(clusterings, weights, control, "GV1")


### * .cl_consensus_partition_GV3

.cl_consensus_partition_GV3 <-
function(clusterings, weights, control)
{
    ## Use a SUMT to solve
    ##   \| Y - M M' \|_F^2 => min
    ## where M is a membership matrix and Y = \sum_b w_b M_b M_b'.

    n <- n_of_objects(clusterings)
    max_n_of_classes <- max(sapply(clusterings, n_of_classes))

    ## Control parameters:
    ## k,
    k <- control$k
    if(is.null(k))
        k <- max_n_of_classes
    ## nruns,
    nruns <- control$nruns
    ## start.
    start <- control$start

    w <- weights / sum(weights)

    comemberships <-
        lapply(clusterings, function(x) {
            ## No need to force a common k here.
            tcrossprod(cl_membership(x))
        })

    Y <- .weighted_sum_of_matrices(comemberships, w, n)

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else {
        if(is.null(nruns)) {
            ## Use nruns only if start is not given.
            nruns <- 1L
        }
        e <- eigen(Y, symmetric = TRUE)
        ## Use M <- U_k \lambda_k^{1/2}, or random perturbations
        ## thereof.
        M <- e$vectors[, seq_len(k), drop = FALSE] *
            rep(sqrt(e$values[seq_len(k)]), each = n)
        m <- c(M)
        start <- c(list(m),
                   replicate(nruns - 1L,
                             m + rnorm(length(m), sd = sd(m) / sqrt(3)),
                             simplify = FALSE))
    }

    y <- c(Y)

    L <- function(m) sum((y - tcrossprod(matrix(m, n))) ^ 2)
    P <- .make_penalty_function_membership(n, k)
    grad_L <- function(m) {
        M <- matrix(m, n)
        4 * c((tcrossprod(M) - Y) %*% M)
    }
    grad_P <- .make_penalty_gradient_membership(n, k)

    out <- sumt(start, L, P, grad_L, grad_P,
                method = control$method, eps = control$eps,
                q = control$q, verbose = control$verbose,
                control = as.list(control$control))

    M <- .stochastify(matrix(out$x, n))
    rownames(M) <- rownames(cl_membership(clusterings[[1L]]))
    meta <- list(objval = L(c(M)))
    M <- .cl_membership_from_memberships(M, k, meta)

    as.cl_partition(M)
}

### * .cl_consensus_partition_soft_symdiff

.cl_consensus_partition_soft_symdiff <-
function(clusterings, weights, control)
{
    ## Use a SUMT to solve
    ##   \sum_b w_b \sum_{ij} | c_{ij}(b) - c_{ij} | => min
    ## where C(b) = comembership(M(b)) and C = comembership(M) and M is
    ## a membership matrix.

    ## Control parameters:
    ## gradient,
    gradient <- control$gradient
    if(is.null(gradient))
        gradient <- TRUE
    ## k,
    k <- control$k
    ## nruns,
    nruns <- control$nruns
    ## start.
    start <- control$start

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else if(is.null(nruns)) {
        ## Use nruns only if start is not given.
        nruns <- 1L
    }
    
    max_n_of_classes <- max(sapply(clusterings, n_of_classes))
    if(is.null(k))
        k <- max_n_of_classes

    B <- length(clusterings)    
    n <- n_of_objects(clusterings)
    
    w <- weights / sum(weights)

    comemberships <-
        lapply(clusterings, function(x) {
            ## No need to force a common k here.
            tcrossprod(cl_membership(x))
        })

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else {
        if(is.null(nruns)) {
            ## Use nruns only if start is not given.
            nruns <- 1L
        }
        ## Try using a rank k "root" of the weighted median of the
        ## comemberships as starting value.
        Y <- apply(array(unlist(comemberships), c(n, n, B)), c(1, 2),
                   weighted_median, w)
        e <- eigen(Y, symmetric = TRUE)
        ## Use M <- U_k \lambda_k^{1/2}, or random perturbations
        ## thereof.
        M <- e$vectors[, seq_len(k), drop = FALSE] *
            rep(sqrt(e$values[seq_len(k)]), each = n)
        m <- c(M)
        start <- c(list(m),
                   replicate(nruns - 1L,
                             m + rnorm(length(m), sd = sd(m) / sqrt(3)),
                             simplify = FALSE))
    }

    L <- function(m) {
        M <- matrix(m, n)
        C_M <- tcrossprod(M)
        ## Note that here (as opposed to hard/symdiff) we take soft
        ## partitions as is without replacing them by their closest hard
        ## partitions.
        sum(w * sapply(comemberships,
                       function(C) sum(abs(C_M - C))))
    }
    P <- .make_penalty_function_membership(n, k)
    if(gradient) {
        grad_L <- function(m) {
            M <- matrix(m, n)
            C_M <- tcrossprod(M)
            .weighted_sum_of_matrices(lapply(comemberships,
                                             function(C)
                                             2 * sign(C_M - C) %*% M),
                                      w, n)
        }
        grad_P <- .make_penalty_gradient_membership(n, k)
    }
    else
        grad_L <- grad_P <- NULL

    out <- sumt(start, L, P, grad_L, grad_P,
                method = control$method, eps = control$eps,
                q = control$q, verbose = control$verbose,
                control = as.list(control$control))

    M <- .stochastify(matrix(out$x, n))
    rownames(M) <- rownames(cl_membership(clusterings[[1L]]))
    meta <- list(objval = L(c(M)))
    M <- .cl_membership_from_memberships(M, k, meta)

    as.cl_partition(M)
}

### * .cl_consensus_partition_hard_symdiff

.cl_consensus_partition_hard_symdiff <-
function(clusterings, weights, control)
{
    ## <NOTE>
    ## This is mostly duplicated from relations.
    ## Once this is on CRAN, we could consider having clue suggest
    ## relations ...
    ## </NOTE>
    
    comemberships <-
        lapply(clusterings,
               function(x) {
                   ## Here, we always turn possibly soft partitions to
                   ## their closest hard partitions.
                   ids <- cl_class_ids(x)
                   outer(ids, ids, `==`)
                   ## (Simpler than using tcrossprod() on
                   ## cl_membership().)
               })

    ## Could also create a relation ensemble from the comemberships and
    ## call relation_consensus().

    B <- relations:::.make_fit_relation_symdiff_B(comemberships,
                                                  weights)
    k <- control$k
    control <- control$control
    ## Note that currently we provide no support for finding *all*
    ## consensus partitions (but allow for specifying the solver).
    control$all <- FALSE
    I <- if(!is.null(k)) {
        ## <NOTE>
        ## We could actually get the memberships directly in this case.
        relations:::fit_relation_LP_E_k(B, k, control)
        ## </NOTE>
    }
    else
        relations:::fit_relation_LP(B, "E", control)

    ids <- relations:::get_class_ids_from_incidence(I)
    names(ids) <- cl_object_names(clusterings)

    as.cl_hard_partition(ids)
}
    
### * .cl_consensus_hierarchy_cophenetic

.cl_consensus_hierarchy_cophenetic <-
function(clusterings, weights, control)
{
    ## d <- .weighted_mean_of_object_dissimilarities(clusterings, weights)
    ## Alternatively:
    ## as.cl_dendrogram(ls_fit_ultrametric(d, control = control))
    control <- c(list(weights = weights), control)
    as.cl_dendrogram(ls_fit_ultrametric(clusterings, control = control))
}

### * .cl_consensus_hierarchy_manhattan

.cl_consensus_hierarchy_manhattan <-
function(clusterings, weights, control)
{
    ## Control parameters:
    ## gradient,
    gradient <- control$gradient
    if(is.null(gradient))
        gradient <- TRUE
    ## nruns,
    nruns <- control$nruns
    ## start.
    start <- control$start

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else if(is.null(nruns)) {
        ## Use nruns only if start is not given.
        nruns <- 1L
    }
    
    w <- weights / sum(weights)
    B <- length(clusterings)
    ultrametrics <- lapply(clusterings, cl_ultrametric)
    
    if(B == 1L)
        return(as.cl_dendrogram(ultrametrics[[1L]]))

    n <- n_of_objects(ultrametrics[[1L]])
    labels <- cl_object_names(ultrametrics[[1L]])

    ## We need to do
    ##
    ##   \sum_b w_b \sum_{i,j} | u_{ij}(b) - u_{ij} | => min
    ##
    ## over all ultrametrics u.  Let's use a SUMT (for which "gradients"
    ## can optionally be switched off) ...
    
    L <- function(d) {
        sum(w * sapply(ultrametrics, function(u) sum(abs(u - d))))
        ## Could also do something like
        ##   sum(w * sapply(ultrametrics, cl_dissimilarity, d,
        ##                  "manhattan"))
    }
    P <- .make_penalty_function_ultrametric(n)
    if(gradient) {
        grad_L <- function(d) {
            ## "Gradient" is \sum_b w_b sign(d - u(b)).
            .weighted_sum_of_vectors(lapply(ultrametrics,
                                            function(u) sign(d - u)),
                                     w)
        }
        grad_P <- .make_penalty_gradient_ultrametric(n)
    } else
        grad_L <- grad_P <- NULL
            
    if(is.null(start)) {
        ## Initialize by "random shaking" of the weighted median of the
        ## ultrametrics.  Any better ideas?
        ## <NOTE>
        ## Using var(x) / 3 is really L2 ...
        ## </NOTE>
        x <- apply(matrix(unlist(ultrametrics), ncol = B), 1, 
                   weighted_median, w)
        start <- replicate(nruns,  
                           x + rnorm(length(x), sd = sd(x) / sqrt(3)),
                           simplify = FALSE)
    }

    out <- sumt(start, L, P, grad_L, grad_P,
                method = control$method, eps = control$eps,
                q = control$q, verbose = control$verbose,
                control = as.list(control$control))

    d <- .ultrametrify(out$x)
    meta <- list(objval = L(d))
    d <- .cl_ultrametric_from_veclh(d, n, labels, meta)
    
    as.cl_dendrogram(d)
}

### * .cl_consensus_hierarchy_majority

.cl_consensus_hierarchy_majority <-
function(clusterings, weights, control)
{
    w <- weights / sum(weights)

    p <- control$p
    if(is.null(p))
        p <- 1 / 2
    else if(!is.numeric(p) || (length(p) != 1) || (p < 1 / 2) || (p > 1))
        stop("Parameter 'p' must be in [1/2, 1].")

    classes <- lapply(clusterings, cl_classes)
    all_classes <- unique(unlist(classes, recursive = FALSE))
    gamma <- double(length = length(all_classes))
    for(i in seq_along(classes))
        gamma <- gamma + w[i] * !is.na(match(all_classes, classes[[i]]))
    ## Rescale to [0, 1].
    gamma <- gamma / max(gamma)

    maj_classes <- if(p == 1) {
        ## Strict consensus tree.
        all_classes[gamma == 1]
    }
    else
        all_classes[gamma > p]
    attr(maj_classes, "labels") <- attr(classes[[1L]], "labels")

    ## <FIXME>
    ## Stop auto-coercing that to dendrograms once we have suitable ways
    ## of representing n-trees.
    as.cl_hierarchy(.cl_ultrametric_from_classes(maj_classes))
    ## </FIXME>
}

### * Utilities

### ** .cl_consensus_method_default

.cl_consensus_method_default <-
function(type)
{
    switch(type, partition = "SE", hierarchy = "euclidean", NULL)
}

### ** .project_to_leading_columns

.project_to_leading_columns <-
function(x, k)
{
    ## For a given matrix stochastic matrix x, return the stochastic
    ## matrix y which has columns from k+1 on all zero which is closest
    ## to x in the Frobenius distance.
    y <- x[, seq_len(k), drop = FALSE]
    y <- cbind(pmax(y + (1 - rowSums(y)) / k, 0),
               matrix(0, nrow(y), ncol(x) - k))
    ## (Use the pmax to ensure that entries remain nonnegative.)
}

### ** .make_X_from_p

.make_X_from_p <-
function(p) {
    ## X matrix corresponding to permutation p as needed for the AO
    ## algorithms.  I.e., x_{ij} = 1 iff j->p(j)=i.
    X <- matrix(0, length(p), length(p))
    i <- seq_along(p)
    X[cbind(p[i], i)] <- 1
    X
}

### ** .n_of_nonzero_columns

## <NOTE>
## Could turn this into n_of_classes.matrix().
.n_of_nonzero_columns <-
function(x)
    sum(colSums(x) > 0)
## </NOTE>

### ** .cscale

## <FIXME>
## Move to utilities eventually ...
.cscale <-
function(A, x)
{
    ## Scale the columns of matrix A by the elements of vector x.
    ## Formally, A %*% diag(x), but faster.
    ## Could also use sweep(A, 2, x, "*")
    rep(x, each = nrow(A)) * A
}
## </FIXME>

## .make_penalty_function_membership

.make_penalty_function_membership <-
function(nr, nc)
    function(m) {
        sum(pmin(m, 0) ^ 2) + sum((rowSums(matrix(m, nr)) - 1) ^ 2)
    }

## .make_penalty_gradient_membership

.make_penalty_gradient_membership <-
function(nr, nc)
    function(m) {
        2 * (pmin(m, 0) + rep.int(rowSums(matrix(m, nr)) - 1, nc))
    }


### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### * n_of_objects

## Get the number of objects in a clustering.

n_of_objects <-
function(x)
    UseMethod("n_of_objects")

### ** Default method.

n_of_objects.default <-
function(x)
    length(cl_class_ids(x))
## (Note that prior to R 2.1.0, kmeans() returned unclassed results,
## hence the best we can do for the *default* method is to look at a
## possibly existing "cluster" component.  Using the class ids incurs
## another round of method dispatch, but avoids code duplication.)

### ** Partitioning methods.

## Package stats: kmeans() (R 2.1.0 or better).
n_of_objects.kmeans <- n_of_objects.default

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
n_of_objects.partition <- n_of_objects.default

## Package cclust: cclust().
n_of_objects.cclust <- n_of_objects.default

## Package e1071: cmeans() gives objects of class "fclust".
n_of_objects.fclust <-
function(x)
    nrow(x$membership)
## Package e1071: cshell().
n_of_objects.cshell <- n_of_objects.fclust
## Package e1071: bclust().
n_of_objects.bclust <- n_of_objects.default

## Package mclust: Mclust().
n_of_objects.Mclust <- n_of_objects.default

### ** Hierarchical methods.

## Package stats: hclust().
n_of_objects.hclust <-
function(x)
    length(x$order)

## Package cluster: agnes() and diana() give objects inheriting from
## class "twins".
n_of_objects.twins <- n_of_objects.hclust
## Package cluster: mona().
n_of_objects.mona <- n_of_objects.hclust

## Package ape: class "phylo".
n_of_objects.phylo <-
function(x)
    length(x$tip.label)

### ** Others.

## Package stats: class "dist".
n_of_objects.dist <-
function(x)
    attr(x, "Size")

## Package clue: Ensembles.
n_of_objects.cl_ensemble <-
function(x)
    attr(x, "n_of_objects")
## Package clue: Memberships.
n_of_objects.cl_membership <- nrow
## Package clue: pclust().
n_of_objects.pclust <- n_of_objects.default
## Package clue: Ultrametrics.
n_of_objects.cl_ultrametric <- n_of_objects.dist
## Package clue: (virtual) class "cl_partition".
n_of_objects.cl_partition <-
function(x)
    .get_property_from_object_or_representation(x, "n_of_objects")
## Package clue: (virtual) class "cl_hierarchy".
n_of_objects.cl_hierarchy <-
function(x)
    .get_property_from_object_or_representation(x, "n_of_objects")

### * cl_object_names

## Determine the names of the objects in a clustering if available; give
## NULL otherwise.  This is in sync with e.g. names() or dimnames(); au
## contraire, cl_object_labels() always gives labels even if no names
## are available.

cl_object_names <- 
function(x)
    UseMethod("cl_object_names")

## ** Default method.

cl_object_names.default <- function(x) names(cl_class_ids(x))

## ** Partitions.

## There is really nothing special we can currently do.
## Most partitioning functions return no information on object names.
## This includes classes
##   stats:      kmeans
##   cba:        ccfkms, rock
##   cclust:     cclust
##   e1071:      bclust
##   flexclust:  kcca
##   kernlab:    specc
##   mclust:     Mclust
## The algorithms for which things "work" all give named class ids.
##   RWeka:      Weka_clusterer
##   cluster:    clara fanny pam
##   e1071:      cclust cshell

## ** Hierarchies.

## Package stats: hclust().
cl_object_names.hclust <- function(x) x$labels

## Package cluster: agnes(), diana() and mona() all return an object
## which has an 'order.lab' component iff "the original observations
## were labelled".  We can use this together the the 'order' component
## to recreate the labels in their original order.  Note that we cannot
## rely on dissimilarity or data components being available.
cl_object_names.twins <-
function(x)
{
    if(!is.null(x$order.lab)) {
        out <- character(length = n_of_objects(x))
        out[x$order] <- x$order.lab
        out
    }
    else
        NULL
}
cl_object_names.mona <- cl_object_names.twins

## Package ape: class "phylo".
cl_object_names.phylo <-
function(x)
    x$tip.label

## ** Others.

## Package stats: class "dist".
## (Raw object dissimilarities.)
cl_object_names.dist <-
function(x)
    attr(x, "Labels")

## Package clue: memberships.
cl_object_names.cl_membership <-
function(x)
    rownames(x)
## Package clue: ultrametrics.
cl_object_names.cl_ultrametric <-
function(x)
    attr(x, "Labels")
## Package clue: (virtual) class "cl_partition".
cl_object_names.cl_partition <-
function(x)
    cl_object_names(.get_representation(x))
## Package clue: (virtual) class "cl_hierarchy".
cl_object_names.cl_hierarchy <-
function(x)
    cl_object_names(.get_representation(x))
## Package clue: ensembles.
cl_object_names.cl_ensemble <-
function(x)
{
    nms <- lapply(x, cl_object_names)
    ind <- which(lengths(nms) > 0L)
    if(any(ind)) nms[[ind[1L]]] else NULL
}

### * cl_object_labels

cl_object_labels <-
function(x)
{
    if(is.null(out <- cl_object_names(x)))
        out <- as.character(seq_len(n_of_objects(x)))
    out
}

### * cl_object_dissimilarities

## Extract object dissimilarities from R objects containing such: this
## includes objects directly inheriting from "dist" as well as
## dendrograms or additive trees.

cl_object_dissimilarities <-
function(x)
{
    ## Keep this in sync with .has_object_dissimilarities().
    if(is.cl_dendrogram(x))
        cl_ultrametric(x)
    else if(inherits(x, "dist"))
        x
    else
        stop("Cannot extract object dissimilarities")
}

.has_object_dissimilarities <-
function(x)
{
    ## Keep this in sync with cl_object_dissimilarities().
    is.cl_dendrogram(x) || inherits(x, "dist")
}

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

cl_prototypes <-
function(x)
    UseMethod("cl_prototypes")

## No default method.

## Package stats: kmeans() (R 2.1.0 or better).
cl_prototypes.kmeans <-
function(x)
    x$centers

## Package cluster: clara() always gives prototypes.
cl_prototypes.clara <-
function(x)
    x$medoids
## Package cluster: fanny() never gives prototypes.
## Package cluster: pam() does not give prototypes if given a
## dissimilarity matrix.
cl_prototypes.pam <-
function(x)
{
    p <- x$medoids
    if(!is.matrix(p))
        stop("Cannot determine prototypes.")
    p
}

## Package cba: ccfkms().
cl_prototypes.ccfkms <- cl_prototypes.kmeans

## Package cclust: cclust().
cl_prototypes.cclust <- cl_prototypes.kmeans

## Package e1071: cmeans() gives objects of class "fclust".
cl_prototypes.fclust <- cl_prototypes.kmeans
## Package e1071: cshell().
cl_prototypes.cshell <- cl_prototypes.kmeans
## Package e1071: bclust().
cl_prototypes.bclust <- cl_prototypes.kmeans

## Package flexclust: kcca() returns objects of S4 class "kcca" which
## extends S4 class "flexclust".
cl_prototypes.kcca <-
function(x)
    methods::slot(x, "centers")

## Package kernlab: specc() and kkmeans() return objects of S4 class
## "specc".
cl_prototypes.specc <-
function(x)
    kernlab::centers(x)

## Package mclust: Mclust().
cl_prototypes.Mclust <-
function(x)
{
    p <- x$mu
    ## For multidimensional models, we get a matrix whose columns are
    ## the means of each group in the best model, and hence needs to be
    ## transposed.
    if(is.matrix(p))
        p <- t(p)
    p
}

## Package clue: cl_pam().
cl_prototypes.cl_pam <-
function(x)
    x$prototypes
## Package clue: (virtual) class "cl_partition".
cl_prototypes.cl_partition <-
function(x)
    cl_prototypes(.get_representation(x))
## Package clue: pclust().
cl_prototypes.pclust <-
function(x)
    x$prototypes

### * n_of_classes

## Get the number of classes in a (hard or soft) partition.

## <NOTE>
## We generally allow for classes to be empty, unlike the current
## version of kmeans().  Package cclust has a version of k-means which
## does not stop in case of empty classes.
## However, we only count NON-EMPTY classes here.
## </NOTE>

n_of_classes <-
function(x)
    UseMethod("n_of_classes")

## Default method.
n_of_classes.default <-
function(x)
    length(unique(cl_class_ids(x)))

## Package stats: kmeans() (R 2.1.0 or better).
n_of_classes.kmeans <- n_of_classes.default

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
n_of_classes.fanny <-
function(x)
    sum(colSums(x$membership, na.rm = TRUE) > 0)
n_of_classes.partition <- n_of_classes.default

## Package cclust: cclust().
n_of_classes.cclust <- n_of_classes.default

## Package e1071: cmeans() gives objects of class "fclust".
n_of_classes.fclust <- n_of_classes.fanny
## Package e1071: cshell().
n_of_classes.cshell <- n_of_classes.fanny
## Package e1071: bclust().
n_of_classes.bclust <- n_of_classes.default

## Package mclust: Mclust().
n_of_classes.Mclust <- n_of_classes.default

## Package clue: Memberships.
n_of_classes.cl_membership <-
function(x)
    attr(x, "n_of_classes")
## Package clue: pclust().
n_of_classes.pclust <-
function(x)
{
    if(is.null(m <- x$membership)) 
        length(unique(cl_class_ids(x)))
    else
        sum(colSums(m, na.rm = TRUE) > 0)
}
## Package clue: (virtual) class "cl_partition".
n_of_classes.cl_partition <-
function(x)
    n_of_classes(.get_representation(x))

### * cl_class_ids

## Get ids of classes in a partition.
## <NOTE>
## Currently, all supported soft partitioning methods provide a softmax
## hard partitioning as well.
## </NOTE>

cl_class_ids <-
function(x)
    UseMethod("cl_class_ids")

## Default method.
cl_class_ids.default <-
function(x)
{
    stop("Cannot infer class ids from given object.")
}

## Package stats: kmeans() (R 2.1.0 or better).
cl_class_ids.kmeans <-
function(x)
    as.cl_class_ids(x$cluster)

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
cl_class_ids.partition <-
function(x)
    as.cl_class_ids(x$clustering)

## Package RWeka: clusterers return objects inheriting from
## "Weka_clusterer".
cl_class_ids.Weka_clusterer <-
function(x)
    as.cl_class_ids(x$class_ids)

## Package cba: ccfkms().
cl_class_ids.ccfkms <-
function(x)
    as.cl_class_ids(as.vector(x$cl))
## Package cba: rockCluster() returns objects of class "rock".
cl_class_ids.rock <-
function(x)
    as.cl_class_ids(as.vector(x$cl))

## Package cclust: cclust().
cl_class_ids.cclust <- cl_class_ids.kmeans

## Package e1071: cmeans() gives objects of class "fclust".
cl_class_ids.fclust <- cl_class_ids.kmeans
## Package e1071: cshell().
cl_class_ids.cshell <- cl_class_ids.kmeans
## Package e1071: bclust().
cl_class_ids.bclust <- cl_class_ids.kmeans

## Package flexclust: kcca() returns objects of S4 class "kcca" which
## extends S4 class "flexclust".
## <NOTE>
## We used to be able to call flexclust::cluster(), but this now only
## has S4 methods for modeltools::clusters() S4 generic.  Let's call this
## one, and hope that flexclust has been loaded ...
## </NOTE>
cl_class_ids.kcca <-
function(x)
    as.cl_class_ids(modeltools::clusters(x))

## Package flexmix: class "flexmix".
## <NOTE>
## We used to be able to call flexmix::cluster(), but this now only has
## S4 methods for modeltools::clusters() S4 generic.  Let's call this
## one, and hope that flexmix has been loaded ...
## </NOTE>
cl_class_ids.flexmix <-
function(x)
    as.cl_class_ids(modeltools::clusters(x))

## Package kernlab: specc() and kkmeans() return objects of S4 class
## "specc".
cl_class_ids.specc <-
function(x)
{
    tmp <- unclass(x)
    as.cl_class_ids(.structure(as.vector(tmp), names = names(tmp)))
}

## Package mclust: Mclust().
cl_class_ids.Mclust <-
function(x)
    as.cl_class_ids(x$classification)

## Package relations: equivalence and preference relations.
cl_class_ids.relation <-
function(x)
    as.cl_class_ids(relations::relation_class_ids(x))

## Package clue: Class ids.
cl_class_ids.cl_class_ids <- identity
## Package clue: Memberships.
cl_class_ids.cl_membership <-
function(x)
    as.cl_class_ids(.structure(max.col(x), names = rownames(x)))
## (Cannot do cl_class_ids.cl_membership <- max.col for generic/method
## consistency.)
## Package clue: cl_pam().
cl_class_ids.cl_pam <- cl_class_ids.kmeans
## Package clue: cl_partition_by_class_ids().
cl_class_ids.cl_partition_by_class_ids <-
function(x)
    .get_representation(x)
## Package clue: kmedoids().
cl_class_ids.kmedoids <- cl_class_ids.kmeans
## Package clue: pclust().
cl_class_ids.pclust <- cl_class_ids.kmeans
## Package clue: (virtual) class "cl_partition".
cl_class_ids.cl_partition <-
function(x)
    cl_class_ids(.get_representation(x))

## Package movMF: class "movMF".
cl_class_ids.movMF <-
function(x)
    as.cl_class_ids(max.col(x$P))

### * as.cl_class_ids

as.cl_class_ids <-
function(x)
{
    ## For the time being, handle only "raw" class ids.
    ## Maybe add methods handling factors lateron (if necessary).
    ## <NOTE>
    ## This could also be used to canonicalize returned class ids
    ## according to the docs (vector of integers with the class ids),
    ## using someting like
    ##   match(ids, unique(ids))
    ## </NOTE>
    .structure(unclass(x), class = "cl_class_ids")
}

### * print.cl_class_ids

print.cl_class_ids <-
function(x, ...)
{
    writeLines("Class ids:")
    print(unclass(x), ...)
    invisible(x)
}

### * cl_class_labels

cl_class_labels <-
function(x)
    UseMethod("cl_class_labels")

### * is.cl_partition

## Determine whether an object is a (generalized) partition.
## Note that this includes both hard and soft partitions, and allows
## sums of memberships of objects to be less than one.

is.cl_partition <-
function(x)
    UseMethod("is.cl_partition")

## Default method.
is.cl_partition.default <- .false

## Package stats: kmeans() (R 2.1.0 or better).
is.cl_partition.kmeans <- .true

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
is.cl_partition.partition <- .true

## Package RWeka: clusterers return objects inheriting from
## "Weka_clusterer".
## (Note that Cobweb internally uses a classification tree, but
## definitely does not expose this structure.)
is.cl_partition.Weka_clusterer <- .true

## Package cba: ccfkms().
is.cl_partition.ccfkms <- .true
## Package cba: rockCluster() returns objects of class "rock".
is.cl_partition.rock <- .true

## Package cclust: cclust().
is.cl_partition.cclust <- .true

## Package e1071: cmeans() gives objects of class "fclust".
is.cl_partition.fclust <- .true
## Package e1071: cshell().
is.cl_partition.cshell <- .true
## Package e1071: bclust().
is.cl_partition.bclust <- .true

## Package flexclust: kcca() returns objects of S4 class "kcca" which
## extends S4 class "flexclust".
is.cl_partition.kcca <- .true

## Package flexmix: class "flexmix".
is.cl_partition.flexmix <- .true

## Package kernlab: specc() and kkmeans() return objects of S4 class
## "specc".
is.cl_partition.specc <- .true

## Package mclust: Mclust().
is.cl_partition.Mclust <- .true

## Package clue: (virtual) class "cl_partition".
## Note that "raw" cl_membership objects are *not* partitions, as they
## are meant for numeric computations.
is.cl_partition.cl_partition <- .true
## Package clue: kmedoids().
is.cl_partition.kmedoids <- .true
## Package clue: pclust().
is.cl_partition.pclust <- .true

## Package movMF: class "movMF".
is.cl_partition.movMF <- .true

### * as.cl_partition

## Note that cl_partition conceptually is a virtual class, so there are
## no prototypes and no cl_partition() creator.

.cl_partition_classes <- "cl_partition"

as.cl_partition <-
function(x)
{
    if(is.cl_partition(x)) {
        if(!inherits(x, "cl_partition"))
            .make_container(x, .cl_partition_classes)
        else
            x
    }
    else
        cl_partition_by_memberships(as.cl_membership(x))
}

### * print.cl_partition

print.cl_partition <-
function(x, ...)
    .print_container(x, "cl_partition", ...)

### * print.cl_partition_by_class_ids

print.cl_partition_by_class_ids <-
function(x, ...)
{
    writeLines(gettextf("A hard partition of %d objects.",
                        n_of_objects(x)))
    print(cl_class_ids(x), ...)
    invisible(x)
}

### * print.cl_partition_by_memberships

print.cl_partition_by_memberships <-
function(x, ...)
{
    writeLines(gettextf("A partition of %d objects.",
                        n_of_objects(x)))
    print(cl_membership(x), ...)
    invisible(x)
}

### * Complex.cl_partition

Complex.cl_partition <-
function(z)
    stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                  .Generic, .Class),
         domain = NA)

### * Math.cl_partition

Math.cl_partition <-
function(x, ...)
    stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                  .Generic, .Class),
         domain = NA)

### * Ops.cl_partition

Ops.cl_partition <-
function(e1, e2)
{
    if(nargs() == 1L)
        stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    ## Only comparisons are supprorted.
    if(!(as.character(.Generic) %in% c("<", "<=", ">", ">=",
                                       "==", "!=")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    ci1 <- cl_class_ids(e1)
    ci2 <- cl_class_ids(e2)
    if(length(ci1) != length(ci2))
        stop("Partitions must have the same number of objects.")
    z <- table(ci1, ci2) > 0
    switch(.Generic,
           "<=" = all(rowSums(z) == 1),
           "<"  = all(rowSums(z) == 1) && any(colSums(z) > 1),
           ">=" = all(colSums(z) == 1),
           ">"  = all(colSums(z) == 1) && any(rowSums(z) > 1),
           "==" = all(rowSums(z) == 1) && all(colSums(z) == 1),
           "!=" = any(rowSums(z) > 1) || any(colSums(z) > 1))

}

### * Summary.cl_partition

Summary.cl_partition <-
function(..., na.rm = FALSE)
{
    ok <- switch(.Generic, max = , min = , range = TRUE, FALSE)
    if(!ok)
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)
    args <- list(...)
    switch(.Generic,
           "min" = cl_meet(cl_ensemble(list = args)),
           "max" = cl_join(cl_ensemble(list = args)),
           "range" = {
               cl_ensemble(min = cl_meet(cl_ensemble(list = args)),
                           max = cl_join(cl_ensemble(list = args)))
           })
}

### * cl_partition_by_class_ids

cl_partition_by_class_ids <-
function(x, labels = NULL)
{
    if(!is.atomic(x))
        stop("Class ids must be atomic.")
    if(is.null(names(x)))
        names(x) <- labels
    ## <FIXME>
    ## Perhaps give the raw class ids more structure?
    ## E.g, class "cl_class_ids"?
    ## Problem is that we used to say about extensibility that all there
    ## is to do for a hard partitioner is to add a cl_class_ids() method
    ## and two predicates, but *not* to have the former give a suitably
    ## classed object.  On the other hand, the recipe would need to be
    ## extended for soft partitioners, for which it would be necessary
    ## to provide a cl_membership() method which really returns an
    ## object of class cl_membership.  Note that we can do this using
    ## as.cl_membership(m), where m is the raw membership matrix.  So
    ## maybe we should ask for using as.cl_class_ids() to coerce raw
    ## class ids ...
    .make_container(as.cl_class_ids(x),
                    c("cl_partition_by_class_ids",
                      .cl_hard_partition_classes),
                    list(n_of_objects = length(x),
                         n_of_classes = length(unique(x))))
    ## </FIXME>
}

### * cl_partition_by_memberships

cl_partition_by_memberships <-
function(x, labels = NULL)
{
    if(!is.matrix(x)
       || any(x < 0, na.rm = TRUE)
       || any(x > 1, na.rm = TRUE))
        stop("Not a valid membership matrix.")
    ## Be nice.
    x <- x / rowSums(x, na.rm = TRUE)
    ## (Note that this does not imply all(rowSums(x) == 1).  If we
    ## wanted to test for this, something like
    ##    .is_stochastic_matrix <- function(x)
    ##       identical(all.equal(rowSums(x), rep(1, nrow(x))), TRUE))
    ## should do.)
    if(is.null(rownames(x)))
        rownames(x) <- labels
    .make_container(as.cl_membership(x),
                    c("cl_partition_by_memberships",
                      .cl_partition_classes),
                    list(n_of_objects = nrow(x)))
}

### * is.cl_hard_partition

## Determine whether an object is a hard partition.

is.cl_hard_partition <-
function(x)
    UseMethod("is.cl_hard_partition")

## Default method.
is.cl_hard_partition.default <- .false

## Package stats: kmeans() (R 2.1.0 or better).
is.cl_hard_partition.kmeans <- .true

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
## <NOTE>
## Of course, fuzzy clustering can also give a hard partition ...
is.cl_hard_partition.fanny <-
function(x)
{
    all(rowSums(cl_membership(x) == 1, na.rm = TRUE) > 0)
}
## </NOTE>
is.cl_hard_partition.partition <- .true

## Package RWeka: clusterers return objects inheriting from
## "Weka_clusterer".
is.cl_hard_partition.Weka_clusterer <- .true

## Package cba: ccfkms().
is.cl_hard_partition.ccfkms <- .true
## Package cba: rockCluster() returns objects of class "rock".
is.cl_hard_partition.rock <- .true

## Package cclust: cclust().
is.cl_hard_partition.cclust <- .true

## Package e1071: cmeans() gives objects of class "fclust".
is.cl_hard_partition.fclust <- is.cl_hard_partition.fanny
## Package e1071: cshell().
is.cl_hard_partition.cshell <- is.cl_hard_partition.fanny
## Package e1071: bclust().
is.cl_hard_partition.bclust <- .true

## Package flexclust: kcca() returns objects of S4 class "kcca" which
## extends S4 class "flexclust".
is.cl_hard_partition.kcca <- .true

## Package flexmix: class "flexmix".
is.cl_hard_partition.flexmix <- is.cl_hard_partition.fanny

## Package kernlab: specc() and kkmeans() return objects of S4 class
## "specc".
is.cl_hard_partition.specc <- .true

## Package mclust: Mclust().
is.cl_hard_partition.Mclust <- is.cl_hard_partition.fanny

## Package clue: (virtual) class "cl_hard_partition".
is.cl_hard_partition.cl_hard_partition <- .true
## Package clue: (virtual) class "cl_partition".
## Note that "raw" cl_membership objects are *not* partitions, as they
## are meant for numeric computations.
## Rather than providing is.cl_hard_partition.cl_membership() we thus
## prefer explicit handling of cl_partition objects with a cl_membership
## representation.
is.cl_hard_partition.cl_partition <-
function(x)
{
    ## If the object has a cl_membership representation ...
    y <- .get_representation(x)
    if(inherits(y, "cl_membership"))
        attr(y, "is_cl_hard_partition")
    ## Other representations, e.g. for "definitely" hard partitions via
    ## vectors of class ids or class labels, or a list of classes, may
    ## be added in future versions.
    ## In any case, this must be kept in sync with what is handled by
    ## as.cl_partition() [which currently runs as.cl_membership() in
    ## case is.cl_partition() gives false].
    else
        is.cl_hard_partition(y)
}
## Package clue: kmedoids().
is.cl_hard_partition.kmedoids <- .true
## Package clue: pclust().
is.cl_hard_partition.pclust <- is.cl_hard_partition.fanny

## Package movMF: class "movMF".
is.cl_hard_partition.movMF <- is.cl_hard_partition.fanny

### * as.cl_hard_partition

.cl_hard_partition_classes <- c("cl_hard_partition", "cl_partition")

as.cl_hard_partition <-
function(x)
{
    if(is.cl_hard_partition(x)) {
        if(!inherits(x, "cl_partition"))
            .make_container(x, .cl_hard_partition_classes)
        else
            x
    }
    else if(is.cl_partition(x)) {
        ## A soft cl_partition ...
        ids <- cl_class_ids(x)
        cl_partition_by_class_ids(ids, names(ids))
    }
    else if(is.matrix(x)) {
        ## A matrix of raw memberships, hopefully ...
        cl_partition_by_class_ids(max.col(x), rownames(x))
    }
    else if(is.atomic(x)) {
        ## A vector of raw class ids, hopefully ...
        cl_partition_by_class_ids(x, names(x))
    }
    else
        stop("Cannot coerce to 'cl_hard_partition'.")
}

### * is.cl_soft_partition

## Determine whether an object is a soft partition.

is.cl_soft_partition <-
function(x)
    is.cl_partition(x) && ! is.cl_hard_partition(x)

### * .maybe_is_proper_soft_partition

## Determine whether an object might be a proper soft partition (in the
## sense that it is a cl_partition but not a cl_hard_partition).
## This is mostly useful when computing fuzziness measures.

.maybe_is_proper_soft_partition <-
function(x)
    UseMethod(".maybe_is_proper_soft_partition")
.maybe_is_proper_soft_partition.default <- .false
.maybe_is_proper_soft_partition.fanny <- .true
.maybe_is_proper_soft_partition.fclust <- .true
.maybe_is_proper_soft_partition.cshell <- .true
.maybe_is_proper_soft_partition.flexmix <- .true
.maybe_is_proper_soft_partition.Mclust <- .true
## See above for why we prefer not to have
## .maybe_is_proper_soft_partition.cl_membership().
## (Although this is an internal generic really only used for making
## cl_fuzziness() computations more efficient, so we could be more
## generous here [perhaps using a slightly different name such as
## .maybe_represents_a_proper_soft_partition()].
.maybe_is_proper_soft_partition.cl_partition <-
function(x)
{
    y <- .get_representation(x)
    if(inherits(y, "cl_membership"))
        !attr(y, "is_cl_hard_partition")
    else
        .maybe_is_proper_soft_partition(y)
}
.maybe_is_proper_soft_partition.pclust <-
function(x)
    x$m > 1


### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### * cl_medoid

cl_medoid <-
function(x, method = "euclidean")
{
    ## <NOTE>
    ## In principle we can get the same using pam(k = 1)$medoids.
    ## </NOTE>

    clusterings <- as.cl_ensemble(x)
    if(!length(clusterings))
        stop("Cannot compute medoid of empty ensemble.")
    dissimilarities <-
        as.matrix(cl_dissimilarity(clusterings, method = method))
    clusterings[[which.min(rowSums(dissimilarities))]]
}

### * cl_pam

cl_pam <-
function(x, k, method = "euclidean", solver = c("pam", "kmedoids"))
{

    clusterings <- as.cl_ensemble(x)
    if(!length(clusterings))
        stop("Cannot compute medoid partition of empty ensemble.")

    ## Actually, we should have at least k distinct elements in the
    ## ensemble ...

    make_cl_pam <- function(class_ids, medoid_ids, medoids, criterion,
                            description)
        .structure(list(cluster = class_ids,
                        medoid_ids = medoid_ids,
                        prototypes = medoids,
                        criterion = criterion,
                        description = description),
                   class = "cl_pam")

    if(k == 1L) {
        ## Simplify matters if a global medoid is sought.
        dissimilarities <-
            cl_dissimilarity(clusterings, method = method)
        description <- attr(dissimilarities, "description")
        dissimilarities <- as.matrix(dissimilarities)
        row_sums <- rowSums(dissimilarities)
        medoid_id <- which.min(row_sums)
        criterion <- row_sums[medoid_id]
        return(make_cl_pam(as.cl_class_ids(seq_along(clusterings)),
                           medoid_id,
                           clusterings[medoid_id],
                           criterion,
                           description))
    }

    solver <- match.arg(solver)

    ## Argh.  We really want to run k-medoids for the unique elements of
    ## the ensemble, but pam() only works for symmetric dissimilarties.
    ## As computing cluster dissimilarities is typically expensive, use
    ## the unique elements for doing so in any case.
    
    values <- unique(clusterings)
    ## Positions of ensemble members in the unique values.
    positions <- match(clusterings, values)

    ## Dissimilarities between unique values.
    dissimilarities <- cl_dissimilarity(values, method = method)
    description <- attr(dissimilarities, "description")
    dissimilarities <- as.matrix(dissimilarities)

    ## For pam(), we need the dissimilarities for all objects.
    if(solver == "pam") {
        dissimilarities <- dissimilarities[positions, positions]
        party <- pam(as.dist(dissimilarities), k)
        class_ids <- cl_class_ids(party)
        medoid_ids <- cl_medoid_ids(party)
        medoids <- clusterings[medoid_ids]
        criterion <- sum(dissimilarities[cbind(seq_along(class_ids),
                                               medoid_ids[class_ids])])
    }
    else {
        ## Counts of unique values.
        counts <- tabulate(positions)

        ## Weigh according to the counts.  Should be straightforward to
        ## add "case weights" as well ...
        dissimilarities <- counts * dissimilarities

        ## Now partition.
        party <- kmedoids(dissimilarities, k)

        ## And build the solution from this ...
        criterion <- party$criterion        
        ## First, things for the unique values.
        medoid_ids <- cl_medoid_ids(party)
        medoids <- values[medoid_ids]
        class_ids <- cl_class_ids(party)
        ## Second, things for all objects.
        class_ids <- class_ids[positions]
        medoid_ids <- match(medoid_ids, positions)
    }

    make_cl_pam(class_ids, medoid_ids, medoids, criterion, description)
}

print.cl_pam <-
function(x, ...)
{
    class_ids <- cl_class_ids(x)
    fmt <- "A k-medoid partition of a cluster ensemble with %d elements into %d
classes (dissimilarity measure: %s)."
    writeLines(c(strwrap(gettextf(fmt,
                                  n_of_objects(x),
                                  n_of_classes(x),
                                  x$description))))
    writeLines(gettext("Class ids:"))
    print(class_ids, ...)
    writeLines(gettext("Criterion:"))
    print(x$criterion, ...)
    invisible(x)
}

### * cl_medoid_ids

## Little helper, internal for the time being ...
cl_medoid_ids <- function(x) UseMethod("cl_medoid_ids")
cl_medoid_ids.cl_pam <- function(x) x$medoid_ids
cl_medoid_ids.kmedoids <- function(x) x$medoid_ids
cl_medoid_ids.clara <- function(x) x$i.med
cl_medoid_ids.pam <- function(x) x$id.med

### * kmedoids

kmedoids <-
function(x, k)
{
    ## <FIXME>
    ## For the time being, 'x' is assumed a dissimilarity object or a
    ## matrix of dissimilarities.
    ## Let's worry about the interface later.
    ## </FIXME>

    x <- as.matrix(x)

    n <- nrow(x)

    ## Use the formulation in Gordon & Vichi (1998), Journal of
    ## Classification, [P4'], page 279, with variables c(vec(X), z), but
    ## with rows and cols interchanged (such that x_{ij} is one iff o_i
    ## has medoid o_j, and z_j is one iff o_j is a medoid).

    make_constraint_mat <- function(n) {
        nsq <- n * n
        rbind(cbind(kronecker(rbind(rep.int(1, n)), diag(1, n)),
                    matrix(0, n, n)),
              cbind(diag(1, nsq),
                    kronecker(diag(1, n), rep.int(-1, n))),
              c(double(nsq), rep.int(1, n)),
              cbind(matrix(0, n, nsq), diag(1, n)))
    }

    make_constraint_dir <- function(n)
        rep.int(c("=", "<=", "=", "<="), c(n, n * n, 1, n))

    make_constraint_rhs <- function(n, k)
        rep.int(c(1, 0, k, 1), c(n, n * n, 1, n))

    ## <NOTE>
    ## We could try a relaxation without integrality constraints first,
    ## which seems to "typically work" (and should be faster).  To test
    ## for integrality, use something like
    ##   if(identical(all.equal(y$solution, round(y$solution)), TRUE))
    ## </NOTE>

    y <- lpSolve::lp("min",
                     c(c(x), double(n)),
                     make_constraint_mat(n),
                     make_constraint_dir(n),
                     make_constraint_rhs(n, k),
                     int.vec = seq_len(n * (n + 1)))

    ## Now get the class ids and medoids.

    ind <- which(matrix(y$solution[seq_len(n * n)], n) > 0,
                 arr.ind = TRUE)
    medoid_ids <- unique(ind[, 2L])
    class_ids <- seq_len(n)
    class_ids[ind[, 1L]] <- match(ind[, 2L], medoid_ids)

    .structure(list(cluster = class_ids,
                    medoid_ids = medoid_ids,
                    criterion = y$objval),
               class = "kmedoids")
}

print.kmedoids <-
function(x, ...)
{
    fmt <- "A k-medoids clustering of %d objects into %d clusters."
    writeLines(gettextf(fmt, n_of_objects(x), n_of_classes(x)))
    writeLines(gettext("Medoid ids:"))
    print(cl_medoid_ids(x), ...)
    writeLines(gettext("Class ids:"))
    print(unclass(cl_class_ids(x)), ...)
    writeLines(gettext("Criterion:"))
    print(x$criterion, ...)
    invisible(x)
}

## <FIXME>
## Maybe add support for "auto" type (class_ids when predicting from a
## hard, memberships when predicting from a soft partition) eventually.
## </FIXME>

cl_predict <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
    UseMethod("cl_predict")

## Default method.
## Should also work for kcca() from package flexclust.
cl_predict.default <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
    .as_cl_class_ids_or_membership(predict(object, newdata, ...), type)

## Package stats: kmeans() (R 2.1.0 or better).
cl_predict.kmeans <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    d <- .rxdist(newdata, object$centers)
    .as_cl_class_ids_or_membership(max.col(-d), type)
}

## Package cluster:
## * fanny() cannot make "new" predictions.
## * clara() gives medoids, and takes metric data using Euclidean or
##   Manhattan dissimilarities (and we can figure out which by looking
##   at the call and the default values).
## * pam() gives medoids, but might have been called with dissimilarity
##   data, so is tricky.  We can always find out which by looking at the
##   medoids: as in the dissimilarity input case this is a vector of
##   class labels, and a matrix with in each row the coordinates of one
##   medoid otherwise.  We then still need to figure out whether
##   Euclidean or Manhattan distances were used by looking at the call
##   and the default values.
## Both pam() and clara() show that the interfaces could be improved to
## accomodate modern needs, e.g., for bagging.

cl_predict.fanny <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    stop("Cannot make new predictions.")
}

cl_predict.clara <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    ## <FIXME>
    ## Add support eventually ...
    if(identical(object$call$stand, TRUE))
        warning("Standardization is currently not supported.")
    ## </FIXME>
    method <- object$call$metric
    if(is.null(method)) {
        ## Not given in the call, hence use default value.
        method <- eval(formals(clara)$metric)[1L]
        ## (Or hard-wire the default value: "euclidean".)
    }
    d <- .rxdist(newdata, object$medoids, method)
    .as_cl_class_ids_or_membership(max.col(-d), type)
}

cl_predict.pam <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    prototypes <- object$medoids
    if(!is.matrix(prototypes))
        stop("Cannot make new predictions.")
    ## <FIXME>
    ## Add support eventually ...
    if(identical(object$call$stand, TRUE))
        warning("Standardization is currently not supported.")
    ## </FIXME>
    method <- object$call$metric
    if(is.null(method)) {
        ## Not given in the call, hence use default value.
        method <- eval(formals(pam)$metric)[1L]
        ## (Or hard-wire the default value: "euclidean".)
    }
    d <- .rxdist(newdata, object$medoids, method)
    .as_cl_class_ids_or_membership(max.col(-d), type)
}

## Package RWeka: clusterers return objects inheriting from
## "Weka_clusterer".
cl_predict.Weka_clusterer <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    .as_cl_class_ids_or_membership(predict(object, newdata = newdata,
                                           type = type, ...),
                                   type)
}

## Package cba: ccfkms().
cl_predict.ccfkms <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    .as_cl_class_ids_or_membership(as.vector(predict(object,
                                                     newdata)$cl),
                                   type)
}
## Package cba: rockCluster() returns objects of class "rock".
## If x is a Rock object, fitted(x) and predict(x, newdata) can result
## in missing classifications, as
##   In the case a 'drop' value greater than zero is specified, all
##   clusters with size equal or less than this value are removed from
##   the classifier.  Especially, 'fitted' uses a threshold of one
##   because for singleton clusters the neighborhood is empty.
cl_predict.rock <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata)) newdata <- object$x
    ids <- as.vector(predict(object, newdata, ...)$cl)
    .as_cl_class_ids_or_membership(ids, type)
}

## Package cclust: cclust().
cl_predict.cclust <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    ## Package cclust provides predict.cclust() which returns (again) an
    ## object of class "cclust", but does not give the labels of the
    ## original data in case no new data are given.
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    .as_cl_class_ids_or_membership(predict(object, newdata), type)
}

## Package e1071: cmeans() gives objects of class "fclust".
cl_predict.fclust <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))

    ## Note that the 'fclust' objects returned by cmeans() do not always
    ## directly contain the information on the fuzzification parameter m
    ## and the distance (Euclidean/Manhattan) employed, so we have to
    ## engineer this from the matched call and the default arguments.
    nms <- names(object$call)
    ## Note that we cannot directly use object$call$m, as this could
    ## give the 'method' argument if 'm' was not given.
    m <- if("m" %in% nms)
        object$call$m
    else {
        ## Not given in the call, hence use default value.
        formals(e1071::cmeans)$m
        ## (Or hard-wire the default value: 2.)
    }
    method <- if("dist" %in% nms)
        object$call$dist
    else {
        ## Not given in the call, hence use default value.
        formals(e1071::cmeans)$dist
        ## (Or hard-wire the default value: "euclidean".)
    }

    d <- .rxdist(newdata, object$centers, method)
    power <- c(m, if(method == "euclidean") 2 else 1)
    M <- .memberships_from_cross_dissimilarities(d, power)
    .as_cl_class_ids_or_membership(M, type)
}

## Package e1071: cshell().
cl_predict.cshell <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))

    ## Not surprisingly, this is rather similar to what we do for fclust
    ## objects.  Only dissimiliraties (and exponents) need to be
    ## computed differently ...
    nms <- names(object$call)
    m <- if("m" %in% nms)
        object$call$m
    else {
        ## Not given in the call, hence use default value.
        formals(e1071::cshell)$m
        ## (Or hard-wire the default value: 2.)
    }
    method <- if("dist" %in% nms)
        object$call$dist
    else {
        ## Not given in the call, hence use default value.
        formals(e1071::cshell)$dist
        ## (Or hard-wire the default value: "euclidean".)
    }

    d <- .rxdist(newdata, object$centers, method)
    d <- sweep(d, 2, object$radius) ^ 2
    M <- .memberships_from_cross_dissimilarities(d, m)
    .as_cl_class_ids_or_membership(M, type)
}

## Package e1071: bclust().
## <NOTE>
## One might argue that it would be better to use the 'dist.method'
## employed for the hierarchical clustering, but it seems that class
## labels ("clusters") are always assigned using Euclidean distances.
cl_predict.bclust <- cl_predict.kmeans
## </NOTE>

## Package flexclust: kcca() returns objects of S4 class "kcca" which
## extends S4 class "flexclust".
cl_predict.kcca <- cl_predict.default

## Package flexmix: class "flexmix".
cl_predict.flexmix <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))
    .as_cl_class_ids_or_membership(modeltools::posterior(object,
                                                         newdata,
                                                         ...),
                                   type)
}

## Package mclust: Mclust().
cl_predict.Mclust <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))

    pred <- predict(object, newdata, ...)
    type <- match.arg(type)
    if(type == "class_ids")
        as.cl_class_ids(pred$classification)
    else
        as.cl_membership(pred$z)
}


## Package movMF: movMF().
cl_predict.movMF <- cl_predict.Weka_clusterer

## Package clue: pclust().
cl_predict.pclust <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
{
    if(is.null(newdata))
        return(.cl_class_ids_or_membership(object, type))

    d <- object$family$D(newdata, object$prototypes)
    power <- c(object$m, object$family$e)
    M <- .memberships_from_cross_dissimilarities(d, power)
    .as_cl_class_ids_or_membership(M, type)
}

## Package clue: (virtual) class "cl_partition".
cl_predict.cl_partition <-
function(object, newdata = NULL,
         type = c("class_ids", "memberships"), ...)
    cl_predict(.get_representation(object), newdata = newdata, type, ...)

## Internal helpers: this looks a bit silly, but makes the rest of the
## code look nicer ...

.cl_class_ids_or_membership <-
function(x, type = c("class_ids", "memberships"))
{
    type <- match.arg(type)

    if(type == "class_ids")
        cl_class_ids(x)
    else
        cl_membership(x)
}

.as_cl_class_ids_or_membership <-
function(x, type = c("class_ids", "memberships"))
{
    type <- match.arg(type)

    if(type == "class_ids") {
        if(is.matrix(x)) {
            ## Same as for cl_class_ids.cl_membership().
            as.cl_class_ids(.structure(max.col(x), names = rownames(x)))
        }
        else
            as.cl_class_ids(x)
    }
    else
        as.cl_membership(x)
}

solve_LSAP <-
function(x, maximum = FALSE)
{
    if(!is.matrix(x) || any(x < 0))
        stop("x must be a matrix with nonnegative entries.")

    nr <- nrow(x)
    nc <- ncol(x)
    if(nr > nc)
        stop("x must not have more rows than columns.")
    if(nc > nr)
        x <- rbind(x, matrix(2 * sum(x), nc - nr, nc))

    if(maximum) x <- max(x) - x

    storage.mode(x) <- "double"
    out <- .C(C_solve_LSAP,
              x,
              as.integer(nc),
              p = integer(nc))$p + 1
    out <- out[seq_len(nr)]
    class(out) <- "solve_LSAP"
    out
}

print.solve_LSAP <-
function(x, ...)
{
    writeLines(c("Optimal assignment:",
                 gsub("x", " ",
                      strwrap(paste(seq_along(x), x,
                                    sep = "x=>x", collapse = ", ")))))
    invisible(x)
}

### * cl_agreement

cl_agreement <-
function(x, y = NULL, method = "euclidean", ...)
{
    ## <NOTE>
    ## This code is repeated from cl_dissimilarity(), mutatis mutandis.
    ## Not really a big surprise ...
    ## </NOTE>
    
    x <- as.cl_ensemble(x)
    is_partition_ensemble <-
        (inherits(x, "cl_partition_ensemble")
         || all(vapply(x, .has_object_memberships, NA)))

    ## Be nice.
    if(is.character(y) || is.function(y)) {
        method <- y
        y <- NULL
    }

    if(is.function(method))
        method_name <- "user-defined method"
    else {
        if(!inherits(method, "cl_agreement_method")) {
            ## Get the method definition and description from the
            ## registry. 
            type <- ifelse(is_partition_ensemble,
                           "partition", "hierarchy")
            method <- get_cl_agreement_method(method, type)
        }
        method_name <- method$description
        method <- method$definition
    }

    if(!is.null(y)) {
        y <- as.cl_ensemble(y)
        is_partition_ensemble_y <-
            (inherits(y, "cl_partition_ensemble")
             || all(vapply(x, .has_object_memberships, NA)))
        if(!identical(is_partition_ensemble, is_partition_ensemble_y))
            stop("Cannot mix partitions and hierarchies.")
        if(n_of_objects(x) != n_of_objects(y))
            stop("All clusterings must have the same number of objects.")
        ## Build a cross-proximity object of cross-agreements.
        d <- matrix(0, length(x), length(y))
        for(j in seq_along(y))
            d[, j] <- sapply(x, method, y[[j]], ...)
        dimnames(d) <- list(names(x), names(y))
        return(cl_cross_proximity(d, method_name,
                                  class = "cl_cross_agreement"))
    }
    
    ## Otherwise, build a proximity object of dissimilarities.
    n <- length(x)
    d <- vector("list", length = n - 1L)
    ind <- seq_len(n)
    while(length(ind) > 1L) {
        j <- ind[1L]
        ind <- ind[-1L]
        d[[j]] <- sapply(x[ind], method, x[[j]], ...)
    }

    ## <NOTE>
    ## We assume that self-agreements are always one ...
    ## </NOTE>
    cl_proximity(unlist(d),
                 method_name,
                 labels = names(x),
                 self = rep.int(1, length(x)),
                 size = n, class = "cl_agreement")
}

### ** .cl_agreement_partition_euclidean

.cl_agreement_partition_euclidean <-
function(x, y)
{
    ## <NOTE>
    ## Upper bound for maximal dissimilarity, maybe improve eventually.
    d_max <- sqrt(2 * n_of_objects(x))
    ## </NOTE>
    1 - .cl_dissimilarity_partition_euclidean(x, y) / d_max
}

### ** .cl_agreement_partition_manhattan

.cl_agreement_partition_manhattan <-
function(x, y)
{
    ## <NOTE>
    ## Upper bound for maximal dissimilarity, maybe improve eventually.
    d_max <- 2 * n_of_objects(x)
    ## </NOTE>
    1 - .cl_dissimilarity_partition_manhattan(x, y) / d_max
}
    
### ** .cl_agreement_partition_Rand

.cl_agreement_partition_Rand <-
function(x, y)
{
    n <- n_of_objects(x)
    
    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)
    x <- table(cl_class_ids(x), cl_class_ids(y))

    ## <NOTE>
    ## The number A of concordant pairs is given by
    ##   A = choose(n,2) + \sum_{i,j} x_{ij}^2
    ##       - (1/2) * (\sum_i x_{i.}^2 + \sum_j x_{.j}^2)
    ##     = choose(n,2) + 2 \sum_{i,j} choose(x_{ij},2)
    ##       - (\sum_i choose(x_{i.},2) + \sum_j choose(x_{.j},2)
    ## with the first version certainly much faster to compute.
    ## </NOTE>
    1 + (sum(x^2) -
         (sum(rowSums(x)^2) + sum(colSums(x)^2)) / 2) / choose(n, 2)
}

### ** .cl_agreement_partition_cRand

.cl_agreement_partition_cRand <-
function(x, y)
{
    if(!is.cl_hard_partition(x) || !is.cl_hard_partition(y))
        stop("Can only handle hard partitions.")

    n <- n_of_objects(x)    
    x <- table(cl_class_ids(x), cl_class_ids(y))

    ## <NOTE>
    ## The basic formula is
    ##   (Sxy - E) / ((Sx. + S.y) / 2 - E)
    ## where
    ##   Sxy = \sum_{i,j} choose(x_{ij}, 2)
    ##   Sx. = \sum_i     choose(x_{i.}, 2)
    ##   S.y = \sum_j     choose(x_{.j}, 2)
    ## and
    ##   E = Sx. * S.y / choose(n, 2)
    ## We replace the bincoefs by the corresponding sums of squares,
    ## getting
    ##   (Txy - F) / ((Tx. + T.y) / 2 - F)
    ## where
    ##   Txy = \sum_{i,j} x_{ij}^2 - n
    ##   Tx. = \sum_i     x_{i.}^2 - n
    ##   T.y = \sum_j     x_{.j}^2 - n
    ## and
    ##   F = Tx. * T.y / (n^2 - n)
    ## </NOTE>
    Txy <- sum(x ^ 2) - n
    Tx. <- sum(rowSums(x) ^ 2) - n
    T.y <- sum(colSums(x) ^ 2) - n
    F <- Tx. * T.y / (n ^ 2 - n)
    (Txy - F) / ((Tx. + T.y) / 2 - F)
}


### ** .cl_agreement_partition_NMI

.cl_agreement_partition_NMI <-
function(x, y)
{
    if(!is.cl_hard_partition(x) || !is.cl_hard_partition(y))
        stop("Can only handle hard partitions.")
    x <- table(cl_class_ids(x), cl_class_ids(y))
    x <- x / sum(x)
    
    m_x <- rowSums(x)
    m_y <- colSums(x)
    y <- outer(m_x, m_y)

    i <- which((x > 0) & (y > 0))
    out <- sum(x[i] * log(x[i] / y[i]))
    e_x <- sum(m_x * log(ifelse(m_x > 0, m_x, 1)))
    e_y <- sum(m_y * log(ifelse(m_y > 0, m_y, 1)))

    out / sqrt(e_x * e_y)
}

### ** .cl_agreement_partition_KP

.cl_agreement_partition_KP <-
function(x, y)
{
    ## Agreement measure due to Katz & Powell (1953, Psychometrika), see
    ## also Messatfa (1992, Journal of Classification).

    n <- n_of_objects(x)    

    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)
    x <- table(cl_class_ids(x), cl_class_ids(y))

    A_xy <- sum(x ^ 2)
    A_x. <- sum(rowSums(x) ^ 2)
    A_.y <- sum(colSums(x) ^ 2)

    (n^2 * A_xy - A_x. * A_.y) /
        sqrt(A_x. * (n^2 - A_x.) * A_.y * (n^2 - A_.y))
}

### ** .cl_agreement_partition_angle

.cl_agreement_partition_angle <-
function(x, y)
{
    ## Maximal angle between the matched memberships.

    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    ## Match classes from conforming memberships.
    ind <- solve_LSAP(crossprod(M_x, M_y), maximum = TRUE)
    sum(M_x * M_y[, ind]) / sqrt(sum(M_x ^ 2) * sum(M_y ^ 2))
}

### ** .cl_agreement_partition_diag

.cl_agreement_partition_diag <-
function(x, y)
{
    ## Maximal co-classification rate.

    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    ## Match classes from conforming memberships.
    ind <- solve_LSAP(crossprod(M_x, M_y), maximum = TRUE)
    sum(M_x * M_y[, ind]) / n_of_objects(x)
}

### ** .cl_agreement_partition_FM

.cl_agreement_partition_FM <-
function(x, y)
{
    ## Fowlkes-Mallows index.

    n <- n_of_objects(x)    

    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)
    x <- table(cl_class_ids(x), cl_class_ids(y))

    (sum(x ^ 2) - n) /
        sqrt((sum(rowSums(x) ^ 2) - n) * (sum(colSums(x) ^ 2) - n))
    
}

### ** .cl_agreement_partition_Jaccard

.cl_agreement_partition_Jaccard <-
function(x, y)
{
    ## Jaccard index.

    n <- n_of_objects(x)    

    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)
    x <- table(cl_class_ids(x), cl_class_ids(y))

    Z <- sum(x ^ 2)

    (Z - n) / (sum(rowSums(x) ^ 2) + sum(colSums(x) ^ 2) - n - Z)
    
}

### ** .cl_agreement_partition_purity

.cl_agreement_partition_purity <-
function(x, y)
{
    ## Purity of classes of x with respect to those of y: relative
    ## fraction of "optimally matched and collapsed" joint class
    ## frequencies, i.e., \sum_i \max_j c_{ij} / n.

    n <- n_of_objects(x)    

    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)
    x <- table(cl_class_ids(x), cl_class_ids(y))

    sum(apply(x, 1L, max)) / n
}

.cl_agreement_partition_PS <-
function(x, y)
{
    ## Prediction Strength as used in Tibshirani and Walter (2005),
    ## "Cluster Validation by Prediction Strength", JCGS.

    ## See Eqn 2.1 in the reference: this is
    ##   min_l rate of different objects in the same class in partition
    ##         A and in class l in partition B,
    ## where the min is taken over all classes l of partition B.

    x <- table(cl_class_ids(x), cl_class_ids(y))
    s <- rowSums(x)

    min((rowSums(x ^ 2) - s) / (s * (s - 1)), na.rm = TRUE)
}   

## Some computations useful for interpreting some of the above.
##
## Consider two hard partitions A and B and write
##   a_{ik} ... indicator of object i in class k for partition A
##   b_{il} ... indicator of object i in class l for partition B
## (so that the a_{ik} and b_{il} are of course the membership matrices
## of the partitions).
##
## Then obviously
##   \sum_i a_{ik} b_{il} = m_{kl}
## is the number of objects in class k for A and in class l for B, and
##   \sum_i a_{ik} = m_{k.} = # objects in class k for A
##   \sum_i b_{il} = m_{.l} = # objects in class l for B
##
## Number of pairs of objects in the same classes for both A and B:
##   \sum_{i, j, k, l} a_{ik} a_{jk} b_{il} b_{jl}
##     = \sum_{k, l} \sum_i a_{ik} b_{il} \sum_j a_{jk} b_{jl}
##     = \sum_{k, l} m_{kl} ^ 2
## This includes the n pairs with identical objects, hence:
## Number of distinct pairs of objects in the same classes for both A
## and B:
##   (\sum_{k, l} m_{kl} ^ 2 - n) / 2
##
## Number of pairs of objects in the same class for A:
##   \sum_{i, j, k} a_{ik} a_{jk}
##     = \sum_k \sum_i a_{ik} \sum_j a_{jk}
##     = \sum_k m_{k.} ^ 2
## Again, this includes the n pairs with identical objects, hence:
## Number of distinct pairs of objects in the same class for A:
##   (\sum_k m_{k.} ^ 2 - n) / 2
##
## Similarly, \sum_l m_{.l} ^ 2 corresponds to the number of pairs of
## objects in the same class for B.
##
## Finally, to get the number of pairs of objects in different classes
## for both A and B, we note that this is the total number of pairs,
## minus the sum of the numbers of those in the same class for A and for
## B, respectively, plus the number of pairs in the same class for both
## A and B.
##
## This makes e.g. the interpretation of some of the Fowlkes-Mallows or
## Rand agreement indices rather straightforward.

### ** .cl_agreement_hierarchy_euclidean

.cl_agreement_hierarchy_euclidean <-
function(x, y)
    1 / (1 + .cl_dissimilarity_hierarchy_euclidean(x, y))

### ** .cl_agreement_hierarchy_manhattan

.cl_agreement_hierarchy_manhattan <-
function(x, y)
    1 / (1 + .cl_dissimilarity_hierarchy_manhattan(x, y))

### ** .cl_agreement_hierarchy_cophenetic

.cl_agreement_hierarchy_cophenetic <-
function(x, y)
{
    ## Cophenetic correlation.
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    cor(cl_object_dissimilarities(x), cl_object_dissimilarities(y))
}

### ** .cl_agreement_hierarchy_angle

.cl_agreement_hierarchy_angle <-
function(x, y)
{
    ## Angle between ultrametrics.
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u_x <- cl_object_dissimilarities(x)
    u_y <- cl_object_dissimilarities(y)
    sum(u_x * u_y) / sqrt(sum(u_x ^ 2) * sum(u_y ^ 2))
}

### ** .cl_agreement_hierarchy_gamma

.cl_agreement_hierarchy_gamma <-
function(x, y)
    1 - .cl_dissimilarity_hierarchy_gamma(x, y)
    
### * [.cl_agreement

"[.cl_agreement" <-
function(x, i, j)
{
    y <- NextMethod("[")
    if(!inherits(y, "cl_agreement")) {
        description <- attr(x, "description")
        return(cl_cross_proximity(y, description = description,
                                  class = "cl_cross_agreement"))
    }
    y
}


### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

cl_meet <-
function(x, y)
{
    ## General case.
    ## x either an ensemble, or x and y two clusterings with the same
    ## number of objects.
    if(!inherits(x, "cl_ensemble")) {
        ## Be nice about error messages.
        if(n_of_objects(x) != n_of_objects(y))
            stop("Arguments 'x' and 'y' must have the same number of objects.")
        x <- cl_ensemble(x, y)
    }

    if(inherits(x, "cl_partition_ensemble"))
        .cl_meet_partition(x)
    else if(inherits(x, "cl_dendrogram_ensemble"))
        .cl_meet_dendrogram(x)
    else if(inherits(x, "cl_hierarchy_ensemble"))
        .cl_meet_hierarchy(x)
    else
        stop("Cannot compute meet of given clusterings.")
}

.cl_meet_partition <-
function(x)    
{
    x <- unique(x)
    if(length(x) == 1L)
        return(cl_partition_by_class_ids(cl_class_ids(x[[1L]])))

    ids <- seq_len(n_of_objects(x[[1L]]))
    ## Cross-classify the objects.
    z <- split(ids, lapply(x, cl_class_ids))
    ## Subscript on the non-empty cells to get adjacent class ids.
    lens <- lengths(z)
    pos <- which(lens > 0)
    ids[unlist(z, use.names = FALSE)] <-
        rep.int(seq_along(z[pos]), lens[pos])
    cl_partition_by_class_ids(ids)
}

.cl_meet_dendrogram <-
function(x)
{
    ## Meet of an ensemble of dendrograms.
    ## We need the maximal ultrametric dominated by the given ones,
    ## which can be obtained by hierarchical clustering with single
    ## linkage on the pointwise minima of the ultrametrics.
    as.cl_dendrogram(hclust(as.dist(do.call(pmin,
                                            lapply(x, cl_ultrametric))),
                            "single"))
}

.cl_meet_hierarchy <-
function(x)
{
    ## Meet of an ensemble of n-trees.
    ## Need to find the classes in *all* n-trees.
    ## Equivalent to computing a strict majority tree.
    .cl_consensus_hierarchy_majority(x, rep.int(1, length(x)),
                                     list(p = 1))
}

cl_join <-
function(x, y)
{
    ## General case.
    ## x either an ensemble, or x and y two clusterings with the same
    ## number of objects.
    if(!inherits(x, "cl_ensemble")) {
        ## Be nice about error messages.
        if(n_of_objects(x) != n_of_objects(y))
            stop("Arguments 'x' and 'y' must have the same number of objects.")
        x <- cl_ensemble(x, y)
    }

    if(inherits(x, "cl_partition_ensemble"))
        .cl_join_partition(x)
    else if(inherits(x, "cl_dendrogram_ensemble"))
        .cl_join_dendrogram(x)
    else if(inherits(x, "cl_hierarchy_ensemble"))
        .cl_join_hierarchy(x)
    else
        stop("Cannot compute join of given clusterings.")
}

.cl_join_partition <-
function(x)
{
    x <- unique(x)
    if(length(x) == 1)
        return(cl_partition_by_class_ids(cl_class_ids(x[[1L]])))

    ## Canonicalize: ensure that class ids are always the integers from
    ## one to the number of classes.
    n <- sapply(x, n_of_classes)
    ids <- mapply(function(p, ncp) match(cl_class_ids(p),
                                         seq_len(ncp)),
                  x, n, SIMPLIFY = FALSE)
    ## Order according to the number of classes.
    ids <- ids[order(n)]

    ## And now incrementally build the join.
    jcids <- ids[[1L]]                  # Class ids of the current join.
    jnc <- length(unique(jcids))        # Number of classes of this.
    for(b in seq.int(from = 2, to = length(x))) {
        z <- table(jcids, ids[[b]])
        ## It is faster to work on the smaller partition, but this
        ## should be ensured by the reordering ...
        ## We need to "join all elements in the same class in at least
        ## one of the partitions".  In the matrix
        ##   C <- (tcrossprod(z) > 0)
        ## entry i,j is true/one iff z_{ik} z_{jk} > 0 for classes
        ## i and j in the current join (ids jcids) and some class k in
        ## the partition with ids[[b]], so that i and j must be joined.
        ## I.e., C indicates which classes need to be joined directly.
        ## We need to determine the transitive closure of this relation,
        ## which can be performed by repeating
        ##   C_{t+1} <- ((C_t %*% C) > 0)
        ## with C_1 = C until C_t does not change.
        C_new <- C_old <- C <- (tcrossprod(z) > 0)
        repeat {
            C_new <- (C_old %*% C) > 0
            if(all(C_new == C_old)) break
            C_old <- C_new
        }
        C <- C_new
        ## This should now have the connected components.
        ## Next, compute the map of the join class ids to the ids of
        ## these components. 
        cnt <- 0
        map <- remaining_ids <- seq_len(jnc)
        while(length(remaining_ids)) {
            cnt <- cnt + 1
            pos <- which(C[remaining_ids[1L], remaining_ids] > 0)
            map[remaining_ids[pos]] <- cnt
            remaining_ids <- remaining_ids[-pos]
        }
        ## And update the join:
        jcids <- map[jcids]
        jnc <- cnt
    }

    cl_partition_by_class_ids(jcids)    
}

.cl_join_dendrogram <-
function(x)
{
    ## Join of an ensemble of dendrograms.
    as.cl_dendrogram(do.call(pmax, lapply(x, cl_ultrametric)))
}

.cl_join_hierarchy <-
function(x)
{
    ## Join of an ensemble of n-trees.
    ## Only exists if the union of all classes of the n-trees is itself
    ## an n-tree (see Barthelemy et al).
    classes <- unique(unlist(lapply(x, cl_classes), recursive = FALSE))
    ## Now check if this is an n-tree.
    ## We must verify that for all classes A and B, their intersection
    ## is A, B, or empty.
    check <- function(A, B) {
        m_AB <- match(A, B)
        m_BA <- match(B, A)
        ((all(is.na(m_AB)) && all(is.na(m_BA)))
         || all(is.finite(m_AB))
         || all(is.finite(m_BA)))
    }
    for(i in seq_along(classes)) {
        A <- classes[[i]]
        for(j in seq_along(classes))
            if(!check(A, classes[[j]]))
                stop("Join of given n-trees does not exist.")
    }
    as.cl_hierarchy(.cl_ultrametric_from_classes(classes))
}

### * cl_ultrametric

cl_ultrametric <-
function(x, size = NULL, labels = NULL) 
{
    if(inherits(x, "cl_hierarchy")) {
        ## <NOTE>
        ## Strictly, not every hierarchy corresponds to an ultrametric.
        ## </NOTE>
        return(cl_ultrametric(.get_representation(x),
                              size = size, labels = labels))
    }
    else if(!inherits(x, "cl_ultrametric")) {
        ## Try using cophenetic().
        ## This starts by coercing to hclust, which has methods for all
        ## currently supported hierarchical classification methods.
        ## To support others, either provide as.hclust methods for
        ## these, or make cl_ultrametric() generic and add methods.
        ## Or use the fact that in R >= 2.1.0, stats::cophenetic() is
        ## generic.
        out <- cophenetic(x)
    }
    else {
        out <- x
        if(is.null(labels))
            labels <- attr(x, "Labels")
    }
    .cl_ultrametric_from_veclh(out, labels = labels, size = size)
}

.cl_ultrametric_from_veclh <-
function(x, size = NULL, labels = NULL, meta = NULL)
{
    if(.non_ultrametricity(x) > 0)
        stop("Not a valid ultrametric.")
    u <- cl_proximity(x, "Ultrametric distances",
                      labels = labels, size = size,
                      class = c("cl_ultrametric", "cl_dissimilarity",
                      "cl_proximity", "dist"))
    if(!is.null(meta))
        attr(u, "meta") <- meta
    u
}

### * as.cl_ultrametric

as.cl_ultrametric <-
function(x)
    UseMethod("as.cl_ultrametric")
as.cl_ultrametric.default <-
function(x)
{
    if(inherits(x, "cl_ultrametric"))
        x
    else if(is.atomic(x))
        .cl_ultrametric_from_veclh(x)
    else
        cl_ultrametric(x)
}
as.cl_ultrametric.matrix <-
function(x)
    .cl_ultrametric_from_veclh(x[row(x) > col(x)],
                               labels = rownames(x))

### * as.dendrogram.cl_ultrametric

as.dendrogram.cl_ultrametric <-
function(object, ...)
    as.dendrogram(as.hclust(object), ...)

### * as.hclust.cl_ultrametric

as.hclust.cl_ultrametric <-
function(x, ...)
{
    ## Hierarchical clustering with single linkage gives the minimal
    ## ultrametric dominated by a dissimilarity, see e.g. Bock (1974,
    ## Theorem 39.2).  Hence, hclust(method = "single") on an
    ## ultrametric gives the hclust representation of the associated
    ## dendrogram.
    hclust(x, "single")
}

### * cophenetic.cl_ultrametric

cophenetic.cl_ultrametric <-
function(x)
    as.dist(x)

### * plot.cl_ultrametric

plot.cl_ultrametric <-
function(x, ...)
    plot(as.dendrogram(x), ...)

### * ls_fit_ultrametric

ls_fit_ultrametric <-
function(x, method = c("SUMT", "IP", "IR"), weights = 1,
         control = list())
{
    if(inherits(x, "cl_ultrametric")) {
        return(.cl_ultrametric_with_meta_added(x, list(objval = 0)))
    } else if(is.cl_ensemble(x) || is.list(x)) {
        ## Might be given a list/ensemble of object dissimilarities.
        ## In this case, compute the suitably weighted average and
        ## proceed.
        if(length(x) == 0L)
            stop("Given ensemble contains no dissimilarities.")
        ## Let's be nice as usual ...
        ind <- !vapply(x, .has_object_dissimilarities, NA)
        if(any(ind))
            x[ind] <- lapply(x[ind], as.dist)
        x <- .weighted_mean_of_object_dissimilarities(x,
                                                      control$weights)
    } else if(!inherits(x, "dist"))
        x <- as.dist(x)

    ## Catch some special cases right away.
    if(attr(x, "Size") <= 2L)
        return(.cl_ultrametric_with_meta_added(as.cl_ultrametric(x),
                                               list(objval = 0)))
    if(.non_ultrametricity(x, max = TRUE) == 0)
        return(.cl_ultrametric_with_meta_added(as.cl_ultrametric(x),
                                               list(objval = 0)))

    ## Handle weights.
    ## This is somewhat tricky ...
    if(is.matrix(weights)) {
        weights <- as.dist(weights)
        if(length(weights) != length(x))
            stop("Argument 'weights' must be compatible with 'x'.")
    }
    else
        weights <- rep_len(weights, length(x))
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")

    method <- match.arg(method)
    switch(method,
           SUMT = .ls_fit_ultrametric_by_SUMT(x, weights, control),
           IP = {
               .ls_fit_ultrametric_by_iterative_projection(x, weights,
                                                           control)
           },
           IR = {
               .ls_fit_ultrametric_by_iterative_reduction(x, weights,
                                                          control)
           })
}

### ** .ls_fit_ultrametric_by_SUMT
           
.ls_fit_ultrametric_by_SUMT <-
function(x, weights = 1, control = list())
{
    ## Fit an ultrametric to a dissimilarity by minimizing euclidean
    ## dissimilarity subject to the ultrametric constraint, using the
    ## sequential algorithm of de Soete (1984) with a slight change: we
    ## try to ensure that what we obtain satisfies the constraints
    ## "exactly" rather than approximately.  We (currently?) do that via
    ## rounding ...

    ## <NOTE>
    ## This fits and hence returns an ultrametric, *not* the hierarchy
    ## corresponding to the ultrametric.
    ## </NOTE>

    w <- weights / sum(weights)

    ## Control parameters:
    ## nruns,
    nruns <- control$nruns
    ## start.
    start <- control$start

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else if(is.null(nruns)) {
        ## Use nruns only if start is not given.
        nruns <- 1L
    }

    ## If x is an ultrametric, or satisfies the ultrametricity
    ## constraints, return it.
    if(inherits(x, "cl_ultrametric")
       || (.non_ultrametricity(x, max = TRUE) == 0))
        return(.cl_ultrametric_with_meta_added(as.cl_ultrametric(x),
                                               list(objval = 0)))

    ## For the time being, use a simple minimizer.

    n <- attr(x, "Size")
    labels <- attr(x, "Labels")

    ## Handle missing values in x along the lines of de Soete (1984):
    ## set the corresponding weights to 0, and impute by the weighted
    ## mean.
    ind <- which(is.na(x))
    if(any(ind)) {
        w[ind] <- 0
        x[ind] <- weighted.mean(x, w, na.rm = TRUE)
    }
    
    ## We follow de Soete's notation, and use the veclh's (vector of
    ## lower half, in S the same as x[lower.tri(x)]) of the respective
    ## proximity objects.

    L <- function(d) sum(w * (x - d) ^ 2)
    P <- .make_penalty_function_ultrametric(n)
    grad_L <- function(d) 2 * w * (d - x)
    grad_P <- .make_penalty_gradient_ultrametric(n)

    if(is.null(start)) {
        ## Initialize by "random shaking".  Use sd() for simplicity.
        start <- replicate(nruns,  
                           x + rnorm(length(x), sd = sd(x) / sqrt(3)),
                           simplify = FALSE)
    }
    
    ## And now ...
    out <- sumt(start, L, P, grad_L, grad_P,
                method = control$method, eps = control$eps,
                q = control$q, verbose = control$verbose,
                control = as.list(control$control))

    d <- .ultrametrify(out$x)
    meta <- list(objval = L(d))

    .cl_ultrametric_from_veclh(d, n, labels, meta)
}

.make_penalty_function_ultrametric <-
function(n)
    function(d) {
        ## Smooth penalty function measuring the extent of violation of
        ## the ultrametricity constraint.  Also ensure nonnegativity ...
        (.non_ultrametricity(.symmetric_matrix_from_veclh(d, n))
         + sum(pmin(d, 0) ^ 2))
    }

.make_penalty_gradient_ultrametric <-
function(n)
    function(d) {
        gr <- matrix(.C(C_deviation_from_ultrametricity_gradient,
                        as.double(.symmetric_matrix_from_veclh(d, n)),
                        as.integer(n),
                        gr = double(n * n))$gr,
                     n, n)
        gr[row(gr) > col(gr)] + 2 * sum(pmin(d, 0))
    }

### ** .ls_fit_ultrametric_by_iterative_projection

## <NOTE>
## Functions
##   .ls_fit_ultrametric_by_iterative_projection()
##   .ls_fit_ultrametric_by_iterative_reduction()
## are really identical apart from the name of the C routine they call.
## (But will this necessarily always be the case in the future?)
## Merge maybe ...
## </NOTE>

.ls_fit_ultrametric_by_iterative_projection <-
function(x, weights = 1, control = list())
{
    if(any(diff(weights) != 0))
        warning("Non-identical weights currently not supported.")

    labels <- attr(x, "Labels")
    n <- attr(x, "Size")

    x <- as.matrix(x)

    ## Control parameters:
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 10000L
    ## nruns,
    nruns <- control$nruns
    ## order,
    order <- control$order
    ## tol,
    tol <- control$tol
    if(is.null(tol))
        tol <- 1e-8
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle order and nruns.
    if(!is.null(order)) {
        if(!is.list(order))
            order <- as.list(order)
        if(!all(vapply(order,
                       function(o) all(sort(o) == seq_len(n)),
                       NA)))
            stop("All given orders must be valid permutations.")
    }
    else {
        if(is.null(nruns))
            nruns <- 1L
        order <- replicate(nruns, sample(n), simplify = FALSE)
    }

    ## <NOTE>
    ## Adjust in case support for non-identical weights is added.
    L <- function(d) sum((x - d) ^ 2)
    ## </NOTE>

    d_opt <- NULL
    v_opt <- Inf
    for(run in seq_along(order)) {
        if(verbose)
            message(gettextf("Iterative projection run: %d", run))
        d <- .C(C_ls_fit_ultrametric_by_iterative_projection,
                as.double(x),
                as.integer(n),
                as.integer(order[[run]] - 1L),
                as.integer(maxiter),
                iter = integer(1L),
                as.double(tol),
                as.logical(verbose))[[1L]]
        v <- L(d)
        if(v < v_opt) {
            v_opt <- v
            d_opt <- d
        }
    }
        
    d <- .ultrametrify(as.dist(matrix(d_opt, n)))
    meta <- list(objval = L(d))
    
    .cl_ultrametric_from_veclh(d, n, labels, meta)
}

### ** .ls_fit_ultrametric_by_iterative_reduction

.ls_fit_ultrametric_by_iterative_reduction <-
function(x, weights = 1, control = list())
{
    if(any(diff(weights) != 0))
        warning("Non-identical weights currently not supported.")

    labels <- attr(x, "Labels")
    n <- attr(x, "Size")

    x <- as.matrix(x)

    ## Control parameters:
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 10000L
    ## nruns,
    nruns <- control$nruns
    ## order,
    order <- control$order
    ## tol,
    tol <- control$tol
    if(is.null(tol))
        tol <- 1e-8
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle order and nruns.
    if(!is.null(order)) {
        if(!is.list(order))
            order <- as.list(order)
        if(!all(vapply(order,
                       function(o) all(sort(o) == seq_len(n)),
                       NA)))
            stop("All given orders must be valid permutations.")
    }
    else {
        if(is.null(nruns))
            nruns <- 1L
        order <- replicate(nruns, sample(n), simplify = FALSE)
    }

    ## <NOTE>
    ## Adjust in case support for non-identical weights is added.
    L <- function(d) sum((x - d) ^ 2)
    ## </NOTE>

    d_opt <- NULL
    v_opt <- Inf
    for(run in seq_along(order)) {
        if(verbose)
            message(gettextf("Iterative reduction run: %d", run))
        d <- .C(C_ls_fit_ultrametric_by_iterative_reduction,
                as.double(x),
                as.integer(n),
                as.integer(order[[run]] - 1L),
                as.integer(maxiter),
                iter = integer(1L),
                as.double(tol),
                as.logical(verbose))[[1L]]
        v <- L(d)
        if(v < v_opt) {
            v_opt <- v
            d_opt <- d
        }
    }

    d <- .ultrametrify(as.dist(matrix(d_opt, n)))
    meta <- list(objval = L(d))
    
    .cl_ultrametric_from_veclh(d, n, labels, meta)
}

### * Ultrametric Target Fitters.

### ** ls_fit_ultrametric_target

ls_fit_ultrametric_target <-
function(x, y, weights = 1)
{
    fitter <- if(identical(weights, 1)) # Default.
        function(x, w) mean(x)
    else
        function(x, w) weighted.mean(x, w)
    distfun <- function(x, u, w) sqrt(sum(w * (x - u) ^ 2))
    .fit_ultrametric_target(x, y, weights, fitter, distfun)
}

### ** l1_fit_ultrametric_target

l1_fit_ultrametric_target <-
function(x, y, weights = 1)
{
    fitter <- if(identical(weights, 1)) # Default.
        function(x, w) median(x)
    else
        function(x, w) weighted_median(x, w)
    distfun <- function(x, u, w) sum(w * abs(x - u))
    .fit_ultrametric_target(x, y, weights, fitter, distfun)
}

### ** .fit_ultrametric_target    

.fit_ultrametric_target <-
function(x, y, w, fitter, distfun = NULL)
{
    w <- .handle_weights_for_ultrametric_target_fitters(w, x)
    ## The documentation says that x should inherit from dist, so coerce
    ## to this if needed but if not a matrix (as we will coerce back to
    ## a matrix right away).
    if(!inherits(x, "dist") && !is.matrix(x))
        x <- as.dist(x)
    x <- as.matrix(x)
    y <- as.hclust(y)
    n <- length(y$order)
    ilist <- vector("list", n)
    out <- matrix(0, n, n)
    mat <- xlist <- wlist <- vector("list", n - 1L)
    for(i in seq_len(n - 1L)) {
        inds <- y$merge[i, ]
        ids1 <- if(inds[1L] < 0) -inds[1L] else ilist[[inds[1L]]]
        ids2 <- if(inds[2L] < 0) -inds[2L] else ilist[[inds[2L]]]
        ilist[[i]] <- c(ids1, ids2)
        mat[[i]] <- cbind(rep.int(ids1,
                                  rep.int(length(ids2), length(ids1))), 
                          rep.int(ids2, length(ids1)))
        xlist[[i]] <- x[mat[[i]]]
        wlist[[i]] <- w[mat[[i]]]
    }
    values <- pava(xlist, wlist, fitter)
    for(i in seq_len(n - 1L))
        out[mat[[i]]] <- values[i]
    rownames(out) <- y$labels
    u <- as.cl_ultrametric(out + t(out))
    if(!is.null(distfun))
        attr(u, "meta") <-
            list(objval = distfun(as.dist(x), u, as.dist(w)))
    u
}

### ** .handle_weights_for_ultrametric_target_fitters

.handle_weights_for_ultrametric_target_fitters <-
function(weights, x)
{
    ## Handle weights for the ultrametric target fitters.
    ## This is somewhat tricky ...
    if(is.matrix(weights)) {
        if(any(dim(weights) != attr(x, "Size")))
            stop("Argument 'weights' must be compatible with 'x'.")
    }
    else
        weights <-
            as.matrix(.dist_from_vector(rep_len(weights, length(x))))
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")
    weights
}

### l1_fit_ultrametric

l1_fit_ultrametric <-
function(x, method = c("SUMT", "IRIP"), weights = 1,
         control = list())
{
    if(inherits(x, "cl_ultrametric"))
        return(.cl_ultrametric_with_meta_added(x, list(objval = 0)))

    if(!inherits(x, "dist"))
        x <- as.dist(x)

    ## Catch some special cases right away.
    if(attr(x, "Size") <= 2L)
        return(.cl_ultrametric_with_meta_added(as.cl_ultrametric(x),
                                               list(objval = 0)))
    if(.non_ultrametricity(x, max = TRUE) == 0)
        return(.cl_ultrametric_with_meta_added(as.cl_ultrametric(x),
                                               list(objval = 0)))

    ## Handle weights.
    ## This is somewhat tricky ...
    if(is.matrix(weights)) {
        weights <- as.dist(weights)
        if(length(weights) != length(x))
            stop("Argument 'weights' must be compatible with 'x'.")
    }
    else
        weights <- rep_len(weights, length(x))
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")

    method <- match.arg(method)
    switch(method,
           SUMT = .l1_fit_ultrametric_by_SUMT(x, weights, control),
           IRIP = .l1_fit_ultrametric_by_IRIP(x, weights, control))
}

### ** .l1_fit_ultrametric_by_SUMT

.l1_fit_ultrametric_by_SUMT <-
function(x, weights = 1, control = list())
{
    ## Try a SUMT with "pseudo-gradients".

    w <- weights / sum(weights)

    ## Control parameters:
    ## gradient,
    gradient <- control$gradient
    if(is.null(gradient))
        gradient <- TRUE
    ## nruns,
    nruns <- control$nruns
    ## start.
    start <- control$start

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else if(is.null(nruns)) {
        ## Use nruns only if start is not given.
        nruns <- 1L
    }

    ## For the time being, use a simple minimizer.

    n <- attr(x, "Size")
    labels <- attr(x, "Labels")

    L <- function(d) sum(w * abs(d - x))
    P <- .make_penalty_function_ultrametric(n)
    if(gradient) {
        grad_L <- function(d) w * sign(d - x)
        grad_P <- .make_penalty_gradient_ultrametric(n)
    } else
        grad_L <- grad_P <- NULL

    if(is.null(start)) {
        ## Initialize by "random shaking".  Use sd() for simplicity.
        start <- replicate(nruns,  
                           x + rnorm(length(x), sd = sd(x) / sqrt(3)),
                           simplify = FALSE)
    }

    ## And now ...
    out <- sumt(start, L, P, grad_L, grad_P,
                method = control$method, eps = control$eps,
                q = control$q, verbose = control$verbose,
                control = as.list(control$control))

    d <- .ultrametrify(out$x)
    meta <- list(objval = L(d))

    .cl_ultrametric_from_veclh(d, n, labels, meta)
}

### ** .l1_fit_ultrametric_by_IRIP

.l1_fit_ultrametric_by_IRIP <-
function(x, weights = 1, control = list())
{
    ## An attempt of implementing "Iteratively Reweighted Iterative
    ## Projection" as described in Smith (2000, 2001), Journal of
    ## Classification.  Note that this suggests using the Iterative
    ## Projection of Hubert and Arabie (1995), which we cannot as we
    ## have not (yet?) implemented this for the weighted case.  Hence,
    ## we use our SUMT least squares ultrametric fitter instead.
    ##
    ## However, we never got this to converge properly ...
    
    w <- weights / sum(weights)
    
    ## Control parameters:
    ## MIN,
    MIN <- control$MIN
    if(is.null(MIN))
        MIN <- 1e-3
    ## (A rather small cut-off which worked best in the cases we tried.)
    ## eps,
    eps <- control$eps
    if(is.null(eps))
        eps <- 1e-6
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 100L
    ## reltol,
    reltol <- control$reltol
    if(is.null(reltol))
        reltol <- 1e-6
    ## start,
    start <- control$start
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    n <- attr(x, "Size")
    labels <- attr(x, "Labels")

    L <- function(d) sum(w * abs(x - d))

    ## Initialize by "random shaking" as for the L2 SUMT, but perhaps we
    ## should not do this?  [Or do it differently?]
    u <- if(is.null(start))
        x + rnorm(length(x), sd = sd(x) / 3)
    else
        start
    ## (No multiple runs for the time being.)
    L_new <- L(u)

    iter <- 1L
    while(iter <= maxiter) {
        if(verbose)
            message(gettextf("Outer iteration: %d", iter))
        L_old <- L_new
        u_old <- u
        weights <- w / pmax(abs(u - x), MIN)
        u <- .ls_fit_ultrametric_by_SUMT(x,
                                         weights = weights,
                                         control = as.list(control$control))
        ## Use some control arguments lateron ...
        L_new <- L(u)
        delta_L <- L_old - L_new
        delta_u <- max(abs(u_old - u))
        if(verbose)
            message(gettextf("Change: u: %g L: %g", delta_u, delta_L))
        if((delta_u < eps) ||
           ((delta_L >= 0) &&
            (delta_L <= reltol * (abs(L_old) + reltol))))
            break
        iter <- iter + 1L
    }

    d <- .ultrametrify(u)
    meta <- list(objval = L(d), status = as.integer(iter == maxiter))

    .cl_ultrametric_from_veclh(d, n, labels, meta)
}

## * ls_fit_sum_of_ultrametrics

ls_fit_sum_of_ultrametrics <-
function(x, nterms = 1, weights = 1, control = list())
{
    if(!inherits(x, "dist"))
        x <- as.dist(x)

    ## We could catch some special cases right away: if x already is an
    ## ultrametric then the fit would be a list with x and nterms - 1
    ## zero ultrametrics ...
    
    ## Control parameters:
    ## eps,
    eps <- control$eps
    if(is.null(eps))
        eps <- 1e-6
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 100L
    ## method,
    method <- control$method
    if(is.null(method))
        method <- "SUMT"
    ## reltol,
    reltol <- control$reltol
    if(is.null(reltol))
        reltol <- 1e-6
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")
    ## Do this at last.
    control <- as.list(control$control)
    ## And be nice ...
    if(identical(method, "SUMT") && is.null(control$nruns))
        control$nruns <- 10L

    L <- function(u)
        sum((x - rowSums(matrix(unlist(u), ncol = nterms))) ^ 2)
    
    ## Init.
    u <- rep.int(list(as.cl_ultrametric(0 * x)), nterms)
    L_new <- L(u)

    ## Loop.
    iter <- 1L
    while(iter <= maxiter) {
        if(verbose)
            message(gettextf("Iteration: %d", iter))
        L_old <- L_new
        delta_u <- 0
        for(i in seq_len(nterms)) {
            if(verbose)
                message(gettextf("Term: %d", i))
            u_old <- u[[i]]
            ## Compute residual r = x - \sum_{j: j \ne i} u(j)
            r <- x - rowSums(matrix(unlist(u[-i]), ncol = nterms - 1L))
            ## Fit residual.
            u[[i]] <- ls_fit_ultrametric(r, method, weights, control)
            ## Accumulate change.
            change <- max(abs(u[[i]] - u_old))
            if(verbose)
                message(gettextf("Change: %g", change))
            delta_u <- max(delta_u, change)
        }
        L_new <- L(u)
        delta_L <- L_old - L_new
        if(verbose)
            message(gettextf("Overall change: u: %g L: %g\n",
                             delta_u, delta_L))
        if((delta_u < eps) ||
           ((delta_L >= 0) &&
            (delta_L <= reltol * (abs(L_old) + reltol))))
            break
        iter <- iter + 1L
    }
    
    .structure(u, objval = L_new, status = as.integer(iter == maxiter))
}

### * as.dist.hclust

## Using hclust() with methods 'median' or 'centroid' typically gives
## reversals and hence not valid hierarchies, i.e., distances which do
## not satisfy the ultrametricity conditions.  The distances can be
## obtained via cophenetic(), but ls_fit_ultrametric() prefers using
## as.dist() [as arguably more appropriate] which in turn can be made to
## "work" by providing as.matrix() methods [bypassing the need to handle
## the extra arguments 'diag' and 'upper' for as.dist()].

as.matrix.hclust <-
function(x, ...)
    as.matrix(cophenetic(x))

### * .non_ultrametricity

.non_ultrametricity <-
function(x, max = FALSE)
{
    if(!is.matrix(x))
        x <- .symmetric_matrix_from_veclh(x)
    .C(C_deviation_from_ultrametricity,
       as.double(x),
       as.integer(nrow(x)),
       fn = double(1L),
       as.logical(max))$fn
}

### * .cl_ultrametric_from_classes

.cl_ultrametric_from_classes <-
function(x)
{
    ## Compute an ultrametric from a hierarchy of classes (i.e., an
    ## n-tree).

    labels <- attr(x, "labels")

    ## Ensure we have no duplicates.
    x <- x[!duplicated(x)]
    
    ## .get_classes_in_hierarchy() orders according to cardinality, but
    ## a consensus method may forget to ...
    x[] <- x[order(lengths(x))]

    ## Get the objects (unique codes in the classes).
    objects <- sort(unique(unlist(x)))
    ## (Could also look at the classes of length 1.)

    ## Recursively compute the heights of the classes.
    heights <- double(length = length(x))
    for(i in which(lengths(x) > 1L)) {
        ## Find the relevant classes.
        j <- sapply(x[seq_len(i - 1L)],
                    function(s) all(s %in% x[[i]]))
        heights[i] <- max(heights[j]) + 1
    }

    ## Next, create an incidence matrix (objects by classes).
    incidences <- sapply(x, function(s) objects %in% s)

    ## Now that we have the heights and incidences, we can compute
    ## distances, using the idea that
    ##     distance(i, j) = min(height(A): A contains i and j)
    n <- length(objects)
    d <- matrix(0, n, n)
    for(i in objects)
        d[i, ] <- heights[apply((rep(incidences[i, ], each = n)
                                 & incidences),
                                1L, which.max)]
    dimnames(d) <- rep.int(list(labels), 2L)

    as.cl_ultrametric(d)
}

### * .cl_ultrametric_with_meta_added

.cl_ultrametric_with_meta_added <-
function(x, meta = NULL)
{
    ## An alternative to adding a 'meta' argument to cl_ultrametric().
    attr(x, "meta") <- meta
    x
}

### .ultrametrify

.ultrametrify <-
function(x)
{
    ## Ensure ultrametricity.
    ## In earlier versions, function
    ## .cl_ultrametric_from_ultrametric_approximation() tried rounding
    ## to non-ultrametric significance, using
    ##   round(x, floor(abs(log10(.non_ultrametricity(x, max = TRUE)))))
    ## which is nice but does not guarantee ultrametricity (and may
    ## result in poorer approximations than what we use now).
    ## Hence, let us use single linkage hierarchical clustering which
    ## gives the best dominated ultrametric approximation.
    cophenetic(hclust(.dist_from_vector(x), "single"))
}

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### * cl_proximity

cl_proximity <-
function(x, description, class = NULL, labels = NULL,
         self = NULL, size = NULL)
{
    ## Similar to as.dist(), in a way.
    ## Currently, as.dist() is not generic, so we cannot provide a
    ## cl_proximity method for it.  Hence, we have our dissimilarities
    ## and ultrametrics extend dist, and we use capitalized names for
    ## the attributes provided for compatibility with dist (Size and
    ## Labels).
    
    if(inherits(x, "dist")) {
        ## Explicitly deal with dist objects.
        ## Useful in particular because cophenetic() returns them.
        out <- x
        if(is.null(size))
            size <- attr(x, "Size")
        if(is.null(labels))
            labels <- attr(x, "Labels")
    }
    else if(inherits(x, "cl_proximity") ||
            !(is.matrix(x) && (nrow(x) == ncol(x))))
        out <- x
    else {
        ## Actually, x should really be a square symmetric matrix.
        ## The "self-proximities" in the main diagonal must be stored
        ## provided there is one non-zero entry.
        self <- diag(x)
        if(all(self == 0)) self <- NULL
        out <- x[row(x) > col(x)]
        if(is.null(labels)) {
            if(!is.null(rownames(x))) 
                labels <- rownames(x)
            else if(!is.null(colnames(x))) 
                labels <- colnames(x)
        }
    }
    if(is.null(size))
        size <- as.integer((sqrt(1 + 8 * length(out)) + 1) / 2)
    attributes(out) <-
        list(Size = size, Labels = labels, description = description,
             self = self)
    class(out) <- unique(c(class, "cl_proximity"))
    out
}

### * names.cl_proximity

names.cl_proximity <-
function(x)
    NULL

### * print.cl_proximity

print.cl_proximity <-
function(x, ...)
{
    description <- attr(x, "description")
    if(length(description) > 0L) {
        ## Could make this generic ...
        kind <- if(inherits(x, "cl_dissimilarity"))
            "Dissimilarities"
        else if(inherits(x, "cl_agreement"))
            "Agreements"
        else
            "Proximities"
        cat(sprintf("%s using %s", kind, description), ":\n", sep = "")
    }
    m <- format(as.matrix(x))
    if(is.null(self <- attr(x, "self")))
        m[row(m) <= col(m)] <- ""
    else
        m[row(m) < col(m)] <- ""
    print(if(is.null(self)) m[-1, -attr(x, "Size")] else m,
          quote = FALSE, right = TRUE, ...)
    invisible(x)
}

### * as.matrix.cl_proximity

as.matrix.cl_proximity <-
function(x, ...)
{
    size <- attr(x, "Size")
    m <- matrix(0, size, size)
    m[row(m) > col(m)] <- x
    m <- m + t(m)
    if(!is.null(self <- attr(x, "self"))) {
        diag(m) <- self
    }
    ## <NOTE>
    ## stats:::as.matrix.dist() provides default dimnames
    ## (seq_len(size)) if no labels are available.
    ## We used to do this too, but ...
    if(!is.null(labels <- attr(x, "Labels")))
        dimnames(m) <- list(labels, labels)
    ## </NOTE>
    m
}

### * [.cl_proximity

"[.cl_proximity" <-
function(x, i, j)
{
    ## Subscripting proximity objects.
    ## Basically matrix-like, but proximity objects are always
    ## "matrices", hence no 'drop' argument.
    ## For double-index subscripting, if i and j are identical,
    ## structure and class are preserved.  Otherwise, a cross-proximity
    ## object is returned (and methods for classes inheriting from
    ## proximity need to readjust the class info as needed).
    ## For single-index subscripting, no attempty is currently made at
    ## preserving structure and class where possible.  (We might also
    ## change this to select objects, i.e., the same rows and columns.)
    size <- attr(x, "Size")
    if(missing(j)) {
        if(missing(i))
            return(x)
        else
            j <- seq_len(size)
    }
    if(missing(i))
        i <- seq_len(size)
    description <- attr(x, "description")
    ## RG's graph:::[.dist avoids as.matrix() in noting that for dist
    ## objects, entry (i,j) is at n(i-1) - i(i-1)/2 + j - i (in the
    ## veclh dist representation).  We could do something similar, but
    ## note that not all proximities have zero diagonals (i.e., NULL
    ## "self" attributes).
    y <- as.matrix(x)[i, j, drop = FALSE]
    if(identical(i, j)) {
        ## Testing using identical() is rather defensive ...
        return(cl_proximity(y, description = description,
                            class = class(x)))
    }
    cl_cross_proximity(y, description = description)
}

### * cl_cross_proximity

cl_cross_proximity <-
function(x, description = NULL, class = NULL)
{
    attr(x, "description") <- description
    class(x) <- c(class, "cl_cross_proximity")
    x
}

### * print.cl_cross_proximity

print.cl_cross_proximity <-
function(x, ...)
{
    description <- attr(x, "description")
    if(length(description) > 0L) {
        ## Could make this generic ...
        kind <- if(inherits(x, "cl_cross_dissimilarity"))
            "Cross-dissimilarities"
        else if(inherits(x, "cl_cross_agreement"))
            "Cross-agreements"
        else
            "Cross-proximities"
        cat(sprintf("%s using %s", kind, description), ":\n", sep = "")
    }
    print(matrix(as.vector(x), nrow = nrow(x), dimnames = dimnames(x)),
          ...)
    invisible(x)
}

### ** print_description_prefix

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

sumt <-
function(x0, L, P, grad_L = NULL, grad_P = NULL, method = NULL,
         eps = NULL, q = NULL, verbose = NULL, control = list())
{
    ## Default values: make it nice for others to call us.

    if(is.null(eps)) eps <- sqrt(.Machine$double.eps)
    if(is.null(method)) method <- "CG"
    if(is.null(q)) q <- 10
    if(is.null(verbose)) verbose <- getOption("verbose")
    
    Phi <- function(rho, x) L(x) + rho * P(x)

    if(is.null(grad_L) || is.null(grad_P)) {
        make_Phi <- function(rho) { function(x) Phi(rho, x) }
        make_grad_Phi <- function(rho) NULL
    }
    else {
        grad_Phi <- function(rho, x) grad_L(x) + rho * grad_P(x)
        make_Phi <- if(method == "nlm") {
            function(rho) {
                function(x)
                .structure(Phi(rho, x), gradient = grad_Phi(rho, x))
            }
        }
        else
            function(rho) {
                function(x)
                Phi(rho, x)
            }
        make_grad_Phi <- function(rho) { function(x) grad_Phi(rho, x) }
    }

    ## <NOTE>
    ## For the penalized minimization, the Newton-type nlm() may be
    ## computationally infeasible (although it works much faster for
    ## fitting ultrametrics to the Phonemes data).
    ## De Soete recommends using Conjugate Gradients.
    ## We provide a simple choice: by default, optim(method = "CG") is
    ## used.  If method is non-null and not "nlm", we use optim() with
    ## this method.  In both cases, control gives the control parameters
    ## for optim().
    ## If method is "nlm", nlm() is used, in which case control is
    ## ignored.  Note that we call nlm() with checking analyticals
    ## turned off, as in some cases (e.g. when fitting ultrametrics) the
    ## penalty function is not even continuous ...
    optimize_with_penalty <- if(method == "nlm")
        function(rho, x)
            nlm(make_Phi(rho), x, check.analyticals = FALSE) $ estimate
    else {
        function(rho, x)
            optim(x, make_Phi(rho), gr = make_grad_Phi(rho),
                  method = method, control = control) $ par
    }
    ## Note also that currently we do not check whether optimization was
    ## "successful" ...
    ## </NOTE>

    ## We currently require that x0 be a *list* of start values, the
    ## length of which gives the number of SUMT runs.  But as always,
    ## let's be nice to users and developers, just in case ...
    if(!is.list(x0))
        x0 <- list(x0)

    v_opt <- Inf
    x_opt <- NULL
    rho_opt <- NULL
    for(run in seq_along(x0)) {
        if(verbose)
            message(gettextf("SUMT run: %d", run))            
        x <- x0[[run]]
        ## <TODO>
        ## Better upper/lower bounds for rho?
        rho <- max(L(x), 0.00001) / max(P(x), 0.00001)
        ## </TODO>
        if(verbose)
            message(gettextf("Iteration: 0 Rho: %g P: %g", rho, P(x)))
        iter <- 1L
        repeat {
            ## <TODO>
            ## Shouldnt't we also have maxiter, just in case ...?
            ## </TODO>
            if(verbose)
                message(gettextf("Iteration: %d Rho: %g P: %g",
                                 iter, rho, P(x)))
            x_old <- x
            x <- optimize_with_penalty(rho, x)
            if(max(abs(x_old - x)) < eps)
                break
            iter <- iter + 1L
            rho <- q * rho
        }
        v <- Phi(rho, x)
        if(v < v_opt) {
            v_opt <- v
            x_opt <- x
            rho_opt <- rho
        }
        if(verbose)
            message(gettextf("Minimum: %g", v_opt))
    }

    .structure(list(x = x_opt, L = L(x_opt), P = P(x_opt), rho = rho_opt,
                    call = match.call()),
               class = "sumt")
}

### * Matrix/vector utilities

### * .dist_from_vector
        
.dist_from_vector <-
function(x, n = NULL, labels = NULL)
{
    ## This might be useful as as.dist.vector, perhaps without the extra
    ## argument n then which we only have for minimal performance gains.
    if(is.null(n))
        n <- as.integer((sqrt(1 + 8 * length(x)) + 1) / 2)
    attr(x, "Size") <- n
    if(!is.null(labels))
        attr(x, "Labels") <- labels
    class(x) <- "dist"
    x
}

### ** .one_entry_per_column

.one_entry_per_column <-
function(x, j)
{
    ## For a matrix x and a vector of column indices j_1, ..., j_n where
    ## n is the number of rows of x, get x[1,j_1], ..., x[n,j_n].

    ## <NOTE>
    ## This used to have
    ##   if(!is.matrix(x))
    ##     stop("Argument 'x' must be a matrix.")
    ## but that will fail for sparse matrix classes.
    ## So let us hope for the best ...
    ## </NOTE>

    x[cbind(seq_len(nrow(x)), j)]
}

".one_entry_per_column<-" <-
function(x, j, value)
{
    ## <NOTE>
    ## This used to have
    ##   if(!is.matrix(x))
    ##     stop("Argument 'x' must be a matrix.")
    ## but that will fail for sparse matrix classes.
    ## So let us hope for the best ...
    ## </NOTE>

    x[cbind(seq_len(nrow(x)), j)] <- value
    x
}

### * .symmetric_matrix_from_veclh

.symmetric_matrix_from_veclh <-
function(x, n = NULL)
{
    ## In essence the same as as.matrix.dist, but without handling the
    ## additional attributes that dist objects might have.
    if(is.null(n))
        n <- as.integer((sqrt(1 + 8 * length(x)) + 1) / 2)
    M <- matrix(0, n, n)
    M[row(M) > col(M)] <- x
    M + t(M)
}

### * .weighted_mean_of_object_dissimilarities

.weighted_mean_of_object_dissimilarities <-
function(x, w = NULL)
{
    w <- if(is.null(w)) {
        rep.int(1, length(x))
    } else {
        rep_len(w, length(x))
    }
    ## (Need the latter because we want w / sum(w) ...)
    dissimilarities <- lapply(x, cl_object_dissimilarities)
    m <- rowSums(mapply(`*`, dissimilarities, w / sum(w)))
    labels <- attr(dissimilarities[[1L]], "Labels")
    .dist_from_vector(m, labels = labels)
}

### ** .weighted_sum_of_matrices

.weighted_sum_of_matrices <-
function(x, w = NULL, nr = NULL)
{
    ## Quite often we need to compute weighted sums \sum_b w_b X_b of
    ## conforming matrices \{ X_b \}.  If x is a list containing the
    ## matrices and w the vector of weights, it seems that one
    ## reasonably efficient way of doing this is the following.
    if(is.null(w)) w <- rep.int(1, length(x))
    if(is.null(nr)) nr <- NROW(x[[1L]])
    matrix(rowSums(mapply(`*`, x, w)), nr)
}

### ** .weighted_sum_of_vectors

.weighted_sum_of_vectors <-
function(x, w = NULL)
{
    ## See above.
    if(is.null(w)) w <- rep.int(1, length(x))
    rowSums(mapply(`*`, x, w))
}

### * Containers

## Creator.

.make_container <-
function(x, classes, properties = NULL)
{
    out <- list(.Data = x, .Meta = properties)
    class(out) <- unique(classes)
    out
}
          
## Getters.

.get_representation <-
function(x)
    x$.Data
.get_properties <-
function(x)
    x$.Meta
.get_property <-
function(x, which)
    x$.Meta[[which]]
.has_property <-
function(x, which)
    which %in% names(x$.Meta)
.get_property_from_object_or_representation <-
function(x, which, getter)
{
    if(.has_property(x, which))
        .get_property(x, which)
    else {
        if(missing(getter)) getter <- get(which)
        getter(.get_representation(x))
    }
}

## Methods (sort of).

.print_container <-
function(x, cls, ...)
{
    writeLines(gettextf("An object of virtual class '%s', with representation:\n",
                        cls))
    print(.get_representation(x), ...)
    invisible(x)
}
    
### * Others

weighted_median <-
function(x, w = 1, na.rm = FALSE)
{
    w <- rep_len(w, length(x))
    if(na.rm && any(ind <- is.na(x))) {
        x <- x[!ind]
        w <- w[!ind]
    }
    if(any(is.na(x)) || !length(x))
        return(NA)
    w <- w / sum(w)    
    ind <- order(x)
    x <- x[ind]
    w <- w[ind]
    x[which.min(x * (cumsum(w) - 0.5) - cumsum(w * x))]
}


### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### * cl_pclust

cl_pclust <-
function(x, k, method = NULL, m = 1, weights = 1, control = list())
{
    ## Partition a cluster ensemble x into (at most) k classes by
    ## minimizing
    ##   \sum_b \sum_j w_b u_{bj}^m d(x_b, p_j) ^ e
    ## for "suitable" prototypes p_1, ..., p_k, where 1 <= m < \infty,
    ## with 1 corresponding to hard (secondary) partitions, and d a
    ## dissimilarity measure (such as Euclidean dissimilarity of
    ## partitions or hierarchies).
    ##
    ## The algorithm works whenever there is a consensus method for
    ## solving
    ##   \sum_b u_{bj}^m d(x_b, p) ^ e => \min_p
    ##
    ## As we refer to consensus methods by their *name* (e.g., 'HBH'),
    ## we rely on the registration mechanism (set_cl_consensus_method())
    ## to provide the required information about d and e.

    clusterings <- as.cl_ensemble(x)

    type <- .cl_ensemble_type(clusterings)

    if(type == "partition") {
        ## Canonicalize by turning into an ensemble of partitions
        ## represented by membership matrices with the same (minimal)
        ## number of columns.
        memberships <-
            lapply(clusterings, cl_membership,
                   max(sapply(clusterings, n_of_classes)))
        clusterings <-
            cl_ensemble(list = lapply(memberships, as.cl_partition))
    }

    if(!inherits(method, "cl_consensus_method")) {
        ## Get required information on d and e from the registry.
        if(is.null(method))
            method <- .cl_consensus_method_default(type)
        method <- get_cl_consensus_method(method, type)
        ## Note that this avoids registry lookup in subsequent calls to
        ## cl_consensus().
        if(is.null(method$exponent))
            stop("No information on exponent in consensus method used.")
        e <- method$exponent
        if(is.null(method$dissimilarity))
            stop("No information on dissimilarity in consensus method used.")
        d <- function(x, y = NULL)
            cl_dissimilarity(x, y, method = method$dissimilarity)
    }

    family <- pclust_family(D = d, C = method$definition, e = e)
    out <- pclust(x, k, family, m, weights, control)

    ## Massage the results a bit.
    dissimilarities <- as.matrix(d(clusterings) ^ e)
    out$call <- match.call()
    out <- .structure(c(out,
                        list(silhouette =
                             silhouette(out$cluster,
                                        dmatrix = dissimilarities),
                             validity =
                             cl_validity(cl_membership(out),
                                         dissimilarities),
                             ## <NOTE>
                             ## Information about d and e is also in the
                             ## family returned, of course.  Trying to be
                             ## nice to users by "directly" providing d
                             ## and e is currently of limited usefulness
                             ## as the pclust representation is not
                             ## directly available to users.
                             d = d,
                             e = e
                             ## </NOTE>
                             )),
                      class = unique(c("cl_pclust", class(out))))

    as.cl_partition(out)
}

print.cl_pclust <-
function(x, ...)
{
    txt <- if(x$m == 1)
        gettextf("A hard partition of a cluster ensemble with %d elements into %d classes.",
                 n_of_objects(x), n_of_classes(x))
    else
        gettextf("A soft partition (degree m = %f) of a cluster ensemble with %d elements into %d classes.",
                x$m, n_of_objects(x), n_of_classes(x))
    writeLines(strwrap(txt))
    NextMethod("print", x, header = FALSE)
    print(x$validity, ...)
    invisible(x)
}

### * pclust

pclust <-
function(x, k, family, m = 1, weights = 1, control = list())
{
    ## A general purpose alternating optimization algorithm for
    ## prototype-based partitioning.
    
    ## For now, assume family specifies three functions:
    ## * A dissimilarity function D() for data and prototypes.
    ## * A consensus function C() for data, weights and control.
    ## * An init function init() of data and k giving an initial object
    ##   of k prototypes.
    ##
    ## We use k as the second argument as this seems to be common
    ## practice for partitioning algorithms.

    ## <NOTE>
    ## We assume that consensus functions can all handle WEIGHTS
    ## (formals: x, weights, control; only used positionally).
    ## <NOTE>
    
    ## <NOTE>
    ## We now allow for arbitrary representations/objects of prototypes.
    ## What is needed are functions to modify a *single* prototype and
    ## subset the prototypes.  By default, list and matrix (with the
    ## usual convention that rows are "objects") representations are
    ## supported.  Otherwise, the family needs to provide suitable
    ## .modify() and .subset() functions.
    ## The approach relies on having the initializer of the family
    ## (init()) return an appropriate object of prototypes.
    ## It would be possible to have default initializers as well to
    ## randomly subset the data (i.e., select elements of lists or rows
    ## of matrices, respectively).
    ## </NOTE>

    ## <NOTE>
    ## The 'prototypes' are not necessarily objects of the same kind as
    ## the data objects.  Therefore, D() is really a 2-argument
    ## cross-dissimilarity function.
    ## It would also be useful to have a way of computing the pairwise
    ## dissimilarities between objects: but this is something different
    ## from D() is objects and prototypes are not of the same kind.
    ## A "clean" solution could consist in specifying the family either
    ## via a (non-symmetric) cross-dissimilarity function X(), or a
    ## symmetric D() which when called with a single argument gives the
    ## pairwise object dissimilarities.
    ## I.e., 
    ##   pclust_family(D = NULL, C, init = NULL, X = NULL, ......)
    ## using
    ## * If both D and X are not given => TROUBLE.
    ## * If only D is given: use for X as well.
    ## * If only X is given: only use as such.
    ## Something for the future ...
    ## </NOTE>

    ## <NOTE>
    ## If people have code for computing cross-dissimilarities for the
    ## data and a *single* prototype (say, xd()), they can easily wrap
    ## into what is needed using
    ##   t(sapply(prototypes, function(p) xd(x, p)))
    ## Assuming symmetry of the dissimilarity, they could also do
    ##   t(sapply(prototypes, xd, x))
    ## </NOTE>

    ## Perhaps check whether 'family' is a feasible/suitable pclust
    ## family (object).
    D <- family$D
    C <- family$C
    e <- family$e
    .modify <- family$.modify
    .subset <- family$.subset

    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 100L
    nruns <- control$nruns
    reltol <- control$reltol
    if(is.null(reltol))
        reltol <- sqrt(.Machine$double.eps)
    start <- control$start
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")
    ## Do this at last ...
    control <- as.list(control$control)

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
        nruns <- length(start)
    } else {
        if(is.null(nruns)) {
            ## Use nruns only if start is not given.
            nruns <- 1L
        }
        start <- replicate(nruns, family$init(x, k), simplify = FALSE)
    }
    
    ## Initialize.
    ## We need to do this here because it is (currently) the only way to
    ## figure out the number B of objects to be partitioned (which is
    ## needed for getting the object weights to the right length).
    prototypes <- start[[1L]]
    dissimilarities <- D(x, prototypes) ^ e
    B <- NROW(dissimilarities)
    ## Also try to figure out (if necessary) how to modify a single
    ## prototype and to subset the prototypes.  Note that we can only
    ## check this *after* prototypes were obtained (and not when the
    ## family object is created).
    if(is.null(.modify)) {
        if(is.list(prototypes))
            .modify <- function(x, i, value) {
                x[[i]] <- value
                x
            }
        else if(is.matrix(prototypes))
            .modify <- function(x, i, value) {
                x[i, ] <- value
                x
            }
        else stop("Cannot determine how to modify prototypes.")
    } else if(!is.function(.modify) ||
            !identical(formals(args(.modify)), c("x", "i", "value")))
        stop("Invalid function to modify prototypes.")
    if(is.null(.subset)) {
        if(is.list(prototypes))
            .subset <- `[`
        else if(is.matrix(prototypes))                 
            .subset <- function(x, i)
                x[i, , drop = FALSE]
        else stop("Cannot determine how to subset prototypes.")
    } else if(!is.function(.subset) ||
            !identical(formals(args(.subset)), c("x", "i")))
        stop("Invalid function to subset prototypes.")

    weights <- rep_len(weights, B)
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")

    ## A little helper.
    .make_unit_weights <- function(B, i) {
        out <- double(B)
        out[i] <- 1
        out
    }

    if(m == 1) {
        ## Hard partitions.
        value <- if(all(weights == 1))
            function(dissimilarities, ids)
                sum(.one_entry_per_column(dissimilarities, ids))
        else
            function(dissimilarities, ids)
                sum(weights *
                    .one_entry_per_column(dissimilarities, ids))
        opt_value <- Inf
        run <- 1L
        if(verbose && (nruns > 1L))
            message(gettextf("Pclust run: %d", run))
        repeat {
            class_ids <- max.col( - dissimilarities )
            old_value <- value(dissimilarities, class_ids)
            if(verbose)
                message(gettextf("Iteration: 0 *** value: %g", old_value))
            iter <- 1L
            while(iter <= maxiter) {
                class_ids_used <- unique(class_ids)
                for(j in class_ids_used)
                    prototypes <-
                        .modify(prototypes, j, 
                                C(x, weights * (class_ids %in% j),
                                  control))
                dissimilarities <- D(x, prototypes) ^ e
                class_ids <- max.col( - dissimilarities )
                ## Try avoiding degenerate solutions.
                if(length(class_ids_used) < k) {
                    ## Find the k - l largest
                    ## object-to-assigned-prototype dissimilarities.
                    o <- order(.one_entry_per_column(dissimilarities,
                                                     class_ids),
                               decreasing = TRUE)
                    ## Find and recompute unused prototypes.
                    unused <- setdiff(seq_len(k), class_ids_used)
                    for(j in seq_along(unused))
                        prototypes <-
                            .modify(prototypes, unused[j],
                                    C(x, .make_unit_weights(B, o[j]),
                                      control))
                    dissimilarities[, unused] <-
                        D(x, .subset(prototypes, unused)) ^ e
                    class_ids <- max.col( - dissimilarities )
                    ## For the time being, do not retry in case the
                    ## solution is still degenerate.
                }
                new_value <- value(dissimilarities, class_ids)
                if(verbose)
                    message(gettextf("Iteration: %d *** value: %g",
                                     iter, new_value))
                if(abs(old_value - new_value)
                   < reltol * (abs(old_value) + reltol))
                    break
                old_value <- new_value
                iter <- iter + 1L
            }
            if(new_value < opt_value) {
                converged <- (iter <= maxiter)
                opt_value <- new_value
                opt_class_ids <- class_ids
                opt_prototypes <- prototypes
            }
            if(run >= nruns) break
            run <- run + 1L
            if(verbose)
                message(gettextf("Pclust run: %d", run))
            prototypes <- start[[run]]
            dissimilarities <- D(x, prototypes) ^ e
        }
        ## We should really have a suitable "sparse matrix" class for
        ## representing the memberships of hard partitions.  For now:
        opt_u <- NULL
        ## opt_u <- matrix(0, B, k)
        ## opt_u[cbind(seq_len(B), opt_class_ids)] <- 1
    }
    else {
        ## Soft partitions.
        value <- if(all(weights == 1))
            function(dissimilarities, u)
                sum(u ^ m * dissimilarities)
        else
            function(dissimilarities, u)
                sum(weights *
                    u ^ m * dissimilarities)
        opt_value <- Inf
        run <- 1L
        if(verbose && (nruns > 1L))
            message(gettextf("Pclust run: %d", run))
        repeat {
            u <- .memberships_from_cross_dissimilarities(dissimilarities,
                                                         m)
            old_value <- value(dissimilarities, u)
            if(verbose)
                message(gettextf("Iteration: 0 *** value: %g", old_value))
            iter <- 1L
            while(iter <= maxiter) {
                ## Update the prototypes.
                ## This amounts to solving, for each j:
                ##   \sum_b w_b u_{bj}^m D(x_b, p) ^ e => \min_p
                ## I.e., p_j is the *weighted* consensus of the x_b with
                ## corresponding weights u_{bj}^m.
                for(j in seq_len(k)) {
                    prototypes <-
                        .modify(prototypes, j,
                                C(x, weights * u[, j] ^ m, control))
                }
                ## Update u.
                dissimilarities <- D(x, prototypes) ^ e
                u <- .memberships_from_cross_dissimilarities(dissimilarities,
                                                             m)
                new_value <- value(dissimilarities, u)
                if(verbose)
                    message(gettextf("Iteration: %d *** value: %g",
                                     iter, new_value))
                if(abs(old_value - new_value)
                   < reltol * (abs(old_value) + reltol))
                    break
                old_value <- new_value
                iter <- iter + 1L
            }
            if(new_value < opt_value) {
                converged <- (iter <= maxiter)                
                opt_value <- new_value
                opt_prototypes <- prototypes
                opt_u <- u
            }
            if(run >= nruns) break
            run <- run + 1L
            if(verbose)
                message(gettextf("Pclust run: %d", run))
            prototypes <- start[[run]]
            dissimilarities <- D(x, prototypes) ^ e
        }
        opt_class_ids <- max.col(opt_u)
        ## Ensure that opt_u is a stochastic matrix.
        opt_u <- pmax(opt_u, 0)
        opt_u <- opt_u / rowSums(opt_u)
        rownames(opt_u) <- rownames(dissimilarities)
        opt_u <- cl_membership(as.cl_membership(opt_u), k)
    }

    names(opt_class_ids) <- rownames(dissimilarities)

    pclust_object(prototypes = opt_prototypes,
                  membership = opt_u,
                  cluster = opt_class_ids,
                  family = family,
                  m = m,
                  value = opt_value,
                  call = match.call(),
                  attributes = list("converged" = converged))
}

print.pclust <-
function(x, header = TRUE, ...)
{
    is_hard <- (x$m == 1)
    class_ids <- cl_class_ids(x)
    if(header) {
        txt <- if(is_hard)
            gettextf("A hard partition of %d objects into %d classes.",
                     length(class_ids), n_of_classes(x))
        else
            gettextf("A soft partition (degree m = %f) of %d objects into %d classes.",
                     x$m, length(class_ids), n_of_classes(x))
        writeLines(strwrap(txt))
    }
    if(is_hard) {
        print(class_ids, ...)
    }
    else {
        writeLines("Class memberships:")
        print(cl_membership(x), ...)
        writeLines("Class ids of closest hard partition:")
        print(unclass(class_ids), ...)
    }
    invisible(x)
}
    

### * pclust_family

pclust_family <-
function(D, C, init = NULL, description = NULL, e = 1,
         .modify = NULL, .subset = NULL)
{
    ## Add checking formals (lengths) eventually ...
    if(is.null(init)) {
        ## Works for list representations ...
        init <- function(x, k) sample(x, k)
    }
    .structure(list(description = description,
                    D = D, C = C, init = init, e = e,
                    .modify = .modify, .subset = .subset),
               class = "pclust_family")
}

### * pclust_object

pclust_object <-
function(prototypes, membership, cluster, family, m = 1, value, ...,
         classes = NULL, attributes = NULL)
{
    out <- c(list(prototypes = prototypes,
                  membership = membership,
                  cluster = cluster,
                  family = family,
                  m = m,
                  value = value),
             list(...))
    attributes(out) <- c(attributes(out), attributes)
    classes <- unique(as.character(classes))
    class(out) <- c(classes[classes != "pclust"], "pclust")
    out
}

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

## Things which must come first in the package code.

### * Internal utilities.

.false <- function(x) FALSE
.true <- function(x) TRUE

## A fast version of structure().
.structure <-
function(x, ...)
    `attributes<-`(x, c(attributes(x), list(...)))

## A slightly polymorphic function, similar to cluster::silhouette() and
## its methods.

cl_validity <-
function(x, ...)
    UseMethod("cl_validity")

cl_validity.default <-
function(x, d, ...)
{
    ## Note that providing methods for classes "cl_partition" and
    ## "cl_hierarchy" is not good enough ...
    out <- list()
    if(.has_object_memberships(x)) {
        v <- .cl_validity_partition_d_a_f(cl_membership(x),
                                          as.matrix(d))
        out <- list("Dissimilarity accounted for" = v)
    }
    else if(.has_object_dissimilarities(x)) {
        x <- cl_object_dissimilarities(x)
        d <- as.dist(d)
        out <- list("Variance accounted for" =
                    .cl_validity_hierarchy_variance_a_f(x, d),
                    "Deviance accounted for" =
                    .cl_validity_hierarchy_deviance_a_f(x, d))
        ## Consider adding e.g. the Agglomerative Coefficient or
        ## Divisive Coeffcient for more than cluster::agnes() and
        ## cluster::diana(), respectively.
    }
    class(out) <- "cl_validity"
    out
}

## Package cluster: agnes().
cl_validity.agnes <-
function(x, ...)
{
    out <- list("Agglomerative coefficient" = x$ac)
    ## According to the docs, agnes objects always have a diss
    ## component, but let's be defensive ...
    if(!is.null(d <- x$diss))
        out <- c(out, cl_validity.default(x, d))
    class(out) <- "cl_validity"
    out
}
## Package cluster: diana().
cl_validity.diana <-
function(x, ...)
{
    out <- list("Divisive coefficient" = x$dc)
    ## According to the docs, diana objects always have a diss
    ## component, but let's be defensive ...
    if(!is.null(d <- x$diss))
        out <- c(out, cl_validity.default(x, d))
    class(out) <- "cl_validity"
    out
}

## Package clue: (virtual) class "cl_partition".
cl_validity.cl_partition <-
function(x, ...)
    cl_validity(.get_representation(x), ...)
## Package clue: class pclust.
## So that this works for all classes extending pclust ...
cl_validity.pclust <-
function(x, ...)
    x$validity

print.cl_validity <-
function(x, ...)
{
    for(nm in names(x))
        cat(nm, ": ", x[[nm]], "\n", sep = "")
    invisible(x)
}

.cl_validity_partition_d_a_f <-
function(m, d)
{
    ## "Dissimilarity accounted for".
    ## Internal function for computing 1 - a / mean(d), where the
    ## "average within dissimilarity" a is given by
    ##   \frac{\sum_{i,j} \sum_k m_{ik}m_{jk} d(i,j)}
    ##        {\sum_{i,j} \sum_k m_{ik}m_{jk}}
    ## where m is the membership matrix and d a *symmetric* matrix of
    ## dissimilarities.
    within_sums <-
        rowSums(sapply(seq_len(ncol(m)),
                       function(k) {
                           z <- m[, k]
                           w <- outer(z, z)
                           c(sum(w * d), sum(w))
                       }))
    average_within_d <- within_sums[1L] / within_sums[2L]
    1 - average_within_d / mean(d)
}

.cl_validity_hierarchy_variance_a_f <-
function(u, d)
{
    ## *Variance accounted for*.
    ## See e.g. Hubert, Arabie, & Meulman (2006), The structural
    ## representation of proximity matrices with MATLAB:
    ## variance_accounted_for = 
    ##   1 - \frac{\sum_{i < j} (d_{ij} - u_{ij}) ^ 2}
    ##            {\sum_{i < j} (d_{ij} - mean(d)) ^ 2}
    ## As this can be arbitrarily negative, we cut at 0.
    
    max(1 - sum((d - u) ^ 2) / sum((d - mean(d)) ^ 2), 0)
}

.cl_validity_hierarchy_deviance_a_f <-
function(u, d)
{
    ## *Deviance accounted for* (i.e., absolute deviation).
    ## See e.g. Smith (2001), Constructing ultrametric and additive
    ## trees based on the ${L}_1$ norm, Journal of Classification.
    ## deviance_accounted_for = 
    ##   1 - \frac{\sum_{i < j} |d_{ij} - u_{ij}|}
    ##            {\sum_{i < j} |d_{ij} - median(d)|}
    ## As this can be arbitrarily negative, we cut at 0.

    max(1 - sum(abs(d - u)) / sum(abs(d - median(d))), 0)
}

## Silhouette methods

silhouette.cl_partition <-
function(x, ...)
    silhouette(.get_representation(x), ...)

silhouette.cl_pclust <-
function(x, ...)
    x$silhouette

cl_boot <-
function(x, B, k = NULL, 
         algorithm = if(is.null(k)) "hclust" else "kmeans",
         parameters = list(), resample = FALSE)
{
    clusterings <- if(!resample) {
        x <- rep.int(list(x), B)        
        eval(as.call(c(list(as.name("lapply"), x, algorithm),
                       if(!is.null(k)) list(k),
                       parameters)))
    }
    else {
        replicate(B,
                  expr = {
                      algorithm <- match.fun(algorithm)
                      ## <NOTE>
                      ## This is not quite perfect.  We have
                      ## cl_predict() to encapsulate the process of
                      ## assigning objects to classes, but for sampling
                      ## from the objects we assume that they correspond
                      ## to the *rows* of 'x'.  Argh.
                      ## </NOTE>
                      ind <- sample(NROW(x), replace = TRUE)
                      train <- if(length(dim(x)) == 2) x[ind, ] else x[ind]
                      out <- eval(as.call(c(list(algorithm, train),
                                            if(!is.null(k)) list(k),
                                            parameters)))
                      as.cl_partition(cl_predict(out, x, "memberships"))
                  },
                  simplify = FALSE)
    }

    cl_ensemble(list = clusterings)
}

## A Pool Adjacent Violators Algorithm framework for minimizing problems
## like
##
##   \sum_i \sum_{J_i} w_{ij} f(y_{ij}, m_i)
##
## under the constraint m_1 <= ... <= m_n with f a convex function in m.
## Note that this formulation allows for repeated data in each block,
## and hence is more general than the usual pava/isoreg ones.  A solver
## for the unconstrained \sum_k w_k f(y_k, m) => min! is needed.
## Typical cases are f(y, m) = |y - m|^p for p = 2 (solved by weighted
## mean) and p = 1 (solved by weighted median), respectively.

## A general design issue is whether weights should be supported or not,
## because in the latter case the solver could be a function of a single
## (data) argument only.  Let's assume the former for the time being.

pava <-
function(x, w = NULL, solver = weighted.mean, merger = c)
{
    n <- length(x)
    if(is.null(w)) {
        w <- if(is.list(x))
            lapply(lengths(x), function(u) rep.int(1, u))
        else
            rep.int(1, n)
    } else if(is.list(x)) 
        w <- as.list(w)
    inds <- as.list(seq_len(n))    
    vals <- mapply(solver, x, w)

    ## Combine blocks i and i + 1.    
    combine <- if(is.list(x)) {
        ## In the repeated data case, we explicitly merge the data (and
        ## weight) lists.
        function(i) {
            ## Merge the data and indices, solve, and put things back
            ## into position i, dropping position i + 1.
            j <- i + 1L
            x[[i]] <<- merger(x[[i]], x[[j]])
            w[[i]] <<- c(w[[i]], w[[j]])
            vals[i] <<- solver(x[[i]], w[[i]])
            inds[[i]] <<- c(inds[[i]], inds[[j]])
            keep <- seq_len(n)[-j]
            x <<- x[keep]
            w <<- w[keep]
            vals <<- vals[keep]
            inds <<- inds[keep]
            n <<- n - 1L
        }
    } else {
        function(i) {
            ## In the "simple" case, merge only indices and values.
            j <- i + 1L
            inds[[i]] <<- c(inds[[i]], inds[[j]])
            vals[i] <<- solver(x[inds[[i]]], w[inds[[i]]])
            keep <- seq_len(n)[-j]
            vals <<- vals[keep]
            inds <<- inds[keep]
            n <<- n - 1L
        }
    }
        
    i <- 1L
    repeat {
        if(i < n) {
            if((vals[i] > vals[i + 1])) {
                combine(i)
                while((i > 1L) && (vals[i - 1L] > vals[i])) {
                    combine(i - 1L)
                    i <- i - 1L
                }
            }
            else
                i <- i + 1L
            }
        else break
    }
    rep.int(vals, lengths(inds))
}


    

cl_bag <-
function(x, B, k = NULL,
         algorithm = "kmeans", parameters = NULL,
         method = "DFBC1", control = NULL)
{
    ## Currently, method 'DFBC1' (Dudoit-Fridlyand BagClust1) is the
    ## only one available, and argument 'control' is ignored.
    
    ## Construct reference partition.
    algorithm <- match.fun(algorithm)
    reference <- eval(as.call(c(list(algorithm, x),
                                if(!is.null(k)) list(k),
                                parameters)))

    ## Construct bootstrap ensemble.
    clusterings <-
        cl_boot(x, B, k, algorithm, parameters, resample = TRUE)

    ## Construct Dudoit-Fridlyand BagClust1 consensus partitions,
    ## suitably generalized ...
    ## <NOTE>
    ## In principle, this could be turned into a "constructive" method
    ## for cl_consensus(), also allowing for weights (straightforward).
    ## E.g.,
    ##    .cl_consensus_partition_DFBC1(clusterings, weights, control)
    ## where either 'control specifies a reference partition, or the
    ## first element of 'clusterings' is taken as such.
    ## </NOTE>
    k <- max(sapply(c(clusterings, reference), n_of_classes))
    M_ref <- cl_membership(reference, k)
    M <- matrix(0, NROW(M_ref), k)
    for(b in seq_len(B)) {
        mem <- cl_membership(clusterings[[b]], k)
        ## Match classes to reference partition.
        ind <- solve_LSAP(crossprod(M_ref, mem), maximum = TRUE)
        M <- M + mem[, ind]
    }
    as.cl_partition(cl_membership(as.cl_membership(M / B), k))
}

cl_fuzziness <-
function(x, method = NULL, normalize = TRUE)
{
    x <- as.cl_ensemble(x)
    out <- double(length(x))
    ## <FIXME>
    ## The docs say that we should only have partitions ...
    attr(out, "description") <- "Fuzziness"
    class(out) <- "cl_fuzziness"
    parties <- vapply(x, is.cl_partition, NA)
    if(!(length(x) || any(parties))) {
        ## Currently, no fuzzy hierarchies ...
        return(out)
    }
    ## </FIXME>

    if(!is.function(method)) {
        builtin_methods <- c("PC", "PE")
        builtin_method_names <-
            c("partition coefficient", "partition entropy")
        if(is.null(method))
            ind <- 1
        else if(is.na(ind <- pmatch(tolower(method),
                                    tolower(builtin_methods))))
            stop(gettextf("Value '%s' is not a valid abbreviation for a fuzziness method.",
                          method),
                 domain = NA)
        method <- paste0(".cl_fuzziness_partition_", builtin_methods[ind])
        method_name <- builtin_method_names[ind]
        if(normalize)
            method_name <- paste("normalized", method_name)
    }
    else
        method_name <- "user-defined method"
    
    out[parties] <- as.numeric(sapply(x[parties], method, normalize))
    attr(out, "description") <- paste("Fuzziness using", method_name)
    out 
}

.cl_fuzziness_partition_PC <-
function(x, normalize = TRUE)
{
    ## Dunn's Partition Coefficient, see also ?fanny.
    ## Note that we normalize differently ...
    if(!.maybe_is_proper_soft_partition(x) && is.cl_hard_partition(x))
        return(1 - normalize)
    pc <- sum(cl_membership(x) ^ 2) / n_of_objects(x)
    if(normalize)
        pc <- (1 - pc) / (1 - 1 / n_of_classes(x))
    pc
}

.cl_fuzziness_partition_PE <-
function(x, normalize = TRUE)
{
    ## Bezdek's Partition Entropy.
    ## Note that we normalize differently ...
    if(!.maybe_is_proper_soft_partition(x) && is.cl_hard_partition(x))
        return(0)
    M <- cl_membership(x)
    pe <- - sum(ifelse(M > 0, M * log(M), 0)) / n_of_objects(x)
    if(normalize)
        pe <- pe / log(n_of_classes(x))
    pe
}

print.cl_fuzziness <-
function(x, ...)
{
    cat(attr(x, "description"), ":\n", sep = "")
    print(as.vector(x), ...)
    invisible(x)
}

### * is.cl_hierarchy

## Determine whether an object is a hierarchy.
## Note that hierarchies are n-trees, which can naturally be represented
## by their classes (as done via cl_classes()) or internal ultrametric
## obtained by assigning height one to all splits (as done by
## .cl_ultrametric_from_classes()). 
## We typically used the latter, but note that this is an *internal*
## reprsentation.
## User-level, cl_dendrogram objects are indexed hierarchies, and
## cl_hierarchy objects are n-trees.  The latter can be "converted" into
## the former (using height one splits) via as.cl_dendrogram().

is.cl_hierarchy <-
function(x)
    UseMethod("is.cl_hierarchy")

## Default method.
is.cl_hierarchy.default <- .false

## Package stats: hclust().
is.cl_hierarchy.hclust <-
function(x)
    !is.unsorted(x$height)

## Package cluster: agnes() and diana() give objects inheriting from
## class "twins".
is.cl_hierarchy.twins <- .true
## Package cluster: mona().
is.cl_hierarchy.mona <- .true

## Package ape: class "phylo".
is.cl_hierarchy.phylo <-
function(x)
    ape::is.ultrametric(x)

## Package clue: (virtual) class "cl_hierarchy".
## Note that "raw" cl_ultrametric objects are *not* hierarchies, as
## these are meant for numeric computations.
## <FIXME>
## Is this really a good idea?
## We can as.hclust() a cl_dendrogram and then it is a cl_hierarchy ...
## </FIXME>
is.cl_hierarchy.cl_hierarchy <- .true

### * as.cl_hierarchy

## Note that cl_hierarchy conceptually is a virtual class, so there are
## no prototypes and no cl_hierarchy() creator.

.cl_hierarchy_classes <- "cl_hierarchy"

as.cl_hierarchy <-
function(x)
{
    if(is.cl_hierarchy(x)) {
        if(!inherits(x, "cl_hierarchy"))
            .make_container(x, .cl_hierarchy_classes)
        else
            x
    }
    else
        .make_container(as.cl_ultrametric(x),
                        .cl_hierarchy_classes)
}

### * print.cl_hierarchy

print.cl_hierarchy <-
function(x, ...)
    .print_container(x, "cl_hierarchy", ...)

### * Complex.cl_hierarchy

## No Complex() for any kind of hierarchy.
Complex.cl_hierarchy <-
function(z)
    stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                  .Generic, .Class),
         domain = NA)

### * Math.cl_hierarchy

## No Math() for any kind of hierarchy.
Math.cl_hierarchy <-
function(x, ...)
    stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                  .Generic, .Class),
         domain = NA)

### * Ops.cl_hierarchy

Ops.cl_hierarchy <-
function(e1, e2)
{
    if(nargs() == 1L)
        stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    ## Only comparisons are supprorted.
    if(!(as.character(.Generic) %in% c("<", "<=", ">", ">=",
                                       "==", "!=")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    if(n_of_objects(e1) != n_of_objects(e2))
        stop("Hierarchies must have the same number of objects.")

    c1 <- cl_classes(e1)
    c2 <- cl_classes(e2)

    switch(.Generic,
           "<=" = all(is.finite(match(c1, c2))),
           "<"  = all(is.finite(match(c1, c2))) && any(is.na(match(c2, c1))),
           ">=" = all(is.finite(match(c2, c1))),
           ">"  = all(is.finite(match(c2, c1))) && any(is.na(match(c1, c2))),
           "==" = all(is.finite(match(c1, c2))) && all(is.finite(match(c2, c1))),
           "!=" = any(is.na(match(c1, c2))) || any(is.na(match(c2, c1))))
}

### * Summary.cl_hierarchy

## <NOTE>
## This is really the same as Summary.cl_partition().
## </NOTE>

Summary.cl_hierarchy <-
function(..., na.rm = FALSE)
{
    ok <- switch(.Generic, max = , min = , range = TRUE, FALSE)
    if(!ok)
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)
    args <- list(...)
    switch(.Generic,
           "min" = cl_meet(cl_ensemble(list = args)),
           "max" = cl_join(cl_ensemble(list = args)),
           "range" = {
               cl_ensemble(min = cl_meet(cl_ensemble(list = args)),
                           max = cl_join(cl_ensemble(list = args)))
           })
}

### * as.hclust.cl_hierarchy

as.hclust.cl_hierarchy <-
function(x, ...)
    as.hclust(.get_representation(x), ...)

### * is.cl_dendrogram

## <FIXME>
## Once we have cl_dendrogram testing, we can simplify cl_hierarchy
## testing.  E.g.,
##   is.cl_hierachy.default <- is.cl_dendrogram
## should be ok, and we can add cl_hierarchy predicates for hierarchies
## which are not dendrograms on top of that.
## </FIXME>

is.cl_dendrogram <-
function(x)
    UseMethod("is.cl_dendrogram")
## Default method.
is.cl_dendrogram.default <- .false
## Package stats: hclust().
is.cl_dendrogram.hclust <-
function(x)
    !is.unsorted(x$height)
## Package cluster: agnes() and diana() give objects inheriting from
## class "twins".
is.cl_dendrogram.twins <- .true
## Package cluster: mona().
is.cl_dendrogram.mona <- .true
## Package ape: class "phylo".
is.cl_dendrogram.phylo <-
function(x)
    ape::is.ultrametric(x)
## (We could also support ape's class "matching" via coercion to class
## "phylo".)
## Package clue: (virtual) class "cl_dendrogram".
is.cl_dendrogram.cl_dendrogram <- .true

### * as.cl_dendrogram

.cl_dendrogram_classes <- c("cl_dendrogram", "cl_hierarchy")

as.cl_dendrogram <-
function(x)
{
    if(is.cl_dendrogram(x)) {
        if(!inherits(x, "cl_dendrogram"))
            .make_container(x, .cl_dendrogram_classes)
        else
            x
    }
    else
        .make_container(as.cl_ultrametric(x),
                        .cl_dendrogram_classes)
}

### * print.cl_dendrogram

print.cl_dendrogram <-
function(x, ...)
    .print_container(x, "cl_dendrogram", ...)

### * plot.cl_dendrogram

plot.cl_dendrogram <-
function(x, ...)
    plot(cl_ultrametric(.get_representation(x)), ...)

### * Group methods for cl_dendrogram objects.

Ops.cl_dendrogram <-
function(e1, e2)
{
    if(nargs() == 1L)
        stop(gettextf("Unary '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    ## Only comparisons are supprorted.
    if(!(as.character(.Generic) %in% c("<", "<=", ">", ">=",
                                       "==", "!=")))
        stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
                      .Generic, .Class),
             domain = NA)

    u1 <- cl_ultrametric(e1)
    u2 <- cl_ultrametric(e2)
    if(length(u1) != length(u2))
        stop("Dendrograms must have the same number of objects.")
    switch(.Generic,
           "<=" = all(u1 <= u2),
           "<"  = all(u1 <= u2) && any(u1 < u2),
           ">=" = all(u1 >= u2),
           ">"  = all(u1 >= u2) && any(u1 > u2),
           "==" = all(u1 == u2),
           "!=" = any(u1 != u2))
}

### * Summary.cl_dendrogram

## <NOTE>
## This is really the same as Summary.cl_hierarchy() ...
## We cannot really call the poset specific internal meet and join
## functions from here as e.g. max(D, H) (D a dendrogram, H an n-tree)
## should use the n-tree poset functions ...
## However, dispatch for cl_dendrogram should not be needed if we also
## dispatch on cl_hierarchy ...
## </NOTE>

## Summary.cl_dendrogram <-
## function(..., na.rm = FALSE)
## {
##     ok <- switch(.Generic, max = , min = , range = TRUE, FALSE)
##     if(!ok)
##         stop(gettextf("Generic '%s' not defined for \"%s\" objects.",
##                       .Generic, .Class))
##     args <- list(...)
##     switch(.Generic,
##            "min" = cl_meet(cl_ensemble(list = args)),
##            "max" = cl_join(cl_ensemble(list = args)),
##            "range" = {
##                cl_ensemble(min = cl_meet(cl_ensemble(list = args)),
##                            max = cl_join(cl_ensemble(list = args)))
##            })
## }

### * as.hclust.cl_dendrogram

## <NOTE>
## This is really the same as as.hclust.cl_hierarchy() ...
## Dispatch for cl_dendrogram should not be needed if we also dispatch
## on cl_hierarchy ...
## </NOTE>

## as.hclust.cl_dendrogram <-
## function(x, ...)
##     as.hclust(.get_representation(x), ...)

### ** cut.cl_dendrogram

## Not perfect as this perhaps return something more "classed" in the
## spirit of clue ...

cut.cl_dendrogram <-
function(x, ...)
    cutree(as.hclust(x), ...)

### * Utilities

## To turn a mona object into a cl_dendrogram, we need to be able to
## compute its associated ultrametric.  Hence, provide a cophenetic()
## method for mona objects ...
cophenetic.mona <-
function(x)
{
    no <- length(x$order)
    ns <- max(x$step) + 1

    m <- matrix(NA, no, no)
    FOO <- function(ind, step, s) {
        if(length(ind) <= 1) return()
        grp <- c(0, cumsum(step == s))
        ind <- split(ind, grp)
        len <- length(ind)
        for(a in seq_len(len)) {
            for(b in seq_len(a - 1L)) {
                ## Need both as we currently cannot assume that the
                ## indices are sorted.  Alternatively, work with the
                ## sequence from one to the number of objects, and
                ## reorder at the end ...
                m[ind[[a]], ind[[b]]] <<- s
                m[ind[[b]], ind[[a]]] <<- s
            }
        }
        ind <- ind[lengths(ind) > 1L]
        pos <- which(step == s)
        step <- split(step[-pos], grp[-1][-pos])
        if(is.null(step)) return()
        for(a in seq_along(ind))
            FOO(ind[[a]], step[[a]], s + 1)
    }
    
    FOO(x$order, x$step, 1)
    m[is.na(m)] <- ns
    m <- ns - m
    rownames(m) <- rownames(x$data)
    as.dist(m)
}

## And while we're at it ...
## (Of course, as.hclust() should really "know" that a cophenetic()
## method is available ...)
as.hclust.mona <-
function(x, ...)
    hclust(cophenetic(x), "single")
    

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

cl_tabulate <-
function(x)
{
    values <- unique(x)
    counts <- tabulate(match(x, values))
    ## Still a bit tricky to create a data frame with a list "column"
    ## which is not protected by I(); otherwise, we oculd simply do
    ##   data.frame(values = I(values), counts = counts)
    out <- data.frame(values = double(length(values)), counts = counts)
    out$values <- values
    out
}

### * ls_fit_addtree

ls_fit_addtree <-
function(x, method = c("SUMT", "IP", "IR"), weights = 1,
         control = list())
{
    if(!inherits(x, "dist"))
        x <- as.dist(x)

    ## Catch some special cases right away.
    if(attr(x, "Size") <= 3L)
        return(as.cl_addtree(x))
    if(.non_additivity(x, max = TRUE) == 0)
        return(as.cl_addtree(x))

    ## Handle argument 'weights'.
    ## This is somewhat tricky ...
    if(is.matrix(weights)) {
        weights <- as.dist(weights)
        if(length(weights) != length(x))
            stop("Argument 'weights' must be compatible with 'x'.")
    }
    else
        weights <- rep_len(weights, length(x))
    if(any(weights < 0))
        stop("Argument 'weights' has negative elements.")
    if(!any(weights > 0))
        stop("Argument 'weights' has no positive elements.")

    method <- match.arg(method)
    switch(method,
           SUMT = .ls_fit_addtree_by_SUMT(x, weights, control),
           IP = {
               .ls_fit_addtree_by_iterative_projection(x, weights,
                                                       control)
           },
           IR = {
               .ls_fit_addtree_by_iterative_reduction(x, weights,
                                                      control)
           })
}

### ** .ls_fit_addtree_by_SUMT

.ls_fit_addtree_by_SUMT <-
function(x, weights = 1, control = list())
{
    ## Control parameters:
    ## gradient,
    gradient <- control$gradient
    if(is.null(gradient))
        gradient <- TRUE
    ## nruns,
    nruns <- control$nruns
    ## start,
    start <- control$start

    ## Handle start values and number of runs.
    if(!is.null(start)) {
        if(!is.list(start)) {
            ## Be nice to users.
            start <- list(start)
        }
    } else if(is.null(nruns)) {
        ## Use nruns only if start is not given.
        nruns <- 1L
    }

    w <- weights / sum(weights)
    n <- attr(x, "Size")
    labels <- attr(x, "Labels")    

    ## Handle missing values in x along the lines of de Soete (1984):
    ## set the corresponding weights to 0, and impute by the weighted
    ## mean.
    ind <- which(is.na(x))
    if(any(ind)) {
        w[ind] <- 0
        x[ind] <- weighted.mean(x, w, na.rm = TRUE)
    }

    L <- function(d) sum(w * (d - x) ^ 2)
    P <- .make_penalty_function_addtree(n)
    if(gradient) {
        grad_L <- function(d) 2 * w * (d - x)
        grad_P <- .make_penalty_gradient_addtree(n)
    }
    else {
        grad_L <- grad_P <- NULL
    }

    if(is.null(start)) {
        ## Initialize by "random shaking".  Use sd() for simplicity.
        start <- replicate(nruns,  
                           x + rnorm(length(x), sd = sd(x) / sqrt(3)),
                           simplify = FALSE)
    }

    ## And now ...
    d <- sumt(start, L, P, grad_L, grad_P,
              method = control$method, eps = control$eps,
              q = control$q, verbose = control$verbose,
              control = as.list(control$control))$x

    ## Round to enforce additivity, and hope for the best ...
    .cl_addtree_from_addtree_approximation(d, n, labels)    

}

.make_penalty_function_addtree <-
function(n)
    function(d) {
        (.non_additivity(.symmetric_matrix_from_veclh(d, n))
         + sum(pmin(d, 0) ^ 2))
    }

.make_penalty_gradient_addtree <-
function(n)
    function(d) {
        gr <- matrix(.C(C_deviation_from_additivity_gradient,
                        as.double(.symmetric_matrix_from_veclh(d, n)),
                        as.integer(n),
                        gr = double(n * n))$gr,
                     n, n)
        gr[row(gr) > col(gr)] + 2 * sum(pmin(d, 0))
    }
    

### ** .ls_fit_addtree_by_iterative_projection

## <NOTE>
## Functions
##   .ls_fit_addtree_by_iterative_projection()
##   .ls_fit_addtree_by_iterative_reduction()
## are really identical apart from the name of the C routine they call.
## (But will this necessarily always be the case in the future?)
## Merge maybe ...
## </NOTE>

.ls_fit_addtree_by_iterative_projection <-
function(x, weights = 1, control = list())
{
    if(any(diff(weights)))
        warning("Non-identical weights currently not supported.")

    labels <- attr(x, "Labels")
    x <- as.matrix(x)
    n <- nrow(x)

    ## Control parameters:
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 10000L
    ## nruns,
    nruns <- control$nruns
    ## order,
    order <- control$order
    ## tol,
    tol <- control$tol
    if(is.null(tol))
        tol <- 1e-8
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle order and nruns.
    if(!is.null(order)) {
        if(!is.list(order))
            order <- as.list(order)
        if(!all(vapply(order,
                       function(o) all(sort(o) == seq_len(n)),
                       NA)))
            stop("All given orders must be valid permutations.")
    }
    else {
        if(is.null(nruns))
            nruns <- 1L
        order <- replicate(nruns, sample(n), simplify = FALSE)
    }

    ind <- lower.tri(x)
    L <- function(d) sum(weights * (x - d)[ind] ^ 2)

    d_opt <- NULL
    v_opt <- Inf
    for(run in seq_along(order)) {
        if(verbose)
            message(gettextf("Iterative projection run: %d", run))
        d <- .C(C_ls_fit_addtree_by_iterative_projection,
                as.double(x),
                as.integer(n),
                as.integer(order[[run]] - 1L),
                as.integer(maxiter),
                iter = integer(1L),
                as.double(tol),
                as.logical(verbose))[[1L]]
        v <- L(d)
        if(v < v_opt) {
            v_opt <- v
            d_opt <- d
        }
    }

    d <- matrix(d_opt, n)
    dimnames(d) <- list(labels, labels)
    .cl_addtree_from_addtree_approximation(as.dist(d))
}

### ** .ls_fit_addtree_by_iterative_reduction

.ls_fit_addtree_by_iterative_reduction <-
function(x, weights = 1, control = list())
{
    if(any(diff(weights)))
        warning("Non-identical weights currently not supported.")

    labels <- attr(x, "Labels")
    x <- as.matrix(x)
    n <- nrow(x)

    ## Control parameters:
    ## maxiter,
    maxiter <- control$maxiter
    if(is.null(maxiter))
        maxiter <- 10000L
    ## nruns,
    nruns <- control$nruns
    ## order,
    order <- control$order
    ## tol,
    tol <- control$tol
    if(is.null(tol))
        tol <- 1e-8
    ## verbose.
    verbose <- control$verbose
    if(is.null(verbose))
        verbose <- getOption("verbose")

    ## Handle order and nruns.
    if(!is.null(order)) {
        if(!is.list(order))
            order <- as.list(order)
        if(!all(vapply(order,
                       function(o) all(sort(o) == seq_len(n)),
                       NA)))
            stop("All given orders must be valid permutations.")
    }
    else {
        if(is.null(nruns))
            nruns <- 1L
        order <- replicate(nruns, sample(n), simplify = FALSE)
    }

    ind <- lower.tri(x)
    L <- function(d) sum(weights * (x - d)[ind] ^ 2)

    d_opt <- NULL
    v_opt <- Inf
    for(run in seq_along(order)) {
        if(verbose)
            message(gettextf("Iterative reduction run: %d", run))
        d <- .C(C_ls_fit_addtree_by_iterative_reduction,
                as.double(x),
                as.integer(n),
                as.integer(order[[run]] - 1L),
                as.integer(maxiter),
                iter = integer(1L),
                as.double(tol),
                as.logical(verbose))[[1L]]
        v <- L(d)
        if(v < v_opt) {
            v_opt <- v
            d_opt <- d
        }
    }

    d <- matrix(d_opt, n)
    dimnames(d) <- list(labels, labels)
    .cl_addtree_from_addtree_approximation(as.dist(d))
}

### * .non_additivity

.non_additivity <-
function(x, max = FALSE)
{
    if(!is.matrix(x))
        x <- .symmetric_matrix_from_veclh(x)
    .C(C_deviation_from_additivity,
       as.double(x),
       as.integer(nrow(x)),
       fn = double(1L),
       as.logical(max))$fn
}

### * ls_fit_centroid

ls_fit_centroid <-
function(x)
{
    ## Fit a centroid additive tree distance along the lines of Carroll
    ## & Pruzansky (1980).  In fact, solving
    ##
    ##    \sum_{i,j: i \ne j} (\delta_{ij} - (g_i + g_j)) ^ 2 => min_g
    ##
    ## gives \sum_{j: j \ne i} (g_i + g_j - \delta_{ij}) = 0, or (also
    ## in Barthemely & Guenoche)
    ##
    ##    (n - 2) g_i + \sum_j g_j = \sum_{j: j \ne i} \delta_{ij}
    ##
    ## which after summing over all i and some manipulations eventually
    ## gives
    ##
    ##    g_i = \frac{1}{n-2} (v_i - m),
    ##
    ##    v_i = \sum_{j: j \ne i} \delta_{ij}
    ##    s   = \frac{1}{2(n-1)} \sum_{i,j: j \ne i} \delta_{ij}

    n <- attr(x, "Size")

    if(n <= 2L)
        return(as.cl_addtree(0 * x))

    x <- as.matrix(x)
    g <- rowSums(x) / (n - 2) - sum(x) / (2 * (n - 1) * (n - 2))
    as.cl_addtree(as.dist(.make_centroid_matrix(g)))
}

.make_centroid_matrix <-
function(g)
{
    y <- outer(g, g, `+`)
    diag(y) <- 0
    y
}

### * as.cl_addtree

as.cl_addtree <-
function(x)
    UseMethod("as.cl_addtree")

as.cl_addtree.default <-
function(x)    
{
    if(inherits(x, "cl_addtree"))
        x
    else if(is.atomic(x) || inherits(x, "cl_ultrametric"))
        .cl_addtree_from_veclh(x)
    else if(is.matrix(x)) {
        ## Should actually check whether the matrix is symmetric, >= 0
        ## and satisfies the 4-point conditions ...
        .cl_addtree_from_veclh(as.dist(x))
    }
    else if(is.cl_dendrogram(x))
        .cl_addtree_from_veclh(cl_ultrametric(x))
    else
        stop("Cannot coerce to 'cl_addtree'.")
}

as.cl_addtree.phylo <-
function(x)
    .cl_addtree_from_veclh(as.dist(cophenetic(x)))
## Phylogenetic trees with edge/branch lengths yield additive tree
## dissimilarities.

### * .cl_addtree_from_veclh

.cl_addtree_from_veclh <-
function(x, size = NULL, labels = NULL)
{
    cl_proximity(x, "Additive tree distances",
                 labels = labels, size = size,
                 class = c("cl_addtree", "cl_dissimilarity",
                 "cl_proximity", "dist"))
}

### * .cl_addtree_from_addtree_approximation

.cl_addtree_from_addtree_approximation <-
function(x, size = NULL, labels = NULL)
{
    ## Turn x into an addtree after possibly rounding to non-additivity
    ## significance (note that this is not guaranteed to work ...).
    mnum <- .non_additivity(x, max = TRUE)
    x <- round(x, floor(abs(log10(mnum))))
    .cl_addtree_from_veclh(x, size = size, labels = labels)
}

### * .decompose_addtree

.decompose_addtree <-
function(x, const = NULL)
{
    ## Decompose an addtree into an ultrametric and a centroid
    ## distance.

    ## If 'const' is not given, we take the root as half way between the
    ## diameter of the addtree, and choose a minimal constant to ensure
    ## non-negativity (but not positivity) of the ultrametric.

    ## As this is all slightly dubious and it is not quite clear how
    ## much positivity we want in the ultrametric of the decomposition,
    ## we keep this hidden.  For plotting addtrees, the choice of the
    ## constant does not seem to matter.

    x <- as.matrix(x)
    n <- nrow(x)

    ## Determine diameter.
    ind <- which.max(x) - 1
    u <- ind %% n + 1
    v <- ind %/% n + 1

    if(!is.null(const))
        g <- pmax(x[u, ], x[v, ]) - const
    else {
        g <- pmax(x[u, ], x[v, ]) - x[u, v] / 2
        u <- x - .make_centroid_matrix(g)
        k <- - min(u)
        g <- g - k / 2
    }
    u <- x - .make_centroid_matrix(g)

    names(g) <- rownames(x)

    ## Ensure a valid ultrametric.
    d <- .ultrametrify(as.dist(u))
    u <- .cl_ultrametric_from_veclh(d, nrow(x), rownames(x))

    ## Note that we return the centroid distances to the root, and not
    ## between the objects (as.dist(.make_centroid_matrix(g))) ...
    list(Ultrametric = as.cl_ultrametric(u), Centroid = g)
}

### * plot.cl_addtree

plot.cl_addtree <-
function(x, ...)
{
    ## Construct a dendrogram-style representation of the addtree with
    ## the root half way between the diameter, and plot.
    y <- .decompose_addtree(x, max(x))
    u <- y$Ultrametric
    g <- y$Centroid
    ## We halve the scale of the ultrametric, and add the maximal g from
    ## the centroid.
    h <- hclust(as.dist(u / 2), "single")
    h$height <- h$height + max(g)
    d <- as.dendrogram(h)
    ## Now modify the heights of the leaves so that the objects giving
    ## the diameter of the addtree end up with height zero.
    g <- max(g) - g
    names(g) <- labels(g)
    d <- dendrapply(d, function(n) {
        if(!is.leaf(n)) return(n)
        attr(n, "height") <- g[attr(n, "label")]
        n
    })
    ## And finally plot
    plot(d, ...)
}


### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### * cl_membership

## Get the class membership matrix from a partition.

## <NOTE>
## We could use sparse matrices for the memberships of hard partitions.
## Not sure if this is really that important, though, as we typically
## use memberships in a context where dense matrices (memberships of
## soft partitions) occur.
## </NOTE>

## <NOTE>
## Currently, the number of classes to be used for the memberships must
## not be less than the number of classes in the partition.  We might
## eventually change this so that "optimal" collapsing of classes is
## performed (but note that optimality needs to be relative to some
## dissimilarity measure) ...
## However, from the discussion of the second method in Gordon and Vichi
## (2001) we note that whereas optimal assignment is "simple", optimal
## collapsing (equivalent to partitioning into an arbitrary number of
## partitions) is of course very hard.
## </NOTE>

cl_membership <-
function(x, k = n_of_classes(x))
{
    if(k < n_of_classes(x))
        stop("k cannot be less than the number of classes in x.")
    UseMethod("cl_membership")
}

## Default method.
cl_membership.default <-
function(x, k = n_of_classes(x))
    .cl_membership_from_class_ids(cl_class_ids(x), k)

## Package stats: kmeans() (R 2.1.0 or better).
cl_membership.kmeans <- cl_membership.default

## Package cluster: clara(), fanny(), and pam() give objects of the
## respective class inheriting from class "partition".
cl_membership.fanny <-
function(x, k = n_of_classes(x))
    .cl_membership_from_memberships(x$membership, k)
cl_membership.partition <- cl_membership.default

## Package cclust: cclust().
cl_membership.cclust <- cl_membership.default

## Package e1071: cmeans() gives objects of class "fclust".
cl_membership.fclust <- cl_membership.fanny
## Package e1071: cshell().
cl_membership.cshell <- cl_membership.fanny
## Package e1071: bclust().
cl_membership.bclust <- cl_membership.default

## Package flexmix: class "flexmix".
## <NOTE>
## We used to be able to call flexmix::posterior(), but this now only
## has S4 methods for modeltools::posterior() S4 generic.  Let's call
## this one, and hope that flexmix has been loaded ...
## </NOTE>
cl_membership.flexmix <-
function(x, k = n_of_classes(x))
    .cl_membership_from_memberships(modeltools::posterior(x), k)

## Package mclust: Mclust().
cl_membership.Mclust <-
function(x, k = n_of_classes(x))
    .cl_membership_from_memberships(x$z, k)

## Package clue: Memberships.
cl_membership.cl_membership <-
function(x, k = n_of_classes(x))
    .cl_membership_from_memberships(x, k)
## (Note: we cannot simply return x in case k equals n_of_classes(x),
## because ncol(x) might be different.)

## Package clue: pclust().
cl_membership.pclust <-
function(x, k = n_of_classes(x))
{
    ## We should really have a suitable "sparse matrix" class for
    ## representing the memberships of hard partitions.  In case we
    ## decide not to fill the membership "slot" for such:
    if(is.null(m <- x$membership))
        .cl_membership_from_class_ids(x$cluster, k)
    else
        .cl_membership_from_memberships(m, k)
}

## Package clue: (virtual) class "cl_partition".
cl_membership.cl_partition <-
function(x, k = n_of_classes(x))
    cl_membership(.get_representation(x), k)

## Package movMF: class "movMF".
cl_membership.movMF <-
function(x, k = n_of_classes(x))
    .cl_membership_from_memberships(x$P, k)

### * .make_cl_membership

## A low-level common creator.

.make_cl_membership <-
function(x, n_of_classes, is_cl_hard_partition, meta = NULL)
{
    attr(x, "n_of_classes") <- n_of_classes
    attr(x, "is_cl_hard_partition") <- is_cl_hard_partition
    attr(x, "meta") <- meta
    class(x) <- "cl_membership"
    x
}

### * .cl_membership_from_class_ids

.cl_membership_from_class_ids <-
function(x, k = NULL, meta = NULL)
{
    x <- factor(x)
    n_of_objects <- length(x)
    n_of_classes <- nlevels(x)
    if(is.null(k))
        k <- n_of_classes
    else if(k < n_of_classes)
        stop("k cannot be less than the number of classes in x.")
    ## <TODO>
    ## Should really use a sparse encoding of this ...
    M <- matrix(0, n_of_objects, k)
    ## (Could also use .one_entry_per_column(M, as.numeric(x)) <- 1 for
    ## the time being.)
    M[cbind(seq_len(n_of_objects), as.numeric(x))] <- 1
    ## But note that we also need to handle NAs ...
    M[is.na(x), ] <- NA
    ## </TODO>
    if(nlevels(x) == k)
        colnames(M) <- levels(x)
    if(!is.null(nm <- names(x)))
        rownames(M) <- nm
    .make_cl_membership(M, n_of_classes, TRUE, meta)
}

### * .cl_membership_from_memberships

.cl_membership_from_memberships <-
function(x, k = NULL, meta = NULL)
{
    ## <NOTE>
    ## Dropping and re-filling of ## zero columns in case k is given may
    ## seem unnecessary, but really canonicalizes by moving zero columns
    ## last ...
    ## </NOTE>
    
    x <- x[ , colSums(x, na.rm = TRUE) > 0, drop = FALSE]
    n_of_classes <- ncol(x)
    if(!is.null(k)) {
        if(k < n_of_classes)
            stop("k cannot be less than the number of classes in x.")
        if(k > n_of_classes) {
            ## Fill up with zero columns.
            x <- cbind(x, matrix(0, nrow(x), k - n_of_classes))
            ## Handle NAs if necessary.
            x[apply(is.na(x), 1, any), ] <- NA
        }
    }
    .make_cl_membership(x, n_of_classes,
                        all(rowSums(x == 1, na.rm = TRUE) > 0),
                        meta)
}

### * as.cl_membership

as.cl_membership <-
function(x)
    UseMethod("as.cl_membership")
as.cl_membership.default <-
function(x)
{
    if(inherits(x, "cl_membership"))
        x
    else if(is.atomic(x))
        .cl_membership_from_class_ids(x)
    else
        cl_membership(x)
}
as.cl_membership.matrix <-
function(x)
    .cl_membership_from_memberships(x)

### * .memberships_from_cross_dissimilarities

.memberships_from_cross_dissimilarities <-
function(d, power = 2)
{
    ## For a given matrix of cross-dissimilarities [d_{bj}], return a
    ## matrix [u_{bj}] such that \sum_{b,j} u_{bj}^p d_{bj}^q => min!
    ## under the constraint that u is a stochastic matrix.
    ## If only one power is given, it is taken as p, with q as 1.
    ## <NOTE>
    ## This returns a plain matrix of membership values and not a
    ## cl_membership object (so that it does not deal with possibly
    ## dropping or re-introducing unused classes).
    ## </NOTE>
    exponent <- if(length(power) == 1L)
        1 / (1 - power)
    else
        power[2L] / (1 - power[1L])
    u <- matrix(0, nrow(d), ncol(d))
    zero_incidences <- !(d > 0)
    n_of_zeroes <- rowSums(zero_incidences)
    if(any(ind <- (n_of_zeroes > 0)))
        u[ind, ] <-
            zero_incidences[ind, , drop = FALSE] / n_of_zeroes[ind]
    if(any(!ind)) {
        ## Compute d_{bj}^e / \sum_k d_{bk}^e without overflow from very
        ## small d_{bj} values.
        d <- exponent * log(d[!ind, , drop = FALSE])
        d <- exp(d - d[cbind(seq_len(nrow(d)), max.col(d))])
        u[!ind, ] <- d / rowSums(d)
    }
    u
}

### * print.cl_membership

print.cl_membership <-
function(x, ...)
{
    writeLines("Memberships:")
    print(matrix(as.vector(x), nrow = nrow(x), dimnames = dimnames(x)),
          ...)
    invisible(x)
}

### .has_object_memberships

## Be nice to users when computing proximities: all measures for
## "partitions" we currently consider really only assume that we can
## compute memberships and/or class ids.

## Note that the cl_membership() default method works for cl_class_ids.

.has_object_memberships <-
function(x)
    (is.cl_partition(x)
     || inherits(x, "cl_membership")
     || inherits(x, "cl_class_ids"))

### * .stochastify

.stochastify <-
function(x)
{
    ## Try to ensure that a stochastic matrix is returned.
    x <- pmax(x, 0)
    x / rowSums(x)
}

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

cl_margin <-
function(x)
{
    if(is.cl_hard_partition(x))
        out <- rep.int(1, n_of_objects(x))
    else if(is.cl_partition(x)) {
        x <- cl_membership(x)
        i <- seq_len(nrow(x))
        j <- cbind(i, max.col(x))
        out <- x[j]
        x[j] <- 0
        out <- out - x[cbind(i, max.col(x))]
    }
    else
        stop("Argument 'x' must be a partition.")

    out
}

### * cl_dissimilarity

cl_dissimilarity <-
function(x, y = NULL, method = "euclidean", ...)
{
    x <- as.cl_ensemble(x)
    is_partition_ensemble <-
        (inherits(x, "cl_partition_ensemble")
         || all(vapply(x, .has_object_memberships, NA)))

    ## Be nice.
    if(is.character(y) || is.function(y)) {
        method <- y
        y <- NULL
    }

    if(is.function(method))
        method_name <- "user-defined method"
    else {
        if(!inherits(method, "cl_dissimilarity_method")) {
            ## Get the method definition and description from the
            ## registry. 
            type <- ifelse(is_partition_ensemble,
                           "partition", "hierarchy")
            method <- get_cl_dissimilarity_method(method, type)
        }
        method_name <- method$description
        method <- method$definition
    }
              
    if(!is.null(y)) {
        y <- as.cl_ensemble(y)
        is_partition_ensemble_y <-
            (inherits(y, "cl_partition_ensemble")
             || all(vapply(x, .has_object_memberships, NA)))
        if(!identical(is_partition_ensemble, is_partition_ensemble_y))
            stop("Cannot mix partitions and hierarchies.")
        if(n_of_objects(x) != n_of_objects(y))
            stop("All clusterings must have the same number of objects.")
        ## Build a cross-proximity object of cross-dissimilarities.
        d <- matrix(0, length(x), length(y))
        for(j in seq_along(y))
            d[, j] <- sapply(x, method, y[[j]], ...)
        dimnames(d) <- list(names(x), names(y))
        return(cl_cross_proximity(d, method_name,
                                  class = "cl_cross_dissimilarity"))
    }
    
    ## Otherwise, build a proximity object of dissimilarities.
    n <- length(x)
    d <- vector("list", length = n - 1L)
    ind <- seq_len(n)
    while(length(ind) > 1L) {
        j <- ind[1L]
        ind <- ind[-1L]
        d[[j]] <- sapply(x[ind], method, x[[j]], ...)
    }
    
    cl_proximity(unlist(d),
                 method_name,
                 labels = names(x),
                 size = n,
                 class = c("cl_dissimilarity", "cl_proximity", "dist"))
}

### ** .cl_dissimilarity_partition_euclidean

.cl_dissimilarity_partition_euclidean <-
function(x, y)
{
    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    ## Match classes from conforming memberships.
    ind <- solve_LSAP(crossprod(M_x, M_y), maximum = TRUE)
    sqrt(sum((M_x - M_y[, ind]) ^ 2))
}

### ### ** .cl_dissimilarity_partition_manhattan

.cl_dissimilarity_partition_manhattan <-
function(x, y)
{
    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    C <- .cxdist(M_x, M_y, "manhattan")
    ind <- solve_LSAP(C)
    sum(C[cbind(seq_along(ind), ind)])
}    

### ** .cl_dissimilarity_partition_comemberships

.cl_dissimilarity_partition_comemberships <-
function(x, y)
{
    ## We used to have the straightforward
    ##   C_x <- tcrossprod(cl_membership(x)) # M_x M_x'
    ##   C_y <- tcrossprod(cl_membership(y)) # M_y M_y'
    ##   sum((C_x - C_y) ^ 2) / n_of_objects(x) ^ 2
    ## But note that
    ##   \| AA' - BB' \|^2
    ##      = tr((AA' - BB')'(AA' - BB')
    ##      = tr(A'A A'A) - 2 tr(A'B B'A) + tr(B'B B'B)
    ##      = \| A'A \|^2 - 2 \| A'B \|^2 + \| B'B \|^2
    ## which can be computed much more efficiently as all involved cross
    ## product matrices are "small" ...
    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    sqrt(sum(crossprod(M_x) ^ 2)
         - 2 * sum(crossprod(M_x, M_y) ^ 2)
         + sum(crossprod(M_y) ^ 2))
}

### ** .cl_dissimilarity_partition_symdiff

.cl_dissimilarity_partition_symdiff <-
function(x, y)
{
    ## Cardinality of the symmetric difference of the partitions
    ## regarded as binary equivalence relations, i.e., the number of
    ## discordant pairs.

    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)

    n <- n_of_objects(x)
    .cl_dissimilarity_partition_Rand(x, y) * choose(n, 2)
}
    
### ** .cl_dissimilarity_partition_Rand

.cl_dissimilarity_partition_Rand <-
function(x, y)    
{
    ## Handle soft partitions using the corresponding hard ones.
    ## (At least, for the time being.)

    1 - .cl_agreement_partition_Rand(x, y)
}

### ** .cl_dissimilarity_partition_GV1

.cl_dissimilarity_partition_GV1 <-
function(x, y)
{    
    k_x <- n_of_classes(x)
    k_y <- n_of_classes(y)
    M_x <- cl_membership(x, k_x)
    M_y <- cl_membership(y, k_y)
    C <- outer(colSums(M_x ^ 2), colSums(M_y ^ 2), `+`) -
        2 * crossprod(M_x, M_y)
    if(k_x < k_y)
        C <- rbind(C, matrix(0, nrow = k_y - k_x, ncol = k_y))
    else if(k_x > k_y)
        C <- cbind(C, matrix(0, nrow = k_x, ncol = k_x - k_y))
    ind <- solve_LSAP(C)
    sqrt(sum(C[cbind(seq_along(ind), ind)]))
    ## (Note that this sum really only includes matched non-dummy
    ## classes.)
}

### ** .cl_dissimilarity_partition_BA_A

.cl_dissimilarity_partition_BA_A <-
function(x, y)
{
    .cl_dissimilarity_partition_manhattan(as.cl_hard_partition(x),
                                          as.cl_hard_partition(y)) / 2
    ## Could to this more efficiently, of course ...
}

### ** .cl_dissimilarity_partition_BA_C

.cl_dissimilarity_partition_BA_C <-
function(x, y)
{
    n_of_classes(x) + n_of_classes(y) - 2 * n_of_classes(cl_join(x, y))
}

### ** .cl_dissimilarity_partition_BA_D

.cl_dissimilarity_partition_BA_D <-
    .cl_dissimilarity_partition_Rand

### ** .cl_dissimilarity_partition_BA_E

.cl_dissimilarity_partition_BA_E <-
function(x, y)
{
    z <- table(cl_class_ids(x), cl_class_ids(y))
    z <- z / sum(z)

    ## Average mutual information between the partitions.
    y <- outer(rowSums(z), colSums(z))
    i <- which((z > 0) & (y > 0))
    I <- sum(z[i] * log(z[i] / y[i]))

    ## Entropy of meet(x, y).
    i <- which(z > 0)
    H <- - sum(z[i] * log(z[i]))

    1 - I / H
}

### ** .cl_dissimilarity_partition_VI

.cl_dissimilarity_partition_VI <-
function(x, y, weights = 1)
{
    ## Variation of information for general "soft clusterings", cf
    ## Section 5.2. in Meila (2002). 
    weights <- rep_len(weights, n_of_objects(x))
    weights <- weights / sum(weights)
    M_x <- cl_membership(x)
    ## Weighted marginal distribution of x:
    m_x <- colSums(weights * M_x)
    M_y <- cl_membership(y)
    ## Weighted marginal distribution of y:
    m_y <- colSums(weights * M_y)
    gamma <- crossprod(weights * M_x, M_y)
    delta <- outer(m_x, m_y)
    ## Entropy of x:
    H_x <- - sum(m_x * log(ifelse(m_x > 0, m_x, 1)))
    ## Entropy of y:
    H_y <- - sum(m_y * log(ifelse(m_y > 0, m_y, 1)))
    ## VI is H_x + H_y minus twice the (weighted) joint information.
    i <- which((gamma > 0) & (delta > 0))
    H_x + H_y - 2 * sum(gamma[i] * log(gamma[i] / delta[i]))
}

### ** .cl_dissimilarity_partition_Mallows

.cl_dissimilarity_partition_Mallows <-
function(x, y, p = 1, alpha = NULL, beta = NULL)
{
    ## Currently, no "real" primal-dual solver for minimum cost flow
    ## problems, and lpSolve::lp.transport() seems to work only for
    ## integer bounds.  Hence, rather than using
    ##
    ##   C <- .cxdist(cl_membership(x), cl_membership(y),
    ##                "minkowski", p) ^ p
    ##   n_x <- nrow(C)
    ##   n_y <- ncol(C)
    ##   if(is.null(alpha))
    ##       alpha <- rep.int(1 / n_x, n_x)
    ##   else {
    ##       alpha <- rep_len(alpha, n_x)
    ##       alpha <- alpha / sum(alpha)
    ##   }
    ##
    ## etc right away, ensure a square cost matrix so that we can have
    ## integer bounds for at least the default case.

    k <- max(n_of_classes(x), n_of_classes(y))
    M_x <- cl_membership(x, k)
    M_y <- cl_membership(y, k)
    C <- .cxdist(M_x, M_y, "minkowski", p) ^ p
    if(is.null(alpha)) alpha <- rep.int(1, k)
    if(is.null(beta)) beta <- rep.int(1, k)
    lpSolve::lp.transport(C, "min",
                          rep.int("==", k), alpha,
                          rep.int("==", k), beta,
                          integers = NULL)$objval ^ (1 / p)
}

### ** .cl_dissimilarity_partition_CSSD

.cl_dissimilarity_partition_CSSD <-
function(x, y, L = NULL, alpha = NULL, beta = NULL, ...)
{
    ## Cluster Similarity Sensitive Distance.
    ## Reference:  D. Zhou, J. Li and H. Zha (2005),
    ##  A new Mallows distance based metric for comparing clusterings.
    ## See .cl_dissimilarity_partition_Mallows() re solving cost flow
    ## problems.

    ## Dissimilarity is defined by minimizing
    ##   \sum_{k,l} (1 - 2 w_{kl} / (alpha_k + beta_l)) L_{kl}
    ## where
    ##   L_{kl} = \sum_i m_{x;ik} m_{y;il} distance(p_{x;k}, p_{y;l})
    ## with m and p the memberships and prototypes, respectively.
    ## If we get matrices of prototypes, use .rxdist; otherwise, the
    ## user needs to specify an L function or matrix.

    k_x <- n_of_classes(x)
    k_y <- n_of_classes(y)
    M_x <- cl_membership(x, k_x)
    M_y <- cl_membership(y, k_y)
    if(!is.matrix(L)) {
        p_x <- cl_prototypes(x)
        p_y <- cl_prototypes(y)
        if(is.matrix(p_x) && is.matrix(p_y) && is.null(L))
            L <- .rxdist(p_x, p_y, ...)
        else if(is.function(L))
            L <- L(p_x, p_y)
        else
            stop("Cannot compute prototype distances.")
    }
    C <- crossprod(M_x, M_y) * L
    if(is.null(alpha)) alpha <- rep.int(1, k_x)
    if(is.null(beta)) beta <- rep.int(1, k_y)
    sum(C) - 2 * lpSolve::lp.transport(C / outer(alpha, beta, `+`),
                                       "max",
                                       rep.int("==", k_x), alpha,
                                       rep.int("==", k_y), beta,
                                       integers = NULL)$objval
}

### ** .cl_dissimilarity_hierarchy_euclidean

.cl_dissimilarity_hierarchy_euclidean <-
function(x, y, weights = 1)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    sqrt(sum(weights * (u - v) ^ 2))
}

### ** .cl_dissimilarity_hierarchy_manhattan

.cl_dissimilarity_hierarchy_manhattan <-
function(x, y, weights = 1)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    sum(weights * abs(u - v))
}

### ** .cl_dissimilarity_hierarchy_cophenetic

.cl_dissimilarity_hierarchy_cophenetic <-
function(x, y)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    1 - cor(u, v) ^ 2
}

### ** .cl_dissimilarity_hierarchy_gamma

.cl_dissimilarity_hierarchy_gamma <-
function(x, y)
{
    ## <NOTE>
    ## This is a dissimilarity measure that works for arbitrary
    ## dissimilarities, see e.g. Bock.
    ## (And the current implementation finally respects this ...)
    ## </NOTE>
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    n <- length(u)
    .C(C_clue_dissimilarity_count_inversions,
       as.double(u),
       as.double(v),
       as.integer(n),
       count = double(1L)) $ count / choose(n, 2)
}

### ** .cl_dissimilarity_hierarchy_symdiff

.cl_dissimilarity_hierarchy_symdiff <-
function(x, y)
{
    ## Cardinality of the symmetric difference of the n-trees when
    ## regarded as sets of subsets (classes) of the set of objects.

    x <- cl_classes(x)
    y <- cl_classes(y)
    sum(is.na(match(x, y))) + sum(is.na(match(y, x)))
}

### ** .cl_dissimilarity_hierarchy_Chebyshev

.cl_dissimilarity_hierarchy_Chebyshev <-
function(x, y)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    max(abs(u - v))
}

### ** .cl_dissimilarity_hierarchy_Lyapunov

.cl_dissimilarity_hierarchy_Lyapunov <-
function(x, y)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    q <- cl_object_dissimilarities(x) / cl_object_dissimilarities(y)
    if(is.matrix(q)) q <- q[lower.tri(q)]
    log(max(q) / min(q))
}

### ** .cl_dissimilarity_hierarchy_BO

.cl_dissimilarity_hierarchy_BO <-
function(x, y, delta, ...)
{
    ## Compute Boorman-Olivier (1973) dendrogram ("valued tree")
    ## dissimilarities of the form
    ##
    ##    m_\delta(T_1, T_2)
    ##      = \int_0^\infty \delta(P_1(\alpha), P_2(\alpha)) d\alpha
    ##
    ## where the trees (dendrograms) are defined as right-continuous
    ## maps from [0, \Infty) to the partition lattice.

    ## We can compute this as follows.  Take the ultrametrics and use
    ## as.hclust() to detemine the heights \alpha_1(k) and \alpha_2(l)
    ## of the splits.  Let \alpha_i be the sequence obtained by
    ## combining these two.  Then
    ##
    ##   m_\delta
    ##     = \sum_{i=0}^{L-1} (\alpha_{i+1} - \alpha_i)
    ##                        \delta(P_1(\alpha_i), P_2(\alpha_i))
    ##
    ## We use cutree() for computing the latter partitions.  As we
    ## already have the hclust representations, we should be able to do
    ## things more efficiently ...

    if(inherits(x, "hclust"))
        t_x <- x
    else if(inherits(x, "cl_ultrametric"))
        t_x <- as.hclust(x)
    else if(is.cl_dendrogram(x))
        t_x <- as.hclust(cl_ultrametric(x))
    else
        return(NA)
    if(inherits(y, "hclust"))
        t_y <- y
    else if(inherits(y, "cl_ultrametric"))
        t_y <- as.hclust(y)
    else if(is.cl_dendrogram(y))
        t_y <- as.hclust(cl_ultrametric(y))
    else
        return(NA)

    if(is.unsorted(t_x$height) || is.unsorted(t_y$height))
        return(NA)
    
    alpha <- sort(unique(c(t_x$height, t_y$height)))
    cuts_x <- cutree(t_x, h = alpha)
    cuts_y <- cutree(t_y, h = alpha)
    deltas <- mapply(cl_dissimilarity,
                     lapply(split(cuts_x, col(cuts_x)),
                            as.cl_partition),
                     lapply(split(cuts_y, col(cuts_y)),
                            as.cl_partition),
                     MoreArgs = list(delta, ...))
    sum(diff(alpha) * deltas[-length(deltas)])
}

### ** .cl_dissimilarity_hierarchy_spectral

.cl_dissimilarity_hierarchy_spectral <-
function(x, y)
{
    if(!.has_object_dissimilarities(x) ||
       !.has_object_dissimilarities(y))
        return(NA)
    u <- cl_object_dissimilarities(x)
    v <- cl_object_dissimilarities(y)
    svd(as.matrix(u - v))$d[1L]
}


### * as.dist.cl_dissimilarity

as.dist.cl_dissimilarity <-
function(m, diag = FALSE, upper = FALSE)
{
    y <- c(m)
    ## Fill non-inherited attributes with default values.
    attributes(y) <- c(attributes(m)[c("Size", "Labels")],
                       Diag = diag, Upper = upper, call = match.call())
    ## (Note that as.dist.default() does not automatically add
    ## 'method'.)
    class(y) <- "dist"
    y
}

### * [.cl_dissimilarity

"[.cl_dissimilarity" <-
function(x, i, j)
{
    y <- NextMethod("[")
    if(!inherits(y, "cl_dissimilarity")) {
        description <- attr(x, "description")
        return(cl_cross_proximity(y, description = description,
                                  class = "cl_cross_dissimilarity"))
    }
    y
}

### .cxdist

.cxdist <-
function(A, B, method = c("euclidean", "manhattan", "minkowski"), ...)
{
    ## Return the column cross distance matrix of A and B.
    ## I.e., the matrix C = [c_{j,k}] with
    ##   c_{j,k} = distance(A[, j], B[, k])
    ## Currently, only Manhattan (L1) distances are provided.
    ## Extensions to Minkowski or even more distances (a la dist())
    ## could be added eventually.

    ## <NOTE>
    ## Possible implementations include
    ##
    ## foo_a <- function(A, B)
    ##   apply(B, 2, function(u) colSums(abs(A - u)))
    ## foo_d <- function(A, B) {
    ##   out <- as.matrix(dist(rbind(t(A), t(B)), "manhattan"))
    ##   dimnames(out) <- NULL
    ##   nc_B <- NCOL(B)
    ##   out[seq(from = NCOL(A) + 1, length.out = nc_B), seq_len(nc_B)]
    ## }
    ## foo_f <- function(A, B) {
    ##   out <- matrix(0, NCOL(A), NCOL(B))
    ##   for(j in seq_len(NCOL(A)))
    ##     for(k in seq_len(NCOL(B)))
    ##       out[j, k] = sum(abs(A[, j] - B[, k]))
    ##   out
    ## }
    ##
    ## The one actually used seems to be the best performer, with the
    ## "for" version a close second (note that "typically", A and B have
    ## much fewer columns than rows).
    ## only few columns

    method <- match.arg(method)
    
    ## Workhorse.
    FOO <- switch(method,
                  "euclidean" =
                  function(M) sqrt(colSums(M ^ 2)),
                  "manhattan" = 
                  function(M) colSums(abs(M)),
                  "minkowski" = {
                      ## Power needs to be given.
                      p <- list(...)[[1L]]
                      function(M)
                          (colSums(abs(M) ^ p)) ^ (1 / p)
                  })
    
    
    out <- matrix(0, NCOL(A), NCOL(B))
    for(k in seq_len(NCOL(B)))
        out[, k] <- FOO(A - B[, k])
    out
}

### .rxdist

.rxdist <-
function(A, B, method = c("euclidean", "manhattan", "minkowski"), ...)
{
    ## Return the row cross distance matrix of A and B.
    ## I.e., the matrix C = [c_{j,k}] with
    ##   c_{j,k} = distance(A[j, ], B[k, ])
    
    ## <NOTE>
    ## Could also do something like
    ##   ind <- seq_len(NROW(B))
    ##   as.matrix(dist(rbind(B, A)))[-ind, ind]
    ## but that is *very* inefficient for the "usual" data by prototype
    ## case (where NROW(B) << NROW(A)).
    ## </NOTE>
    
    ## No fancy pmatching for methods for the time being.
    method <- match.arg(method)
    
    ## Workhorse: Full A, single row of b.
    FOO <- switch(method,
                  "euclidean" =
                  function(A, b) sqrt(rowSums(sweep(A, 2, b) ^ 2)),
                  "manhattan" = 
                  function(A, b) rowSums(abs(sweep(A, 2, b))),
                  "minkowski" = {
                      ## Power needs to be given.
                      p <- list(...)[[1L]]
                      function(A, b)
                          (rowSums(abs(sweep(A, 2, b)) ^ p)) ^ (1 / p)
                  })
                      
    out <- matrix(0, NROW(A), NROW(B))
    ## Be nice: thanks to Wael Salem ZRAFI <zrafi.w.s@gmail.com> for
    ## suggesting this improvement.
    if(!is.null(cnA <- colnames(A)) &&
       !is.null(cnB <- colnames(B)) &&
       !identical(cnA, cnB))
        A <- A[, cnB, drop = FALSE]
    for(k in seq_len(NROW(B)))
        out[, k] <- FOO(A, B[k, ])
    out
}

### Local variables: ***
### mode: outline-minor ***
### outline-regexp: "### [*]+" ***
### End: ***

### <NOTE>
### At least currently, all registries are meant and used for all types
### of clusterings (for the time being, partitions and hierarchies)
### simultaneously.
### </NOTE>

### * Internal stuff.

.make_db_key <- function(name, type) paste(type, name, sep = "_")

### * General-purpose stuff.

### <FIXME>
### This currently insists on a given type: maybe it should simply list
### everything split according to type.  But hey, it's internal stuff
### anyway (at least for the time being ...)
### </FIXME>
get_methods_from_db <- function(db, type) {
    type <- match.arg(type, c("partition", "hierarchy"))
    
    pattern <- sprintf("^%s_", type)
    sub(pattern, "",
        grep(pattern, objects(db), value = TRUE))
}

get_method_from_db <- function(db, type, name, msg) {
    ## <NOTE>
    ## Keep 'msg' here so that gettext()ing could work ...
    ## </NOTE>
    type <- match.arg(type, c("partition", "hierarchy"))

    db_keys <- objects(db)
    ind <- pmatch(.make_db_key(tolower(name), type), tolower(db_keys))
    if(is.na(ind))
        stop(msg, call. = FALSE, domain = NA)
    
    db[[db_keys[ind]]]
}

put_method_into_db <- function(db, type, name, value) {
    type <- match.arg(type, c("partition", "hierarchy"))
    db[[.make_db_key(name, type)]] <- value
}

### * Consensus Method Registration.

cl_consensus_methods_db <- new.env()

get_cl_consensus_methods <- function(type)
    get_methods_from_db(cl_consensus_methods_db, type)
get_cl_consensus_method <- function(name, type) {
    get_method_from_db(cl_consensus_methods_db, type, name,
                       gettextf("Invalid consensus method '%s'.", name))
}
                       
set_cl_consensus_method <-
function(name, type, definition, ...)
{
    ## Register a @code{type} consensus method called @code{name} with
    ## definition @code{definition}.  Provide more information where
    ## appropriate, e.g., @code{dissimilarity} d and @code{exponent} e
    ## for methods minimizing \sum_b d(x_b, x) ^ e.

    put_method_into_db(cl_consensus_methods_db, type, name,
                       .structure(c(list(definition = definition),
                                    list(...)),
                                  class = "cl_consensus_method"))
}

set_cl_consensus_method("DWH", "partition",
                        .cl_consensus_partition_DWH,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("soft/euclidean", "partition",
                        .cl_consensus_partition_soft_euclidean,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("SE", "partition",
                        .cl_consensus_partition_soft_euclidean,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("hard/euclidean", "partition",
                        .cl_consensus_partition_hard_euclidean,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("HE", "partition",
                        .cl_consensus_partition_hard_euclidean,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("soft/manhattan", "partition",
                        .cl_consensus_partition_soft_manhattan,
                        dissimilarity = "manhattan",
                        exponent = 1)
set_cl_consensus_method("SM", "partition",
                        .cl_consensus_partition_soft_manhattan,
                        dissimilarity = "manhattan",
                        exponent = 1)
set_cl_consensus_method("hard/manhattan", "partition",
                        .cl_consensus_partition_hard_manhattan,
                        dissimilarity = "manhattan",
                        exponent = 1)
set_cl_consensus_method("HM", "partition",
                        .cl_consensus_partition_hard_manhattan,
                        dissimilarity = "manhattan",
                        exponent = 1)
set_cl_consensus_method("GV1", "partition",
                        .cl_consensus_partition_GV1,
                        dissimilarity = "GV1",
                        exponent = 2)
set_cl_consensus_method("GV3", "partition",
                        .cl_consensus_partition_GV3,
                        dissimilarity = "comemberships",
                        exponent = 2)
set_cl_consensus_method("soft/symdiff", "partition",
                        .cl_consensus_partition_soft_symdiff,
                        dissimilarity = "symdiff",
                        exponent = 1)
set_cl_consensus_method("hard/symdiff", "partition",
                        .cl_consensus_partition_hard_symdiff,
                        dissimilarity = "symdiff",
                        exponent = 1)

set_cl_consensus_method("cophenetic", "hierarchy",
                        .cl_consensus_hierarchy_cophenetic,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("euclidean", "hierarchy",
                        .cl_consensus_hierarchy_cophenetic,
                        dissimilarity = "euclidean",
                        exponent = 2)
set_cl_consensus_method("manhattan", "hierarchy",
                        .cl_consensus_hierarchy_manhattan,
                        dissimilarity = "manhattan",
                        exponent = 1)
set_cl_consensus_method("majority", "hierarchy",
                        .cl_consensus_hierarchy_majority,
                        dissimilarity = "symdiff",
                        exponent = 1)

### * Dissimilarity Method Registration.

cl_dissimilarity_methods_db <- new.env()

get_cl_dissimilarity_methods <- function(type)
    get_methods_from_db(cl_dissimilarity_methods_db, type)
get_cl_dissimilarity_method <- function(name, type)
    get_method_from_db(cl_dissimilarity_methods_db, type, name,
                       gettextf("Invalid dissimilarity method '%s'.",
                                name))
set_cl_dissimilarity_method <-
    function(name, type, definition, description, ...)
    put_method_into_db(cl_dissimilarity_methods_db, type, name,
                       .structure(c(list(definition = definition,
                                         description = description),
                                    list(...)),
                                  class = "cl_dissimilarity_method"))

set_cl_dissimilarity_method("euclidean", "partition",
                            .cl_dissimilarity_partition_euclidean,
                            "minimal Euclidean membership distance")
set_cl_dissimilarity_method("manhattan", "partition",
                            .cl_dissimilarity_partition_manhattan,
                            "minimal Manhattan membership distance")
set_cl_dissimilarity_method("comemberships", "partition",
                            .cl_dissimilarity_partition_comemberships,
                            "Euclidean comembership distance")
set_cl_dissimilarity_method("symdiff", "partition",
                            .cl_dissimilarity_partition_symdiff,
                            "symmetric difference distance")
set_cl_dissimilarity_method("Rand", "partition",
                            .cl_dissimilarity_partition_Rand,
                            "Rand distance")
set_cl_dissimilarity_method("GV1", "partition",
                            .cl_dissimilarity_partition_GV1,
                            "Gordon-Vichi Delta_1 dissimilarity")
set_cl_dissimilarity_method("BA/A", "partition",
                            .cl_dissimilarity_partition_BA_A,
                            "Boorman/Arabie minimum element moves distance")
set_cl_dissimilarity_method("BA/C", "partition",
                            .cl_dissimilarity_partition_BA_C,
                            "Boorman/Arabie minimum lattice moves distance")
set_cl_dissimilarity_method("BA/D", "partition",
                            .cl_dissimilarity_partition_BA_D,
                            "Boorman/Arabie pair-bonds distance")
set_cl_dissimilarity_method("BA/E", "partition",
                            .cl_dissimilarity_partition_BA_E,
                            "Boorman/Arabie normalized information distance")
set_cl_dissimilarity_method("VI", "partition",
                            .cl_dissimilarity_partition_VI,
                            "Variation of information")
set_cl_dissimilarity_method("Mallows", "partition",
                            .cl_dissimilarity_partition_Mallows,
                            "Mallows dissimilarity")
set_cl_dissimilarity_method("CSSD", "partition",
                            .cl_dissimilarity_partition_CSSD,
                            "Cluster Similarity Sensitive Distance")

set_cl_dissimilarity_method("euclidean", "hierarchy",
                            .cl_dissimilarity_hierarchy_euclidean,
                            "Euclidean ultrametric distance")
set_cl_dissimilarity_method("manhattan", "hierarchy",
                            .cl_dissimilarity_hierarchy_manhattan,
                            "Manhattan ultrametric distance")
set_cl_dissimilarity_method("cophenetic", "hierarchy",
                            .cl_dissimilarity_hierarchy_cophenetic,
                            "cophenetic correlations")
set_cl_dissimilarity_method("gamma", "hierarchy",
                            .cl_dissimilarity_hierarchy_gamma,
                            "rate of inversions")
set_cl_dissimilarity_method("symdiff", "hierarchy",
                            .cl_dissimilarity_hierarchy_symdiff,
                            "symmetric difference distance")
set_cl_dissimilarity_method("Chebyshev", "hierarchy",
                            .cl_dissimilarity_hierarchy_Chebyshev,
                            "Chebyshev distance")
set_cl_dissimilarity_method("Lyapunov", "hierarchy",
                            .cl_dissimilarity_hierarchy_Lyapunov,
                            "Lyapunov distance")
set_cl_dissimilarity_method("BO", "hierarchy",
                            .cl_dissimilarity_hierarchy_BO,
                            "Boorman/Olivier m_delta tree distance")
set_cl_dissimilarity_method("spectral", "hierarchy",
                            .cl_dissimilarity_hierarchy_spectral,
                            "spectral ultrametric distance")

### * Agreement Method Registration.

cl_agreement_methods_db <- new.env()
get_cl_agreement_methods <-
function(type)
    get_methods_from_db(cl_agreement_methods_db, type)
get_cl_agreement_method <-
function(name, type)
    get_method_from_db(cl_agreement_methods_db, type, name,
                       gettextf("Invalid agreement method '%s'.",
                                name))
set_cl_agreement_method <-
function(name, type, definition, description, ...)
    put_method_into_db(cl_agreement_methods_db, type, name,
                       .structure(c(list(definition = definition,
                                         description = description),
                                    list(...)),
                                  class = "cl_agreement_method"))

set_cl_agreement_method("euclidean", "partition",
                        .cl_agreement_partition_euclidean,
                        "minimal euclidean membership distance")
set_cl_agreement_method("manhattan", "partition",
                        .cl_agreement_partition_manhattan,
                        "minimal manhattan membership distance")
set_cl_agreement_method("Rand", "partition",
                        .cl_agreement_partition_Rand,
                        "Rand index")
set_cl_agreement_method("cRand", "partition",
                        .cl_agreement_partition_cRand,
                        "corrected Rand index")
set_cl_agreement_method("NMI", "partition",
                        .cl_agreement_partition_NMI,
                        "normalized mutual information")
set_cl_agreement_method("KP", "partition",
                        .cl_agreement_partition_KP,
                        "Katz-Powell index")
set_cl_agreement_method("angle", "partition",
                        .cl_agreement_partition_angle,
                        "maximal angle between memberships")
set_cl_agreement_method("diag", "partition",
                        .cl_agreement_partition_diag,
                        "maximal co-classification rate")
set_cl_agreement_method("FM", "partition",
                        .cl_agreement_partition_FM,
                        "Fowlkes-Mallows index")
set_cl_agreement_method("Jaccard", "partition",
                        .cl_agreement_partition_Jaccard,
                        "Jaccard index")
set_cl_agreement_method("purity", "partition",
                        .cl_agreement_partition_purity,
                        "purity")
set_cl_agreement_method("PS", "partition",
                        .cl_agreement_partition_PS,
                        "Prediction Strength")

set_cl_agreement_method("euclidean", "hierarchy",
                        .cl_agreement_hierarchy_euclidean,
                        "euclidean ultrametric distance")
set_cl_agreement_method("manhattan", "hierarchy",
                        .cl_agreement_hierarchy_manhattan,
                        "manhattan ultrametric distance")
set_cl_agreement_method("cophenetic", "hierarchy",
                        .cl_agreement_hierarchy_cophenetic,
                        "cophenetic correlations")
set_cl_agreement_method("angle", "hierarchy",
                        .cl_agreement_hierarchy_angle,
                        "angle between ultrametrics")
set_cl_agreement_method("gamma", "hierarchy",
                        .cl_agreement_hierarchy_gamma,
                        "rate of inversions")

cl_classes <-
function(x)
    UseMethod("cl_classes")

cl_classes.default <-
function(x)
{
    ## Be nice to users ...
    if(is.cl_partition(x))
        cl_classes(as.cl_partition(x))
    else if(is.cl_dendrogram(x))
        cl_classes(as.cl_dendrogram(x))
    else
        stop("Can only determine classes of partitions or hierarchies.")
}

cl_classes.cl_partition <-
function(x)
{
    n <- n_of_objects(x)
    out <- split(seq_len(n), cl_class_ids(x))
    class(out) <- c("cl_classes_of_partition_of_objects",
                    "cl_classes_of_objects")
    attr(out, "n_of_objects") <- n
    attr(out, "labels") <- cl_object_labels(x)
    out
}

cl_classes.cl_hierarchy <-
function(x)
{
    ## Assume a valid hierarchy/dendrogram.
    x <- as.hclust(x)
    n <- n_of_objects(x)
    labels <- seq_len(n)
    ## Only use the "maximal" partitions for each height (relevant in
    ## case of non-binary trees).
    groups <- cutree(x, h = unique(c(0, x$height)))
    ## Give a list with the (unique) sets of numbers of the objects.
    ## Note that objects may already be merged at height zero.
    out <- unique(unlist(c(as.list(labels),
                           lapply(split(groups, col(groups)),
                                  function(k) split(labels, k))),
                         recursive = FALSE,
                         use.names = FALSE))
    ## Preserve labels if possible, and re-order according to
    ## cardinality.
    out <- out[order(lengths(out))]
    class(out) <- c("cl_classes_of_hierarchy_of_objects",
                    "cl_classes_of_objects")
    attr(out, "n_of_objects") <- n
    attr(out, "labels") <- cl_object_labels(x)
    out
}

## Be nice to users of ultrametric fitters ... which should really fit
## dendrograms (which inherit from hierarchies).
cl_classes.cl_ultrametric <- cl_classes.cl_hierarchy

print.cl_classes_of_partition_of_objects <-
function(x, ...)
{
    labels <- attr(x, "labels")
    y <- lapply(x, function(i) paste(labels[i], collapse = ", "))
    writeLines(formatDL(names(x), sprintf("{%s}", unlist(y)),
                        style = "list", ...))
    invisible(x)
}

print.cl_classes_of_hierarchy_of_objects <-
function(x, ...)
{
    labels <- attr(x, "labels")
    y <- lapply(x, function(i) paste(labels[i], collapse = ", "))
    y <- strwrap(sprintf("{%s},", unlist(y)), exdent = 2)
    y[length(y)] <- sub(",$", "", y[length(y)])
    writeLines(y)
    invisible(x)
}

#! /bin/bash -e
########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

source ${virtualenvwrapper_path}

operation_mkvirtualenv()
{
    workon ${virtualenv_name} 2> /dev/null || mkvirtualenv ${virtualenv_name} || true
    local virtualenv_location=${VIRTUAL_ENV}
    deactivate
    ctx instance runtime-properties virtualenv_location ${virtualenv_location}
}

operation_${operation}

#! /bin/bash -e
########
# Copyright (c) 2016 GigaSpaces Technologies Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
############

print_git_branch_state()
{
    GIT_PS1_SHOWCOLORHINTS=true
    GIT_PS1_SHOWDIRTYSTATE=true
    GIT_PS1_SHOWSTASHSTATE=true
    GIT_PS1_SHOWUNTRACKEDFILES=true
    GIT_PS1_SHOWUPSTREAM=auto
    source {{git_prompt_path}}
    cd {{repo_location}}
    __git_ps1 "" "" "%s" || true
    export PS1
    export __git_ps1_branch_name
    program='''
import os
import sys
output = os.environ["PS1"]
output = output.replace("\\[", "")
output = output.replace("\\]", "")
branch_name = os.environ.get("__git_ps1_branch_name")
if branch_name:
    output = output.replace("${__git_ps1_branch_name}", branch_name)
sys.stdout.write(output)
'''
    output=$(python -c "$program")
    echo -e "${output}"
}

print_git_branch_state

\documentclass[fleqn]{article}
\usepackage[round,longnamesfirst]{natbib}
\usepackage{graphicx,keyval,hyperref,doi}

\newcommand\argmin{\mathop{\mathrm{arg min}}}
\newcommand\trace{\mathop{\mathrm{tr}}}
\newcommand\R{{\mathbb{R}}}
\newcommand{\pkg}[1]{{\normalfont\fontseries{b}\selectfont #1}}
\newcommand{\sQuote}[1]{`{#1}'}
\newcommand{\dQuote}[1]{``{#1}''}
\let\code=\texttt
\newcommand{\file}[1]{\sQuote{\textsf{#1}}}
\newcommand{\class}[1]{\code{"#1"}}

\SweaveOpts{strip.white=true}

\AtBeginDocument{\setkeys{Gin}{width=0.6\textwidth}}

\date{2007-06-28}
\title{A CLUE for CLUster Ensembles}
\author{Kurt Hornik}
%% \VignetteIndexEntry{CLUster Ensembles}

\sloppy{}

\begin{document}
\maketitle

\begin{abstract}
  Cluster ensembles are collections of individual solutions to a given
  clustering problem which are useful or necessary to consider in a wide
  range of applications.  The R package~\pkg{clue} provides an
  extensible computational environment for creating and analyzing
  cluster ensembles, with basic data structures for representing
  partitions and hierarchies, and facilities for computing on these,
  including methods for measuring proximity and obtaining consensus and
  ``secondary'' clusterings.
\end{abstract}

<<echo=FALSE>>=
options(width = 60)
library("clue")
@ %

\section{Introduction}
\label{sec:introduction}

\emph{Cluster ensembles} are collections of clusterings, which are all
of the same ``kind'' (e.g., collections of partitions, or collections of
hierarchies), of a set of objects.  Such ensembles can be obtained, for
example, by varying the (hyper)parameters of a ``base'' clustering
algorithm, by resampling or reweighting the set of objects, or by
employing several different base clusterers.

Questions of ``agreement'' in cluster ensembles, and obtaining
``consensus'' clusterings from it, have been studied in several
scientific communities for quite some time now.  A special issue of the
Journal of Classification was devoted to ``Comparison and Consensus of
Classifications'' \citep{cluster:Day:1986} almost two decades ago.  The
recent popularization of ensemble methods such as Bayesian model
averaging \citep{cluster:Hoeting+Madigan+Raftery:1999}, bagging
\citep{cluster:Breiman:1996} and boosting
\citep{cluster:Friedman+Hastie+Tibshirani:2000}, typically in a
supervised leaning context, has also furthered the research interest in
using ensemble methods to improve the quality and robustness of cluster
solutions.  Cluster ensembles can also be utilized to aggregate base
results over conditioning or grouping variables in multi-way data, to
reuse existing knowledge, and to accommodate the needs of distributed
computing, see e.g.\ \cite{cluster:Hornik:2005a} and
\cite{cluster:Strehl+Ghosh:2003a} for more information.

Package~\pkg{clue} is an extension package for R~\citep{cluster:R:2005}
providing a computational environment for creating and analyzing cluster
ensembles.  In Section~\ref{sec:structures+algorithms}, we describe the
underlying data structures, and the functionality for measuring
proximity, obtaining consensus clusterings, and ``secondary''
clusterings.  Four examples are discussed in Section~\ref{sec:examples}.
Section~\ref{sec:outlook} concludes the paper.

A previous version of this manuscript was published in the \emph{Journal
  of Statistical Software} \citep{cluster:Hornik:2005b}.


\section{Data structures and algorithms}
\label{sec:structures+algorithms}

\subsection{Partitions and hierarchies}

Representations of clusterings of objects greatly vary across the
multitude of methods available in R packages.  For example, the class
ids (``cluster labels'') for the results of \code{kmeans()} in base
package~\pkg{stats}, \code{pam()} in recommended
package~\pkg{cluster}~\citep{cluster:Rousseeuw+Struyf+Hubert:2005,
  cluster:Struyf+Hubert+Rousseeuw:1996}, and \code{Mclust()} in
package~\pkg{mclust}~\citep{cluster:Fraley+Raftery+Wehrens:2005,
  cluster:Fraley+Raftery:2003}, are available as components named
\code{cluster}, \code{clustering}, and \code{classification},
respectively, of the R objects returned by these functions.  In many
cases, the representations inherit from suitable classes.  (We note that
for versions of R prior to 2.1.0, \code{kmeans()} only returned a
``raw'' (unclassed) result, which was changed alongside the development
of \pkg{clue}.)

We deal with this heterogeneity of representations by providing getters
for the key underlying data, such as the number of objects from which a
clustering was obtained, and predicates, e.g.\ for determining whether
an R object represents a partition of objects or not.  These getters,
such as \code{n\_of\_objects()}, and predicates are implemented as S3
generics, so that there is a \emph{conceptual}, but no formal class
system underlying the predicates.  Support for classed representations
can easily be added by providing S3 methods.

\subsubsection{Partitions}

The partitions considered in \pkg{clue} are possibly soft (``fuzzy'')
partitions, where for each object~$i$ and class~$j$ there is a
non-negative number~$\mu_{ij}$ quantifying the ``belongingness'' or
\emph{membership} of object~$i$ to class~$j$, with $\sum_j \mu_{ij} =
1$.  For hard (``crisp'') partitions, all $\mu_{ij}$ are in $\{0, 1\}$.
We can gather the $\mu_{ij}$ into the \emph{membership matrix} $M =
[\mu_{ij}]$, where rows correspond to objects and columns to classes.
The \emph{number of classes} of a partition, computed by function
\code{n\_of\_classes()}, is the number of $j$ for which $\mu_{ij} > 0$
for at least one object~$i$.  This may be less than the number of
``available'' classes, corresponding to the number of columns in a
membership matrix representing the partition.

The predicate functions \code{is.cl\_partition()},
\code{is.cl\_hard\_partition()}, and \code{is.cl\_soft\_partition()} are
used to indicate whether R objects represent partitions of objects of
the respective kind, with hard partitions as characterized above (all
memberships in $\{0, 1\}$).  (Hence, ``fuzzy clustering'' algorithms can
in principle also give a hard partition.)  \code{is.cl\_partition()} and
\code{is.cl\_hard\_partition()} are generic functions;
\code{is.cl\_soft\_partition()} gives true iff \code{is.cl\_partition()}
is true and \code{is.cl\_hard\_partition()} is false.

For R objects representing partitions, function \code{cl\_membership()}
computes an R object with the membership values, currently always as a
dense membership matrix with additional attributes.  This is obviously
rather inefficient for computations on hard partitions; we are planning
to add ``canned'' sparse representations (using the vector of class ids)
in future versions.  Function \code{as.cl\_membership()} can be used for
coercing \dQuote{raw} class ids (given as atomic vectors) or membership
values (given as numeric matrices) to membership objects.

Function \code{cl\_class\_ids()} determines the class ids of a
partition.  For soft partitions, the class ids returned are those of the
\dQuote{nearest} hard partition obtained by taking the class ids of the
(first) maximal membership values.  Note that the cardinality of the set
of the class ids may be less than the number of classes in the (soft)
partition.

Many partitioning methods are based on \emph{prototypes} (``centers'').
In typical cases, these are points~$p_j$ in the same feature space the
measurements~$x_i$ on the objects~$i$ to be partitioned are in, so that
one can measure distance between objects and prototypes, and e.g.\
classify objects to their closest prototype.  Such partitioning methods
can also induce partitions of the entire feature space (rather than
``just'' the set of objects to be partitioned).  Currently, package
\pkg{clue} has only minimal support for this ``additional'' structure,
providing a \code{cl\_prototypes()} generic for extracting the
prototypes, and is mostly focused on computations on partitions which
are based on their memberships.

Many algorithms resulting in partitions of a given set of objects can be
taken to induce a partition of the underlying feature space for the
measurements on the objects, so that class memberships for ``new''
objects can be obtained from the induced partition.  Examples include
partitions based on assigning objects to their ``closest'' prototypes,
or providing mixture models for the distribution of objects in feature
space.  Package~\pkg{clue} provides a \code{cl\_predict()} generic for
predicting the class memberships of new objects (if possible).

Function \code{cl\_fuzziness()} computes softness (fuzziness) measures
for (ensembles) of partitions.  Built-in measures are the partition
coefficient
\label{PC}
and partition entropy \citep[e.g.,][]{cluster:Bezdek:1981}, with an
option to normalize in a way that hard partitions and the ``fuzziest''
possible partition (where all memberships are the same) get fuzziness
values of zero and one, respectively.  Note that this normalization
differs from ``standard'' ones in the literature.

In the sequel, we shall also use the concept of the \emph{co-membership
  matrix} $C(M) = M M'$, where $'$ denotes matrix transposition, of a
partition.  For hard partitions, an entry $c_{ij}$ of $C(M)$ is 1 iff
the corresponding objects $i$ and $j$ are in the same class, and 0
otherwise.

\subsubsection{Hierarchies}

The hierarchies considered in \pkg{clue} are \emph{total indexed
  hierarchies}, also known as \emph{$n$-valued trees}, and hence
correspond in a one-to-one manner to \emph{ultrametrics} (distances
$u_{ij}$ between pairs of objects $i$ and $j$ which satisfy the
ultrametric constraint $u_{ij} = \max(u_{ik}, u_{jk})$ for all triples
$i$, $j$, and $k$).  See e.g.~\citet[Page~69--71]{cluster:Gordon:1999}.

Function \code{cl\_ultrametric(x)} computes the associated ultrametric
from an R object \code{x} representing a hierarchy of objects.  If
\code{x} is not an ultrametric, function \code{cophenetic()} in base
package~\pkg{stats} is used to obtain the ultrametric (also known as
cophenetic) distances from the hierarchy, which in turn by default calls
the S3 generic \code{as.hclust()} (also in \pkg{stats}) on the
hierarchy.  Support for classes which represent hierarchies can thus be
added by providing \code{as.hclust()} methods for this class.  In
R~2.1.0 or better (again as part of the work on \pkg{clue}),
\code{cophenetic} is an S3 generic as well, and one can also more
directly provide methods for this if necessary.

In addition, there is a generic function \code{as.cl\_ultrametric()}
which can be used for coercing \emph{raw} (non-classed) ultrametrics,
represented as numeric vectors (of the lower-half entries) or numeric
matrices, to ultrametric objects.  Finally, the generic predicate
function \code{is.cl\_hierarchy()} is used to determine whether an R
object represents a hierarchy or not.

Ultrametric objects can also be coerced to classes~\class{dendrogram}
and \class{hclust} (from base package~\pkg{stats}), and hence in
particular use the \code{plot()} methods for these classes.  By default,
plotting an ultrametric object uses the plot method for dendrograms.

Obtaining a hierarchy on a given set of objects can be thought of as
transforming the pairwise dissimilarities between the objects (which
typically do not yet satisfy the ultrametric constraints) into an
ultrametric.  Ideally, this ultrametric should be as close as possible
to the dissimilarities.  In some important cases, explicit solutions are
possible (e.g., ``standard'' hierarchical clustering with single or
complete linkage gives the optimal ultrametric dominated by or
dominating the dissimilarities, respectively).  On the other hand, the
problem of finding the closest ultrametric in the least squares sense is
known to be NP-hard
\citep{cluster:Krivanek+Moravek:1986,cluster:Krivanek:1986}.  One
important class of heuristics for finding least squares fits is based on
iterative projection on convex sets of constraints
\citep{cluster:Hubert+Arabie:1995}.

\label{SUMT}
Function \code{ls\_fit\_ultrametric()} follows
\cite{cluster:DeSoete:1986} to use an SUMT \citep[Sequential
Unconstrained Minimization Technique;][]{cluster:Fiacco+McCormick:1968}
approach in turn simplifying the suggestions in
\cite{cluster:Carroll+Pruzansky:1980}.  Let $L(u)$ be the function to be
minimized over all $u$ in some constrained set $\mathcal{U}$---in our
case, $L(u) = \sum (d_{ij}-u_{ij})^2$ is the least squares criterion,
and $\mathcal{U}$ is the set of all ultrametrics $u$.  One iteratively
minimizes $L(u) + \rho_k P(u)$, where $P(u)$ is a non-negative function
penalizing violations of the constraints such that $P(u)$ is zero iff $u
\in \mathcal{U}$.  The $\rho$ values are increased according to the rule
$\rho_{k+1} = q \rho_k$ for some constant $q > 1$, until convergence is
obtained in the sense that e.g.\ the Euclidean distance between
successive solutions $u_k$ and $u_{k+1}$ is small enough.  Optionally,
the final $u_k$ is then suitably projected onto $\mathcal{U}$.

For \code{ls\_fit\_ultrametric()}, we obtain the starting value $u_0$ by
\dQuote{random shaking} of the given dissimilarity object, and use the
penalty function $P(u) = \sum_{\Omega} (u_{ij} - u_{jk}) ^ 2$, were
$\Omega$ contains all triples $i, j, k$ for which $u_{ij} \le
\min(u_{ik}, u_{jk})$ and $u_{ik} \ne u_{jk}$, i.e., for which $u$
violates the ultrametric constraints.  The unconstrained minimizations
are carried out using either \code{optim()} or \code{nlm()} in base
package~\pkg{stats}, with analytic gradients given in
\cite{cluster:Carroll+Pruzansky:1980}.  This ``works'', even though we
note however that $P$ is not even a continuous function, which seems to
have gone unnoticed in the literature!  (Consider an ultrametric $u$ for
which $u_{ij} = u_{ik} < u_{jk}$ for some $i, j, k$ and define
$u(\delta)$ by changing the $u_{ij}$ to $u_{ij} + \delta$.  For $u$,
both $(i,j,k)$ and $(j,i,k)$ are in the violation set $\Omega$, whereas
for all $\delta$ sufficiently small, only $(j,i,k)$ is the violation set
for $u(\delta)$.  Hence, $\lim_{\delta\to 0} P(u(\delta)) = P(u) +
(u_{ij} - u_{ik})^2$.  This shows that $P$ is discontinuous at all
non-constant $u$ with duplicated entries.  On the other hand, it is
continuously differentiable at all $u$ with unique entries.)  Hence, we
need to turn off checking analytical gradients when using \code{nlm()}
for minimization.

The default optimization using conjugate gradients should work
reasonably well for medium to large size problems.  For \dQuote{small}
ones, using \code{nlm()} is usually faster.  Note that the number of
ultrametric constraints is of the order $n^3$, suggesting to use the
SUMT approach in favor of \code{constrOptim()} in \pkg{stats}.  It
should be noted that the SUMT approach is a heuristic which can not be
guaranteed to find the global minimum.  Standard practice would
recommend to use the best solution found in \dQuote{sufficiently many}
replications of the base algorithm.

\subsubsection{Extensibility}

The methods provided in package~\pkg{clue} handle the partitions and
hierarchies obtained from clustering functions in the base R
distribution, as well as packages
\pkg{RWeka}~\citep{cluster:Hornik+Hothorn+Karatzoglou:2006},
\pkg{cba}~\citep{cluster:Buchta+Hahsler:2005},
\pkg{cclust}~\citep{cluster:Dimitriadou:2005}, \pkg{cluster},
\pkg{e1071}~\citep{cluster:Dimitriadou+Hornik+Leisch:2005},
\pkg{flexclust}~\citep{cluster:Leisch:2006a},
\pkg{flexmix}~\citep{cluster:Leisch:2004},
\pkg{kernlab}~\citep{cluster:Karatzoglou+Smola+Hornik:2004},
and \pkg{mclust} (and of course, \pkg{clue} itself).

Extending support to other packages is straightforward, provided that
clusterings are instances of classes.  Suppose e.g.\ that a package has
a function \code{glvq()} for ``generalized'' (i.e., non-Euclidean)
Learning Vector Quantization which returns an object of
class~\class{glvq}, in turn being a list with component
\code{class\_ids} containing the class ids.  To integrate this into the
\pkg{clue} framework, all that is necessary is to provide the following
methods.
<<>>=
cl_class_ids.glvq <-
function(x)
    as.cl_class_ids(x$class_ids)
is.cl_partition.glvq <-
function(x)
    TRUE
is.cl_hard_partition.glvq <-
function(x)
    TRUE
@ % $

\subsection{Cluster ensembles}

Cluster ensembles are realized as lists of clusterings with additional
class information.  All clusterings in an ensemble must be of the same
``kind'' (i.e., either all partitions as known to
\code{is.cl\_partition()}, or all hierarchies as known to
\code{is.cl\_hierarchy()}, respectively), and have the same number of
objects.  If all clusterings are partitions, the list realizing the
ensemble has class~\class{cl\_partition\_ensemble} and inherits from
\class{cl\_ensemble}; if all clusterings are hierarchies, it has
class~\class{cl\_hierarchy\_ensemble} and inherits from
\class{cl\_ensemble}.  Empty ensembles cannot be categorized according
to the kind of clusterings they contain, and hence only have
class~\class{cl\_ensemble}.

Function \code{cl\_ensemble()} creates a cluster ensemble object from
clusterings given either one-by-one, or as a list passed to the
\code{list} argument.  As unclassed lists could be used to represent
single clusterings (in particular for results from \code{kmeans()} in
versions of R prior to 2.1.0), we prefer not to assume that an unnamed
given list is a list of clusterings.  \code{cl\_ensemble()} verifies
that all given clusterings are of the same kind, and all have the same
number of objects.  (By the notion of cluster ensembles, we should in
principle verify that the clusterings come from the \emph{same} objects,
which of course is not always possible.)

The list representation makes it possible to use \code{lapply()} for
computations on the individual clusterings in (i.e., the components of)
a cluster ensemble.

Available methods for cluster ensembles include those for subscripting,
\code{c()}, \code{rep()}, \code{print()}, and \code{unique()}, where the
last is based on a \code{unique()} method for lists added in R~2.1.1,
and makes it possible to find unique and duplicated elements in cluster
ensembles.  The elements of the ensemble can be tabulated using
\code{cl\_tabulate()}.

Function \code{cl\_boot()} generates cluster ensembles with bootstrap
replicates of the results of applying a \dQuote{base} clustering
algorithm to a given data set.  Currently, this is a rather
simple-minded function with limited applicability, and mostly useful for
studying the effect of (uncontrolled) random initializations of
fixed-point partitioning algorithms such as \code{kmeans()} or
\code{cmeans()} in package~\pkg{e1071}.  To study the effect of varying
control parameters or explicitly providing random starting values, the
respective cluster ensemble has to be generated explicitly (most
conveniently by using \code{replicate()} to create a list \code{lst} of
suitable instances of clusterings obtained by the base algorithm, and
using \code{cl\_ensemble(list = lst)} to create the ensemble).
Resampling the training data is possible for base algorithms which can
predict the class memberships of new data using \code{cl\_predict}
(e.g., by classifying the out-of-bag data to their closest prototype).
In fact, we believe that for unsupervised learning methods such as
clustering, \emph{reweighting} is conceptually superior to resampling,
and have therefore recently enhanced package~\pkg{e1071} to provide an
implementation of weighted fuzzy $c$-means, and package~\pkg{flexclust}
contains an implementation of weighted $k$-means.  We are currently
experimenting with interfaces for providing ``direct'' support for
reweighting via \code{cl\_boot()}.

\subsection{Cluster proximities}

\subsubsection{Principles}

Computing dissimilarities and similarities (``agreements'') between
clusterings of the same objects is a key ingredient in the analysis of
cluster ensembles.  The ``standard'' data structures available for such
proximity data (measures of similarity or dissimilarity) are
classes~\class{dist} and \class{dissimilarity} in package~\pkg{cluster}
(which basically, but not strictly, extends \class{dist}), and are both
not entirely suited to our needs.  First, they are confined to
\emph{symmetric} dissimilarity data.  Second, they do not provide enough
reflectance.  We also note that the Bioconductor
package~\pkg{graph}~\citep{cluster:Gentleman+Whalen:2005} contains an
efficient subscript method for objects of class~\class{dist}, but
returns a ``raw'' matrix for row/column subscripting.

For package~\pkg{clue}, we use the following approach.  There are
classes for symmetric and (possibly) non-symmetric proximity data
(\class{cl\_proximity} and \class{cl\_cross\_proximity}), which, in
addition to holding the numeric data, also contain a description
``slot'' (attribute), currently a character string, as a first
approximation to providing more reflectance.  Internally, symmetric
proximity data are store the lower diagonal proximity values in a
numeric vector (in row-major order), i.e., the same way as objects of
class~\class{dist}; a \code{self} attribute can be used for diagonal
values (in case some of these are non-zero).  Symmetric proximity
objects can be coerced to dense matrices using \code{as.matrix()}.  It
is possible to use 2-index matrix-style subscripting for symmetric
proximity objects; unless this uses identical row and column indices, it
results in a non-symmetric proximity object.

This approach ``propagates'' to classes for symmetric and (possibly)
non-symmetric cluster dissimilarity and agreement data (e.g.,
\class{cl\_dissimilarity} and \class{cl\_cross\_dissimilarity} for
dissimilarity data), which extend the respective proximity classes.

Ultrametric objects are implemented as symmetric proximity objects with
a dissimilarity interpretation so that self-proximities are zero, and
inherit from classes~\class{cl\_dissimilarity} and
\class{cl\_proximity}.

Providing reflectance is far from optimal.  For example, if \code{s} is
a similarity object (with cluster agreements), \code{1 - s} is a
dissimilarity one, but the description is preserved unchanged.  This
issue could be addressed by providing high-level functions for
transforming proximities.

\label{synopsis}
Cluster dissimilarities are computed via \code{cl\_dissimilarity()} with
synopsis \code{cl\_dissimilarity(x, y = NULL, method = "euclidean")},
where \code{x} and \code{y} are cluster ensemble objects or coercible to
such, or \code{NULL} (\code{y} only).  If \code{y} is \code{NULL}, the
return value is an object of class~\class{cl\_dissimilarity} which
contains the dissimilarities between all pairs of clusterings in
\code{x}.  Otherwise, it is an object of
class~\class{cl\_cross\_dissimilarity} with the dissimilarities between
the clusterings in \code{x} and the clusterings in \code{y}.  Formal
argument \code{method} is either a character string specifying one of
the built-in methods for computing dissimilarity, or a function to be
taken as a user-defined method, making it reasonably straightforward to
add methods.

Function \code{cl\_agreement()} has the same interface as
\code{cl\_dissimilarity()}, returning cluster similarity objects with
respective classes~\class{cl\_agreement} and
\class{cl\_cross\_agreement}.  Built-in methods for computing
dissimilarities may coincide (in which case they are transforms of each
other), but do not necessarily do so, as there typically are no
canonical transformations.  E.g., according to needs and scientific
community, agreements might be transformed to dissimilarities via $d = -
\log(s)$ or the square root thereof
\citep[e.g.,][]{cluster:Strehl+Ghosh:2003b}, or via $d = 1 - s$.

\subsubsection{Partition proximities}

When assessing agreement or dissimilarity of partitions, one needs to
consider that the class ids may be permuted arbitrarily without changing
the underlying partitions.  For membership matrices~$M$, permuting class
ids amounts to replacing $M$ by $M \Pi$, where $\Pi$ is a suitable
permutation matrix.  We note that the co-membership matrix $C(M) = MM'$
is unchanged by these transformations; hence, proximity measures based
on co-occurrences, such as the Katz-Powell
\citep{cluster:Katz+Powell:1953} or Rand \citep{cluster:Rand:1971}
indices, do not explicitly need to adjust for possible re-labeling.  The
same is true for measures based on the ``confusion matrix'' $M'
\tilde{M}$ of two membership matrices $M$ and $\tilde{M}$ which are
invariant under permutations of rows and columns, such as the Normalized
Mutual Information (NMI) measure introduced in
\cite{cluster:Strehl+Ghosh:2003a}.  Other proximity measures need to
find permutations so that the classes are optimally matched, which of
course in general requires exhaustive search through all $k!$ possible
permutations, where $k$ is the (common) number of classes in the
partitions, and thus will typically be prohibitively expensive.
Fortunately, in some important cases, optimal matchings can be
determined very efficiently.  We explain this in detail for
``Euclidean'' partition dissimilarity and agreement (which in fact is
the default measure used by \code{cl\_dissimilarity()} and
\code{cl\_agreement()}).

Euclidean partition dissimilarity
\citep{cluster:Dimitriadou+Weingessel+Hornik:2002} is defined as
\begin{displaymath}
  d(M, \tilde{M}) = \min\nolimits_\Pi \| M - \tilde{M} \Pi \|
\end{displaymath}
where the minimum is taken over all permutation matrices~$\Pi$,
$\|\cdot\|$ is the Frobenius norm (so that $\|Y\|^2 = \trace(Y'Y)$), and
$n$ is the (common) number of objects in the partitions.  As $\| M -
\tilde{M} \Pi \|^2 = \trace(M'M) - 2 \trace(M'\tilde{M}\Pi) +
\trace(\Pi'\tilde{M}'\tilde{M}\Pi) = \trace(M'M) - 2
\trace(M'\tilde{M}\Pi) + \trace(\tilde{M}'\tilde{M})$, we see that
minimizing $\| M - \tilde{M} \Pi \|^2$ is equivalent to maximizing
$\trace(M'\tilde{M}\Pi) = \sum_{i,k}{\mu_{ik}\tilde{\mu}}_{i,\pi(k)}$,
which for hard partitions is the number of objects with the same label
in the partitions given by $M$ and $\tilde{M}\Pi$.  Finding the optimal
$\Pi$ is thus recognized as an instance of the \emph{linear sum
  assignment problem} (LSAP, also known as the weighted bipartite graph
matching problem).  The LSAP can be solved by linear programming, e.g.,
using Simplex-style primal algorithms as done by
function~\code{lp.assign()} in
package~\pkg{lpSolve}~\citep{cluster:Buttrey:2005}, but primal-dual
algorithms such as the so-called Hungarian method can be shown to find
the optimum in time $O(k^3)$
\citep[e.g.,][]{cluster:Papadimitriou+Steiglitz:1982}.  Available
published implementations include TOMS 548
\citep{cluster:Carpaneto+Toth:1980}, which however is restricted to
integer weights and $k < 131$.  One can also transform the LSAP into a
network flow problem, and use e.g.~RELAX-IV
\citep{cluster:Bertsekas+Tseng:1994} for solving this, as is done in
package~\pkg{optmatch}~\citep{cluster:Hansen:2005}.  In
package~\pkg{clue}, we use an efficient C implementation of the
Hungarian algorithm kindly provided to us by Walter B\"ohm, which has
been found to perform very well across a wide range of problem sizes.

\cite{cluster:Gordon+Vichi:2001} use a variant of Euclidean
dissimilarity (``GV1 dissimilarity'') which is based on the sum of the
squared difference of the memberships of matched (non-empty) classes
only, discarding the unmatched ones (see their Example~2).  This results
in a measure which is discontinuous over the space of soft partitions
with arbitrary numbers of classes.

The partition agreement measures ``angle'' and ``diag'' (maximal cosine
of angle between the memberships, and maximal co-classification rate,
where both maxima are taken over all column permutations of the
membership matrices) are based on solving the same LSAP as for Euclidean
dissimilarity.

Finally, Manhattan partition dissimilarity is defined as the minimal sum
of the absolute differences of $M$ and all column permutations of
$\tilde{M}$, and can again be computed efficiently by solving an LSAP.

For hard partitions, both Manhattan and squared Euclidean dissimilarity
give twice the \emph{transfer distance}
\citep{cluster:Charon+Denoeud+Guenoche:2006}, which is the minimum
number of objects that must be removed so that the implied partitions
(restrictions to the remaining objects) are identical.  This is also
known as the \emph{$R$-metric} in \cite{cluster:Day:1981}, i.e., the
number of augmentations and removals of single objects needed to
transform one partition into the other, and the
\emph{partition-distance} in \cite{cluster:Gusfield:2002}.

Note when assessing proximity that agreements for soft partitions are
always (and quite often considerably) lower than the agreements for the
corresponding nearest hard partitions, unless the agreement measures are
based on the latter anyways (as currently done for Rand, Katz-Powell,
and NMI).

Package~\pkg{clue} provides additional agreement measures, such as the
Jaccard and Fowles-Mallows \citep[quite often incorrectly attributed to
\cite{cluster:Wallace:1983}]{cluster:Fowlkes+Mallows:1983a} indices, and
dissimilarity measures such as the ``symdiff'' and Rand distances (the
latter is proportional to the metric of \cite{cluster:Mirkin:1996}) and
the metrics discussed in \cite{cluster:Boorman+Arabie:1972}.  One could
easily add more proximity measures, such as the ``Variation of
Information'' \citep{cluster:Meila:2003}.  However, all these measures
are rigorously defined for hard partitions only.  To see why extensions
to soft partitions are far from straightforward, consider e.g.\ measures
based on the confusion matrix.  Its entries count the cardinality of
certain intersections of sets.
\label{fuzzy} In a fuzzy context for soft partitions, a natural
generalization would be using fuzzy cardinalities (i.e., sums of
memberships values) of fuzzy intersections instead.  There are many
possible choices for the latter, with the product of the membership
values (corresponding to employing the confusion matrix also in the
fuzzy case) one of them, but the minimum instead of the product being
the ``usual'' choice.  A similar point can be made for co-occurrences of
soft memberships.  We are not aware of systematic investigations of
these extension issues.

\subsubsection{Hierarchy proximities}

Available built-in dissimilarity measures for hierarchies include
\emph{Euclidean} (again, the default measure used by
\code{cl\_dissimilarity()}) and Manhattan dissimilarity, which are
simply the Euclidean (square root of the sum of squared differences) and
Manhattan (sum of the absolute differences) dissimilarities between the
associated ultrametrics.  Cophenetic dissimilarity is defined as $1 -
c^2$, where $c$ is the cophenetic correlation coefficient
\citep{cluster:Sokal+Rohlf:1962}, i.e., the Pearson product-moment
correlation between the ultrametrics.  Gamma dissimilarity is the rate
of inversions between the associated ultrametrics $u$ and $v$ (i.e., the
rate of pairs $(i,j)$ and $(k,l)$ for which $u_{ij} < u_{kl}$ and
$v_{ij} > v_{kl}$).  This measure is a linear transformation of
Kruskal's~$\gamma$.  Finally, symdiff dissimilarity is the cardinality
of the symmetric set difference of the sets of classes (hierarchies in
the strict sense) induced by the dendrograms.

Associated agreement measures are obtained by suitable transformations
of the dissimilarities~$d$; for Euclidean proximities, we prefer to use
$1 / (1 + d)$ rather than e.g.\ $\exp(-d)$.

One should note that whereas cophenetic and gamma dissimilarities are
invariant to linear transformations, Euclidean and Manhattan ones are
not.  Hence, if only the relative ``structure'' of the dendrograms is of
interest, these dissimilarities should only be used after transforming
the ultrametrics to a common range of values (e.g., to $[0,1]$).

\subsection{Consensus clusterings}

Consensus clusterings ``synthesize'' the information in the elements of
a cluster ensemble into a single clustering.  There are three main
approaches to obtaining consensus clusterings
\citep{cluster:Hornik:2005a,cluster:Gordon+Vichi:2001}: in the
\emph{constructive} approach, one specifies a way to construct a
consensus clustering.  In the \emph{axiomatic} approach, emphasis is on
the investigation of existence and uniqueness of consensus clusterings
characterized axiomatically.  The \emph{optimization} approach
formalizes the natural idea of describing consensus clusterings as the
ones which ``optimally represent the ensemble'' by providing a criterion
to be optimized over a suitable set $\mathcal{C}$ of possible consensus
clusterings.  If $d$ is a dissimilarity measure and $C_1, \ldots, C_B$
are the elements of the ensemble, one can e.g.\ look for solutions of
the problem
\begin{displaymath}
  \sum\nolimits_{b=1}^B w_b d(C, C_b) ^ p
  \Rightarrow \min\nolimits_{C \in \mathcal{C}},
\end{displaymath}
for some $p \ge 0$, i.e., as clusterings~$C^*$ minimizing weighted
average dissimilarity powers of order~$p$.  Analogously, if a similarity
measure is given, one can look for clusterings maximizing weighted
average similarity powers.  Following \cite{cluster:Gordon+Vichi:1998},
an above $C^*$ is referred to as (weighted) \emph{median} or
\emph{medoid} clustering if $p = 1$ and the optimum is sought over the
set of all possible base clusterings, or the set $\{ C_1, \ldots, C_B
\}$ of the base clusterings, respectively.  For $p = 2$, we have
\emph{least squares} consensus clusterings (generalized means).

For computing consensus clusterings, package~\pkg{clue} provides
function \code{cl\_consensus()} with synopsis \code{cl\_consensus(x,
  method = NULL, weights = 1, control = list())}.  This allows (similar
to the functions for computing cluster proximities, see
Section~\ref{synopsis} on Page~\pageref{synopsis}) argument
\code{method} to be a character string specifying one of the built-in
methods discussed below, or a function to be taken as a user-defined
method (taking an ensemble, the case weights, and a list of control
parameters as its arguments), again making it reasonably straightforward
to add methods.  In addition, function~\code{cl\_medoid()} can be used
for obtaining medoid partitions (using, in principle, arbitrary
dissimilarities).  Modulo possible differences in the case of ties, this
gives the same results as (the medoid obtained by) \code{pam()} in
package~\pkg{cluster}.

If all elements of the ensemble are partitions, package~\pkg{clue}
provides algorithms for computing soft least squares consensus
partitions for weighted Euclidean, GV1 and co-membership
dissimilarities.  Let $M_1, \ldots, M_B$ and $M$ denote the membership
matrices of the elements of the ensemble and their sought least squares
consensus partition, respectively.  For Euclidean dissimilarity, we need
to find
\begin{displaymath}
  \sum_b w_b \min\nolimits_{\Pi_b} \| M - M_b \Pi_b \|^2
  \Rightarrow \min\nolimits_M
\end{displaymath}
over all membership matrices (i.e., stochastic matrices) $M$, or
equivalently,
\begin{displaymath}
  \sum_b w_b \| M - M_b \Pi_b \|^2 
  \Rightarrow \min\nolimits_{M, \Pi_1, \ldots, \Pi_B}
\end{displaymath}
over all $M$ and permutation matrices $\Pi_1, \ldots, \Pi_B$.  Now fix
the $\Pi_b$ and let $\bar{M} = s^{-1} \sum_b w_b M_b \Pi_b$ be the
weighted average of the $M_b \Pi_b$, where $s = \sum_b w_b$.  Then
\begin{eqnarray*}
  \lefteqn{\sum_b w_b \| M - M_b \Pi_b \|^2} \\
  &=& \sum_b w_b (\|M\|^2 - 2 \trace(M' M_b \Pi_b) + \|M_b\Pi_b\|^2) \\
  &=& s \|M\|^2 - 2 s \trace(M' \bar{M}) + \sum_b w_b \|M_b\|^2 \\
  &=& s (\|M - \bar{M}\|^2) + \sum_b w_b \|M_b\|^2 - s \|\bar{M}\|^2
\end{eqnarray*}
Thus, as already observed in
\cite{cluster:Dimitriadou+Weingessel+Hornik:2002} and
\cite{cluster:Gordon+Vichi:2001}, 
for fixed permutations $\Pi_b$ the optimal soft $M$ is given by
$\bar{M}$.  The optimal permutations can be found by minimizing $- s
\|\bar{M}\|^2$, or equivalently, by maximizing
\begin{displaymath}
  s^2 \|\bar{M}\|^2
  = \sum_{\beta, b} w_\beta w_b \trace(\Pi_\beta'M_\beta'M_b\Pi_b).
\end{displaymath}
With $U_{\beta,b} = w_\beta w_b M_\beta' M_b$ we can rewrite the above
as
\begin{displaymath}
  \sum_{\beta, b} w_\beta w_b \trace(\Pi_\beta'M_\beta'M_b\Pi_b) 
  = \sum_{\beta,b} \sum_{j=1}^k [U_{\beta,b}]_{\pi_\beta(j), \pi_b(j)}
  =: \sum_{j=1}^k c_{\pi_1(j), \ldots, \pi_B(j)}
\end{displaymath}
This is an instance of the \emph{multi-dimensional assignment problem}
(MAP), which, contrary to the LSAP, is known to be NP-hard \citep[e.g.,
via reduction to 3-DIMENSIONAL MATCHING,][]{cluster:Garey+Johnson:1979},
and can e.g.\ be approached using randomized parallel algorithms
\citep{cluster:Oliveira+Pardalos:2004}.  Branch-and-bound approaches
suggested in the literature
\citep[e.g.,][]{cluster:Grundel+Oliveira+Pardalos:2005} are
unfortunately computationally infeasible for ``typical'' sizes of
cluster ensembles ($B \ge 20$, maybe even in the hundreds).

Package~\pkg{clue} provides two heuristics for (approximately) finding
the soft least squares consensus partition for Euclidean dissimilarity.
Method \code{"DWH"} of function \code{cl\_consensus()} is an extension
of the greedy algorithm in
\cite{cluster:Dimitriadou+Weingessel+Hornik:2002} which is based on a
single forward pass through the ensemble which in each step chooses the
``locally'' optimal $\Pi$.  Starting with $\tilde{M}_1 = M_1$,
$\tilde{M}_b$ is obtained from $\tilde{M}_{b-1}$ by optimally matching
$M_b \Pi_b$ to this, and taking a weighted average of $\tilde{M}_{b-1}$
and $M_b \Pi_b$ in a way that $\tilde{M}_b$ is the weighted average of
the first~$b$ $M_\beta \Pi_\beta$.  This simple approach could be
further enhanced via back-fitting or several passes, in essence
resulting in an ``on-line'' version of method \code{"SE"}.  This, in
turn, is a fixed-point algorithm, which iterates between updating $M$ as
the weighted average of the current $M_b \Pi_b$, and determining the
$\Pi_b$ by optimally matching the current $M$ to the individual $M_b$.
Finally, method \code{"GV1"} implements the fixed-point algorithm for
the ``first model'' in \cite{cluster:Gordon+Vichi:2001}, which gives
least squares consensus partitions for GV1 dissimilarity.

In the above, we implicitly assumed that all partitions in the ensemble
as well as the sought consensus partition have the same number of
classes.  The more general case can be dealt with through suitable
``projection'' devices.

When using co-membership dissimilarity, the least squares consensus
partition is determined by minimizing
\begin{eqnarray*}
  \lefteqn{\sum_b w_b \|MM' - M_bM_b'\|^2} \\
  &=& s \|MM' - \bar{C}\|^2 + \sum_b w_b \|M_bM_b'\|^2 - s \|\bar{C}\|^2
\end{eqnarray*}
over all membership matrices~$M$, where now $\bar{C} = s^{-1} \sum_b
C(M_b) = s^{-1} \sum_b M_bM_b'$ is the weighted average co-membership
matrix of the ensemble.  This corresponds to the ``third model'' in
\cite{cluster:Gordon+Vichi:2001}.  Method \code{"GV3"} of function
\code{cl\_consensus()} provides a SUMT approach (see Section~\ref{SUMT}
on Page~\pageref{SUMT}) for finding the minimum.  We note that this
strategy could more generally be applied to consensus problems of the
form
\begin{displaymath}
  \sum_b w_b \|\Phi(M) - \Phi(M_b)\|^2 \Rightarrow \min\nolimits_M,
\end{displaymath}
which are equivalent to minimizing $\|\Phi(B) - \bar{\Phi}\|^2$, with
$\bar{\Phi}$ the weighted average of the $\Phi(M_b)$.  This includes
e.g.\ the case where generalized co-memberships are defined by taking
the ``standard'' fuzzy intersection of co-incidences, as discussed in
Section~\ref{fuzzy} on Page~\pageref{fuzzy}.

Package~\pkg{clue} currently does not provide algorithms for obtaining
\emph{hard} consensus partitions, as e.g.\ done in
\cite{cluster:Krieger+Green:1999} using Rand proximity.  It seems
``natural'' to extend the methods discussed above to include a
constraint on softness, e.g., on the partition coefficient PC (see
Section~\ref{PC} on Page~\pageref{PC}).  For Euclidean dissimilarity,
straightforward Lagrangian computations show that the constrained minima
are of the form $\bar{M}(\alpha) = \alpha \bar{M} + (1 - \alpha) E$,
where $E$ is the ``maximally soft'' membership with all entries equal to
$1/k$, $\bar{M}$ is again the weighted average of the $M_b\Pi_b$ with
the $\Pi_b$ solving the underlying MAP, and $\alpha$ is chosen such that
$PC(\bar{M}(\alpha))$ equals a prescribed value.  As $\alpha$ increases
(even beyond one), softness of the $\bar{M}(\alpha)$ decreases.
However, for $\alpha^* > 1 / (1 - k\mu^*)$, where $\mu^*$ is the minimum
of the entries of $\bar{M}$, the $\bar{M}(\alpha)$ have negative
entries, and are no longer feasible membership matrices.  Obviously, the
non-negativity constraints for the $\bar{M}(\alpha)$ eventually put
restrictions on the admissible $\Pi_b$ in the underlying MAP.  Thus,
such a simple relaxation approach to obtaining optimal hard partitions
is not feasible.

For ensembles of hierarchies, \code{cl\_consensus()} provides a built-in
method (\code{"cophenetic"}) for approximately minimizing average
weighted squared Euclidean dissimilarity
\begin{displaymath}
  \sum_b w_b \| U - U_b \|^2 \Rightarrow \min\nolimits_U
\end{displaymath}
over all ultrametrics~$U$, where $U_1, \ldots, U_B$ are the ultrametrics
corresponding to the elements of the ensemble.  This is of course
equivalent to minimizing $\| U - \bar{U} \|^2$, where $\bar{U} = s^{-1}
\sum_b w_b U_b$ is the weighted average of the $U_b$.  The SUMT approach
provided by function \code{ls\_fit\_ultrametric()} (see
Section~\ref{SUMT} on Page~\pageref{SUMT}) is employed for finding the
sought weighted least squares consensus hierarchy.  

In addition, method \code{"majority"} obtains a consensus hierarchy from
an extension of the majority consensus tree of
\cite{cluster:Margush+McMorris:1981}, which minimizes $L(U) = \sum_b w_b
d(U_b, U)$ over all ultrametrics~$U$, where $d$ is the symmetric
difference dissimilarity.

Clearly, the available methods use heuristics for solving hard
optimization problems, and cannot be guaranteed to find a global
optimum.  Standard practice would recommend to use the best solution
found in ``sufficiently many'' replications of the methods.

Alternative recent approaches to obtaining consensus partitions include
``Bagged Clustering'' \citep[provided by \code{bclust()} in
package~\pkg{e1071}]{cluster:Leisch:1999}, the ``evidence accumulation''
framework of \cite{cluster:Fred+Jain:2002}, the NMI optimization and
graph-partitioning methods in \cite{cluster:Strehl+Ghosh:2003a},
``Bagged Clustering'' as in \cite{cluster:Dudoit+Fridlyand:2003}, and
the hybrid bipartite graph formulation of
\cite{cluster:Fern+Brodley:2004}.  Typically, these approaches are
constructive, and can easily be implemented based on the infrastructure
provided by package~\pkg{clue}.  Evidence accumulation amounts to
standard hierarchical clustering of the average co-membership matrix.
Procedure~BagClust1 of \cite{cluster:Dudoit+Fridlyand:2003} amounts to
computing $B^{-1} \sum_b M_b\Pi_b$, where each $\Pi_b$ is determined by
optimal Euclidean matching of $M_b$ to a fixed reference membership
$M_0$.  In the corresponding ``Bagged Clustering'' framework, $M_0$ and
the $M_b$ are obtained by applying the base clusterer to the original
data set and bootstrap samples from it, respectively.  This is
implemented as method \code{"DFBC1"} of \code{cl\_bag()} in
package~\pkg{clue}.  Finally, the approach of
\cite{cluster:Fern+Brodley:2004} solves an LSAP for an asymmetric cost
matrix based on object-by-all-classes incidences.

\subsection{Cluster partitions}

To investigate the ``structure'' in a cluster ensemble, an obvious idea
is to start clustering the clusterings in the ensemble, resulting in
``secondary'' clusterings \citep{cluster:Gordon+Vichi:1998,
  cluster:Gordon:1999}.  This can e.g.\ be performed by using
\code{cl\_dissimilarity()} (or \code{cl\_agreement()}) to compute a
dissimilarity matrix for the ensemble, and feed this into a
dissimilarity-based clustering algorithm (such as \code{pam()} in
package~\pkg{cluster} or \code{hclust()} in package~\pkg{stats}).  (One
can even use \code{cutree()} to obtain hard partitions from hierarchies
thus obtained.)  If prototypes (``typical clusterings'') are desired for
partitions of clusterings, they can be determined post-hoc by finding
suitable consensus clusterings in the classes of the partition, e.g.,
using \code{cl\_consensus()} or \code{cl\_medoid()}.

Package~\pkg{clue} additionally provides \code{cl\_pclust()} for direct
prototype-based partitioning based on minimizing criterion functions of
the form $\sum w_b u_{bj}^m d(x_b, p_j)^e$, the sum of the case-weighted
membership-weighted $e$-th powers of the dissimilarities between the
elements~$x_b$ of the ensemble and the prototypes~$p_j$, for suitable
dissimilarities~$d$ and exponents~$e$.  (The underlying feature spaces
are that of membership matrices and ultrametrics, respectively, for
partitions and hierarchies.)

Parameter~$m$ must not be less than one and controls the softness of the
obtained partitions, corresponding to the \dQuote{fuzzification
  parameter} of the fuzzy $c$-means algorithm.  For $m = 1$, a
generalization of the Lloyd-Forgy variant \citep{cluster:Lloyd:1957,
  cluster:Forgy:1965, cluster:Lloyd:1982} of the $k$-means algorithm is
used, which iterates between reclassifying objects to their closest
prototypes, and computing new prototypes as consensus clusterings for
the classes.  \citet{cluster:Gaul+Schader:1988} introduced this
procedure for \dQuote{Clusterwise Aggregation of Relations} (with the
same domains), containing equivalence relations, i.e., hard partitions,
as a special case.  For $m > 1$, a generalization of the fuzzy $c$-means
recipe \citep[e.g.,][]{cluster:Bezdek:1981} is used, which alternates
between computing optimal memberships for fixed prototypes, and
computing new prototypes as the suitably weighted consensus clusterings
for the classes.

This procedure is repeated until convergence occurs, or the maximal
number of iterations is reached.  Consensus clusterings are computed
using (one of the methods provided by) \code{cl\_consensus}, with
dissimilarities~$d$ and exponent~$e$ implied by method employed, and
obtained via a registration mechanism.  The default methods compute
Least Squares Euclidean consensus clusterings, i.e., use Euclidean
dissimilarity~$d$ and $e = 2$.

\section{Examples}
\label{sec:examples}

\subsection{Cassini data}

\cite{cluster:Dimitriadou+Weingessel+Hornik:2002} and
\cite{cluster:Leisch:1999} use Cassini data sets to illustrate how e.g.\
suitable aggregation of base $k$-means results can reveal underlying
non-convex structure which cannot be found by the base algorithm.  Such
data sets contain points in 2-dimensional space drawn from the uniform
distribution on 3 structures, with the two ``outer'' ones banana-shaped
and the ``middle'' one a circle, and can be obtained by
function~\code{mlbench.cassini()} in
package~\pkg{mlbench}~\citep{cluster:Leisch+Dimitriadou:2005}.
Package~\pkg{clue} contains the data sets \code{Cassini} and
\code{CKME}, which are an instance of a 1000-point Cassini data set, and
a cluster ensemble of 50 $k$-means partitions of the data set into three
classes, respectively.

The data set is shown in Figure~\ref{fig:Cassini}.
<<Cassini-data,eval=FALSE>>=
data("Cassini")
plot(Cassini$x, col = as.integer(Cassini$classes),
     xlab = "", ylab = "")
@ % $
\begin{figure}
  \centering
<<fig=TRUE,echo=FALSE>>=
<<Cassini-data>>
@ %
  \caption{The Cassini data set.}
  \label{fig:Cassini}
\end{figure}
Figure~\ref{fig:CKME} gives a dendrogram of the Euclidean
dissimilarities of the elements of the $k$-means ensemble.
<<CKME,eval=FALSE>>=
data("CKME")
plot(hclust(cl_dissimilarity(CKME)), labels = FALSE)
@ %
\begin{figure}
  \centering
<<fig=TRUE,echo=FALSE>>=
<<CKME>>
@ %
  \caption{A dendrogram of the Euclidean dissimilarities of 50 $k$-means
    partitions of the Cassini data into 3 classes.}
  \label{fig:CKME}
\end{figure}
We can see that there are large groups of essentially identical
$k$-means solutions.  We can gain more insight by inspecting
representatives of these three groups, or by computing the medoid of the
ensemble
<<>>=
m1 <- cl_medoid(CKME)
table(Medoid = cl_class_ids(m1), "True Classes" = Cassini$classes)
@ % $
and inspecting it (Figure~\ref{fig:Cassini-medoid}):
<<Cassini-medoid,eval=FALSE>>=
plot(Cassini$x, col = cl_class_ids(m1), xlab = "", ylab = "")
@ % $
\begin{figure}
  \centering
<<fig=TRUE,echo=FALSE>>=
<<Cassini-medoid>>
@ %
  \caption{Medoid of the Cassini $k$-means ensemble.}
  \label{fig:Cassini-medoid}
\end{figure}
Flipping this solution top-down gives a second ``typical'' partition.
We see that the $k$-means base clusterers cannot resolve the underlying
non-convex structure.  For the least squares consensus of the ensemble,
we obtain
<<>>=
set.seed(1234)
m2 <- cl_consensus(CKME)
@ %
where here and below we set the random seed for reproducibility, noting
that one should really use several replicates of the consensus
heuristic.  This consensus partition has confusion matrix
<<>>=
table(Consensus = cl_class_ids(m2), "True Classes" = Cassini$classes)
@ % $
and class details as displayed in Figure~\ref{fig:Cassini-mean}:
<<Cassini-mean,eval=FALSE>>=
plot(Cassini$x, col = cl_class_ids(m2), xlab = "", ylab = "")
@ % $
\begin{figure}
  \centering
<<fig=TRUE,echo=FALSE>>=
<<Cassini-mean>>
@ %
  \caption{Least Squares Consensus of the Cassini $k$-means ensemble.}
  \label{fig:Cassini-mean}
\end{figure}
This has drastically improved performance, and almost perfect recovery
of the two outer shapes.  In fact,
\cite{cluster:Dimitriadou+Weingessel+Hornik:2002} show that almost
perfect classification can be obtained by suitable combinations of
different base clusterers ($k$-means, fuzzy $c$-means, and unsupervised
fuzzy competitive learning).

\subsection{Gordon-Vichi macroeconomic data}

\citet[Table~1]{cluster:Gordon+Vichi:2001} provide soft partitions of 21
countries based on macroeconomic data for the years 1975, 1980, 1985,
1990, and 1995.  These partitions were obtained using fuzzy $c$-means on
measurements of the following variables: the annual per capita gross
domestic product (GDP) in USD (converted to 1987 prices); the percentage
of GDP provided by agriculture; the percentage of employees who worked
in agriculture; and gross domestic investment, expressed as a percentage
of the GDP.

Table~5 in \cite{cluster:Gordon+Vichi:2001} gives 3-class consensus
partitions obtained by applying their models 1, 2, and 3 and the
approach in \cite{cluster:Sato+Sato:1994}.

The partitions and consensus partitions are available in data sets
\code{GVME} and \code{GVME\_Consensus}, respectively.  We compare the
results of \cite{cluster:Gordon+Vichi:2001} using GV1 dissimilarities
(model 1) to ours as obtained by \code{cl\_consensus()} with method
\code{"GV1"}.

<<>>=
data("GVME")
GVME
set.seed(1)
m1 <- cl_consensus(GVME, method = "GV1",
                   control = list(k = 3, verbose = TRUE))
@ %
This results in a soft partition with average squared GV1 dissimilarity
(the criterion function to be optimized by the consensus partition) of
<<>>=
mean(cl_dissimilarity(GVME, m1, "GV1") ^ 2)
@ %
We compare this to the consensus solution given in
\cite{cluster:Gordon+Vichi:2001}:
<<>>=
data("GVME_Consensus")
m2 <- GVME_Consensus[["MF1/3"]]
mean(cl_dissimilarity(GVME, m2, "GV1") ^ 2)
table(CLUE = cl_class_ids(m1), GV2001 = cl_class_ids(m2))
@ %
Interestingly, we are able to obtain a ``better'' solution, which
however agrees with the one reported on the literature with respect to
their nearest hard partitions.

For the 2-class consensus partition, we obtain
<<>>=
set.seed(1)
m1 <- cl_consensus(GVME, method = "GV1",
                   control = list(k = 2, verbose = TRUE))
@ 
which is slightly better than the solution reported in
\cite{cluster:Gordon+Vichi:2001}
<<>>=
mean(cl_dissimilarity(GVME, m1, "GV1") ^ 2)
m2 <- GVME_Consensus[["MF1/2"]]
mean(cl_dissimilarity(GVME, m2, "GV1") ^ 2)
@ 
but in fact agrees with it apart from rounding errors:
<<>>=
max(abs(cl_membership(m1) - cl_membership(m2)))
@ 
It is interesting to compare these solutions to the Euclidean 2-class
consensus partition for the GVME ensemble:
<<>>=
m3 <- cl_consensus(GVME, method = "GV1",
                   control = list(k = 2, verbose = TRUE))
@ 
This is markedly different from the GV1 consensus partition
<<>>=
table(GV1 = cl_class_ids(m1), Euclidean = cl_class_ids(m3))
@ 
with countries
<<>>=
rownames(m1)[cl_class_ids(m1) != cl_class_ids(m3)]
@ %
classified differently, being with the ``richer'' class for the GV1 and
the ``poorer'' for the Euclidean consensus partition.  (In fact, all
these countries end up in the ``middle'' class for the 3-class GV1
consensus partition.)

\subsection{Rosenberg-Kim kinship terms data}

\cite{cluster:Rosenberg+Kim:1975} describe an experiment where perceived
similarities of the kinship terms were obtained from six different
``sorting'' experiments.  In one of these, 85 female undergraduates at
Rutgers University were asked to sort 15 English terms into classes ``on
the basis of some aspect of meaning''.  These partitions were printed in
\citet[Table~7.1]{cluster:Rosenberg:1982}.  Comparison with the original
data indicates that the partition data have the ``nephew'' and ``niece''
columns interchanged, which is corrected in data set \code{Kinship82}.

\citet[Table~6]{cluster:Gordon+Vichi:2001} provide consensus partitions
for these data based on their models 1--3 (available in data set
\code{Kinship82\_Consensus}).  We compare their results using
co-membership dissimilarities (model 3) to ours as obtained by
\code{cl\_consensus()} with method \code{"GV3"}.

<<>>=
data("Kinship82")
Kinship82
set.seed(1)
m1 <- cl_consensus(Kinship82, method = "GV3",
                   control = list(k = 3, verbose = TRUE))
@ %
This results in a soft partition with average co-membership
dissimilarity (the criterion function to be optimized by the consensus
partition) of
<<>>=
mean(cl_dissimilarity(Kinship82, m1, "comem") ^ 2)
@ %
Again, we compare this to the corresponding consensus solution given in
\cite{cluster:Gordon+Vichi:2001}:
<<>>=
data("Kinship82_Consensus")
m2 <- Kinship82_Consensus[["JMF"]]
mean(cl_dissimilarity(Kinship82, m2, "comem") ^ 2)
@ %
Interestingly, again we obtain a (this time only ``slightly'') better
solution, with
<<>>=
cl_dissimilarity(m1, m2, "comem")
table(CLUE = cl_class_ids(m1), GV2001 = cl_class_ids(m2))
@ %
indicating that the two solutions are reasonably close, even though
<<>>=
cl_fuzziness(cl_ensemble(m1, m2))
@ %
shows that the solution found by \pkg{clue} is ``softer''.

\subsection{Miller-Nicely consonant phoneme confusion data}

\cite{cluster:Miller+Nicely:1955} obtained the data on the auditory
confusions of 16 English consonant phonemes by exposing female subjects
to a series of syllables consisting of one of the consonants followed by
the vowel `a' under 17 different experimental conditions.  Data set
\code{Phonemes} provides consonant misclassification probabilities
(i.e., similarities) obtained from aggregating the six so-called
flat-noise conditions in which only the speech-to-noise ratio was varied
into a single matrix of misclassification frequencies.

These data are used in \cite{cluster:DeSoete:1986} as an illustration of
the SUMT approach for finding least squares optimal fits to
dissimilarities by ultrametrics.  We can reproduce this analysis as
follows.

<<>>=
data("Phonemes")
d <- as.dist(1 - Phonemes)
@ %
(Note that the data set has the consonant misclassification
probabilities, i.e., the similarities between the phonemes.)
<<>>=
u <- ls_fit_ultrametric(d, control = list(verbose = TRUE))
@ %
This gives an ultrametric~$u$ for which Figure~\ref{fig:Phonemes} plots
the corresponding dendrogram, ``basically'' reproducing Figure~1
in \cite{cluster:DeSoete:1986}.
<<Phonemes,eval=FALSE>>=
plot(u)
@ %
\begin{figure}
  \centering
<<fig=TRUE,echo=FALSE>>=
<<Phonemes>>
@ %
  \caption{Dendrogram for least squares fit to the Miller-Nicely 
    consonant phoneme confusion data.}
  \label{fig:Phonemes}
\end{figure}

We can also compare the least squares fit obtained to that of other
hierarchical clusterings of $d$, e.g.\ those obtained by
\code{hclust()}.  The ``optimal''~$u$ has Euclidean dissimilarity
<<>>=
round(cl_dissimilarity(d, u), 4)
@ %
to $d$.  For the \code{hclust()} results, we get
<<>>=
hclust_methods <- c("ward", "single", "complete", "average", "mcquitty")
hens <- cl_ensemble(list = lapply(hclust_methods,
                                  function(m) hclust(d, m)))
names(hens) <- hclust_methods
round(sapply(hens, cl_dissimilarity, d), 4)
@ %
which all exhibit greater Euclidean dissimilarity to $d$ than $u$.  (We
exclude methods \code{"median"} and \code{"centroid"} as these do not
yield valid hierarchies.)  We can also compare the ``structure'' of the
different hierarchies, e.g.\ by looking at the rate of inversions
between them:
<<>>=
ahens <- c(L2opt = cl_ensemble(u), hens)
round(cl_dissimilarity(ahens, method = "gamma"), 2)
@ %

\section{Outlook}
\label{sec:outlook}

Package~\pkg{clue} was designed as an \emph{extensible} environment for
computing on cluster ensembles.  It currently provides basic data
structures for representing partitions and hierarchies, and facilities
for computing on these, including methods for measuring proximity and
obtaining consensus and ``secondary'' clusterings.

Many extensions to the available functionality are possible and in fact
planned (some of these enhancements were already discussed in more
detail in the course of this paper).
\begin{itemize}
 \item Provide mechanisms to generate cluster ensembles based on
  reweighting (assuming base clusterers allowing for case weights) the
  data set.
 \item Explore recent advances (e.g., parallelized random search) in
  heuristics for solving the multi-dimensional assignment problem.
 \item Add support for \emph{additive trees}
  \citep[e.g.,][]{cluster:Barthelemy+Guenoche:1991}.
 \item Add heuristics for finding least squares fits based on iterative
  projection on convex sets of constraints, see e.g.\
  \cite{cluster:Hubert+Arabie+Meulman:2006} and the accompanying MATLAB
  code available at \url{http://cda.psych.uiuc.edu/srpm_mfiles/} for
  using these methods (instead of SUMT approaches) to fit ultrametrics
  and additive trees to proximity data.
 \item Add an ``$L_1$ View''.  Emphasis in \pkg{clue}, in particular for
  obtaining consensus clusterings, is on using Euclidean dissimilarities
  (based on suitable least squares distances); arguably, more ``robust''
  consensus solutions should result from using Manhattan dissimilarities
  (based on absolute distances).  Adding such functionality necessitates
  developing the corresponding structure theory for soft Manhattan
  median partitions.  Minimizing average Manhattan dissimilarity between
  co-memberships and ultrametrics results in constrained $L_1$
  approximation problems for the weighted medians of the co-memberships
  and ultrametrics, respectively, and could be approached by employing
  SUMTs analogous to the ones used for the $L_2$ approximations.
 \item Provide heuristics for obtaining \emph{hard} consensus
  partitions.
 \item Add facilities for tuning hyper-parameters (most prominently, the
  number of classes employed) and ``cluster validation'' of partitioning
  algorithms, as recently proposed by
  \cite{cluster:Roth+Lange+Braun:2002},
  \cite{cluster:Lange+Roth+Braun:2004},
  \cite{cluster:Dudoit+Fridlyand:2002},
  and \cite{cluster:Tibshirani+Walther:2005}.
\end{itemize}
We are hoping to be able to provide many of these extensions in the near
future.

\subsubsection*{Acknowledgments}

We are grateful to Walter B\"ohm for providing efficient C code for
solving assignment problems.


{\small
  \bibliographystyle{abbrvnat}
  \bibliography{cluster}
}

\end{document}

\name{l1_fit_ultrametric}
\alias{l1_fit_ultrametric}
\title{Least Absolute Deviation Fit of Ultrametrics to Dissimilarities}
\description{
  Find the ultrametric with minimal absolute distance (Manhattan
  dissimilarity) to a given dissimilarity object.
}
\usage{
l1_fit_ultrametric(x, method = c("SUMT", "IRIP"), weights = 1,
                   control = list())
}
\arguments{
  \item{x}{a dissimilarity object inheriting from or coercible to class
    \code{"\link{dist}"}.}
  \item{method}{a character string indicating the fitting method to be
    employed.  Must be one of \code{"SUMT"} (default) or \code{"IRIP"},
    or a unique abbreviation thereof.}
  \item{weights}{a numeric vector or matrix with non-negative weights
    for obtaining a weighted least squares fit.  If a matrix, its
    numbers of rows and columns must be the same as the number of
    objects in \code{x}, and the lower diagonal part is used.
    Otherwise, it is recycled to the number of elements in \code{x}.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\value{
  An object of class \code{"\link{cl_ultrametric}"} containing the
  fitted ultrametric distances.
}
\details{
  The problem to be solved is minimizing
  \deqn{L(u) = \sum_{i,j} w_{ij} |x_{ij} - u_{ij}|}
  over all \eqn{u} satisfying the ultrametric constraints (i.e., for all
  \eqn{i, j, k}, \eqn{u_{ij} \le \max(u_{ik}, u_{jk})}).  This problem
  is known to be NP hard
  \bibcitep{Křivánek+Morávek:1986}.

  We provide two heuristics for solving this problem.

  Method \code{"SUMT"} implements a \acronym{SUMT} (Sequential
  Unconstrained Minimization Technique, see \code{\link{sumt}}) approach
  using the sign function for the gradients of the absolute value
  function.

  Available control parameters are \code{method}, \code{control},
  \code{eps}, \code{q}, and \code{verbose}, which have the same roles as
  for \code{\link{sumt}}, and the following.

  \describe{
    \item{\code{nruns}}{an integer giving the number of runs to be
      performed.  Defaults to 1.}
    \item{\code{start}}{a single dissimilarity, or a list of
      dissimilarities to be employed as starting values.}
  }

  Method \code{"IRIP"} implements a variant of the Iteratively
  Reweighted Iterative Projection approach of \bibcitet{Smith:2001},
  which
  attempts to solve the \eqn{L_1} problem via a sequence of weighted
  \eqn{L_2} problems, determining \eqn{u(t+1)} by minimizing the
  criterion function
  \deqn{\sum_{i,j} w_{ij}
    (x_{ij} - u_{ij})^2 / \max(|x_{ij} - u_{ij}(t)|, m)}
  with \eqn{m} a \dQuote{small} non-zero value to avoid zero divisors.
  We use the \acronym{SUMT} method of \code{\link{ls_fit_ultrametric}}
  for solving the weighted least squares problems.

  Available control parameters are as follows.
  \describe{
    \item{\code{maxiter}}{an integer giving the maximal number of
      iteration steps to be performed.
      Defaults to 100.}
    \item{\code{eps}}{a nonnegative number controlling the iteration,
      which stops when the maximal change in \eqn{u} is less than
      \code{eps}.
      Defaults to \eqn{10^{-6}}.}
    \item{\code{reltol}}{the relative convergence tolerance.  Iteration
      stops when the relative change in the \eqn{L_1} criterion is less
      than \code{reltol}.
      Defaults to \eqn{10^{-6}}.}
    \item{\code{MIN}}{the cutoff \eqn{m}.  Defaults to \eqn{10^{-3}}.}
    \item{\code{start}}{a dissimilarity object to be used as the
      starting value for \eqn{u}.}
    \item{\code{control}}{a list of control parameters to be used by the
      method of \code{\link{ls_fit_ultrametric}} employed for solving
      the weighted \eqn{L_2} problems.}
  }
    
  One may need to adjust the default control parameters to achieve
  convergence.

  It should be noted that all methods are heuristics which can not be
  guaranteed to find the global minimum.  
}
\seealso{
  \code{\link{cl_consensus}} for computing least absolute deviation
  (Manhattan) consensus hierarchies;
  \code{\link{ls_fit_ultrametric}}.
}
\references{
  \bibshow{Křivánek+Morávek:1986, Smith:2001}
}  
\keyword{cluster}
\keyword{optimize}

\name{n_of_classes}
\alias{n_of_classes}
\alias{cl_class_ids}
\alias{as.cl_class_ids}
\title{Classes in a Partition}
\description{Determine the number of classes and the class ids in a
  partition of objects.}
\usage{
n_of_classes(x)
cl_class_ids(x)
as.cl_class_ids(x)
}
\arguments{
  \item{x}{an object representing a (hard or soft) partition (for
    \code{n_of_classes} and \code{cl_class_ids}), or raw class ids (for
    \code{as.cl_class_ids}).}
}
\value{
  For \code{n_of_classes}, an integer giving the number of classes in
  the partition.
  
  For \code{cl_class_ids}, a vector of integers with the corresponding
  class ids.  For soft partitions, the class ids returned are those of
  the \emph{nearest hard partition} obtained by taking the class ids of
  the (first) maximal membership values.
}
\details{
  These function are generic functions.

  The methods provided in package \pkg{clue} handle the partitions
  obtained from clustering functions in the base R distribution, as well
  as packages \pkg{RWeka}, \pkg{cba}, \pkg{cclust}, \pkg{cluster},
  \pkg{e1071}, \pkg{flexclust}, \pkg{flexmix}, \pkg{kernlab},
  \pkg{mclust}, \pkg{movMF} and \pkg{skmeans} (and of course, \pkg{clue}
  itself).

  Note that the number of classes is taken as the number of distinct
  class ids actually used in the partition, and may differ from the
  number of columns in a membership matrix representing the partition.

  \code{as.cl_class_ids} can be used for coercing \dQuote{raw} class
  ids (given as atomic vectors) to class id objects.
}
\seealso{
  \code{\link{is.cl_partition}}
}
\examples{
data("Cassini")
party <- kmeans(Cassini$x, 3)
n_of_classes(party)
## A simple confusion matrix:
table(cl_class_ids(party), Cassini$classes)
## For an "oversize" membership matrix representation:
n_of_classes(cl_membership(party, 6))
}
\keyword{cluster}

\name{cl_validity}
\alias{cl_validity}
\alias{cl_validity.default}
\title{Validity Measures for Partitions and Hierarchies}
\description{
  Compute validity measures for partitions and hierarchies, attempting
  to measure how well these clusterings capture the underlying structure
  in the data they were obtained from.
}
\usage{
cl_validity(x, ...)
\method{cl_validity}{default}(x, d, ...)
}
\arguments{
  \item{x}{an object representing a partition or hierarchy.}
  \item{d}{a dissimilarity object from which \code{x} was obtained.}
  \item{\dots}{arguments to be passed to or from methods.}
}
\value{
  A list of class \code{"cl_validity"} with the computed validity
  measures.
}
\details{
  \code{cl_validity} is a generic function.

  For partitions, its default method gives the \dQuote{dissimilarity
    accounted for}, defined as \eqn{1 - a_w / a_t}, where \eqn{a_t} is
  the average total dissimilarity, and the \dQuote{average within
    dissimilarity} \eqn{a_w} is given by
  \deqn{\frac{\sum_{i,j} \sum_k m_{ik}m_{jk} d_{ij}}{
        \sum_{i,j} \sum_k m_{ik}m_{jk}}}{%
        \sum_{i,j} \sum_k m_{ik}m_{jk} d_{ij} /
        \sum_{i,j} \sum_k m_{ik}m_{jk}}
  where \eqn{d} and \eqn{m} are the dissimilarities and memberships,
  respectively, and the sums are over all pairs of objects and all
  classes.

  For hierarchies, the validity measures computed by default are
  \dQuote{variance accounted for}
  (VAF, e.g., \bibcitet{Hubert+Arabie+Meulman:2006})
  and \dQuote{deviance accounted for}
  (DEV, e.g., \bibcitet{Smith:2001}).
  If \code{u} is the ultrametric corresponding to the hierarchy \code{x}
  and \code{d} the dissimilarity \code{x} was obtained from, these
  validity measures are given by
  \deqn{\mathrm{VAF} =
    \max\left(0, 1 - \frac{\sum_{i,j} (d_{ij} - u_{ij})^2}{
    \sum_{i,j} (d_{ij} - \mathrm{mean}(d)) ^ 2}\right)}{
    max(0, 1 - sum_{i,j} (d_{ij} - u_{ij})^2 /
    sum_{i,j} (d_{ij} - mean(d))^2)}
  and
  \deqn{\mathrm{DEV} =
    \max\left(0, 1 - \frac{\sum_{i,j} |d_{ij} - u_{ij}|}{
    \sum_{i,j} |d_{ij} - \mathrm{median}(d)|}\right)}{
    max(0, 1 - sum_{i,j} |d_{ij} - u_{ij}| /
    sum_{i,j} |d_{ij} - median(d)|)}
  respectively.  Note that VAF and DEV are not invariant under rescaling
  \code{u}, and may be \dQuote{arbitrarily small} (i.e., 0 using the
  above definitions) even though \code{u} and \code{d} are
  \dQuote{structurally close} in some sense.
  
  For the results of using \code{\link[cluster]{agnes}} and
  \code{\link[cluster]{diana}}, the agglomerative and divisive
  coefficients are provided in addition to the default ones.
}
\references{
  \bibshow{Hubert+Arabie+Meulman:2006, Smith:2001}
}  
\seealso{
  \code{\link[fpc]{cluster.stats}} in package \pkg{fpc} for a variety of
  cluster validation statistics;
  \code{\link[e1071]{fclustIndex}} in package \pkg{e1071} for several
  fuzzy cluster indexes;
  \code{\link[cclust]{clustIndex}} in package \pkg{cclust};
  \code{\link[cluster]{silhouette}} in package \pkg{cluster}.
}
\keyword{cluster}

\name{cl_object_names}
\alias{cl_object_names}
\title{Find Object Names}
\description{
  Find the names of the objects from which a taxonomy (partition or
  hierarchy) or proximity was obtained.
}
\usage{
cl_object_names(x)
}
\arguments{
  \item{x}{an \R object representing a taxonomy or proximity.}
}
\value{
  A character vector of length \code{\link{n_of_objects}(x)} in case the
  names of the objects could be determined, or \code{NULL}.
}
\details{
  This is a generic function.

  The methods provided in package \pkg{clue} handle the partitions and
  hierarchies obtained from clustering functions in the base R
  distribution, as well as packages \pkg{RWeka}, \pkg{ape}, \pkg{cba},
  \pkg{cclust}, \pkg{cluster}, \pkg{e1071}, \pkg{flexclust},
  \pkg{flexmix}, \pkg{kernlab}, \pkg{mclust}, \pkg{movMF} and
  \pkg{skmeans} (and of course, \pkg{clue} itself), in as much as
  possible.

  There is also a method for object dissimilarities which inherit from
  class \code{"\link{dist}"}.
}
\keyword{cluster}

\name{cl_pam}
\alias{cl_pam}
\title{K-Medoids Partitions of Clusterings}
\description{
  Compute \eqn{k}-medoids partitions of clusterings.
}
\usage{
cl_pam(x, k, method = "euclidean", solver = c("pam", "kmedoids"))
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies, or something
    coercible to that (see \code{\link{cl_ensemble}}).}
  \item{k}{an integer giving the number of classes to be used in the
    partition.}
  \item{method}{a character string or a function, as for argument
    \code{method} of function \code{\link{cl_dissimilarity}}.}
  \item{solver}{a character string indicating the \eqn{k}-medoids solver
    to be employed.  May be abbreviated.  If \code{"pam"} (default), the
    Partitioning Around Medoids
    \bibcitep{|Kaufman+Rousseeuw:1990|Chapter 2}
    heuristic \code{\link[cluster]{pam}} of package \pkg{cluster} is
    used.  Otherwise, the exact algorithm of \code{\link{kmedoids}} is
    employed.}
}
\value{
  An object of class \code{"cl_pam"} representing the obtained
  \dQuote{secondary} partition, which is a list with the following
  components.
  \item{cluster}{the class ids of the partition.}
  \item{medoid_ids}{the indices of the medoids.}
  \item{prototypes}{a cluster ensemble with the \eqn{k} prototypes
    (medoids).}
  \item{criterion}{the value of the criterion function of the
    partition.}
  \item{description}{a character string indicating the dissimilarity
    method employed.}
}  
\details{
  An optimal \eqn{k}-medoids partition of the given cluster ensemble is
  defined as a partition of the objects \eqn{x_i} (the elements of the
  ensemble) into \eqn{k} classes \eqn{C_1, \ldots, C_k} such that the
  criterion function
  \eqn{L = \sum_{l=1}^k \min_{j \in C_l} \sum_{i \in C_l} d(x_i, x_j)} 
  is minimized.

  Such secondary partitions \bibcitep{e.g.|Gordon+Vichi:1998|}
  are obtained by
  computing the dissimilarities \eqn{d} of the objects in the ensemble
  for the given dissimilarity method, and applying a dissimilarity-based
  \eqn{k}-medoids solver to \eqn{d}.
}
\references{
  \bibshow{Kaufman+Rousseeuw:1990, Gordon+Vichi:1998}
}
\seealso{
  \code{\link{cl_pclust}} for more general prototype-based partitions of
  clusterings.
}
\examples{
data("Kinship82")
party <- cl_pam(Kinship82, 3, "symdiff")
## Compare results with tables 5 and 6 in Gordon & Vichi (1998).
party
lapply(cl_prototypes(party), cl_classes)
table(cl_class_ids(party))
}
\keyword{cluster}

\name{kmedoids}
\alias{kmedoids}
\title{K-Medoids Clustering}
\description{
  Compute a \eqn{k}-medoids partition of a dissimilarity object.
}
\usage{
kmedoids(x, k)
}
\arguments{
  \item{x}{a dissimilarity object inheriting from class
    \code{"\link{dist}"}, or a square matrix of pairwise
    object-to-object dissimilarity values.}
  \item{k}{an integer giving the number of classes to be used in the
    partition.}
}
\value{
  An object of class \code{"kmedoids"} representing the obtained
  partition, which is a list with the following components.
  \item{cluster}{the class ids of the partition.}
  \item{medoid_ids}{the indices of the medoids.}
  \item{criterion}{the value of the criterion function of the
    partition.}
}
\details{
  Let \eqn{d} denote the pairwise object-to-object dissimilarity matrix
  corresponding to \code{x}.  A \eqn{k}-medoids partition of \code{x} is
  defined as a partition of the numbers from 1 to \eqn{n}, the number of
  objects in \code{x}, into \eqn{k} classes \eqn{C_1, \ldots, C_k} such
  that the criterion function
  \eqn{L = \sum_l \min_{j \in C_l} \sum_{i \in C_l} d_{ij}}
  is minimized.

  This is an NP-hard optimization problem.  PAM (Partitioning Around
  Medoids, see \bibcitet{|Kaufman+Rousseeuw:1990|Chapter 2})
  is a very popular
  heuristic for obtaining optimal \eqn{k}-medoids partitions, and
  provided by \code{\link[cluster]{pam}} in package \pkg{cluster}.

  \code{kmedoids} is an exact algorithm based on a binary linear
  programming formulation of the optimization problem
  \bibcitep{e.g.|Gordon+Vichi:1998|[P4']},
  using \code{\link[lpSolve]{lp}} from package
  \pkg{lpSolve} as solver.  Depending on available hardware resources
  (the number of constraints of the program is of the order \eqn{n^2}),
  it may not be possible to obtain a solution.
}
\references{
  \bibshow{Kaufman+Rousseeuw:1990, Gordon+Vichi:1998}
}  
\keyword{cluster}
\keyword{optimize}

\name{cl_prototypes}
\alias{cl_prototypes}
\title{Partition Prototypes}
\description{
  Determine prototypes for the classes of an R object representing a
  partition.
}
\usage{
cl_prototypes(x)
}
\arguments{
  \item{x}{an R object representing a partition of objects.}
}
\details{
  Many partitioning methods are based on prototypes (\dQuote{centers},
  \dQuote{centroids}, \dQuote{medoids}, \dots).  In typical cases, these
  are points in the feature space for the measurements on the objects to
  be partitioned, such that one can quantify the distance between the
  objects and the prototypes, and, e.g., classify objects to their
  closest prototype.

  This is a generic function.  The methods provided in package
  \pkg{clue} handle the partitions obtained from clustering functions in
  the base R distribution, as well as packages \pkg{cba}, \pkg{cclust},
  \pkg{cluster}, \pkg{e1071}, \pkg{flexclust}, \pkg{kernlab}, and
  \pkg{mclust} (and of course, \pkg{clue} itself).
}
\examples{
## Show how prototypes ("centers") vary across k-means runs on
## bootstrap samples from the Cassini data.
data("Cassini")
nr <- NROW(Cassini$x)
out <- replicate(50,
                 { kmeans(Cassini$x[sample(nr, replace = TRUE), ], 3) },
                 simplify = FALSE)
## Plot the data points in light gray, and the prototypes found.
plot(Cassini$x, col = gray(0.8))
points(do.call("rbind", lapply(out, cl_prototypes)), pch = 19)
}
\keyword{cluster}

\name{CKME}
\alias{CKME}
\title{Cassini Data Partitions Obtained by K-Means}
\description{
  A cluster ensemble of 50 \eqn{k}-means partitions of the Cassini data
  into three classes.
}
\usage{data("CKME")}
\format{
  A cluster ensemble of 50 (\eqn{k}-means) partitions.
}
\details{
  The ensemble was generated via
  \preformatted{
    require("clue")
    data("Cassini")
    set.seed(1234)
    CKME <- cl_boot(Cassini$x, 50, 3)
  }
}
\keyword{datasets}

\name{cl_classes}
\alias{cl_classes}
\title{Cluster Classes}
\description{
  Extract the classes in a partition or hierarchy.
}
\usage{
cl_classes(x)
}
\arguments{
  \item{x}{an R object representing a partition or hierarchy of
    objects.}
}
\value{
  A list inheriting from \code{"cl_classes_of_objects"} of vectors
  indicating the classes.
}
\details{
  For partitions, the classes are the equivalence classes
  (\dQuote{clusters}) of the partition; for soft partitions, the classes
  of the nearest hard partition are used.

  For hierarchies represented by trees, the classes are the sets of
  objects corresponding to (joined at or split by) the nodes of the
  tree.
}
\keyword{cluster}

\name{hierarchy}
\alias{cl_hierarchy}		% class ...
\alias{is.cl_hierarchy}
\alias{as.cl_hierarchy}
\alias{cl_dendrogram}		% class ...
\alias{is.cl_dendrogram}
\alias{as.cl_dendrogram}
\alias{plot.cl_dendrogram}
\title{Hierarchies}
\description{
  Determine whether an R object represents a hierarchy of objects, or
  coerce to an R object representing such.}
\usage{
is.cl_hierarchy(x)
is.cl_dendrogram(x)

as.cl_hierarchy(x)
as.cl_dendrogram(x)
}
\arguments{
  \item{x}{an R object.}
}
\value{
  For the testing functions, a logical indicating whether the given
  object represents a clustering of objects of the respective kind.

  For the coercion functions, a container object inheriting from
  \code{"cl_hierarchy"}, with a suitable representation of the hierarchy
  given by \code{x}.
}
\details{
  These functions are generic functions.

  The methods provided in package \pkg{clue} handle the partitions and
  hierarchies obtained from clustering functions in the base R
  distribution, as well as packages \pkg{RWeka}, \pkg{ape}, \pkg{cba},
  \pkg{cclust}, \pkg{cluster}, \pkg{e1071}, \pkg{flexclust},
  \pkg{flexmix}, \pkg{kernlab}, \pkg{mclust}, \pkg{movMF} and
  \pkg{skmeans} (and of course, \pkg{clue} itself).

  The hierarchies considered by \pkg{clue} are \emph{\eqn{n}-trees}
  (hierarchies in the strict sense) and \emph{dendrograms} (also known
  as valued \eqn{n}-trees or total indexed hierarchies), which are
  represented by the virtual classes \code{"cl_hierarchy"} and
  \code{"cl_dendrogram"} (which inherits from the former),
  respectively.

  \eqn{n}-trees on a set \eqn{X} of objects correspond to collections
  \eqn{H} of subsets of \eqn{X}, usually called \emph{classes} of the
  hierarchy, which satisfy the following properties:
  \itemize{
    \item \eqn{H} contains all singletons with objects of \eqn{X},
    \eqn{X} itself, but not the empty set;
    \item The intersection of two sets \eqn{A} and \eqn{B} in \eqn{H} is
    either empty or one of the sets.
  }

  The classes of a hierarchy can be obtained by
  \code{\link{cl_classes}}.

  Dendrograms are \eqn{n}-trees where additionally a height \eqn{h} is
  associated with each of the classes, so that for two classes \eqn{A}
  and \eqn{B} with non-empty intersection we have \eqn{h(A) \le h(B)}
  iff \eqn{A} is a subset of \eqn{B}.  For each pair of objects one can
  then define \eqn{u_{ij}} as the height of the smallest class
  containing both \eqn{i} and \eqn{j}: this results in a dissimilarity
  on \eqn{X} which satisfies the ultrametric (3-point) conditions
  \eqn{u_{ij} \le \max(u_{ik}, u_{jk})} for all triples \eqn{(i, j, k)}
  of objects.  Conversely, an ultrametric dissimilarity induces a unique
  dendrogram.

  The ultrametric dissimilarities of a dendrogram can be obtained by
  \code{\link{cl_ultrametric}}.

  \code{as.cl_hierarchy} returns an object of class
  \code{"cl_hierarchy"} \dQuote{containing} the given object \code{x} if
  this already represents a hierarchy (i.e., \code{is.cl_hierarchy(x)}
  is true), or the ultrametric obtained from \code{x} via
  \code{\link{as.cl_ultrametric}}.

  \code{as.cl_dendrogram} returns an object which has class
  \code{"cl_dendrogram"} and inherits from \code{"cl_hierarchy"},
  and contains \code{x} if it represents a dendrogram (i.e.,
  \code{is.cl_dendrogram(x)} is true), or the ultrametric obtained from
  \code{x}.
  
  Conceptually, hierarchies and dendrograms are \emph{virtual} classes,
  allowing for a variety of representations.

  There are group methods for comparing dendrograms and computing their
  minimum, maximum, and range based on the meet and join operations, see
  \code{\link{cl_meet}}.  There is also a \code{plot} method.
}
\examples{
hcl <- hclust(dist(USArrests))
is.cl_dendrogram(hcl)
is.cl_hierarchy(hcl)
}
\keyword{cluster}

\name{cl_predict}
\alias{cl_predict}
\title{Predict Memberships}
\description{
  Predict class ids or memberships from R objects representing
  partitions.
}
\usage{
cl_predict(object, newdata = NULL,
           type = c("class_ids", "memberships"), ...)
}
\arguments{
  \item{object}{an R object representing a partition of objects.}
  \item{newdata}{an optional data set giving the objects to make
    predictions for.  This must be of the same \dQuote{kind} as the data
    set employed for obtaining the partition.  If omitted, the original
    data are used.}
  \item{type}{a character string indicating whether class ids or
    memberships should be returned. May be abbreviated.}
  \item{\dots}{arguments to be passed to and from methods.}
}
\value{
  Depending on \code{type}, an object of class \code{"cl_class_ids"}
  with the predicted class ids, or of class \code{"cl_membership"} with
  the matrix of predicted membership values.
}
\details{
  Many algorithms resulting in partitions of a given set of objects can
  be taken to induce a partition of the underlying feature space for the
  measurements on the objects, so that class memberships for
  \dQuote{new} objects can be obtained from the induced partition.
  Examples include partitions based on assigning objects to their
  \dQuote{closest} prototypes, or providing mixture models for the
  distribution of objects in feature space.

  This is a generic function.  The methods provided in package
  \pkg{clue} handle the partitions obtained from clustering functions in
  the base R distribution, as well as packages \pkg{RWeka}, \pkg{cba},
  \pkg{cclust}, \pkg{cluster}, \pkg{e1071}, \pkg{flexclust},
  \pkg{flexmix}, \pkg{kernlab}, \pkg{mclust}, \pkg{movMF} and
  \pkg{skmeans} (and of course, \pkg{clue} itself).
}
\examples{
## Run kmeans on a random subset of the Cassini data, and predict the
## memberships for the "test" data set.
data("Cassini")
nr <- NROW(Cassini$x)
ind <- sample(nr, 0.9 * nr, replace = FALSE)
party <- kmeans(Cassini$x[ind, ], 3)
table(cl_predict(party, Cassini$x[-ind, ]),
      Cassini$classes[-ind])
}
\keyword{cluster}

\name{cl_bag}
\alias{cl_bag}
\title{Bagging for Clustering}
\description{
  Construct partitions of objects by running a base clustering algorithm
  on bootstrap samples from a given data set, and \dQuote{suitably}
  aggregating these primary partitions.
}
\usage{
cl_bag(x, B, k = NULL, algorithm = "kmeans", parameters = NULL, 
       method = "DFBC1", control = NULL)
}
\arguments{
  \item{x}{the data set of objects to be clustered, as appropriate for
    the base clustering algorithm.}
  \item{B}{an integer giving the number of bootstrap replicates.}
  \item{k}{\code{NULL} (default), or an integer giving the number of
    classes to be used for a partitioning base algorithm.}
  \item{algorithm}{a character string or function specifying the base
    clustering algorithm.}
  \item{parameters}{a named list of additional arguments to be passed to
    the base algorithm.}
  \item{method}{a character string indicating the bagging method to
    use.  Currently, only method \code{"DFBC1"} is available, which
    implements algorithm \emph{BagClust1} in
    \bibcitet{Dudoit+Fridlyand:2003}.}
  \item{control}{a list of control parameters for the aggregation.
    Currently, not used.}
}
\value{
  An R object representing a partition of the objects given in \code{x}.
}
\details{
  Bagging for clustering is really a rather general conceptual framework
  than a specific algorithm.  If the primary partitions generated in the
  bootstrap stage form a cluster ensemble (so that class memberships of
  the objects in \code{x} can be obtained), consensus methods for
  cluster ensembles (as implemented, e.g., in \code{\link{cl_consensus}}
  and \code{\link{cl_medoid}}) can be employed for the aggregation
  stage.  In particular, (possibly new) bagging algorithms can easily be
  realized by directly running \code{\link{cl_consensus}} on the results
  of \code{\link{cl_boot}}.

  In BagClust1, aggregation proceeds by generating a reference partition
  by running the base clustering algorithm on the whole given data set,
  and averaging the ensemble memberships after optimally matching them
  to the reference partition (in fact, by minimizing Euclidean
  dissimilarity, see \code{\link{cl_dissimilarity}}).

  If the base clustering algorithm yields prototypes, aggregation can be
  based on clustering these.  This is the idea underlying the
  \dQuote{Bagged Clustering} algorithm introduced in
  \bibcitet{Leisch:1999} and implemented by function
  \code{\link[e1071]{bclust}} in package \pkg{e1071}.
}
\references{
  \bibshow{Dudoit+Fridlyand:2003, Leisch:1999}
}
\examples{
set.seed(1234)
## Run BagClust1 on the Cassini data.
data("Cassini")
party <- cl_bag(Cassini$x, 50, 3)
plot(Cassini$x, col = cl_class_ids(party), xlab = "", ylab = "")
## Actually, using fuzzy c-means as a base learner works much better:
if(require("e1071", quietly = TRUE)) {
    party <- cl_bag(Cassini$x, 20, 3, algorithm = "cmeans")
    plot(Cassini$x, col = cl_class_ids(party), xlab = "", ylab = "")
}
}
\keyword{cluster}

\name{fit_ultrametric_target}
\alias{ls_fit_ultrametric_target}
\alias{l1_fit_ultrametric_target}
\title{Fit Dissimilarities to a Hierarchy}
\description{
  Find the ultrametric from a target equivalence class of hierarchies
  which minimizes weighted Euclidean or Manhattan dissimilarity to a
  given dissimilarity object.
}
\usage{
ls_fit_ultrametric_target(x, y, weights = 1)
l1_fit_ultrametric_target(x, y, weights = 1)
}
\arguments{
  \item{x}{a dissimilarity object inheriting from class
    \code{"\link{dist}"}.}
  \item{y}{a target hierarchy.}
  \item{weights}{a numeric vector or matrix with non-negative weights
    for obtaining a weighted fit.  If a matrix, its numbers of rows and
    columns must be the same as the number of objects in \code{x}.
    Otherwise, it is recycled to the number of elements in \code{x}.}
}
\value{
  An object of class \code{"\link{cl_ultrametric}"} containing the
  optimal ultrametric distances.
}
\details{
  The target equivalence class consists of all dendrograms for which the
  corresponding \eqn{n}-trees are the same as the one corresponding to
  \code{y}.  I.e., all splits are the same as for \code{y}, and
  optimization is over the height of the splits.

  The criterion function to be optimized over all ultrametrics from the
  equivalence class is \eqn{\sum w_{ij} |x_{ij} - u_{ij}|^p}, where
  \eqn{p = 2} in the Euclidean and \eqn{p = 1} in the Manhattan case,
  respectively.

  The optimum can be computed as follows.  Suppose split \eqn{s} joins
  object classes \eqn{A} and \eqn{B}.  As the ultrametric
  dissimilarities of all objects in \eqn{A} to all objects in \eqn{B}
  must be the same value, say, \eqn{u_{A,B} = u_s}, the contribution
  from the split to the criterion function is of the form
  \eqn{f_s(u_s) = \sum_{i \in A, j \in B} w_{ij} |x_{ij} - u_s|^p}.
  We need to minimize \eqn{\sum_s f_s(u_s)} under the constraint that
  the \eqn{u_s} form a non-decreasing sequence, which is accomplished by
  using the Pool Adjacent Violator Algorithm (\acronym{PAVA}) using the
  weighted mean (\eqn{p = 2}) or weighted median (\eqn{p = 1}) for
  solving the blockwise optimization problems.
}
\seealso{
  \code{\link{ls_fit_ultrametric}} for finding the ultrametric
  minimizing Euclidean dissimilarity (without fixing the splits).
}
\examples{
data("Phonemes")
## Note that the Phonemes data set has the consonant misclassification
## probabilities, i.e., the similarities between the phonemes.
d <- as.dist(1 - Phonemes)
## Find the maximal dominated and miminal dominating ultrametrics by
## hclust() with single and complete linkage:
y1 <- hclust(d, "single")
y2 <- hclust(d, "complete")
## Note that these are quite different:
cl_dissimilarity(y1, y2, "gamma")
## Now find the L2 optimal members of the respective dendrogram
## equivalence classes.
u1 <- ls_fit_ultrametric_target(d, y1)
u2 <- ls_fit_ultrametric_target(d, y2)
## Compute the L2 optimal ultrametric approximation to d.
u <- ls_fit_ultrametric(d)
## And compare ...
cl_dissimilarity(cl_ensemble(Opt = u, Single = u1, Complete = u2), d)
## The solution obtained via complete linkage is quite close:
cl_agreement(u2, u, "cophenetic")
}
\keyword{cluster}
\keyword{optimize}

\name{ls_fit_ultrametric}
\encoding{UTF-8}
\alias{ls_fit_ultrametric}
\title{Least Squares Fit of Ultrametrics to Dissimilarities}
\description{
  Find the ultrametric with minimal square distance (Euclidean
  dissimilarity) to given dissimilarity objects.
}
\usage{
ls_fit_ultrametric(x, method = c("SUMT", "IP", "IR"), weights = 1,
                   control = list())
}
\arguments{
  \item{x}{a dissimilarity object inheriting from or coercible to class
    \code{"\link{dist}"}, or an ensemble of such objects.} 
  \item{method}{a character string indicating the fitting method to be
    employed.  Must be one of \code{"SUMT"} (default), \code{"IP"}, or
    \code{"IR"}, or a unique abbreviation thereof.}
  \item{weights}{a numeric vector or matrix with non-negative weights
    for obtaining a weighted least squares fit.  If a matrix, its
    numbers of rows and columns must be the same as the number of
    objects in \code{x}, and the lower diagonal part is used.
    Otherwise, it is recycled to the number of elements in \code{x}.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\value{
  An object of class \code{"\link{cl_ultrametric}"} containing the
  fitted ultrametric distances.
}
\details{
  For a single dissimilarity object \code{x}, the problem to be solved
  is minimizing
  \deqn{L(u) = \sum_{i,j} w_{ij} (x_{ij} - u_{ij})^2}
  over all \eqn{u} satisfying the ultrametric constraints (i.e., for all
  \eqn{i, j, k}, \eqn{u_{ij} \le \max(u_{ik}, u_{jk})}).  This problem
  is known to be NP hard \bibcitep{Křivánek+Morávek:1986}.

  For an ensemble of dissimilarity objects, the criterion function is
  \deqn{L(u) = \sum_b w_b \sum_{i,j} w_{ij} (x_{ij}(b) - u_{ij})^2,}
  where \eqn{w_b} is the weight given to element \eqn{x_b} of the
  ensemble and can be specified via control parameter \code{weights}
  (default: all ones).  This problem reduces to the above basic problem
  with \eqn{x} as the \eqn{w_b}-weighted mean of the \eqn{x_b}.

  We provide three heuristics for solving the basic problem.

  Method \code{"SUMT"} implements the \acronym{SUMT}
  \bibcitep{Sequential Unconstrained Minimization Technique|Fiacco+McCormick:1968|}
  approach of \bibcitet{De_Soete:1984a} which in turn simplifies the suggestions
  in \bibcitet{Carroll+Pruzansky:1980}.  (See \code{\link{sumt}} for more
  information on the \acronym{SUMT} approach.)  We then use a final
  single linkage hierarchical clustering step to ensure that the
  returned object exactly satisfies the ultrametric constraints.  The
  starting value \eqn{u_0} is obtained by \dQuote{random shaking} of the
  given dissimilarity object (if not given).  If there are missing
  values in \code{x}, i.e., the given dissimilarities are
  \emph{incomplete}, we follow a suggestion of \bibcitet{De_Soete:1984a}, imputing
  the missing values by the weighted mean of the non-missing ones, and
  setting the corresponding weights to zero.

  Available control parameters are \code{method}, \code{control},
  \code{eps}, \code{q}, and \code{verbose}, which have the same roles as
  for \code{\link{sumt}}, and the following.
  
  \describe{
    \item{\code{nruns}}{an integer giving the number of runs to be
      performed.  Defaults to 1.}
    \item{\code{start}}{a single dissimilarity, or a list of
      dissimilarities to be employed as starting values.}
  }

  The default optimization using conjugate gradients should work
  reasonably well for medium to large size problems.  For \dQuote{small}
  ones, using \code{nlm} is usually faster.  Note that the number of
  ultrametric constraints is of the order \eqn{n^3}, where \eqn{n} is
  the number of objects in the dissimilarity object, suggesting to use
  the \acronym{SUMT} approach in favor of
  \code{\link[stats]{constrOptim}}.

  If starting values for the \acronym{SUMT} are provided via
  \code{start}, the number of starting values gives the number of runs
  to be performed, and control option \code{nruns} is ignored.
  Otherwise, \code{nruns} starting values are obtained by random shaking
  of the dissimilarity to be fitted.  In the case of multiple
  \acronym{SUMT} runs, the (first) best solution found is returned.

  Method \code{"IP"} implements the Iterative Projection approach of
  \bibcitet{Hubert+Arabie:1995}.  This iteratively projects the current
  dissimilarities to the closed convex set given by the ultrametric
  constraints (3-point conditions) for a single index triple \eqn{(i, j,
    k)}, in fact replacing the two largest values among \eqn{d_{ij},
    d_{ik}, d_{jk}} by their mean.  The following control parameters can
  be provided via the \code{control} argument.

  \describe{
    \item{\code{nruns}}{an integer giving the number of runs to be
      performed.  Defaults to 1.}
    \item{\code{order}}{a permutation of the numbers from 1 to the
      number of objects in \code{x}, specifying the order in which the
      ultrametric constraints are considered, or a list of such
      permutations.}
    \item{\code{maxiter}}{an integer giving the maximal number of
      iterations to be employed.}
    \item{\code{tol}}{a double indicating the maximal convergence
      tolerance.  The algorithm stops if the total absolute change in
      the dissimilarities in an iteration is less than \code{tol}.}
    \item{\code{verbose}}{a logical indicating whether to provide some
      output on minimization progress.  Defaults to
      \code{getOption("verbose")}.}
  }

  If permutations are provided via \code{order}, the number of these
  gives the number of runs to be performed, and control option
  \code{nruns} is ignored.  Otherwise, \code{nruns} randomly generated
  orders are tried.  In the case of multiple runs, the (first) best
  solution found is returned.

  Non-identical weights and incomplete dissimilarities are currently not
  supported.

  Method \code{"IR"} implements the Iterative Reduction approach
  suggested by \bibcitet{Roux:1988}, see also
  \bibcitet{Barthélemy+Guénoche:1991}.
  This is similar to the Iterative Projection method, but modifies the
  dissimilarities between objects proportionally to the aggregated
  change incurred from the ultrametric projections.  Available control
  parameters are identical to those of method \code{"IP"}.

  Non-identical weights and incomplete dissimilarities are currently not
  supported.

  It should be noted that all methods are heuristics which can not be
  guaranteed to find the global minimum.  Standard practice would
  recommend to use the best solution found in \dQuote{sufficiently many}
  replications of the base algorithm.
}
\references{
  \bibinfo{De_Soete:1984a}{year}{1984a}
  \bibinfo{De_Soete:1984b}{year}{1984b}  
  \bibshow{Barthélemy+Guénoche:1991,
    Carroll+Pruzansky:1980,
    Fiacco+McCormick:1968,
    Hubert+Arabie:1995,
    Křivánek+Morávek:1986,
    Roux:1988,
    De_Soete:1984a,
    De_Soete:1984b}
}
\seealso{
  \code{\link{cl_consensus}} for computing least squares (Euclidean)
  consensus hierarchies by least squares fitting of average ultrametric
  distances;
  \code{\link{l1_fit_ultrametric}}.
}
\examples{
## Least squares fit of an ultrametric to the Miller-Nicely consonant
## phoneme confusion data.
data("Phonemes")
## Note that the Phonemes data set has the consonant misclassification
## probabilities, i.e., the similarities between the phonemes.
d <- as.dist(1 - Phonemes)
u <- ls_fit_ultrametric(d, control = list(verbose = TRUE))
## Cophenetic correlation:
cor(d, u)
## Plot:
plot(u)
## ("Basically" the same as Figure 1 in de Soete (1984b).)
}
\keyword{cluster}
\keyword{optimize}

\name{cl_membership}
\alias{cl_membership}
\alias{as.cl_membership}
\title{Memberships of Partitions}
\description{
  Compute the memberships values for objects representing partitions.
}
\usage{
cl_membership(x, k = n_of_classes(x))
as.cl_membership(x)
}
\arguments{
  \item{x}{an R object representing a partition of objects (for
    \code{cl_membership}) or raw memberships or class ids (for
    \code{as.cl_membership}).}
  \item{k}{an integer giving the number of columns (corresponding to
    class ids) to be used in the membership matrix.  Must not be less,
    and default to, the number of classes in the partition.}
}
\value{
  An object of class \code{"cl_membership"} with the matrix of
  membership values.
}
\details{
  \code{cl_membership} is a generic function.

  The methods provided in package \pkg{clue} handle the partitions
  obtained from clustering functions in the base R distribution, as well
  as packages \pkg{RWeka}, \pkg{cba}, \pkg{cclust}, \pkg{cluster},
  \pkg{e1071}, \pkg{flexclust}, \pkg{flexmix}, \pkg{kernlab},
  \pkg{mclust}, \pkg{movMF} and \pkg{skmeans} (and of course, \pkg{clue}
  itself).

  \code{as.cl_membership} can be used for coercing \dQuote{raw} class
  ids (given as atomic vectors) or membership values (given as numeric
  matrices) to membership objects.
}
\seealso{
  \code{\link{is.cl_partition}}
}
\examples{
## Getting the memberships of a single soft partition.
d <- dist(USArrests)
hclust_methods <-
    c("ward", "single", "complete", "average", "mcquitty")
hclust_results <- lapply(hclust_methods, function(m) hclust(d, m))
names(hclust_results) <- hclust_methods 
## Now create an ensemble from the results.
hens <- cl_ensemble(list = hclust_results)
## And add the results of agnes and diana.
require("cluster")
hens <- c(hens, list(agnes = agnes(d), diana = diana(d)))
## Create a dissimilarity object from this.
d1 <- cl_dissimilarity(hens)
## And compute a soft partition.
party <- fanny(d1, 2)
round(cl_membership(party), 5)
## The "nearest" hard partition to this:
as.cl_hard_partition(party)
## (which has the same class ids as cl_class_ids(party)).

## Extracting the memberships from the elements of an ensemble of
## partitions.
pens <- cl_boot(USArrests, 30, 3)
pens
mems <- lapply(pens, cl_membership)
## And turning these raw memberships into an ensemble of partitions.
pens <- cl_ensemble(list = lapply(mems, as.cl_partition))
pens
pens[[length(pens)]]
}
\keyword{cluster}

\name{cl_agreement}
\alias{cl_agreement}
\title{Agreement Between Partitions or Hierarchies}
\description{Compute the agreement between (ensembles) of partitions or
  hierarchies.
}
\usage{
cl_agreement(x, y = NULL, method = "euclidean", \dots)
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies and dissimilarities,
    or something coercible to that (see \code{\link{cl_ensemble}}).}
  \item{y}{\code{NULL} (default), or as for \code{x}.}
  \item{method}{a character string specifying one of the built-in
    methods for computing agreement, or a function to be taken as
    a user-defined method.  If a character string, its lower-cased
    version is matched against the lower-cased names of the available
    built-in methods using \code{\link{pmatch}}.  See \bold{Details} for
    available built-in methods.}
  \item{\dots}{further arguments to be passed to methods.}
}
\value{
  If \code{y} is \code{NULL}, an object of class \code{"cl_agreement"}
  containing the agreements between the all pairs of components of
  \code{x}.  Otherwise, an object of class \code{"cl_cross_agreement"}
  with the agreements between the components of \code{x} and the
  components of \code{y}.
}
\details{
  If \code{y} is given, its components must be of the same kind as those
  of \code{x} (i.e., components must either all be partitions, or all be
  hierarchies or dissimilarities).

  If all components are partitions, the following built-in methods for
  measuring agreement between two partitions with respective membership
  matrices \eqn{u} and \eqn{v} (brought to a common number of columns)
  are available:
  
  \describe{
    \item{\code{"euclidean"}}{\eqn{1 - d / m}, where \eqn{d} is the
      Euclidean dissimilarity of the memberships, i.e., the square root
      of the minimal sum of the squared differences of \eqn{u} and all
      column permutations of \eqn{v}, and \eqn{m} is an upper bound for
      the maximal Euclidean dissimilarity.
      See \bibcitet{Dimitriadou+Weingessel+Hornik:2002}.}
    \item{\code{"manhattan"}}{\eqn{1 - d / m}, where \eqn{d} is the
      Manhattan dissimilarity of the memberships, i.e., the minimal
      sum of the absolute differences of \eqn{u} and all column
      permutations of \eqn{v}, and \eqn{m} is an upper bound for the
      maximal Manhattan dissimilarity.}
    \item{\code{"Rand"}}{the Rand index (the rate of distinct pairs of
      objects both in the same class or both in different classes in
      both partitions), see \bibcitet{Rand:1971} or
      \bibcitet{|R:Gordon:1999|page 198}.
      For soft partitions, (currently) the Rand index of the
      corresponding nearest hard partitions is used.}
    \item{\code{"cRand"}}{the Rand index corrected for agreement by
      chance, see \bibcitet{Hubert+Arabie:1985} or
      \bibcitet{|R:Gordon:1999|page 198}.
      Can only be used for hard partitions.}
    \item{\code{"NMI"}}{Normalized Mutual Information, see
      \bibcitet{Strehl+Ghosh:2002}.
      For soft partitions, (currently) the NMI of the
      corresponding nearest hard partitions is used.} 
    \item{\code{"KP"}}{the Katz-Powell index, i.e., the product-moment
      correlation coefficient between the elements of the co-membership
      matrices \eqn{C(u) = u u'} and \eqn{C(v)}, respectively,
      see \bibcitet{Katz+Powell:1953}.  For soft partitions, (currently) the
      Katz-Powell index of the corresponding nearest hard partitions is
      used.  (Note that for hard partitions, the \eqn{(i,j)} entry of
      \eqn{C(u)} is one iff objects \eqn{i} and \eqn{j} are in the same
      class.)}
    \item{\code{"angle"}}{the maximal cosine of the angle between the
      elements of \eqn{u} and all column permutations of \eqn{v}.}
    \item{\code{"diag"}}{the maximal co-classification rate, i.e., the
      maximal rate of objects with the same class ids in both
      partitions after arbitrarily permuting the ids.}
    \item{\code{"FM"}}{the index of \bibcitet{Fowlkes+Mallows:1983}, i.e.,
      the ratio \eqn{N_{xy} / \sqrt{N_x N_y}}{N_xy / sqrt(N_x N_y)} of
      the number \eqn{N_{xy}}{N_xy} of distinct pairs of objects in the
      same class in both partitions and the geometric mean of the
      numbers \eqn{N_x} and \eqn{N_y} of distinct pairs of objects in
      the same class in partition \eqn{x} and partition \eqn{y},
      respectively.
      For soft partitions, (currently) the Fowlkes-Mallows index of the
      corresponding nearest hard partitions is used.}
    \item{\code{"Jaccard"}}{the Jaccard index, i.e., the ratio of the
      numbers of distinct pairs of objects in the same class in both
      partitions and in at least one partition, respectively.
      For soft partitions, (currently) the Jaccard index of the
      corresponding nearest hard partitions is used.}
    \item{\code{"purity"}}{the purity of the classes of \code{x} with
      respect to those of \code{y}, i.e.,
      \eqn{\sum_j \max_i n_{ij} / n},
      where \eqn{n_{ij}} is the joint frequency of objects in class
      \eqn{i} for \code{x} and in class \eqn{j} for \code{y}, and
      \eqn{n} is the total number of objects.}
    \item{\code{"PS"}}{Prediction Strength, see
      \bibcitet{Tibshirani+Walther:2005}: the minimum, over all classes
      \eqn{j} of \code{y}, of the maximal rate of objects in the same
      class for \code{x} and in class \eqn{j} for \code{y}.}
  }

  If all components are hierarchies, available built-in methods for
  measuring agreement between two hierarchies with respective
  ultrametrics \eqn{u} and \eqn{v} are as follows.

  \describe{
    \item{\code{"euclidean"}}{\eqn{1 / (1 + d)}, where \eqn{d} is the
      Euclidean dissimilarity of the ultrametrics (i.e., the square root
      of the sum of the squared differences of \eqn{u} and \eqn{v}).}
    \item{\code{"manhattan"}}{\eqn{1 / (1 + d)}, where \eqn{d} is the
      Manhattan dissimilarity of the ultrametrics (i.e., the sum of the
      absolute differences of \eqn{u} and \eqn{v}).}
    \item{\code{"cophenetic"}}{The cophenetic correlation coefficient.
      (I.e., the product-moment correlation of the ultrametrics.)}
    \item{\code{"angle"}}{the cosine of the angle between the
      ultrametrics.}
    \item{\code{"gamma"}}{\eqn{1 - d}, where \eqn{d} is the rate of
      inversions between the associated ultrametrics (i.e., the rate of
      pairs \eqn{(i,j)} and \eqn{(k,l)} for which \eqn{u_{ij} < u_{kl}}
      and \eqn{v_{ij} > v_{kl}}).  (This agreement measure is a linear
      transformation of Kruskal's \eqn{\gamma}{gamma}.)}
  }

  The measures based on ultrametrics also allow computing agreement with
  \dQuote{raw} dissimilarities on the underlying objects (R objects
  inheriting from class \code{"dist"}).

  If a user-defined agreement method is to be employed, it must be a
  function taking two clusterings as its arguments.

  Symmetric agreement objects of class \code{"cl_agreement"} are
  implemented as symmetric proximity objects with self-proximities
  identical to one, and inherit from class \code{"cl_proximity"}.  They
  can be coerced to dense square matrices using \code{as.matrix}.  It is
  possible to use 2-index matrix-style subscripting for such objects;
  unless this uses identical row and column indices, this results in a
  (non-symmetric agreement) object of class \code{"cl_cross_agreement"}.
}
\references{
  \bibshow{Dimitriadou+Weingessel+Hornik:2002,
    Fowlkes+Mallows:1983,
    R:Gordon:1999,
    Hubert+Arabie:1985,
    Katz+Powell:1953,
    Rand:1971,
    Strehl+Ghosh:2002,
    Tibshirani+Walther:2005}
}
\seealso{
  \code{\link{cl_dissimilarity}};
  \code{\link[e1071]{classAgreement}} in package \pkg{e1071}.
}
\examples{
## An ensemble of partitions.
data("CKME")
pens <- CKME[1 : 20]		# for saving precious time ...
summary(c(cl_agreement(pens)))
summary(c(cl_agreement(pens, method = "Rand")))
summary(c(cl_agreement(pens, method = "diag")))
cl_agreement(pens[1:5], pens[6:7], method = "NMI")
## Equivalently, using subscripting.
cl_agreement(pens, method = "NMI")[1:5, 6:7]

## An ensemble of hierarchies.
d <- dist(USArrests)
hclust_methods <-
    c("ward", "single", "complete", "average", "mcquitty")
hclust_results <- lapply(hclust_methods, function(m) hclust(d, m))
names(hclust_results) <- hclust_methods 
hens <- cl_ensemble(list = hclust_results)
summary(c(cl_agreement(hens)))
## Note that the Euclidean agreements are *very* small.
## This is because the ultrametrics differ substantially in height:
u <- lapply(hens, cl_ultrametric)
round(sapply(u, max), 3)
## Rescaling the ultrametrics to [0, 1] gives:
u <- lapply(u, function(x) (x - min(x)) / (max(x) - min(x)))
shens <- cl_ensemble(list = lapply(u, as.cl_dendrogram))
summary(c(cl_agreement(shens)))
## Au contraire ...
summary(c(cl_agreement(hens, method = "cophenetic")))
cl_agreement(hens[1:3], hens[4:5], method = "gamma")
}
\keyword{cluster}

\name{Kinship82}
\alias{Kinship82}
\title{Rosenberg-Kim Kinship Terms Partition Data}
\description{
  Partitions of 15 kinship terms given by 85 female undergraduates at
  Rutgers University who were asked to sort the terms into classes
  \dQuote{on the basis of some aspect of meaning}.
}
\usage{data("Kinship82")}
\format{
  A cluster ensemble of 85 hard partitions of the 15 kinship terms.
}
\details{
  \bibcitet{Rosenberg+Kim:1975} describe an experiment where perceived
  similarities of the kinship terms were obtained from six different
  \dQuote{sorting} experiments.  These \dQuote{original} Rosenberg-Kim
  kinship terms data were published in
  \bibcitet{Arabie+Carroll+DeSarbo:1987}, and are also contained in file
  \file{indclus.data} in the shell archive
  \url{https://netlib.org/mds/indclus.shar}. 

  For one of the experiments, partitions of the terms were printed in
  \bibcitet{Rosenberg:1982}.
  Comparison with the original data indicates that
  the partition data have the \dQuote{nephew} and \dQuote{niece} columns
  interchanged, which is corrected in the data set at hand.
}
\source{
  Table 7.1 in \bibcitet{Rosenberg:1982}, with the \dQuote{nephew} and 
  \dQuote{niece} columns interchanged.
}
\references{
  \bibshow{Arabie+Carroll+DeSarbo:1987,
    Rosenberg+Kim:1975,
    Rosenberg:1982}
}
\keyword{datasets}

\name{sumt}
\alias{sumt}
\title{Sequential Unconstrained Minimization Technique}
\description{
  Solve constrained optimization problems via the Sequential
  Unconstrained Minimization Technique (\acronym{SUMT}).
}
\usage{
sumt(x0, L, P, grad_L = NULL, grad_P = NULL, method = NULL,
     eps = NULL, q = NULL, verbose = NULL, control = list())
}
\arguments{
  \item{x0}{a list of starting values, or a single starting value.}
  \item{L}{a function to minimize.}
  \item{P}{a non-negative penalty function such that \eqn{P(x)} is zero
    iff the constraints are satisfied.}
  \item{grad_L}{a function giving the gradient of \code{L}, or
    \code{NULL} (default).}
  \item{grad_P}{a function giving the gradient of \code{P}, or
    \code{NULL} (default).}
  \item{method}{a character string, or \code{NULL}.  If not given,
    \code{"CG"} is used.  If equal to \code{"nlm"}, minimization is
    carried out using \code{\link[stats]{nlm}}.  Otherwise,
    \code{\link[stats]{optim}} is used with \code{method} as the given
    method.}
  \item{eps}{the absolute convergence tolerance.  The algorithm stops if
    the (maximum) distance between successive \code{x} values is
    less than \code{eps}.

    Defaults to \code{sqrt(.Machine$double.eps)}.}
  \item{q}{a double greater than one controlling the growth of the
    \eqn{\rho_k} as described in \bold{Details}.
    
    Defaults to 10.}
  \item{verbose}{a logical indicating whether to provide some output on
    minimization progress.
    
    Defaults to \code{getOption("verbose")}.}
  \item{control}{a list of control parameters to be passed to the
    minimization routine in case \code{optim} is used.}
}
\details{
  The Sequential Unconstrained Minimization Technique is a heuristic for
  constrained optimization.  To minimize a function \eqn{L} subject to
  constraints, one employs a non-negative function \eqn{P} penalizing
  violations of the constraints, such that \eqn{P(x)} is zero iff \eqn{x}
  satisfies the constraints.  One iteratively minimizes \eqn{L(x) +
    \rho_k P(x)}, where the \eqn{\rho} values are increased according to
  the rule \eqn{\rho_{k+1} = q \rho_k} for some constant \eqn{q > 1},
  until convergence is obtained in the sense that the Euclidean distance
  between successive solutions \eqn{x_k} and \eqn{x_{k+1}} is small
  enough.  Note that the \dQuote{solution} \eqn{x} obtained does not
  necessarily satisfy the constraints, i.e., has zero \eqn{P(x)}.  Note
  also that there is no guarantee that global (approximately)
  constrained optima are found.  Standard practice would recommend to
  use the best solution found in \dQuote{sufficiently many} replications
  of the algorithm.

  The unconstrained minimizations are carried out by either
  \code{\link[stats]{optim}} or \code{\link[stats]{nlm}}, using analytic
  gradients if both \code{grad_L} and \code{grad_P} are given, and
  numeric ones otherwise.

  If more than one starting value is given, the solution with the
  minimal augmented criterion function value is returned.
}
\value{
  A list inheriting from class \code{"sumt"}, with components \code{x},
  \code{L}, \code{P}, and \code{rho} giving the solution obtained, the
  value of the criterion and penalty function at \code{x}, and the final
  \eqn{\rho} value used in the augmented criterion function. 
}

\references{
  \bibshow{Fiacco+McCormick:1968}
}
\keyword{optimize}

\name{solve_LSAP}
\encoding{UTF-8}
\alias{solve_LSAP}
\title{Solve Linear Sum Assignment Problem}
\description{
  Solve the linear sum assignment problem using the Hungarian method.
}
\usage{
solve_LSAP(x, maximum = FALSE)
}
\arguments{
  \item{x}{a matrix with nonnegative entries and at least as many
    columns as rows.}
  \item{maximum}{a logical indicating whether to minimize of maximize
    the sum of assigned costs.}
}
\details{
  If \eqn{nr} and \eqn{nc} are the numbers of rows and columns of
  \code{x}, \code{solve_LSAP} finds an optimal \emph{assignment} of rows
  to columns, i.e., a one-to-one map \code{p} of the numbers from 1 to
  \eqn{nr} to the numbers from 1 to \eqn{nc} (a permutation of these
  numbers in case \code{x} is a square matrix) such that
  \eqn{\sum_{i=1}^{nr} x[i, p[i]]} is minimized or maximized.

  This assignment can be found using a linear program (and package
  \pkg{lpSolve} provides a function \code{lp.assign} for doing so), but
  typically more efficiently and provably in polynomial time
  \eqn{O(n^3)} using primal-dual methods such as the so-called Hungarian
  method (see the references).
}
\value{
  An object of class \code{"solve_LSAP"} with the optimal assignment of
  rows to columns.
}
\references{
  \bibshow{Papadimitriou+Steiglitz:1982}
}
\author{
  Walter Böhm \email{Walter.Boehm@wu.ac.at} kindly provided C code
  implementing the Hungarian method.
}
\examples{
x <- matrix(c(5, 1, 4, 3, 5, 2, 2, 4, 4), nrow = 3)
solve_LSAP(x)
solve_LSAP(x, maximum = TRUE)
## To get the optimal value (for now):
y <- solve_LSAP(x)
sum(x[cbind(seq_along(y), y)])
}
\keyword{optimize}

\name{Phonemes}
\alias{Phonemes}
\title{Miller-Nicely Consonant Phoneme Confusion Data}
\description{
  Miller-Nicely data on the auditory confusion of 16 consonant
  phonemes.
}
\usage{data("Phonemes")}
\format{
  A symmetric matrix of the misclassification probabilities of 16
  English consonant phonemes.
}
\details{
  \bibcitet{Miller+Nicely:1955}
  obtained the confusions by exposing female
  subjects to a series of syllables consisting of one of the 16
  consonants followed by the vowel \samp{a} under 17 different
  experimental conditions.  The data provided are obtained from
  aggregating the six so-called flat-noise conditions in which only the
  speech-to-noise ratio was varied into a single matrix of
  misclassification frequencies.
}
\source{
  The data set is also contained in file \file{mapclus.data} in the
  shell archive \url{https://netlib.org/mds/mapclus.shar}.
}
\references{
  \bibshow{Miller+Nicely:1955}
}
\keyword{datasets}

\name{cl_dissimilarity}
\encoding{UTF-8}
\alias{cl_dissimilarity}
\title{Dissimilarity Between Partitions or Hierarchies}
\description{Compute the dissimilarity between (ensembles) of partitions
  or hierarchies.}
\usage{
cl_dissimilarity(x, y = NULL, method = "euclidean", \dots)
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies and dissimilarities,
    or something coercible to that (see \code{\link{cl_ensemble}}).}
  \item{y}{\code{NULL} (default), or as for \code{x}.}
  \item{method}{a character string specifying one of the built-in
    methods for computing dissimilarity, or a function to be taken as
    a user-defined method.  If a character string, its lower-cased
    version is matched against the lower-cased names of the available
    built-in methods using \code{\link{pmatch}}.  See \bold{Details} for
    available built-in methods.}
  \item{\dots}{further arguments to be passed to methods.}
}
\value{
  If \code{y} is \code{NULL}, an object of class
  \code{"cl_dissimilarity"} containing the dissimilarities between all
  pairs of components of \code{x}.  Otherwise, an object of class
  \code{"cl_cross_dissimilarity"} with the dissimilarities between the
  components of \code{x} and the components of \code{y}.
}
\details{
  If \code{y} is given, its components must be of the same kind as those
  of \code{x} (i.e., components must either all be partitions, or all be
  hierarchies or dissimilarities).

  If all components are partitions, the following built-in methods for
  measuring dissimilarity between two partitions with respective
  membership matrices \eqn{u} and \eqn{v} (brought to a common number of
  columns) are available:

  \describe{
    \item{\code{"euclidean"}}{the Euclidean dissimilarity of the
      memberships, i.e., the square root of the minimal sum of the
      squared differences of \eqn{u} and all column permutations of
      \eqn{v}.  See \bibcitet{Dimitriadou+Weingessel+Hornik:2002}.}
    \item{\code{"manhattan"}}{the Manhattan dissimilarity of the
      memberships, i.e., the minimal sum of the absolute differences of
      \eqn{u} and all column permutations of \eqn{v}.}
    \item{\code{"comemberships"}}{the Euclidean dissimilarity of the
      elements of the co-membership matrices \eqn{C(u) = u u'} and
      \eqn{C(v)}, i.e., the square root of the sum of the squared
      differences of \eqn{C(u)} and \eqn{C(v)}.}
    \item{\code{"symdiff"}}{the cardinality of the symmetric set
      difference of the sets of co-classified pairs of distinct objects
      in the partitions.  I.e., the number of distinct pairs of objects
      in the same class in exactly one of the partitions.
      (Alternatively, the cardinality of the symmetric set difference
      between the (binary) equivalence relations corresponding to the
      partitions.)  For soft partitions, (currently) the symmetric set
      difference of the corresponding nearest hard partitions is used.}
    \item{\code{"Rand"}}{the Rand distance, i.e., the rate of distinct
      pairs of objects in the same class in exactly one of the
      partitions.  (Related to the Rand index \eqn{a} via the linear
      transformation \eqn{d = (1 - a) / 2}.)  For soft partitions,
      (currently) the Rand distance of the corresponding nearest hard
      partitions is used.}
    \item{\code{"GV1"}}{the square root of the dissimilarity
      \eqn{\Delta_1}{Delta_1} used for the first model in
      \bibcitet{Gordon+Vichi:2001},
      i.e., the square root of the minimal sum of the
      squared differences of the \emph{matched} non-zero columns of
      \eqn{u} and \eqn{v}.}
    \item{\code{"BA/\var{d}"}}{distance measures for hard partitions
      discussed in \bibcitet{Boorman+Arabie:1972}, with \var{d} one of
      \samp{A}, \samp{C}, \samp{D}, or \samp{E}.  For soft partitions,
      the distances of the corresponding nearest hard partitions are
      used.

      \code{"BA/A"} is the minimum number of single element moves (move
      from one class to another or a new one) needed to transform one
      partition into the other.  Introduced in \bibcitet{Rubin:1967}.

      \code{"BA/C"} is the minimum number of lattice moves for
      transforming one partition into the other, where partitions are
      said to be connected by a lattice move if one is \emph{just} finer
      than the other (i.e., there is no other partition between them) in
      the partition lattice (see \code{\link{cl_meet}}).  Equivalently,
      with \eqn{z} the join of \code{x} and \code{y} and \eqn{S} giving
      the number of classes, this can be written as \eqn{S(x) + S(y) - 2
	S(z)}.  Attributed to David Pavy.
	
      \code{"BA/D"} is the \dQuote{pair-bonds} distance, which can be
      defined as \eqn{S(x) + S(y) - 2 S(z)}, with \eqn{z} the meet of
      \code{x} and \code{y} and \eqn{S} the \emph{supervaluation} (i.e.,
      non-decreasing with respect to the partial order on the partition
      lattice) function \eqn{\sum_i (n_i (n_i - 1)) / (n (n - 1))},
      where the \eqn{n_i} are the numbers of objects in the respective
      classes of the partition (such that \eqn{n_i (n_i - 1) / 2} are the
      numbers of pair bonds in the classes), and \eqn{n} the total
      number of objects.

      \code{"BA/E"} is the normalized information distance, defined as
      \eqn{1 - I / H}, where \eqn{I} is the average mutual information
      between the partitions, and \eqn{H} is the average entropy of the
      meet \eqn{z} of the partitions.
      Introduced in \bibcitet{Rajski:1961}.

      (Boorman and Arabie also discuss a distance measure (\eqn{B})
      based on the minimum number of set moves needed to transform one
      partition into the other, which, differently from the \eqn{A} and
      \eqn{C} distance measures is hard to compute \bibcitep{Day:1981}
      and (currently) not provided.)}
    \item{\code{"VI"}}{Variation of Information, see
      \bibcitet{Meila:2003}.  If
      \code{\dots} has an argument named \code{weights}, it is taken to
      specify case weights.}
    \item{\code{"Mallows"}}{the Mallows-type distance by
      \bibcitet{Zhou+Li+Zha:2005},
      which is related to the Monge-Kantorovich mass
      transfer problem, and given as the \eqn{p}-th root of the minimal
      value of the transportation problem \eqn{\sum w_{jk} \sum_i
      |u_{ij} - v_{ik}| ^ p} with constraints \eqn{w_{jk} \ge 0},
      \eqn{\sum_j w_{jk} = \alpha_j}, \eqn{\sum_k w_{jk} = \beta_k},
      where \eqn{\sum_j \alpha_j = \sum_k \beta_k}.  The parameters
      \eqn{p}, \eqn{\alpha} and \eqn{\beta} all default to one (in this
      case, the Mallows distance coincides with the Manhattan
      dissimilarity), and can be specified via additional arguments
      named \code{p}, \code{alpha}, and \code{beta}, respectively.}
    \item{\code{"CSSD"}}{the Cluster Similarity Sensitive Distance of
      \bibcitet{Zhou+Li+Zha:2005},
      which is given as the minimal value of
      \eqn{\sum_{k,l} (1 - 2 w_{kl} / (\alpha_k + \beta_l)) L_{kl}},
      where \eqn{L_{kl} = \sum_i u_{ik} v_{il} d(p_{x;k}, p_{y;l})} with
      \eqn{p_{x;k}} and \eqn{p_{y;l}} the prototype of the \eqn{k}-th
      class of \code{x} and the \eqn{l}-th class of \code{y},
      respectively, \eqn{d} is the distance between these, and the
      \eqn{w_{kl}} as for Mallows distance.  If prototypes are matrices,
      the Euclidean distance between these is used as default.  Using
      the additional argument \code{L}, one can give a matrix of
      \eqn{L_{kl}} values, or the function \eqn{d}.  Parameters
      \eqn{\alpha} and \eqn{\beta} all default to one, and can be
      specified via additional arguments named \code{alpha} and
      \code{beta}, respectively.}
  }

  For hard partitions, both Manhattan and squared Euclidean
  dissimilarity give twice the \emph{transfer distance}
  \bibcitep{Charon+Denoeud+Guénoche:2006},
  which is the minimum number of objects that must be removed so
  that the implied partitions (restrictions to the remaining objects)
  are identical.  This is also known as the \emph{\eqn{R}-metric} in
  \bibcitet{Day:1981},
  i.e., the number of augmentations and removals of single
  objects needed to transform one partition into the other, and the
  \emph{partition-distance} in \bibcitet{Gusfield:2002}, and equals
  twice the number of single element moves distance of Boorman and
  Arabie. 

  For hard partitions, the pair-bonds (Boorman-Arabie \eqn{D}) distance
  is identical to the Rand distance, and can also be written as the
  Manhattan distance between the co-membership matrices corresponding to
  the partitions, or equivalently, their symdiff distance, normalized by
  \eqn{n (n - 1)}.

  If all components are hierarchies, available built-in methods for
  measuring dissimilarity between two hierarchies with respective
  ultrametrics \eqn{u} and \eqn{v} are as follows.

  \describe{
    \item{\code{"euclidean"}}{the Euclidean dissimilarity of the
      ultrametrics (i.e., the square root of the sum of the squared
      differences of \eqn{u} and \eqn{v}).}
    \item{\code{"manhattan"}}{the Manhattan dissimilarity of the
      ultrametrics (i.e., the sum of the absolute differences of \eqn{u}
      and \eqn{v}).}
    \item{\code{"cophenetic"}}{\eqn{1 - c^2}, where \eqn{c} is the
      cophenetic correlation coefficient (i.e., the product-moment
      correlation of the ultrametrics).}
    \item{\code{"gamma"}}{the rate of inversions between the
      ultrametrics (i.e., the rate of pairs \eqn{(i,j)} and \eqn{(k,l)}
      for which \eqn{u_{ij} < u_{kl}} and \eqn{v_{ij} > v_{kl}}).}
    \item{\code{"symdiff"}}{the cardinality of the symmetric set
      difference of the sets of classes (hierarchies in the strict
      sense) induced by the dendrograms.  I.e., the number of sets of
      objects obtained by a split in exactly one of the hierarchies.}
    \item{\code{"Chebyshev"}}{the Chebyshev (maximal) dissimilarity of
      the ultrametrics (i.e., the maximum of the absolute differences of
      \eqn{u} and \eqn{v}).}
    \item{\code{"Lyapunov"}}{the logarithm of the product of the
      maximal and minimal ratios of the ultrametrics.  This is also
      known as the \dQuote{Hilbert projective metric} on the cone
      represented by the ultrametrics
      \bibcitep{e.g.|Jardine+Sibson:1971|page 107},
      and only defined for \emph{strict} ultrametrics (which
      are strictly positive for distinct objects).}
    \item{\code{"BO"}}{the \eqn{m_\delta} family of tree metrics by
      \bibcitet{Boorman+Olivier:1973}, which are of the form
      \eqn{m_\delta = \int_0^\infty \delta(p(h), q(h)) dh}, where
      \eqn{p(h)} and
      \eqn{q(h)} are the hard partitions obtaining by cutting the trees
      (dendrograms) at height \eqn{h}, and \eqn{\delta} is a suitably
      dissimilarity measure for partitions.  In particular, when taking
      \eqn{\delta} as symdiff or Rand dissimilarity, \eqn{m_\delta} is
      the Manhattan dissimilarity of the hierarchies.

      If \code{\dots} has an argument named \code{delta} it is taken to
      specify the partition dissimilarity \eqn{\delta} to be employed.}
    \item{\code{"spectral"}}{the spectral norm (2-norm) of the
      differences of the ultrametrics, suggested in
      \bibcitet{Mérigot+Durbec+Gaertner:2010}.}
  }

  The measures based on ultrametrics also allow computing dissimilarity
  with \dQuote{raw} dissimilarities on the underlying objects (R objects
  inheriting from class \code{"dist"}).

  If a user-defined dissimilarity method is to be employed, it must be a
  function taking two clusterings as its arguments.

  Symmetric dissimilarity objects of class \code{"cl_dissimilarity"} are
  implemented as symmetric proximity objects with self-proximities
  identical to zero, and inherit from class \code{"cl_proximity"}.  They
  can be coerced to dense square matrices using \code{as.matrix}.  It
  is possible to use 2-index matrix-style subscripting for such objects;
  unless this uses identical row and column indices, this results in a
  (non-symmetric dissimilarity) object of class
  \code{"cl_cross_dissimilarity"}.

  Symmetric dissimilarity objects also inherit from class
  \code{"\link{dist}"} (although they currently do not \dQuote{strictly}
  extend this class), thus making it possible to use them directly for
  clustering algorithms based on dissimilarity matrices of this class,
  see the examples.
}
\references{
  \bibshow{Boorman+Arabie:1972,
    Boorman+Olivier:1973,
    Charon+Denoeud+Guénoche:2006,
    Day:1981,
    Dimitriadou+Weingessel+Hornik:2002,
    Gordon+Vichi:2001,
    Gusfield:2002,
    Jardine+Sibson:1971,
    Meila:2003,
    Mérigot+Durbec+Gaertner:2010,
    Rajski:1961,
    Rubin:1967,
    Zhou+Li+Zha:2005}
}  
\seealso{
  \code{\link{cl_agreement}}
}
\examples{
## An ensemble of partitions.
data("CKME")
pens <- CKME[1 : 30]
diss <- cl_dissimilarity(pens)
summary(c(diss))
cl_dissimilarity(pens[1:5], pens[6:7])
## Equivalently, using subscripting.
diss[1:5, 6:7]
## Can use the dissimilarities for "secondary" clustering
## (e.g. obtaining hierarchies of partitions):
hc <- hclust(diss)
plot(hc)

## Example from Boorman and Arabie (1972).
P1 <- as.cl_partition(c(1, 2, 2, 2, 3, 3, 2, 2))
P2 <- as.cl_partition(c(1, 1, 2, 2, 3, 3, 4, 4))
cl_dissimilarity(P1, P2, "BA/A")
cl_dissimilarity(P1, P2, "BA/C")

## Hierarchical clustering.
d <- dist(USArrests)
x <- hclust(d)
cl_dissimilarity(x, d, "cophenetic")
cl_dissimilarity(x, d, "gamma")
}
\keyword{cluster}

\name{cl_pclust}
\alias{cl_pclust}
\title{Prototype-Based Partitions of Clusterings}
\description{
  Compute prototype-based partitions of a cluster ensemble by minimizing
  \eqn{\sum w_b u_{bj}^m d(x_b, p_j)^e}, the sum of the case-weighted and
  membership-weighted \eqn{e}-th powers of the dissimilarities between
  the elements \eqn{x_b} of the ensemble and the prototypes \eqn{p_j},
  for suitable dissimilarities \eqn{d} and exponents \eqn{e}.
}
\usage{
cl_pclust(x, k, method = NULL, m = 1, weights = 1,
          control = list())
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies, or something
    coercible to that (see \code{\link{cl_ensemble}}).}
  \item{k}{an integer giving the number of classes to be used in the
    partition.}
  \item{method}{the consensus method to be employed, see
    \code{\link{cl_consensus}}.}
  \item{m}{a number not less than 1 controlling the softness of the
    partition (as the \dQuote{fuzzification parameter} of the fuzzy
    \eqn{c}-means algorithm).  The default value of 1 corresponds to
    hard partitions obtained from a generalized \eqn{k}-means problem;
    values greater than one give partitions of increasing softness
    obtained from a generalized fuzzy \eqn{c}-means problem.}
  \item{weights}{a numeric vector of non-negative case weights.
    Recycled to the number of elements in the ensemble given by \code{x}
    if necessary.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\value{
  An object of class \code{"cl_partition"} representing the obtained 
  \dQuote{secondary} partition by an object of class \code{"cl_pclust"},
  which is a list containing at least the following components.
  \item{prototypes}{a cluster ensemble with the \eqn{k} prototypes.}
  \item{membership}{an object of class \code{"\link{cl_membership}"}
    with the membership values \eqn{u_{bj}}.}
  \item{cluster}{the class ids of the nearest hard partition.}
  \item{silhouette}{Silhouette information for the partition, see
    \code{\link[cluster]{silhouette}}.} 
  \item{validity}{precomputed validity measures for the partition.}
  \item{m}{the softness control argument.}
  \item{call}{the matched call.}
  \item{d}{the dissimilarity function \eqn{d = d(x, p)} employed.}
  \item{e}{the exponent \eqn{e} employed.}
}
\details{
  Partitioning is performed using \code{\link{pclust}} via a family
  constructed from \code{method}.  The dissimilarities \eqn{d} and
  exponent \eqn{e} are implied by the consensus method employed, and
  inferred via a registration mechanism currently only made available to
  built-in consensus methods.  The default methods compute Least Squares
  Euclidean consensus clusterings, i.e., use Euclidean dissimilarity
  \eqn{d} and \eqn{e = 2}.

  For \eqn{m = 1}, the partitioning procedure was introduced by
  \bibcitet{Gaul+Schader:1988}
  for \dQuote{Clusterwise Aggregation of Relations} (with
  the same domains), containing equivalence relations, i.e., hard
  partitions, as a special case.

  Available control parameters are as for \code{\link{pclust}}.

  The fixed point approach employed is a heuristic which cannot be
  guaranteed to find the global minimum (as this is already true for the
  computation of consensus clusterings).  Standard practice would
  recommend to use the best solution found in \dQuote{sufficiently many}
  replications of the base algorithm.
}
\references{
  \bibshow{Bezdek:1981, Gaul+Schader:1988}
}
\examples{
## Use a precomputed ensemble of 50 k-means partitions of the
## Cassini data.
data("CKME")
CKME <- CKME[1 : 30]		# for saving precious time ...
diss <- cl_dissimilarity(CKME)
hc <- hclust(diss)
plot(hc)
## This suggests using a partition with three classes, which can be
## obtained using cutree(hc, 3).  Could use cl_consensus() to compute
## prototypes as the least squares consensus clusterings of the classes,
## or alternatively:
set.seed(123)
x1 <- cl_pclust(CKME, 3, m = 1)
x2 <- cl_pclust(CKME, 3, m = 2)
## Agreement of solutions.
cl_dissimilarity(x1, x2)
table(cl_class_ids(x1), cl_class_ids(x2))
}
\keyword{cluster}

\name{cl_consensus}
\alias{cl_consensus}
\title{Consensus Partitions and Hierarchies}
\description{
  Compute the consensus clustering of an ensemble of partitions or
  hierarchies.
}
\usage{
cl_consensus(x, method = NULL, weights = 1, control = list())
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies, or something
    coercible to that (see \code{\link{cl_ensemble}}).}
  \item{method}{a character string specifying one of the built-in
    methods for computing consensus clusterings, or a function to be
    taken as a user-defined method, or \code{NULL} (default value).  If
    a character string, its lower-cased version is matched against the
    lower-cased names of the available built-in methods using
    \code{\link{pmatch}}.  See \bold{Details} for available built-in
    methods and defaults.}
  \item{weights}{a numeric vector with non-negative case weights.
    Recycled to the number of elements in the ensemble given by \code{x}
    if necessary.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\value{
  The consensus partition or hierarchy.
}
\details{
  Consensus clusterings \dQuote{synthesize} the information in the
  elements of a cluster ensemble into a single clustering, often by
  minimizing a criterion function measuring how dissimilar consensus
  candidates are from the (elements of) the ensemble (the so-called
  \dQuote{optimization approach} to consensus clustering).  

  The most popular criterion functions are of the form \eqn{L(x) = \sum
    w_b d(x_b, x)^p}, where \eqn{d} is a suitable dissimilarity measure
  (see \code{\link{cl_dissimilarity}}), \eqn{w_b} is the case weight
  given to element \eqn{x_b} of the ensemble, and \eqn{p \ge 1}.  If
  \eqn{p = 1} and minimization is over all possible base clusterings, a
  consensus solution is called a \emph{median} of the ensemble; if
  minimization is restricted to the elements of the ensemble, a
  consensus solution is called a \emph{medoid} (see
  \code{\link{cl_medoid}}).  For \eqn{p = 2}, we obtain \emph{least
    squares} consensus partitions and hierarchies (generalized means).
  See also \bibcitet{R:Gordon:1999} for more information.

  If all elements of the ensemble are partitions, the built-in consensus
  methods compute consensus partitions by minimizing a criterion of the
  form \eqn{L(x) = \sum w_b d(x_b, x)^p} over all hard or soft
  partitions \eqn{x} with a given (maximal) number \eqn{k} of classes.
  Available built-in methods are as follows.
  
  \describe{
    \item{\code{"SE"}}{a fixed-point algorithm for obtaining \emph{soft}
      least squares Euclidean consensus partitions (i.e., for minimizing
      \eqn{L} with Euclidean dissimilarity \eqn{d} and \eqn{p =	2} over
      all soft partitions with a given maximal number of classes).

      This iterates between individually matching all partitions to the
      current approximation to the consensus partition, and computing
      the next approximation as the membership matrix closest to a
      suitable weighted average of the memberships of all partitions
      after permuting their columns for the optimal matchings of class
      ids.

      The following control parameters are available for this method.
      
      \describe{
	\item{\code{k}}{an integer giving the number of classes to be
	  used for the least squares consensus partition.
	  By default, the maximal number of classes in the ensemble is
	  used.}
	\item{\code{maxiter}}{an integer giving the maximal number of
	  iterations to be performed.
	  Defaults to 100.}
	\item{\code{nruns}}{an integer giving the number of runs to be
	  performed.  Defaults to 1.}
	\item{\code{reltol}}{the relative convergence tolerance.
	  Defaults to \code{sqrt(.Machine$double.eps)}.}
	\item{\code{start}}{a matrix with number of rows equal to the
	  number of objects of the cluster ensemble, and \eqn{k}
	  columns, to be used as a starting value, or a list of such
	  matrices.  By default, suitable random membership matrices are
	  used.}
	\item{\code{verbose}}{a logical indicating whether to provide
	  some output on minimization progress.
	  Defaults to \code{getOption("verbose")}.}
      }
      In the case of multiple runs, the first optimum found is returned.

      This method can also be referred to as \code{"soft/euclidean"}.
    }

    \item{\code{"GV1"}}{the fixed-point algorithm for the \dQuote{first
	model} in \bibcitet{Gordon+Vichi:2001} for minimizing \eqn{L}
      with \eqn{d} being GV1 dissimilarity and \eqn{p = 2} over all soft
      partitions with a given maximal number of classes.

      This is similar to \code{"SE"}, but uses GV1 rather than Euclidean
      dissimilarity.

      Available control parameters are the same as for \code{"SE"}.
    }
    
    \item{\code{"DWH"}}{an extension of the greedy algorithm in
      \bibcitet{Dimitriadou+Weingessel+Hornik:2002} for (approximately)
      obtaining soft least squares Euclidean consensus partitions.
      The reference provides some structure theory relating finding
      the consensus partition to an instance of the multiple assignment
      problem, which is known to be NP-hard, and suggests a simple
      heuristic based on successively matching an individual partition
      \eqn{x_b} to the current approximation to the consensus partition,
      and compute the memberships of the next approximation as a
      weighted average of those of the current one and of \eqn{x_b}
      after permuting its columns for the optimal matching of class
      ids.

      The following control parameters are available for this method.
      
      \describe{
	\item{\code{k}}{an integer giving the number of classes to be
	  used for the least squares consensus partition.  By default,
	  the maximal number of classes in the ensemble is used.}
	\item{\code{order}}{a permutation of the integers from 1 to the
	  size of the ensemble, specifying the order in which the
	  partitions in the ensemble should be aggregated.  Defaults to
	  using a random permutation (unlike the reference, which does
	  not permute at all).}
      }
    }

    \item{\code{"HE"}}{a fixed-point algorithm for obtaining \emph{hard}
      least squares Euclidean consensus partitions (i.e., for minimizing
      \eqn{L} with Euclidean dissimilarity \eqn{d} and \eqn{p =	2} over
      all hard partitions with a given maximal number of classes.)

      Available control parameters are the same as for \code{"SE"}.

      This method can also be referred to as \code{"hard/euclidean"}.
    }
    
    \item{\code{"SM"}}{a fixed-point algorithm for obtaining \emph{soft}
      median Manhattan consensus partitions (i.e., for minimizing
      \eqn{L} with Manhattan dissimilarity \eqn{d} and \eqn{p =	1} over 
      all soft partitions with a given maximal number of classes).

      Available control parameters are the same as for \code{"SE"}.

      This method can also be referred to as \code{"soft/manhattan"}.
    }

    \item{\code{"HM"}}{a fixed-point algorithm for obtaining \emph{hard}
      median Manhattan consensus partitions (i.e., for minimizing
      \eqn{L} with Manhattan dissimilarity \eqn{d} and \eqn{p =	1} over 
      all hard partitions with a given maximal number of classes).

      Available control parameters are the same as for \code{"SE"}.

      This method can also be referred to as \code{"hard/manhattan"}.
    }

    \item{\code{"GV3"}}{a \acronym{SUMT} algorithm for the \dQuote{third
      model} in \bibcitet{Gordon+Vichi:2001} for minimizing \eqn{L} with
      \eqn{d} being co-membership dissimilarity and \eqn{p = 2}.  (See
      \code{\link{sumt}} for more information on the \acronym{SUMT}
      approach.)  This optimization problem is equivalent to finding the
      membership matrix \eqn{m} for which the sum of the squared
      differences between \eqn{C(m) = m m'} and the weighted average
      co-membership matrix \eqn{\sum_b w_b C(m_b)} of the partitions is
      minimal.

      Available control parameters are \code{method}, \code{control},
      \code{eps}, \code{q}, and \code{verbose}, which have the same
      roles as for \code{\link{sumt}}, and the following.
      
      \describe{
	\item{\code{k}}{an integer giving the number of classes to be
	  used for the least squares consensus partition.  By default,
	  the maximal number of classes in the ensemble is used.}
	\item{\code{nruns}}{an integer giving the number of runs to be
	  performed.  Defaults to 1.}
	\item{\code{start}}{a matrix with number of rows equal to the
	  size of the cluster ensemble, and \eqn{k} columns, to be used
	  as a starting value, or a list of such matrices.  By default,
	  a membership based on a rank \eqn{k} approximation to the
	  weighted average co-membership matrix is used.}
      }
      In the case of multiple runs, the first optimum found is returned.
      
    }

    \item{\code{"soft/symdiff"}}{a \acronym{SUMT} approach for
      minimizing \eqn{L = \sum w_b d(x_b, x)} over all soft partitions
      with a given maximal number of classes, where \eqn{d} is the
      Manhattan dissimilarity of the co-membership matrices (coinciding
      with symdiff partition dissimilarity in the case of hard
      partitions).

      Available control parameters are the same as for \code{"GV3"}.
    }

    \item{\code{"hard/symdiff"}}{an exact solver for minimizing
      \eqn{L = \sum w_b d(x_b, x)} over all hard partitions (possibly
      with a given maximal number of classes as specified by the control
      parameter \code{k}), where \eqn{d} is symdiff partition
      dissimilarity (so that soft partitions in the ensemble are
      replaced by their closest hard partitions), or equivalently, Rand
      distance or pair-bonds (Boorman-Arabie \eqn{D}) distance.  The
      consensus solution is found via mixed linear or quadratic
      programming.
    }
  }
  
  By default, method \code{"SE"} is used for ensembles of partitions.
  
  If all elements of the ensemble are hierarchies, the following
  built-in methods for computing consensus hierarchies are available.
  
  \describe{
    \item{\code{"euclidean"}}{an algorithm for minimizing
      \eqn{L(x) = \sum w_b d(x_b, x) ^ 2} over all dendrograms, where
      \eqn{d} is Euclidean dissimilarity.  This is equivalent to finding
      the best least squares ultrametric approximation of the weighted
      average \eqn{d = \sum w_b u_b} of the ultrametrics \eqn{u_b} of
      the hierarchies \eqn{x_b}, which is attempted by calling
      \code{\link{ls_fit_ultrametric}} on \eqn{d} with appropriate
      control parameters.

      This method can also be referred to as \code{"cophenetic"}.
    }

    \item{\code{"manhattan"}}{a \acronym{SUMT} for minimizing
      \eqn{L = \sum w_b d(x_b, x)} over all dendrograms, where \eqn{d}
      is Manhattan dissimilarity.

      Available control parameters are the same as for
      \code{"euclidean"}.
    }
    
    \item{\code{"majority"}}{a hierarchy obtained from an extension of
      the majority consensus tree of \bibcitet{Margush+McMorris:1981},
      which minimizes \eqn{L(x) = \sum w_b d(x_b, x)} over all dendrograms,
      where \eqn{d} is the symmetric difference dissimilarity.  The
      unweighted \eqn{p}-majority tree is the \eqn{n}-tree (hierarchy in
      the strict sense) consisting of all subsets of objects contained
      in more than \eqn{100 p} percent of the \eqn{n}-trees \eqn{T_b}
      induced by the dendrograms, where \eqn{1/2 \le p < 1} and
      \eqn{p = 1/2} (default) corresponds to the standard majority tree.
      In the weighted case, it consists of all subsets \eqn{A} for which
      \eqn{\sum_{b: A \in T_b} w_b > W p}, where \eqn{W = \sum_b w_b}.
      We also allow for \eqn{p = 1}, which gives the \emph{strict
	consensus tree} consisting of all subsets contained in each of
      the \eqn{n}-trees.  The majority dendrogram returned is a
      representation of the majority tree where all splits have height
      one.

      The fraction \eqn{p} can be specified via the control parameter
      \code{p}.
    }
  }

  By default, method \code{"euclidean"} is used for ensembles of
  hierarchies.

  If a user-defined consensus method is to be employed, it must be a
  function taking the cluster ensemble, the case weights, and a list of
  control parameters as its arguments, with formals named \code{x},
  \code{weights}, and \code{control}, respectively.

  Most built-in methods use heuristics for solving hard optimization
  problems, and cannot be guaranteed to find a global minimum.  Standard
  practice would recommend to use the best solution found in
  \dQuote{sufficiently many} replications of the methods.
}
\references{
  \bibshow{Dimitriadou+Weingessel+Hornik:2002,
    Gordon+Vichi:2001,
    R:Gordon:1999,
    Margush+McMorris:1981}
}
\seealso{
  \code{\link{cl_medoid}},
  \code{\link[ape]{consensus}}
}
\examples{
## Consensus partition for the Rosenberg-Kim kinship terms partition
## data based on co-membership dissimilarities.
data("Kinship82")
m1 <- cl_consensus(Kinship82, method = "GV3",
                   control = list(k = 3, verbose = TRUE))
## (Note that one should really use several replicates of this.)
## Value for criterion function to be minimized:
sum(cl_dissimilarity(Kinship82, m1, "comem") ^ 2)
## Compare to the consensus solution given in Gordon & Vichi (2001).
data("Kinship82_Consensus")
m2 <- Kinship82_Consensus[["JMF"]]
sum(cl_dissimilarity(Kinship82, m2, "comem") ^ 2)
## Seems we get a better solution ...
## How dissimilar are these solutions?
cl_dissimilarity(m1, m2, "comem")
## How "fuzzy" are they?
cl_fuzziness(cl_ensemble(m1, m2))
## Do the "nearest" hard partitions fully agree?
cl_dissimilarity(as.cl_hard_partition(m1),
                 as.cl_hard_partition(m2))

## Consensus partition for the Gordon and Vichi (2001) macroeconomic
## partition data based on Euclidean dissimilarities.
data("GVME")
set.seed(1)
## First, using k = 2 classes.
m1 <- cl_consensus(GVME, method = "GV1",
                   control = list(k = 2, verbose = TRUE))
## (Note that one should really use several replicates of this.)
## Value of criterion function to be minimized:
sum(cl_dissimilarity(GVME, m1, "GV1") ^ 2)
## Compare to the consensus solution given in Gordon & Vichi (2001).
data("GVME_Consensus")
m2 <- GVME_Consensus[["MF1/2"]]
sum(cl_dissimilarity(GVME, m2, "GV1") ^ 2)
## Seems we get a slightly  better solution ...
## But note that
cl_dissimilarity(m1, m2, "GV1")
## and that the maximal deviation of the memberships is
max(abs(cl_membership(m1) - cl_membership(m2)))
## so the differences seem to be due to rounding.
## Do the "nearest" hard partitions fully agree?
table(cl_class_ids(m1), cl_class_ids(m2))

## And now for k = 3 classes.
m1 <- cl_consensus(GVME, method = "GV1",
                   control = list(k = 3, verbose = TRUE))
sum(cl_dissimilarity(GVME, m1, "GV1") ^ 2)
## Compare to the consensus solution given in Gordon & Vichi (2001).
m2 <- GVME_Consensus[["MF1/3"]]
sum(cl_dissimilarity(GVME, m2, "GV1") ^ 2)
## This time we look much better ...
## How dissimilar are these solutions?
cl_dissimilarity(m1, m2, "GV1")
## Do the "nearest" hard partitions fully agree?
table(cl_class_ids(m1), cl_class_ids(m2))
}
\keyword{cluster}

\name{n_of_objects}
\alias{n_of_objects}
\title{Number of Objects in a Partition or Hierarchy}
\description{Determine the number of objects from which a partition or
  hierarchy was obtained.}
\usage{
n_of_objects(x)
}
\arguments{
  \item{x}{an \R object representing a (hard of soft) partition or a
    hierarchy of objects, or dissimilarities between objects.}
}
\value{
  An integer giving the number of objects.
}
\details{
  This is a generic function.

  The methods provided in package \pkg{clue} handle the partitions and
  hierarchies obtained from clustering functions in the base R
  distribution, as well as packages \pkg{RWeka}, \pkg{ape}, \pkg{cba},
  \pkg{cclust}, \pkg{cluster}, \pkg{e1071}, \pkg{flexclust},
  \pkg{flexmix}, \pkg{kernlab}, \pkg{mclust}, \pkg{movMF} and
  \pkg{skmeans} (and of course, \pkg{clue} itself).

  There is also a method for object dissimilarities which inherit from
  class \code{"\link{dist}"}.
}
\seealso{
  \code{\link{is.cl_partition}},
  \code{\link{is.cl_hierarchy}}
}
\examples{
data("Cassini")
pcl <- kmeans(Cassini$x, 3)
n_of_objects(pcl)
hcl <- hclust(dist(USArrests))
n_of_objects(hcl)
}
\keyword{cluster}

\name{Kinship82_Consensus}
\alias{Kinship82_Consensus}
\title{Gordon-Vichi Kinship82 Consensus Partition Data}
\description{
  The soft (\dQuote{fuzzy}) consensus partitions for the Rosenberg-Kim
  kinship terms partition data given in \bibcitet{Gordon+Vichi:2001}.
}
\usage{data("Kinship82_Consensus")}
\format{
  A named cluster ensemble of three soft partitions of the 15 kinship
  terms into three classes.
}
\details{
  The elements of the ensemble are named \code{"MF1"}, \code{"MF2"}, and
  \code{"JMF"}, and correspond to the consensus partitions obtained by
  applying models 1, 2, and 3 in \bibcitet{Gordon+Vichi:2001} to the
  kinship   terms partition data in \bibcitet{Rosenberg:1982}, which are
  available as data set \code{\link{Kinship82}}.
}
\source{
  Table 6 in \bibcitet{Gordon+Vichi:2001}.
}
\references{
  \bibshow{Gordon+Vichi:2001, Rosenberg:1982}
}
\examples{
## Load the consensus partitions.
data("Kinship82_Consensus")
## Fuzziness using the Partition Coefficient.
cl_fuzziness(Kinship82_Consensus)
## (Corresponds to 1 - F in the source.)
## Dissimilarities:
cl_dissimilarity(Kinship82_Consensus)
cl_dissimilarity(Kinship82_Consensus, method = "comem")
}
\keyword{datasets}

\name{GVME_Consensus}
\alias{GVME_Consensus}
\title{Gordon-Vichi Macroeconomic Consensus Partition Data}
\description{
  The soft (\dQuote{fuzzy}) consensus partitions for the macroeconomic
  partition data given in \bibcitet{Gordon+Vichi:2001}.
}
\usage{data("GVME_Consensus")}
\format{
  A named cluster ensemble of eight soft partitions of 21 countries
  terms into two or three classes.
}
\details{
  The elements of the ensemble are consensus partitions for the
  macroeconomic partition data in \bibcitet{Gordon+Vichi:2001}, which are
  available as data set \code{\link{GVME}}.  Element names are of the
  form \code{"\var{m}/\var{k}"}, where \var{m} indicates the consensus
  method employed (one of \samp{MF1}, \samp{MF2}, \samp{JMF}, and
  \samp{S&S}, corresponding to the application of models 1, 2, and 3
  in \bibcitet{Gordon+Vichi:2001} and the approach in
  \bibcitet{Sato+Sato:1994},
  respectively), and \var{k} denotes the number classes (2 or 3).
}
\source{
  Tables 4 and 5 in \bibcitet{Gordon+Vichi:2001}.
}
\references{
  \bibshow{Gordon+Vichi:2001, Sato+Sato:1994}
}
\examples{
## Load the consensus partitions.
data("GVME_Consensus")
## Pick the partitions into 2 classes.
GVME_Consensus_2 <- GVME_Consensus[1 : 4]
## Fuzziness using the Partition Coefficient.
cl_fuzziness(GVME_Consensus_2)
## (Corresponds to 1 - F in the source.)
## Dissimilarities:
cl_dissimilarity(GVME_Consensus_2)
cl_dissimilarity(GVME_Consensus_2, method = "comem")
}
\keyword{datasets}

\name{partition}
\alias{cl_partition}		% class ...
\alias{is.cl_partition}
\alias{is.cl_hard_partition}
\alias{is.cl_soft_partition}
\alias{cl_hard_partition}	% class ...
\alias{as.cl_partition}
\alias{as.cl_hard_partition}
\title{Partitions}
\description{
  Determine whether an R object represents a partition of objects, or
  coerce to an R object representing such.}
\usage{
is.cl_partition(x)
is.cl_hard_partition(x)
is.cl_soft_partition(x)

as.cl_partition(x)
as.cl_hard_partition(x)
}
\arguments{
  \item{x}{an R object.}
}
\value{
  For the testing functions, a logical indicating whether the given
  object represents a clustering of objects of the respective kind.

  For the coercion functions, a container object inheriting from
  \code{"cl_partition"}, with a suitable representation of the partition
  given by \code{x}.
}
\details{
  \code{is.cl_partition} and \code{is.cl_hard_partition} are generic
  functions.

  The methods provided in package \pkg{clue} handle the partitions
  obtained from clustering functions in the base R distribution, as well
  as packages \pkg{RWeka}, \pkg{cba}, \pkg{cclust}, \pkg{cluster},
  \pkg{e1071}, \pkg{flexclust}, \pkg{flexmix}, \pkg{kernlab},
  \pkg{mclust}, \pkg{movMF} and \pkg{skmeans} (and of course, \pkg{clue}
  itself).

  \code{is.cl_soft_partition} gives true iff \code{is.cl_partition} is
  true and \code{is.cl_hard_partition} is false.

  \code{as.cl_partition} returns an object of class
  \code{"cl_partition"} \dQuote{containing} the given object \code{x} if
  this already represents a partition (i.e., \code{is.cl_partition(x)}
  is true), or the memberships obtained from \code{x} via
  \code{\link{as.cl_membership}}.

  \code{as.cl_hard_partition(x)} returns an object which has class
  \code{"cl_hard_partition"} and inherits from \code{"cl_partition"},
  and contains \code{x} if it already represents a hard partition (i.e.,
  provided that \code{is.cl_hard_partition(x)} is true), or the class
  ids obtained from \code{x}, using \code{x} if this is an atomic vector
  of raw class ids, or, if \code{x} represents a soft partition or is a
  raw matrix of membership values, using the class ids of the
  \emph{nearest hard partition}, defined by taking the class ids of the
  (first) maximal membership values.

  Conceptually, partitions and hard partitions are \emph{virtual}
  classes, allowing for a variety of representations.

  There are group methods for comparing partitions and computing their
  minimum, maximum, and range based on the meet and join operations, see
  \code{\link{cl_meet}}.
}
\examples{
data("Cassini")
pcl <- kmeans(Cassini$x, 3)
is.cl_partition(pcl)
is.cl_hard_partition(pcl)
is.cl_soft_partition(pcl)
}
\keyword{cluster}

\name{ls_fit_sum_of_ultrametrics}
\alias{ls_fit_sum_of_ultrametrics}
\title{Least Squares Fit of Sums of Ultrametrics to Dissimilarities}
\description{
  Find a sequence of ultrametrics with sum minimizing square distance
  (Euclidean dissimilarity) to a given dissimilarity object.
}
\usage{
ls_fit_sum_of_ultrametrics(x, nterms = 1, weights = 1,
                           control = list())
}
\arguments{
  \item{x}{a dissimilarity object inheriting from or coercible to class
    \code{"\link{dist}"}.}
  \item{nterms}{an integer giving the number of ultrametrics to be
    fitted.}
  \item{weights}{a numeric vector or matrix with non-negative weights
    for obtaining a weighted least squares fit.  If a matrix, its
    numbers of rows and columns must be the same as the number of
    objects in \code{x}, and the lower diagonal part is used.
    Otherwise, it is recycled to the number of elements in \code{x}.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\details{
  The problem to be solved is minimizing the criterion function
  \deqn{L(u(1), \dots, u(n)) =
    \sum_{i,j} w_{ij} \left(x_{ij} - \sum_{k=1}^n u_{ij}(k)\right)^2}{
    L(u(1), \dots, u(n)) =
    \sum_{i,j} w_{ij} \left(x_{ij} - \sum_{k=1}^n u_{ij}(k)\right)^2}
  over all \eqn{u(1), \ldots, u(n)} satisfying the ultrametric
  constraints.

  We provide an implementation of the iterative heuristic suggested in
  \bibcitet{Carroll+Pruzansky:1980} which in each step \eqn{t} sequentially
  refits the \eqn{u(k)} as the least squares ultrametric fit to the
  \dQuote{residuals} \eqn{x - \sum_{l \ne k} u(l)} using
  \code{\link{ls_fit_ultrametric}}.

  Available control parameters include
  \describe{
    \item{\code{maxiter}}{an integer giving the maximal number of
      iteration steps to be performed.
      Defaults to 100.}
    \item{\code{eps}}{a nonnegative number controlling the iteration,
      which stops when the maximal change in all \eqn{u(k)} is less than
      \code{eps}.
      Defaults to \eqn{10^{-6}}.}
    \item{\code{reltol}}{the relative convergence tolerance.  Iteration
      stops when the relative change in the criterion function is less
      than \code{reltol}.
      Defaults to \eqn{10^{-6}}.}
    \item{\code{method}}{a character string indicating the fitting
      method to be employed by the individual least squares fits.}
    \item{\code{control}}{a list of control parameters to be used by the
      method of \code{\link{ls_fit_ultrametric}} employed.  By default,
      if the \acronym{SUMT} method method is used, 10 inner
      \acronym{SUMT} runs are performed for each refitting.}
  }

  It should be noted that the method used is a heuristic which can not
  be guaranteed to find the global minimum.
}
\value{
  A list of objects of class \code{"\link{cl_ultrametric}"} containing
  the fitted ultrametric distances.
}
\references{
  \bibshow{Carroll+Pruzansky:1980}
}
\keyword{cluster}
\keyword{optimize}

\name{cl_medoid}
\alias{cl_medoid}
\title{Medoid Partitions and Hierarchies}
\description{
  Compute the medoid of an ensemble of partitions or hierarchies, i.e.,
  the element of the ensemble minimizing the sum of dissimilarities to
  all other elements.
}
\usage{
cl_medoid(x, method = "euclidean")
}
\arguments{
  \item{x}{an ensemble of partitions or hierarchies, or something
    coercible to that (see \code{\link{cl_ensemble}}).}
  \item{method}{a character string or a function, as for argument
    \code{method} of function \code{\link{cl_dissimilarity}}.}
}
\value{
  The medoid partition or hierarchy.
}
\details{
  Medoid clusterings are special cases of \dQuote{consensus} clusterings
  characterized as the solutions of an optimization problem.
  See \bibcitet{R:Gordon:1999} for more information.
  
  The dissimilarities \code{d} for determining the medoid are obtained
  by calling \code{cl_dissimilarity} with arguments \code{x} and
  \code{method}.  The medoid can then be found as the (first) row index
  for which the row sum of \code{as.matrix(d)} is minimal.  Modulo
  possible differences in the case of ties, this gives the same results
  as (the medoid obtained by) \code{\link[cluster]{pam}} in package
  \pkg{cluster}.
}
\references{
  \bibshow{R:Gordon:1999}
}  
\seealso{
  \code{\link{cl_consensus}}
}
\examples{
## An ensemble of partitions.
data("CKME")
pens <- CKME[1 : 20]
m1 <- cl_medoid(pens)
diss <- cl_dissimilarity(pens)
require("cluster")
m2 <- pens[[pam(diss, 1)$medoids]]
## Agreement of medoid consensus partitions.
cl_agreement(m1, m2)
## Or, more straightforwardly:
table(cl_class_ids(m1), cl_class_ids(m2))
}
\keyword{cluster}

\name{lattice}
\encoding{UTF-8}
\alias{cl_meet}
\alias{cl_join}
\alias{Ops.cl_partition}
\alias{Summary.cl_partition}
\alias{Ops.cl_dendrogram}
\alias{Ops.cl_hierarchy}
\alias{Summary.cl_hierarchy}
\title{Cluster Lattices}
\description{
  Computations on the lattice of all (hard) partitions, or the lattice
  of all dendrograms, or the meet semilattice of all hierarchies
  (\eqn{n}-trees) of/on a set of objects: meet, join, and comparisons.
}
\usage{
cl_meet(x, y)
cl_join(x, y)
}
\arguments{
  \item{x}{an ensemble of partitions or dendrograms or hierarchies, or
    an R object representing a partition or dendrogram or hierarchy.}
  \item{y}{an R object representing a partition or dendrogram or
    hierarchy.  Ignored if \code{x} is an ensemble.}
}
\details{
  For a given finite set of objects \eqn{X}, the set \eqn{H(X)} of all
  (hard) partitions of \eqn{X} can be partially ordered by defining a
  partition \eqn{P} to be \dQuote{finer} than a partition \eqn{Q}, i.e.,
  \eqn{P \le Q}, if each class of \eqn{P} is contained in some class of
  \eqn{Q}.  With this partial order, \eqn{H(X)} becomes a bounded
  \dfn{lattice}, with intersection and union of two elements given by
  their greatest lower bound (\dfn{meet}) and their least upper bound
  (\dfn{join}), respectively.

  Specifically, the meet of two partitions computed by \code{cl_meet} is
  the partition obtained by intersecting the classes of the partitions;
  the classes of the join computed by \code{cl_join} are obtained by
  joining all elements in the same class in at least one of the
  partitions.  Obviously, the least and greatest elements of the
  partition lattice are the partitions where each object is in a single
  class (sometimes referred to as the \dQuote{splitter} partition) or in
  the same class (the \dQuote{lumper} partition), respectively.  Meet
  and join of an arbitrary number of partitions can be defined
  recursively.

  In addition to computing the meet and join, the comparison operations
  corresponding to the above partial order as well as \code{min},
  \code{max}, and \code{range} are available at least for R objects
  representing partitions inheriting from \code{"\link{cl_partition}"}.
  The summary methods give the meet and join of the given partitions
  (for \code{min} and \code{max}), or a partition ensemble with the meet
  and join (for \code{range}).

  If the partitions specified by \code{x} and \code{y} are soft
  partitions, the corresponding nearest hard partitions are used.
  Future versions may optionally provide suitable \dQuote{soft} (fuzzy)
  extensions for computing meets and joins.

  The set of all dendrograms on \eqn{X} can be ordered using pointwise
  inequality of the associated ultrametric dissimilarities: i.e., if
  \eqn{D} and \eqn{E} are the dendrograms with ultrametrics \eqn{u} and
  \eqn{v}, respectively, then \eqn{D \le E} if \eqn{u_{ij} \le v_{ij}}
  for all pairs \eqn{(i, j)} of objects.  This again yields a lattice
  (of dendrograms).  The join of \eqn{D} and \eqn{E} is the dendrogram
  with ultrametrics given by \eqn{\max(u_{ij}, v_{ij})} (as this gives
  an ultrametric); the meet is the dendrogram with the maximal
  ultrametric dominated by \eqn{\min(u_{ij}, v_{ij})}, and can be
  obtained by applying single linkage hierarchical clustering to the
  minima.

  The set of all hierarchies on \eqn{X} can be ordered by set-wise
  inclusion of the classes: i.e., if \eqn{H} and \eqn{G} are two
  hierarchies, then \eqn{H \le G} if all classes of \eqn{H} are also
  classes of \eqn{G}.  This yields a meet semilattice, with meet given
  by the classes contained in both hierarchies.  The join only exists if
  the union of the classes is a hierarchy.

  In each case, a modular semilattice is obtained, which allows for a
  natural metrization via least element (semi)lattice move distances,
  see \bibcitet{Barthélemy+Leclerc+Monjardet:1986}.
  These latticial metrics
  are given by the BA/C (partitions), Manhattan (dendrograms), and
  symdiff (hierarchies) dissimilarities, respectively (see
  \code{\link{cl_dissimilarity}}).
}
\value{
  For \code{cl_meet} and \code{cl_join}, an object of class
  \code{"\link{cl_partition}"} or \code{"\link{cl_dendrogram}"} with the
  class ids or ultrametric dissimilarities of the meet and join of the
  partitions or dendrograms, respectively.
}
\references{
  \bibshow{Barthélemy+Leclerc+Monjardet:1986}
}
\examples{
## Two simple partitions of 7 objects.
A <- as.cl_partition(c(1, 1, 2, 3, 3, 5, 5))
B <- as.cl_partition(c(1, 2, 2, 3, 4, 5, 5))
## These disagree on objects 1-3, A splits objects 4 and 5 into
## separate classes.  Objects 6 and 7 are always in the same class.
(A <= B) || (B <= A)
## (Neither partition is finer than the other.)
cl_meet(A, B)
cl_join(A, B)
## Meeting with the lumper (greatest) or joining with the splitter
## (least) partition does not make a difference: 
C_lumper <- as.cl_partition(rep(1, n_of_objects(A)))
cl_meet(cl_ensemble(A, B, C_lumper))
C_splitter <- as.cl_partition(seq_len(n_of_objects(A)))
cl_join(cl_ensemble(A, B, C_splitter))
## Another way of computing the join:
range(A, B, C_splitter)$max
}
\keyword{cluster}

\name{cl_fuzziness}
\alias{cl_fuzziness}
\title{Partition Fuzziness}
\description{
  Compute the fuzziness of partitions.
}
\usage{
cl_fuzziness(x, method = NULL, normalize = TRUE)
}
\arguments{
  \item{x}{a cluster ensemble of partitions, or an R object coercible to
    such.}
  \item{method}{a character string indicating the fuzziness measure to
    be employed, or \code{NULL} (default), or a function to be taken as
    a user-defined method.  Currently available built-in methods are
    \code{"PC"} (Partition Coefficient) and \code{"PE"} (Partition
    Entropy), with the default corresponding to the first one.  If
    \code{method} is a character string, its lower-cased version is
    matched against the lower-cased names of the available built-in
    methods using \code{\link{pmatch}}.}
  \item{normalize}{a logical indicating whether the fuzziness measure
    should be normalized in a way that hard partitions have value 0, and
    \dQuote{completely fuzzy} partitions (where for all objects, all
    classes get the same membership) have value 1.}
}
\details{
  If \eqn{m} contains the membership values of a partition, the
  (unnormalized) Partition Coefficient and Partition Entropy are given
  by \eqn{\sum_{n,i} m_{n,i}^2} and \eqn{\sum_{n,i} H(m_{n,i})},
  respectively, where \eqn{H(u) = u \log u - (1-u) \log(1-u)}{u log(u)
    - (1-u) log(1-u)}.

  Note that the normalization used here is different from the
  normalizations typically found in the literature.

  If a user-defined fuzziness method is to be employed, is must be a
  function taking a matrix of membership values and a logical to
  indicate whether normalization is to be performed as its arguments (in
  that order; argument names are not used).
}
\value{
  An object of class \code{"cl_fuzziness"} giving the fuzziness
  values.
}
\references{
  \bibshow{Bezdek:1981}
}
\seealso{
  Function \code{\link[e1071]{fclustIndex}} in package \pkg{e1071},
  which also computes several other \dQuote{fuzzy cluster indexes}
  (typically based on more information than just the membership
  values).
}
\examples{
if(require("e1071", quietly = TRUE)) {
    ## Use an on-line version of fuzzy c-means from package e1071 if
    ## available.
    data("Cassini")
    pens <- cl_boot(Cassini$x, B = 15, k = 3, algorithm = "cmeans",
                    parameters = list(method = "ufcl"))
    pens
    summary(cl_fuzziness(pens, "PC"))
    summary(cl_fuzziness(pens, "PE"))
}
}
\keyword{cluster}

\name{Cassini}
\alias{Cassini}
\title{Cassini Data}
\description{
  A Cassini data set with 1000 points in 2-dimensional space which are
  drawn from the uniform distribution on 3 structures.  The two outer
  structures are banana-shaped; the \dQuote{middle} structure in between
  them is a circle.
}
\usage{data("Cassini")}
\format{
  A classed list with components
  \describe{
    \item{\code{x}}{a matrix with 1000 rows and 2 columns giving the
      coordinates of the points.}
    \item{\code{classes}}{a factor indicating which structure the
      respective points belong to.}
  }
}
\details{
  Instances of Cassini data sets can be created using function
  \code{\link[mlbench]{mlbench.cassini}} in package \pkg{mlbench}.
  The data set at hand was obtained using
  \preformatted{
    library("mlbench")
    set.seed(1234)
    Cassini <- mlbench.cassini(1000)
  }
}
\examples{
data("Cassini")
op <- par(mfcol = c(1, 2))
## Plot the data set:
plot(Cassini$x, col = as.integer(Cassini$classes),
     xlab = "", ylab = "")
## Create a "random" k-means partition of the data:
set.seed(1234)
party <- kmeans(Cassini$x, 3)
## And plot that.
plot(Cassini$x, col = cl_class_ids(party),
     xlab = "", ylab = "")
## (We can see the problem ...)
par(op)
}
\keyword{datasets}

\name{cl_boot}
\alias{cl_boot}
\title{Bootstrap Resampling of Clustering Algorithms}
\description{
  Generate bootstrap replicates of the results of applying a
  \dQuote{base} clustering algorithm to a given data set.
}
\usage{
cl_boot(x, B, k = NULL,
        algorithm = if (is.null(k)) "hclust" else "kmeans", 
        parameters = list(), resample = FALSE)
}
\arguments{
  \item{x}{the data set of objects to be clustered, as appropriate for
    the base clustering algorithm.}
  \item{B}{an integer giving the number of bootstrap replicates.}
  \item{k}{\code{NULL} (default), or an integer giving the number of
    classes to be used for a partitioning base algorithm.}
  \item{algorithm}{a character string or function specifying the base
    clustering algorithm.}
  \item{parameters}{a named list of additional arguments to be passed to
    the base algorithm.}
  \item{resample}{a logical indicating whether the data should be
    resampled in addition to \dQuote{sampling from the algorithm}.
    If resampling is used, the class memberships of the objects given in
    \code{x} are predicted from the results of running the base
    algorithm on bootstrap samples of \code{x}.}
}
\value{
  A cluster ensemble of length \eqn{B}, with either (if resampling is
  not used, default) the results of running the base algorithm on the
  given data set, or (if resampling is used) the memberships for the
  given data predicted from the results of running the base algorithm on
  bootstrap samples of the data.
}
\details{
  This is a rather simple-minded function with limited applicability,
  and mostly useful for studying the effect of (uncontrolled) random
  initializations of fixed-point partitioning algorithms such as
  \code{\link[stats]{kmeans}} or \code{\link[e1071]{cmeans}}, see the
  examples.  To study the effect of varying control parameters or
  explicitly providing random starting values, the respective cluster
  ensemble has to be generated explicitly (most conveniently by using
  \code{\link{replicate}} to create a list \code{lst} of suitable
  instances of clusterings obtained by the base algorithm, and using
  \code{cl_ensemble(list = lst)} to create the ensemble).
}
\examples{
## Study e.g. the effect of random kmeans() initializations.
data("Cassini")
pens <- cl_boot(Cassini$x, 15, 3)
diss <- cl_dissimilarity(pens)
summary(c(diss))
plot(hclust(diss))
}
\keyword{cluster}

\name{pclust}
\alias{pclust}
\alias{pclust_family}
\alias{pclust_object}
\title{Prototype-Based Partitioning}
\description{
  Obtain prototype-based partitions of objects by minimizing the criterion
  \eqn{\sum w_b u_{bj}^m d(x_b, p_j)^e}, the sum of the case-weighted and
  membership-weighted \eqn{e}-th powers of the dissimilarities between
  the objects \eqn{x_b} and the prototypes \eqn{p_j}, for suitable
  dissimilarities \eqn{d} and exponents \eqn{e}.
}
\usage{
pclust(x, k, family, m = 1, weights = 1, control = list())
pclust_family(D, C, init = NULL, description = NULL, e = 1,
              .modify = NULL, .subset = NULL)
pclust_object(prototypes, membership, cluster, family, m = 1,
              value, ..., classes = NULL, attributes = NULL)
}
\arguments{
  \item{x}{the object to be partitioned.}
  \item{k}{an integer giving the number of classes to be used in the
    partition.}
  \item{family}{an object of class \code{"pclust_family"} as generated
    by \code{pclust_family}, containing the information about \eqn{d}
    and \eqn{e}.}
  \item{m}{a number not less than 1 controlling the softness of the
    partition (as the \dQuote{fuzzification parameter} of the fuzzy
    \eqn{c}-means algorithm).  The default value of 1 corresponds to
    hard partitions obtained from a generalized \eqn{k}-means problem;
    values greater than one give partitions of increasing softness
    obtained from a generalized fuzzy \eqn{c}-means problem.}
  \item{weights}{a numeric vector of non-negative case weights.
    Recycled to the number of elements given by \code{x} if necessary.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
  \item{D}{a function for computing dissimilarities \eqn{d} between
    objects and prototypes.}
  \item{C}{a \sQuote{consensus} function with formals \code{x},
    \code{weights} and \code{control} for computing a consensus
    prototype \eqn{p} minimizing \eqn{\sum_b w_b d(x_b, p) ^ e}.}
  \item{init}{a function with formals \code{x} and \code{k} initializing
    an object with \eqn{k} prototypes from the object \code{x} to be
    partitioned.}
  \item{description}{a character string describing the family.}
  \item{e}{a number giving the exponent \eqn{e} of the criterion.}
  \item{.modify}{a function with formals \code{x}, \code{i} and
    \code{value} for modifying a single prototype,
    or \code{NULL} (default).}
  \item{.subset}{a function with formals \code{x} and \code{i} for
    subsetting prototypes,
    or \code{NULL} (default).}
  \item{prototypes}{an object representing the prototypes of the
    partition.}
  \item{membership}{an object of class \code{"\link{cl_membership}"}
    with the membership values \eqn{u_{bj}}.}
  \item{cluster}{the class ids of the nearest hard partition.}
  \item{value}{the value of the criterion to be minimized.}
  \item{...}{further elements to be included in the generated pclust
    object.}
  \item{classes}{a character vector giving further classes to be given
    to the generated pclust object in addition to \code{"pclust"}, or
    \code{NULL} (default).}
  \item{attributes}{a list of attributes, or \code{NULL} (default).}
}
\value{
  \code{pclust} returns the partition found as an object of class
  \code{"pclust"} (as obtained by calling \code{pclust_object}) which in
  addition to the \emph{default} components contains \code{call} (the 
  matched call) and a \code{converged} attribute indicating convergence
  status (i.e., whether the maximal number of iterations was reached).

  \code{pclust_family} returns an object of class
  \code{"pclust_family"}, which is a list with components corresponding
  to the formals of \code{pclust_family}.

  \code{pclust_object} returns an object inheriting from class
  \code{"pclust"}, which is a list with components corresponding
  to the formals (up to and including \code{...}) of
  \code{pclust_object}, and additional classes and attributes specified
  by \code{classes} and \code{attributes}, respectively.
}
\details{
  For \eqn{m = 1}, a generalization of the Lloyd-Forgy variant of the
  \eqn{k}-means algorithm is used, which iterates between reclassifying
  objects to their closest prototypes (according to the dissimilarities
  given by \code{D}), and computing new prototypes as the consensus for
  the classes (using \code{C}).

  For \eqn{m > 1}, a generalization of the fuzzy \eqn{c}-means recipe
  (e.g., \bibcitet{Bezdek:1981}) is used, which alternates between computing
  optimal memberships for fixed prototypes, and computing new prototypes
  as the suitably weighted consensus clusterings for the classes.

  This procedure is repeated until convergence occurs, or the maximal
  number of iterations is reached.

  Currently, no local improvement heuristics are provided.

  It is possible to perform several runs of the procedure via control
  arguments \code{nruns} or \code{start} (the default is to perform a
  single run), in which case the first partition with the smallest
  value of the criterion is returned.

  The dissimilarity and consensus functions as well as the exponent
  \eqn{e} are specified via \code{family}.  In principle, arbitrary
  representations of objects to be partitioned and prototypes (which do
  not necessarily have to be \dQuote{of the same kind}) can be employed.
  In addition to \code{D} and \code{C}, what is needed are means to
  obtain an initial collection of \eqn{k} prototypes (\code{init}), to
  modify a single prototype (\code{.modify}), and subset the prototypes
  (\code{.subset}).  By default, list and (currently, only dense) matrix
  (with the usual convention that the rows correspond to the objects)
  are supported.  Otherwise, the family has to provide the functions
  needed. 

  Available control parameters are as follows.
  \describe{
    \item{\code{maxiter}}{an integer giving the maximal number of
      iterations to be performed.
      Defaults to 100.}
    \item{\code{nruns}}{an integer giving the number of runs to be
      performed.
      Defaults to 1.} 
    \item{\code{reltol}}{the relative convergence tolerance.
      Defaults to \code{sqrt(.Machine$double.eps)}.}
    \item{\code{start}}{a list of prototype objects to be used as
      starting values.}
    \item{\code{verbose}}{a logical indicating whether to provide
      some output on minimization progress.
      Defaults to \code{getOption("verbose")}.}
    \item{\code{control}}{control parameters to be used in the consensus
      function.}
  }

  The fixed point approach employed is a heuristic which cannot be
  guaranteed to find the global minimum, in particular if \code{C} is
  not an exact minimizer.  Standard practice would recommend to use the
  best solution found in \dQuote{sufficiently many} replications of the
  base algorithm.
}
\references{
  \bibshow{Bezdek:1981}
}
\seealso{
  \code{\link[stats]{kmeans}},
  \code{\link[e1071]{cmeans}}.
}  

\name{cl_ensemble}
\alias{cl_ensemble}
\alias{as.cl_ensemble}
\alias{is.cl_ensemble}
\title{Cluster Ensembles}
\description{Creation and manipulation of cluster ensembles.}
\usage{
cl_ensemble(..., list = NULL)
as.cl_ensemble(x)
is.cl_ensemble(x)
}
\arguments{
  \item{\dots}{R objects representing clusterings of or dissimilarities
    between the same objects.}
  \item{list}{a list of R objects as in \code{\dots}.}
  \item{x}{for \code{as.cl_ensemble}, an R object as in \code{\dots};
    for \code{is.cl_ensemble}, an arbitrary R object.}
}
\details{
  \code{cl_ensemble} creates \dQuote{cluster ensembles}, which are
  realized as lists of clusterings (or dissimilarities) with additional
  class information, always inheriting from \code{"cl_ensemble"}.  All
  elements of the ensemble must have the same number of objects.

  If all elements are partitions, the ensemble has class
  \code{"cl_partition_ensemble"};
  if all elements are dendrograms, it has class
  \code{"cl_dendrogram_ensemble"} and inherits from
  \code{"cl_hierarchy_ensemble"};
  if all elements are hierarchies (but not always dendrograms), it has
  class \code{"cl_hierarchy_ensemble"}.
  Note that empty or \dQuote{mixed} ensembles cannot be categorized
  according to the kind of elements they contain, and hence only have
  class \code{"cl_ensemble"}.

  The list representation makes it possible to use \code{lapply} for
  computations on the individual clusterings in (i.e., the components
  of) a cluster ensemble.

  Available methods for cluster ensembles include those for
  subscripting, \code{c}, \code{rep}, and \code{print}.  There is also a
  \code{plot} method for ensembles for which all elements can be plotted
  (currently, additive trees, dendrograms and ultrametrics).
}
\value{
  \code{cl_ensemble} returns a list of the given clusterings or
  dissimilarities, with additional class information (see
  \bold{Details}).
}
\examples{
d <- dist(USArrests)
hclust_methods <-
    c("ward", "single", "complete", "average", "mcquitty")
hclust_results <- lapply(hclust_methods, function(m) hclust(d, m))
names(hclust_results) <- hclust_methods 
## Now create an ensemble from the results.
hens <- cl_ensemble(list = hclust_results)
hens
## Subscripting.
hens[1 : 3]
## Replication.
rep(hens, 3)
## Plotting.
plot(hens, main = names(hens))
## And continue to analyze the ensemble, e.g.
round(cl_dissimilarity(hens, method = "gamma"), 4)
}
\keyword{cluster}

\name{GVME}
\alias{GVME}
\title{Gordon-Vichi Macroeconomic Partition Ensemble Data}
\description{
  Soft partitions of 21 countries based on macroeconomic data for the
  years 1975, 1980, 1985, 1990, and 1995.
}
\usage{data("GVME")}
\format{
  A named cluster ensemble of 5 soft partitions of 21 countries into 2
  or 3 classes.  The names are the years to which the partitions
  correspond.
}
\details{
  The partitions were obtained using fuzzy \eqn{c}-means on measurements
  of the following variables: the annual per capita gross domestic
  product (GDP) in USD (converted to 1987 prices); the percentage of GDP
  provided by agriculture; the percentage of employees who worked in
  agriculture; and gross domestic investment, expressed as a percentage
  of the GDP.
  See \bibcitet{|Gordon+Vichi:2001|page 230}, for more details.
}
\source{
  Table 1 in \bibcitet{Gordon+Vichi:2001}.
}
\references{
  \bibshow{Gordon+Vichi:2001}
}
\keyword{datasets}

\name{cl_ultrametric}
\alias{cl_ultrametric}
\alias{as.cl_ultrametric}
\title{Ultrametrics of Hierarchies}
\description{
  Compute the ultrametric distances for objects representing (total
  indexed) hierarchies.
}
\usage{
cl_ultrametric(x, size = NULL, labels = NULL)
as.cl_ultrametric(x)
}
\arguments{
  \item{x}{an R object representing a (total indexed) hierarchy of
    objects.}
  \item{size}{an integer giving the number of objects in the hierarchy.}
  \item{labels}{a character vector giving the names of the objects in
    the hierarchy.}
}
\value{
  An object of class \code{"cl_ultrametric"} containing the ultrametric
  distances.
}  
\details{
  If \code{x} is not an ultrametric or a hierarchy with an ultrametric
  representation, \code{cl_ultrametric} uses
  \code{\link[stats]{cophenetic}} to obtain the ultrametric (also known
  as cophenetic) distances from the hierarchy, which in turn by default
  calls the S3 generic \code{\link[stats]{as.hclust}} on the hierarchy.
  Support for a class which represents hierarchies can thus be added by 
  providing \code{as.hclust} methods for this class.  In R 2.1.0 or
  better, \code{cophenetic} is an S3 generic as well, and one can also
  more directly provide methods for this if necessary.

  \code{as.cl_ultrametric} is a generic function which can be used for
  coercing \emph{raw} (non-classed) ultrametrics, represented as numeric
  vectors (of the lower-half entries) or numeric matrices, to
  ultrametric objects.

  Ultrametric objects are implemented as symmetric proximity objects
  with a dissimilarity interpretation so that self-proximities are zero,
  and inherit from classes \code{"\link{cl_dissimilarity}"} and
  \code{"cl_proximity"}.  See section \bold{Details} in the
  documentation for \code{\link{cl_dissimilarity}} for implications.

  Ultrametric objects can also be coerced to classes
  \code{"\link[stats]{dendrogram}"} and
  \code{"\link[stats]{hclust}"}, and hence in particular use the
  \code{plot} methods for these classes.  By default, plotting an
  ultrametric object uses the plot method for dendrograms.
}
\seealso{
  \code{\link{is.cl_hierarchy}}
}
\examples{
hc <- hclust(dist(USArrests))
u <- cl_ultrametric(hc)
## Subscripting.
u[1 : 5, 1 : 5]
u[1 : 5, 6 : 7]
## Plotting.
plot(u)
}
\keyword{cluster}

\name{ls_fit_addtree}
\encoding{UTF-8}
\alias{ls_fit_addtree}
\alias{ls_fit_centroid}
\title{Least Squares Fit of Additive Tree Distances to Dissimilarities}
\description{
  Find the additive tree distance or centroid distance minimizing least
  squares distance (Euclidean dissimilarity) to a given dissimilarity
  object.
}
\usage{
ls_fit_addtree(x, method = c("SUMT", "IP", "IR"), weights = 1,
               control = list())
ls_fit_centroid(x)
}
\arguments{
  \item{x}{a dissimilarity object inheriting from class
    \code{"\link{dist}"}.}
  \item{method}{a character string indicating the fitting method to be
    employed.  Must be one of \code{"SUMT"} (default), \code{"IP"}, or
    \code{"IR"}, or a unique abbreviation thereof.}
  \item{weights}{a numeric vector or matrix with non-negative weights
    for obtaining a weighted least squares fit.  If a matrix, its
    numbers of rows and columns must be the same as the number of
    objects in \code{x}, and the lower diagonal part is used.
    Otherwise, it is recycled to the number of elements in \code{x}.}
  \item{control}{a list of control parameters.  See \bold{Details}.}
}
\value{
  An object of class \code{"cl_addtree"} containing the optimal additive
  tree distances.
}
\details{
  See \code{\link{as.cl_addtree}} for details on additive tree distances
  and centroid distances.
  
  With \eqn{L(d) = \sum w_{ij} (x_{ij} - d_{ij})^2}, the problem to be
  solved by \code{ls_fit_addtree} is minimizing \eqn{L} over all
  additive tree distances \eqn{d}.  This problem is known to be NP
  hard.

  We provide three heuristics for solving this problem.

  Method \code{"SUMT"} implements the \acronym{SUMT}
  \bibcitep{Sequential Unconstrained Minimization Technique|Fiacco+McCormick:1968|}
  approach of \bibcitet{De_Soete:1983}.
  Incomplete dissimilarities are currently not supported.

  Methods \code{"IP"} and \code{"IR"} implement the Iterative
  Projection and Iterative Reduction approaches of
  \bibcitet{Hubert+Arabie:1995} and \bibcitet{Roux:1988}, respectively.
  Non-identical weights and
  incomplete dissimilarities are currently not supported.

  See \code{\link{ls_fit_ultrametric}} for details on these methods and
  available control parameters.

  It should be noted that all methods are heuristics which can not be
  guaranteed to find the global minimum.  Standard practice would
  recommend to use the best solution found in \dQuote{sufficiently many}
  replications of the base algorithm.

  \code{ls_fit_centroid} finds the centroid distance \eqn{d} minimizing
  \eqn{L(d)} (currently, only for the case of identical weights).  This
  optimization problem has a closed-form solution.
}
\references{
  \bibshow{Fiacco+McCormick:1968,
    Hubert+Arabie:1995,
    Roux:1988,
    De_Soete:1983}
}
\keyword{cluster}
\keyword{optimize}

\name{cl_margin}
\alias{cl_margin}
\title{Membership Margins}
\description{
  Compute the \emph{margin} of the memberships of a partition, i.e., the
  difference between the largest and second largest membership values of
  the respective objects.
}
\usage{
cl_margin(x)
}
\arguments{
  \item{x}{an \R object representing a partition of objects.}
}
\details{
  For hard partitions, the margins are always 1.

  For soft partitions, the margins may be taken as an indication of the
  \dQuote{sureness} of classifying an object to the class with maximum
  membership value.
}
\examples{
data("GVME")
## Look at the classes obtained for 1980:
split(cl_object_names(GVME[["1980"]]), cl_class_ids(GVME[["1980"]]))
## Margins:
x <- cl_margin(GVME[["1980"]])
## Add names, and sort:
names(x) <- cl_object_names(GVME[["1980"]])
sort(x)
## Note the "uncertainty" of assigning Egypt to the "intermediate" class
## of nations.
}
\keyword{cluster}

\name{addtree}
\encoding{UTF-8}
\alias{as.cl_addtree}
\title{Additive Tree Distances}
\description{
  Objects representing additive tree distances.
}
\usage{
as.cl_addtree(x)
}
\arguments{
  \item{x}{an R object representing additive tree distances.}
}
\value{
  An object of class \code{"cl_addtree"} containing the additive
  tree distances.
}
\details{
  Additive tree distances are object dissimilarities \eqn{d} satisfying
  the so-called \emph{additive tree conditions}, also known as
  \emph{four-point conditions} \eqn{d_{ij} + d_{kl} \le \max(d_{ik} +
    d_{jl}, d_{il} + d_{jk})} for all quadruples \eqn{i, j, k, l}.
  Equivalently, for each such quadruple, the largest two values of the
  sums \eqn{d_{ij} + d_{kl}}, \eqn{d_{ik} + d_{jl}}, and \eqn{d_{il} +
    d_{jk}} must be equal.
  Centroid distances are additive tree distances where the inequalities
  in the four-point conditions are strengthened to equalities (such that
  all three sums are equal), and can be represented as \eqn{d_{ij} = g_i
    + g_j}, i.e., as sums of distances from a \dQuote{centroid}.
  See, e.g., \bibcitet{Barthélemy+Guénoche:1991} for more details on
  additive tree distances.

  \code{as.cl_addtree} is a generic function.  Its default method can
  handle objects representing ultrametric distances and raw additive
  distance matrices.  In addition, there is a method for coercing
  objects of class \code{"\link[ape:as.phylo]{phylo}"} from package
  \pkg{ape}.

  Functions \code{\link{ls_fit_addtree}} and
  \code{\link{ls_fit_centroid}} can be used to find the additive tree
  distance or centroid distance minimizing least squares distance
  (Euclidean dissimilarity) to a given dissimilarity object.

  There is a \code{\link{plot}} method for additive tree distances.
}
\references{
  \bibshow{Barthélemy+Guénoche:1991}
}
\keyword{cluster}

\name{cl_tabulate}
\alias{cl_tabulate}
\title{Tabulate Vector Objects}
\description{Tabulate the unique values in vector objects.}
\usage{
cl_tabulate(x)
}
\arguments{
  \item{x}{a vector.}
}
\value{
  A data frame with components:
  \item{values}{the unique values.}
  \item{counts}{an integer vector with the number of times each of the
    unique values occurs in \code{x}.}
}
\examples{
data("Kinship82")
tab <- cl_tabulate(Kinship82)
## The counts:
tab$counts
## The most frequent partition:
tab$values[[which.max(tab$counts)]]
}
\keyword{utilities}
